/* alloset.c is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */

#include "biogeo.h"
#include "tessellation.h"

static SFLAG CompareSets (BITFIELD* nodeAsetA, BITFIELD* nodeAsetB, BITFIELD* nodeBsetA, BITFIELD* nodeBsetB);
static SFLAG IsNodeComp (PHYLONODE* nodeA, PHYLONODE* nodeB);

void DestroyAlloSet (ALLOPATRY* alloSet) {
 if (alloSet -> triangles != NULL)
   DestroyTriangleList (alloSet -> triangles);
 g_list_free (alloSet -> nodeList);
 g_slice_free1 (GETBITFIELDSIZE, alloSet -> setA);
 g_slice_free1 (GETBITFIELDSIZE, alloSet -> setB);
 g_slice_free1 (GETBITFIELDSIZE, alloSet -> baseA);
 g_slice_free1 (GETBITFIELDSIZE, alloSet -> baseB);
}

SFLAG IsNodeComp (PHYLONODE* nodeA, PHYLONODE* nodeB) {
 SFLAG toRet;
 SCOUNTER timeDiff;

 toRet = TRUE;
 /* Check if nodes are independent */
 if (nodeA -> tree == nodeB -> tree) {
   if (IsDesc (nodeA -> node, nodeB -> node))
     toRet = FALSE;
   else if (IsDesc (nodeB -> node, nodeA -> node))
     toRet = FALSE;
 }
 /* Check if nodes fit the time constraint */
 if (params . timeLag > 0) {
   timeDiff = (nodeA -> minAge >= nodeB -> minAge) ? nodeA -> minAge - nodeB -> minAge : nodeB -> minAge - nodeA -> minAge;
   if (timeDiff > params . timeLag)
     toRet = FALSE;
 }
 return toRet;
}

SFLAG CompareSets (BITFIELD* nodeAsetA, BITFIELD* nodeAsetB, BITFIELD* nodeBsetA, BITFIELD* nodeBsetB) {
 SFLAG toRet;

 toRet = FALSE;
 if (BitFieldAnd (nodeAsetA, nodeBsetA)) {
   if (BitFieldAnd (nodeAsetB, nodeBsetB) && (!(BitFieldAnd (nodeAsetA, nodeBsetB))) && (!(BitFieldAnd (nodeAsetB, nodeBsetA))))
     toRet = TRUE;
 }
 else if (BitFieldAnd (nodeAsetA, nodeBsetB)) {
   if (BitFieldAnd (nodeAsetB, nodeBsetA) && (!(BitFieldAnd (nodeAsetA, nodeBsetA))) && (!(BitFieldAnd (nodeAsetB, nodeBsetB))))
     toRet = TRUE;
 }
 return toRet;
}

void MakeCommonBarriers () {
 GList* navTree;
 GList* nuList;
 GList* navComp;
 GNode* navNode;
 PHYLOTREE* tree;
 PHYLONODE* node;
 PHYLONODE* compNode;
 SCOUNTER numNodes, i, j, k;
 GPtrArray* nodeArray;
 BITFIELD* tmpField;
 RECONS* actRec;

 /* Checks */
 tree = GETTREES -> data;
 if (tree -> recCons != NULL) {
   g_string_append (outLog, "\n\nConsensus of reconstructions");
   MakeConsensus ();
   return;
 }
 RemoveAlloSets ();
 MakeConsensus ();

 /* Initialization */
 numNodes = 0;
 tmpField = CreateBitField ();
 for (navTree = GETTREES; navTree != NULL; navTree = g_list_next (navTree)) {
   tree = (PHYLOTREE*) navTree -> data;
   node = (PHYLONODE*) tree -> root -> data;
   numNodes += node -> rec -> numVics;
 }
 nodeArray = g_ptr_array_sized_new (numNodes);
 for (navTree = GETTREES; navTree != NULL; navTree = g_list_next (navTree)) {
   tree = (PHYLOTREE*) navTree -> data;
   for (j = 0; j < tree -> nodeArray -> len; ++ j) {
     navNode = (GNode*) g_ptr_array_index (tree -> nodeArray, j);
     node = (PHYLONODE*) navNode -> data;
     if (node -> rec -> nodeFlag & VIC_NODE) {
       g_ptr_array_add (nodeArray, node);
       node -> alloSet = g_slice_alloc (sizeof (ALLOPATRY));
       node -> alloSet -> setA = CreateBitField ();
       node -> alloSet -> setB = CreateBitField ();
       CopyBitField (node -> alloSet -> setA, node -> rec -> setA -> rec -> cells);
       CopyBitField (node -> alloSet -> setB, node -> rec -> setB -> rec -> cells);
       BitFieldRemove (node -> alloSet -> setA, node -> rec -> setB -> rec -> cells);
       BitFieldRemove (node -> alloSet -> setB, node -> rec -> setA -> rec -> cells);
       node -> alloSet -> baseA = CreateBitField ();
       node -> alloSet -> baseB = CreateBitField ();
       CopyBitField (node -> alloSet -> baseA, node -> alloSet -> setA);
       CopyBitField (node -> alloSet -> baseB, node -> alloSet -> setB);
       node -> alloSet -> nodeList = NULL;
       node -> alloSet -> numNodes = 0;
       node -> alloSet -> triangles = NULL;
     }
   }
 }

 /* Compare nodes */
 for (j = 0; j < (nodeArray -> len - 1); ++ j) {
   node = (PHYLONODE*) g_ptr_array_index (nodeArray, j);
   for (k = j + 1; k < nodeArray -> len; ++ k) {
     compNode = (PHYLONODE*) g_ptr_array_index (nodeArray, k);
     if (IsNodeComp (node, compNode) == FALSE) continue;
     if (CompareSets (node -> alloSet -> setA, node -> alloSet -> setB, compNode -> alloSet -> setA, compNode -> alloSet -> setB)
                                     == FALSE) continue;
     nuList = InsertListElement (node -> alloSet -> nodeList, compNode);
     if (node -> alloSet -> nodeList == NULL)
       node -> alloSet -> nodeList = nuList;
     ++ node -> alloSet -> numNodes;
     nuList = InsertListElement (compNode -> alloSet -> nodeList, node);
     if (compNode -> alloSet -> nodeList == NULL)
       compNode -> alloSet -> nodeList = nuList;
     ++ compNode -> alloSet -> numNodes;
   }
 }
 /* Add distros */

 for (j = 0; j < nodeArray -> len; ++ j) {
   node = (PHYLONODE*) g_ptr_array_index (nodeArray, j);
   if (node -> alloSet -> numNodes == 0) continue;
   tree = node -> tree;
   ++ tree -> suppNodes;
   for (navComp = node -> alloSet -> nodeList; navComp != NULL; navComp = g_list_next (navComp)) {
     compNode = (PHYLONODE*) navComp -> data;
     if (BitFieldAnd (node -> alloSet -> baseA, compNode -> alloSet -> baseA)) {
       BitFieldORsum (node -> alloSet -> setA, compNode -> alloSet -> baseA);
       BitFieldORsum (node -> alloSet -> setB, compNode -> alloSet -> baseB);
     }
     else {
       BitFieldORsum (node -> alloSet -> setA, compNode -> alloSet -> baseB);
       BitFieldORsum (node -> alloSet -> setB, compNode -> alloSet -> baseA);
     }
   }
   CopyBitField (tmpField, node -> alloSet -> setA);
   BitFieldRemove (node -> alloSet -> setA, node -> alloSet -> setB);
   BitFieldRemove (node -> alloSet -> setB, tmpField);
 }
 g_slice_free1 (GETBITFIELDSIZE, tmpField);
 g_ptr_array_free (nodeArray, TRUE);

 /* Write the output */
 i = 0;
 for (navTree = GETTREES; navTree != NULL; navTree = g_list_next (navTree)) {
   tree = (PHYLOTREE*) navTree -> data;
   actRec = (RECONS*) g_ptr_array_index (tree -> recCons, 0);
   g_string_append_printf (outLog, "\n\nTree %ld, (%s)\nDisjunt sister pairs: %ld\nSupported disjunctions: %ld\n",
                           i, tree -> name, actRec -> numVics, tree -> suppNodes);
   ++ i;
 }
}
/* array.c is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */

#include "ghead.h"

INTARRAY* CreateIntArray (SCOUNTER size) {
 INTARRAY* nu;

 nu = (INTARRAY*) g_slice_alloc (sizeof (INTARRAY));
 nu -> max = size;
 nu -> len = 0;
 nu -> data = (SINTEGER*) g_malloc (sizeof (SINTEGER) * size);
 return nu;
}

void DestroyIntArray (INTARRAY* array) {
 g_free (array -> data);
 g_slice_free1 (sizeof (INTARRAY), array);
}

void ShuffleIntArray (INTARRAY* array) {
 SINTEGER temp;
 SCOUNTER i;
 SCOUNTER act;

 for (i = 0; i < array -> len; ++ i) {
   act = g_random_int_range (0, array -> len);
   temp = array -> data [i];
   array -> data [i] = array -> data [act];
   array -> data [act] = temp;
 }
}

SCOUNTER SearchIndexValue (INTARRAY* array, SINTEGER val) {
 SCOUNTER i;
 for (i = 0; i < array -> len; ++ i)
   if (array -> data [i] == val) return i;
 return 0;
}

void CopyPtrArray (GPtrArray* dest, GPtrArray* src) {
 SCOUNTER i;

 for (i = 0; i < src -> len; ++ i)
   g_ptr_array_add (dest, g_ptr_array_index (src, i));
}

void CopyPtrArrayPart (GPtrArray* dest, GPtrArray* src, SCOUNTER start, SCOUNTER end) {
 SCOUNTER i, max;

 max = (end > src -> len) ? src -> len : end;
 for (i = start; i < max; ++ i)
   g_ptr_array_add (dest, g_ptr_array_index (src, i));
}

void ShufflePtrArray (GPtrArray* array) {
 gpointer temp;
 SCOUNTER i;
 SCOUNTER act;

 for (i = 0; i < array -> len; ++ i) {
   act = g_random_int_range (0, array -> len);
   temp = g_ptr_array_index (array, i);
   array -> pdata [i] = array -> pdata [act];
   array -> pdata [act] = temp;
 }
}

void ShuffleSinglePtrArray (GPtrArray* array, SCOUNTER pos) {
 gpointer temp;
 SCOUNTER act;

 if (pos >= array -> len) return;
 act = g_random_int_range (0, array -> len);
 temp = g_ptr_array_index (array, pos);
 array -> pdata [pos] = array -> pdata [act];
 array -> pdata [act] = temp;
}
/* biogeo.c is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */

#include "biogeo.h"
#include "bitfield.h"
#include "tessellation.h"
#include <glib-object.h>

void StartBioGeo () {
 trees = NULL;
 lastTree = NULL;
 numTrees = 0;

 spp = NULL;
 lastSpp = NULL;
 numSpp = 0;

 geoRefs = NULL;
 lastGeoRef = NULL;
 numGeoRefs = 0;

 maxAzimuth = -1.00;
 minAzimuth = 361.00;
 maxInclination = -1.00;
 minInclination = 181.00;
 worldStart = -180.000000;

 params . costNoVic = 1;
 params . costRemov = 1;
 params . costPartRemov = 1;
 params . costTimeLag = 0;
 params . extra = 0;
 params . overlap = 0;
 params . timeLag = 0;
 params . maxTime = 0;
 params . timeDiff = 0;
 params . mode = USE_INTS;
 params . search = NORMAL_SEARCH;
 params . algorithm = PAGE_SEARCH | FLIP_NODES | FIRST_SECTOR;
 params . its = 100;
 params . acceptEqual = 50;
 params . sectSize = 20;
 params . generations = 1000;
 params . abandon = 20;
}

void DestroyBioGeo () {
 GList* navTree;

 if (trees != NULL) {
   DestroyRecs ();
   for (navTree = trees; navTree != NULL; navTree = g_list_next (navTree))
     DestroyTree ((PHYLOTREE*) navTree -> data);
   g_list_free (trees);
 }
 DestroyTermList (spp);
 DestroyGeoList (geoRefs);
 StartBioGeo ();
}

void DestroyTree (PHYLOTREE* tree) {
 GNode* actNode;
 GNode* nuAct;
 PHYLONODE* node;

 if (tree == NULL) return;
 g_ptr_array_free (tree -> nodeArray, TRUE);
 actNode = tree -> root;
 /* Deletes the tree without recursive function callings */
 while (actNode != NULL) {
   if (g_node_first_child (actNode) != NULL) {
     actNode = g_node_last_child (actNode);
     continue;
   }
   if (g_node_prev_sibling (actNode) != NULL)
     nuAct = g_node_prev_sibling (actNode);
   else
     nuAct = actNode -> parent;
   g_node_unlink (actNode);
   node = actNode -> data;
   if (node -> layout != NULL)
     g_object_unref (node -> layout);
   g_free (node -> name);
   g_slice_free1 (sizeof (PHYLONODE), node);
   g_node_destroy (actNode);
   actNode = nuAct;
 }
 g_free (tree -> name);
 g_slice_free1 (sizeof (PHYLOTREE), tree);
}

PHYLOTREE* CreateTree (gchar* name, SCOUNTER id) {
 PHYLOTREE* nu;

 nu = g_slice_alloc (sizeof (PHYLOTREE));
 if (name == NULL)
   nu -> name = g_strdup_printf ("unnamed tree %ld", id);
 else
   nu -> name = g_strdup ("unnamed tree");
 nu -> id = id;
 nu -> root = NULL;
 nu -> list = NULL;
 nu -> nodeArray = g_ptr_array_new ();
 nu -> recOR = NULL;
 nu -> recCons = NULL;
 nu -> recs = nu -> actRec = NULL;
 nu -> stored = 0;
 nu -> suppNodes = 0;
 nu -> maxTreeAge = NO_AGE;
 nu -> x = 0;
 nu -> y = 0;
 return nu;
}

PHYLOTREE* GetTreeFromID (SCOUNTER id) {
 GList* navTree;
 PHYLOTREE* toRet;

 if (trees == NULL) return NULL;
 for (navTree = trees; navTree != NULL; navTree = g_list_next (navTree)) {
   toRet = (PHYLOTREE*) navTree -> data;
   if (toRet -> id == id) return toRet;
 }
 return NULL;
}

GNode* CreateNode (PHYLOTREE* tree, GNode* anc, GNode* lastSibling, SCOUNTER id) {
 PHYLONODE* nu;
 GNode* toRet;

 nu = g_slice_alloc (sizeof (PHYLONODE));
 nu -> name = NULL;
 nu -> term = NULL;
 nu -> tree = tree;
 nu -> withData = 0;
 nu -> fill = 30000;
 nu -> numTerms = 0;
 nu -> level = 0;
 nu -> nested = 0;
 nu -> minAge = NO_AGE;
 nu -> constrained = FALSE;
 nu -> ageX = -1;
 nu -> minX = -1;
 nu -> maxX = -1;
 nu -> y = -1;
 nu -> startY = -1;
 nu -> endY = -1;
 nu -> size = 0;
 nu -> alloSet = NULL;
 toRet = g_node_new (nu);
 nu -> node = toRet;
 nu -> layout = NULL;
 AddChildren (toRet, anc, lastSibling);
 nu -> id = id;
 nu -> rec = NULL;
 return toRet;
}

void RemoveAlloSets () {
 GList* navTree;
 PHYLOTREE* actTree;
 GNode* node;
 PHYLONODE* actNode;
 SCOUNTER i;

 if (trees != NULL) {
   for (navTree = trees; navTree != NULL; navTree = g_list_next (navTree)) {
     actTree = (PHYLOTREE*) navTree -> data;
     for (i = 0; i < actTree -> nodeArray -> len; ++ i) {
       node = (GNode*) g_ptr_array_index (actTree -> nodeArray, i);
       actNode = (PHYLONODE*) node -> data;
       if (actNode -> alloSet != NULL) {
         DestroyAlloSet (actNode -> alloSet);
         g_slice_free1 (sizeof (ALLOPATRY), actNode -> alloSet);
         actNode -> alloSet = NULL;
       }
     }
   }
 }
}

void PrepareTree (PHYLOTREE* tree) {
 GNode* navNode;
 GNode* navAnc;
 PHYLONODE* actNode;
 PHYLONODE* ancNode;
 SCOUNTER y;
 SCOUNTER maxX;

 y = 0;
 actNode = (PHYLONODE*) tree -> root -> data;
 maxX = actNode -> level - 1;
 for (navNode = tree -> root; navNode != NULL; ) {
   actNode = (PHYLONODE*) navNode -> data;
   if ((actNode -> startY == -1) && (g_node_first_child (navNode) != NULL)) {
     navNode = g_node_first_child (navNode);
     continue;
   }
   actNode -> minX = actNode -> nested;
   if ((tree -> maxTreeAge > 0) && (tree -> maxTreeAge != NO_AGE))
     actNode -> ageX = tree -> maxTreeAge - actNode -> minAge;
   if (g_node_first_child (navNode) == NULL) {
     actNode -> maxX = maxX;
     actNode -> y = actNode -> startY = actNode -> endY = y;
     y += 2;
     for (navAnc = navNode -> parent; navAnc != NULL; navAnc = navAnc -> parent) {
       ancNode = (PHYLONODE*) navAnc -> data;
     }
   }
   else {
     actNode -> maxX = maxX - (actNode -> level - 1);
     actNode -> y = actNode -> startY + ((actNode -> endY - actNode -> startY) / 2);
   }
   if (navNode -> parent != NULL) {
     ancNode = (PHYLONODE*) navNode -> parent -> data;
     if (ancNode -> startY == -1)
       ancNode -> startY = actNode -> y;
     ancNode -> endY = actNode -> y;
   }
   if (g_node_next_sibling (navNode) == NULL)
     navNode = navNode -> parent;
   else
     navNode = g_node_next_sibling (navNode);
   continue;
 }
}

void SortNodes (SFLAG mode) {
 GList* navTree;
 PHYLOTREE* tree;
 PHYLONODE* actNode;
 PHYLONODE* descNode;
 PHYLONODE* sibNode;
 GNode* node;
 GNode* desc;
 GNode* sib;
 GNode* tmp;
 SCOUNTER i;
 SFLAG swap;
 SFLAG change;

 for (navTree = GETTREES; navTree != NULL; navTree = g_list_next (navTree)) {
   tree = (PHYLOTREE*) navTree -> data;
   change = FALSE;
   for (i = 0; i < tree -> nodeArray -> len; ++ i) {
     node = (GNode*) g_ptr_array_index (tree -> nodeArray, i);
     actNode = node -> data;
     if (g_node_first_child (node) == NULL) continue;
     for (desc = g_node_first_child (node); ; desc = g_node_next_sibling (desc)) {
       sib = g_node_next_sibling (desc);
       if (sib == NULL) break;
       descNode = (PHYLONODE*) desc -> data;
       for (; sib != NULL; sib = g_node_next_sibling (sib)) {
         swap = FALSE;
         sibNode = (PHYLONODE*) sib -> data;
         switch (mode) {
          case TREE_TERMS_UP: if (sibNode -> numTerms < descNode -> numTerms)
                 swap = TRUE;
               break;
          case TREE_TERMS_DOWN: if (sibNode -> numTerms > descNode -> numTerms)
                 swap = TRUE;
               break;
          case TREE_TERMS_MIX: if (((actNode -> nested % 2) == 0) && (sibNode -> numTerms < descNode -> numTerms))
                 swap = TRUE;
               else if (((actNode -> nested % 2) != 0) && (sibNode -> numTerms > descNode -> numTerms))
                 swap = TRUE;
               break;
          case TREE_TERMS_DEF: if (sibNode -> id < descNode -> id)
                 swap = TRUE;
               break;
         }
         if (swap == TRUE) {
           if (desc -> next == sib) {
             desc -> next = sib -> next;
             sib -> prev = desc -> prev;
             desc -> prev = sib;
             sib -> next = desc;
             if (sib -> prev != NULL)
               sib -> prev -> next = sib;
             if (desc -> next != NULL)
               desc -> next -> prev = desc;
           }
           else {
             if (desc -> prev != NULL)
               desc -> prev -> next = sib;
             desc -> next -> prev = sib;
             if (sib -> next != NULL)
               sib -> next -> prev = desc;
             sib -> prev -> next = desc;
             tmp = desc -> prev;
             desc -> prev = sib -> prev;
             sib -> prev = tmp;
             tmp = desc -> next;
             desc -> next = sib -> next;
             sib -> next = tmp;
           }
           if (desc == g_node_first_child (node))
             node -> children = sib;
           descNode = sibNode;
           tmp = desc;
           desc = sib;
           sib = tmp;
           change = TRUE;
         }
       }
     }
   }
   if (change == TRUE) {
     for (i = 0; i < tree -> nodeArray -> len; ++ i) {
       node = (GNode*) g_ptr_array_index (tree -> nodeArray, i);
       actNode = (PHYLONODE*) node -> data;
       actNode -> startY = -1;
     }
     PrepareTree (tree);
   }
 }
}

PHYLONODE* GetNodeFromID (PHYLOTREE* tree, SCOUNTER id) {
 SCOUNTER i;
 GNode* node;
 PHYLONODE* actNode;

 if (tree == NULL) return NULL;
 for (i = 0; i < tree -> nodeArray -> len; ++ i) {
   node = (GNode*) g_ptr_array_index (tree -> nodeArray, i);
   actNode = (PHYLONODE*) node -> data;
   if (actNode -> id == id) return actNode;
 }
 return NULL;
}

TERMINAL* CreateTerminal (GNode* node) {
 TERMINAL* nuTerm;

 nuTerm = g_slice_alloc (sizeof (TERMINAL));
 nuTerm -> name = NULL;
 nuTerm -> node = node;
 nuTerm -> list = NULL;
 nuTerm -> firstGeo = NULL;
 nuTerm -> numGeo = 0;
 nuTerm -> obs = NULL;
 nuTerm -> synonyms = NULL;
 return nuTerm;
}

void DestroyTermList (GList* list) {
 GList* nav;
 GList* navSyn;
 TERMINAL* term;

 if (list == NULL) return;
 if (g_list_previous (list) != NULL)
   g_list_previous (list) -> next = NULL;
 for (nav = list; nav != NULL; nav = g_list_next (nav)) {
   term = (TERMINAL*) nav -> data;
   if (term -> obs != NULL)
     g_slice_free1 (GETBITFIELDSIZE, term -> obs);
   if (term -> synonyms != NULL) {
     for (navSyn = term -> synonyms; navSyn != NULL; navSyn = g_list_next (navSyn))
       g_free (navSyn -> data);
     g_list_free (term -> synonyms);
   }
   g_slice_free1 (sizeof (TERMINAL), term);
 }
 g_list_free (list);
}

TERMINAL* GetTermFromName (gchar* name) {
 GList* nav;
 GList* navSyn;
 TERMINAL* term;
 gchar* syn;

 for (nav = spp; nav != NULL; nav = g_list_next (nav)) {
   term = (TERMINAL*) nav -> data;
   if (term -> name != NULL) {
     if (!g_strcmp0 (name, term -> name)) return term;
   }
 }
 for (nav = spp; nav != NULL; nav = g_list_next (nav)) {
   term = (TERMINAL*) nav -> data;
   for (navSyn = term -> synonyms; navSyn != NULL; navSyn = g_list_next (navSyn)) {
     syn = (gchar*) navSyn -> data;
     if (!g_strcmp0 (name, syn)) return term;
   }
 }
 return NULL;
}

gchar* AddSynonim (TERMINAL* actTerm, gchar* syn) {
 GList* nav;
 GList* nuSyn;
 gchar* toAdd;
 gchar* probe;
 SINTEGER cmp;

 if (actTerm -> synonyms == NULL) {
   toAdd = g_strdup (syn);
   nuSyn = InsertListElement (NULL, toAdd);
   actTerm -> synonyms = nuSyn;
   return toAdd;
 }
 toAdd = NULL;
 for (nav = actTerm -> synonyms; nav != NULL; nav = g_list_next (nav)) {
   probe = (gchar*) nav -> data;
   cmp = g_strcmp0 (syn, probe);
   if (cmp == 0) return probe;
   else if (cmp < 0) {
     toAdd = g_strdup (syn);
     if (nav == actTerm -> synonyms) {
       nuSyn = InsertListElement (NULL, toAdd);
       nuSyn -> next = actTerm -> synonyms;
       actTerm -> synonyms -> prev = nuSyn;
       actTerm -> synonyms = nuSyn;
       break;
     }
     InsertListElement (g_list_previous (nav), toAdd);
     break;
   }
   else if (g_list_next (nav) == NULL) {
     toAdd = g_strdup (syn);
     InsertListElement (nav, toAdd);
     break;
   }
 }
 return toAdd;
}

GList* CreateGeoRef (GList* last, GList* term, SREAL lon, SREAL lat) {
 GList* toRet;
 GEOREF* nuGeo;

 nuGeo = g_slice_alloc (sizeof (GEOREF));
 nuGeo -> lon = lon;
 nuGeo -> lat = lat;
 nuGeo -> actCell = -1;
 nuGeo -> azimuth = FROMLONTOAZM (lon);
 nuGeo -> inclination = FROMLATTOINC (lat);
 nuGeo -> sp = term;
 nuGeo -> actX = -1;
 nuGeo -> actY = -1;
 toRet = InsertListElement (last, nuGeo);
 nuGeo -> list = toRet;
 nuGeo -> name = NULL;
 nuGeo -> collec = NULL;
 nuGeo -> catalog = NULL;
 nuGeo -> url = NULL;
 nuGeo -> reference = NULL;
 nuGeo -> identifier = NULL;
 nuGeo -> sampleDate = NULL;
 nuGeo -> collector = NULL;
 return toRet;
}

void DestroyGeoList (GList* toDel) {
 GList* navGeo;
 GEOREF* actGeo;

 if (toDel == NULL) return;
 if (g_list_previous (toDel) != NULL)
   g_list_previous (toDel) -> next = NULL;
 for (navGeo = toDel; navGeo != NULL; navGeo = g_list_next (navGeo)) {
   actGeo = (GEOREF*) navGeo -> data;
   if (actGeo -> collec != NULL)
     g_free (actGeo -> collec);
   if (actGeo -> catalog != NULL)
     g_free (actGeo -> catalog);
   if (actGeo -> url != NULL)
     g_free (actGeo -> url);
   if (actGeo -> reference != NULL)
     g_free (actGeo -> reference);
   if (actGeo -> identifier != NULL)
     g_free (actGeo -> identifier);
   if (actGeo -> sampleDate != NULL)
     g_free (actGeo -> sampleDate);
 }
 DestroyList (toDel, sizeof (GEOREF));
}

RECONS* CreateRecons (PHYLONODE* node) {
 RECONS* toRet;

 toRet = (RECONS*) g_slice_alloc (sizeof (RECONS));
 toRet -> node = node;
 toRet -> cells = CreateBitField ();
 toRet -> obs = CreateBitField ();
 toRet -> setA = toRet -> setB = toRet -> removSet = NULL;
 toRet -> numVics = toRet -> numNoVics = toRet -> numRemovNodes = 0;
 toRet -> cost = 0;
 toRet -> numCells = toRet -> numObs = 0;
 toRet -> overlap = 0;
 toRet -> active = 0;
 toRet -> timeLag = 0;
 toRet -> maxAge = NO_AGE;
 toRet -> nodeFlag = 0;
 toRet -> triangles = NULL;
 return toRet;
}

void DestroySingleRec (GPtrArray* recArray) {
 SCOUNTER i;
 RECONS* actRec;
 PHYLONODE* node;

 for (i = 0; i < recArray -> len; ++ i) {
   actRec = (RECONS*) g_ptr_array_index (recArray, i);
   node = actRec -> node;
   if (node -> rec == actRec)
     node -> rec = NULL;
   if (actRec -> triangles != NULL)
     DestroyTriangleList (actRec -> triangles);
   g_slice_free1 (GETBITFIELDSIZE, actRec -> cells);
   g_slice_free1 (GETBITFIELDSIZE, actRec -> obs);
   g_slice_free1 (sizeof (RECONS), actRec);
 }
 g_ptr_array_free (recArray, TRUE);
}

void DestroyRecs () {
 GList* navTree;
 PHYLOTREE* tree;

 for (navTree = trees; navTree != NULL; navTree = g_list_next (navTree)) {
   tree = (PHYLOTREE*) navTree -> data;
   if (tree -> recOR != NULL) {
     DestroySingleRec (tree -> recOR);
     tree -> recOR = NULL;
   }
   if (tree -> recs != NULL) {
     DestroyRecList (tree -> recs);
     tree -> recs = NULL;
   }
   if (tree -> recCons != NULL) {
     DestroySingleRec (tree -> recCons);
     tree -> recCons = NULL;
   }
   tree -> actRec = NULL;
   tree -> stored = 0;
 }
}

void DestroyRecList (GList* recList) {
 GList* navRec;
 for (navRec = recList; navRec != NULL; navRec = g_list_next (navRec))
   DestroySingleRec ((GPtrArray*) navRec -> data);
 g_list_free (recList);
}

void FilterTree (PHYLOTREE* actTree, SREAL extra) {
 RECONS* actRec;
 SREAL orCost;
 SCOUNTER numRecs;
 GList* navRec;
 GList* nextRec;
 GPtrArray* tmpRec;

 if (actTree -> recs == NULL) return;
 /* Search for the best reconstruction */
 actRec = (RECONS*) g_ptr_array_index (actTree -> recOR, 0);
 orCost = actRec -> cost;
 if ((orCost < actTree -> best) && (extra = 0)) {
   /* All recosntructions are worst than actual or */
   DestroyRecList (actTree -> recs);
   actTree -> recs = NULL;
   actTree -> stored = 0;
   return;
 }
 numRecs = 0;
 for (navRec = actTree -> recs; navRec != NULL; navRec = nextRec) {
   nextRec = g_list_next (navRec);
   tmpRec = (GPtrArray*) navRec -> data;
   actRec = (RECONS*) g_ptr_array_index (tmpRec, 0);
   if (actRec -> cost > (actTree -> best + extra)) {
     if (navRec == actTree -> recs)
       actTree -> recs = nextRec;
     DestroySingleRec (tmpRec);
     DestroyLink (navRec);
   }
   else
     ++ numRecs;
 }
 actTree -> stored = numRecs;
}

void FilterRecs (SREAL extra) {
 GList* navTree;
 PHYLOTREE* tree;
 SCOUNTER j;

 j = 0;
 for (navTree = GETTREES; navTree != NULL; navTree = g_list_next (navTree)) {
   tree = (PHYLOTREE*) navTree -> data;
   ++ j;
   FilterTree (tree, extra);
   if (tree -> recs != NULL) {
     g_string_append_printf (outLog,"\n\nTree %ld (%s)\nBest score: %0.4f", j - 1, tree -> name, tree -> best);
     if (extra > 0)
       g_string_append_printf (outLog, "\nKeeping reconstructions between %.4f-%.4f", tree -> best, tree -> best + extra);
     else
       g_string_append_printf (outLog, "\nKeeping best reconstructions");
     g_string_append_printf (outLog, "\n%ld reconstructions keeped", tree -> stored);
   }
 }
}

GEOREF* AddGeoRef (SREAL lon, SREAL lat, GNode* actNode) {
 PHYLONODE* node;
 TERMINAL* term;
 GEOREF* geo;
 GList* nuGeo;

 node = (PHYLONODE*) actNode -> data;
 term = node -> term;
 if (term -> numGeo == 0) {
   nuGeo = CreateGeoRef (GETLASTGEOREF, term -> list, lon, lat);
   term -> firstGeo = nuGeo;
   if (GETGEOREF == NULL)
     SETGEOREF (nuGeo);
   SETLASTGEOREF (nuGeo);
 }
 else {
   nuGeo = CreateGeoRef (term -> firstGeo, term -> list, lon, lat);
   if (GETLASTGEOREF == term -> firstGeo)
     SETLASTGEOREF (nuGeo);
 }
 INCREASENUMGEOREFS (1);
 ++ term -> numGeo;
 geo = (GEOREF*) nuGeo -> data;
 UpdateLimits (geo);
 UpdateWithData (term);
 return geo;
}

void DeleteGeoRef (GEOREF* toDel) {
 GList* actGeo;
 TERMINAL* term;

 actGeo = toDel -> list;
 if (GETLASTGEOREF == actGeo)
   SETLASTGEOREF (g_list_previous (actGeo));
 if (GETGEOREF == actGeo)
   SETGEOREF (g_list_next (actGeo));
 term = (TERMINAL*) toDel -> sp -> data;
 if (term -> firstGeo == actGeo) {
   if (term -> numGeo > 1)
     term -> firstGeo = g_list_next (actGeo);
   else
     term -> firstGeo = NULL;
 }
 -- term -> numGeo;
 DECREASENUMGEOREFS (1);
 if (toDel -> collec != NULL)
   g_free (toDel -> collec);
 if (toDel -> catalog != NULL)
   g_free (toDel -> catalog);
 if (toDel -> url != NULL)
   g_free (toDel -> url);
 if (toDel -> reference != NULL)
   g_free (toDel -> reference);
 if (toDel -> identifier != NULL)
   g_free (toDel -> identifier);
 if (toDel -> sampleDate != NULL)
   g_free (toDel -> sampleDate);
 DestroyLinkElement (actGeo, sizeof (GEOREF));
 UpdateWithData (term);
}

void UpdateLimits (GEOREF* geo) {
 if (GETMAXAZM < geo -> azimuth)
   SETMAXAZM (geo -> azimuth);
 if (GETMINAZM > geo -> azimuth)
   SETMINAZM (geo -> azimuth);
 if (GETMAXINC < geo -> inclination)
   SETMAXINC (geo -> inclination);
 if (GETMININC > geo -> inclination)
   SETMININC (geo -> inclination);
}

SFLAG SetNodeAge (GNode* node, SCOUNTER age) {
 GNode* navNode;
 PHYLONODE* actNode;
 PHYLONODE* actNav;
 SCOUNTER i;
 PHYLOTREE* tree;

 actNode = (PHYLONODE*) node -> data;
 tree = actNode -> tree;
 if ((age < actNode -> minAge) && (g_node_first_child (node) != NULL)) return FALSE;
 actNode -> minAge = age;
 actNode -> ageX = tree -> maxTreeAge - actNode -> minAge;
 for (navNode = node -> parent; navNode != NULL; navNode = navNode -> parent) {
   actNav = (PHYLONODE*) navNode -> data;
   if (actNav -> minAge < actNode -> minAge) {
     actNav -> minAge = actNode -> minAge;
     actNav -> ageX = tree -> maxTreeAge - actNav -> minAge;
   }
   else break;
   actNode = actNav;
 }
 if ((age != NO_AGE) && (age > tree -> maxTreeAge))
   tree -> maxTreeAge = age;
 else return TRUE;
 if ((age > params . maxTime) && (age != NO_AGE)) {
   params . maxTime = age;
   params . timeDiff = params . maxTime - params . timeLag;
 }
 for (i = 0; i < tree -> nodeArray -> len; ++ i) {
   navNode = (GNode*) g_ptr_array_index (tree -> nodeArray, i);
   actNode = (PHYLONODE*) navNode -> data;
   actNode -> ageX = tree -> maxTreeAge - actNode -> minAge;
 }
 return TRUE;
}

SCOUNTER GetNumTermsWithGeo () {
 SCOUNTER num;
 GList* nav;
 TERMINAL* actTerm;

 for (nav = GETSPP, num = 0; nav != NULL; nav = g_list_next (nav)) {
   actTerm = (TERMINAL*) nav -> data;
   if (actTerm -> numGeo > 0)
     ++ num;
 }
 return num;
}

void UpdateWithData (TERMINAL* actTerm) {
 GNode* navNode;
 PHYLONODE* actNode;

 navNode = actTerm -> node;
 actNode = (PHYLONODE*) navNode -> data;
 if ((actTerm -> numGeo > 0) && (actNode -> withData == 0)) {
   for (; navNode != NULL; navNode = navNode -> parent) {
     actNode = (PHYLONODE*) navNode -> data;
     ++ actNode -> withData;
   }
 }
 else if ((actTerm -> numGeo == 0) && (actNode -> withData == 1))  {
   for (; navNode != NULL; navNode = navNode -> parent) {
     actNode = (PHYLONODE*) navNode -> data;
     ++ actNode -> withData;
   }
 }
}

void PutFillValue (PHYLONODE* node, SCOUNTER fillValue) {
 GNode* navDesc;
 GNode* actNode;

 node -> fill = fillValue;
 actNode = node -> node;
 for (navDesc = g_node_first_child (actNode); navDesc != NULL; navDesc = g_node_next_sibling (navDesc))
   PutFillValue ((PHYLONODE*) navDesc -> data, fillValue);
}
/* bitfield.c is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */

#include "bitfield.h"

static SCOUNTER numBits [65536];
static SCOUNTER maxNumFields = 0;

static void AddBits (BITFIELD act);

void AddBits (BITFIELD act) {
 SCOUNTER count, i;
 BITFIELD bit, trgt;

 count = numBits [act] + 1;
 for (i = 0; i < BITS_ON_FIELD; ++ i) {
   bit = 1 << i;
   trgt = bit | act;
   if (numBits [trgt] == 0) {
     numBits [trgt] = count;
     AddBits (trgt);
   }
 }
}

void StartBitFields (SCOUNTER bits) {
 SCOUNTER i;
 static SFLAG bitCountReady = FALSE;

 if (bits == 0) {
   EndBitFields ();
   return;
 }
 maxNumFields = (bits / BITS_ON_FIELD) + 1;
 maxFieldSize = sizeof (BITFIELD) * maxNumFields;
 if (bitCountReady == FALSE) {
   for (i = 0; i < 65536; ++ i)
     numBits [i] = 0;
   AddBits (0);
 }
 bitCountReady = TRUE;
}

void EndBitFields () {
 maxNumFields = 0;
 maxFieldSize = 0;
}

BITFIELD* CreateBitField () {
 BITFIELD* nu;
 SCOUNTER i;

 nu = (BITFIELD*) g_slice_alloc (maxFieldSize);
 for (i = 0; i < maxNumFields; ++ i)
   nu [i] = 0;
 return nu;
}

void ResetBitField (BITFIELD* a) {
 SCOUNTER i;
 for (i = 0; i < maxNumFields; ++ i)
   a [i] = 0;
}

void PutBitOn (BITFIELD* a, SCOUNTER bit) {
 SCOUNTER actField;
 SCOUNTER actBit;

 actField = bit / BITS_ON_FIELD;
 actBit = bit % BITS_ON_FIELD;
 a [actField] |= (1 << actBit);
}

SFLAG IsBitOn (BITFIELD* a, SCOUNTER bit) {
 SCOUNTER actField;
 SCOUNTER actBit;

 actField = bit / BITS_ON_FIELD;
 actBit = 1 << (bit % BITS_ON_FIELD);
 return (a [actField] & actBit) ? TRUE : FALSE;
}

void CopyBitField (BITFIELD* dest, BITFIELD* src) {
 SCOUNTER i;
 for (i = 0; i < maxNumFields; ++ i)
   dest [i] = src [i];
}

SFLAG BitFieldAnd (BITFIELD* a1, BITFIELD* a2) {
 SCOUNTER i;
 for (i = 0; i < maxNumFields; ++ i) {
   if (a1 [i] & a2 [i]) return TRUE;
 }
 return FALSE;
}

void BitFieldORsum (BITFIELD* dest, BITFIELD* src) {
 SCOUNTER i;
 for (i = 0; i < maxNumFields; ++ i)
   dest [i] |= src [i];
}

void BitFieldANDsum (BITFIELD* dest, BITFIELD* src) {
 SCOUNTER i;
 for (i = 0; i < maxNumFields; ++ i)
   dest [i] &= src [i];
}

void BitFieldRemove (BITFIELD* dest, BITFIELD* toRemov) {
 SCOUNTER i;
 for (i = 0; i < maxNumFields; ++ i)
   dest [i] &= ~ toRemov [i];
}

SCOUNTER CountBits (BITFIELD* a) {
 SCOUNTER num;
 SCOUNTER i;

 num = 0;
 for (i = 0; i < maxNumFields; ++ i)
   num += numBits [a [i]];
 return num;
}

SCOUNTER BitFieldCommon (BITFIELD* a, BITFIELD* b) {
 SCOUNTER num;
 SCOUNTER i;

 num = 0;
 for (i = 0; i < maxNumFields; ++ i)
   num += numBits [a [i] & b [i]];
 return num;
}

void BitFieldAddNotRemoved (BITFIELD* dest, BITFIELD* src, BITFIELD* toRemov) {
 SCOUNTER i;
 for (i = 0; i < maxNumFields; ++ i)
   dest [i] |= (src [i] & (~ toRemov [i]));
}
/* dialogs.c is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */

#include "wins.h"
#include "dialogs.h"
#include "out.h"
#include "parser.h"
#include "biogeo.h"
#include "grid.h"
#include <string.h>

GRIDDIALOG* GridDialog (SREAL size, SINTEGER fill, SFLAG editable) {
 GRIDDIALOG* toRet;
 GtkWidget* separator;
 gchar* text;

 toRet = g_slice_alloc (sizeof (GRIDDIALOG));
 toRet -> dialog = gtk_dialog_new_with_buttons ("Grid settings", GTK_WINDOW (GETMAINWIN),
                       GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR,
                       GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, NULL);

 toRet -> sizeEntry = CreateEntry ("Grid Size", GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));

 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);

 toRet -> fillEntry = CreateEntry ("Maximum fill", GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));
 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);

 toRet -> mooreRadio = gtk_radio_button_new_with_mnemonic (NULL, "_Moore neightborhood");
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), toRet -> mooreRadio, TRUE, TRUE, 2);
 gtk_widget_show (toRet -> mooreRadio);
 toRet -> vonneRadio = gtk_radio_button_new_with_mnemonic_from_widget (GTK_RADIO_BUTTON (toRet -> mooreRadio),
                                          "_Von Neumann neighborhood");
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), toRet -> vonneRadio, TRUE, TRUE, 2);
 gtk_widget_show (toRet -> vonneRadio);

 text = g_strdup_printf ("%.4f", size);
 CheckRealStr (text);
 gtk_entry_set_text (GTK_ENTRY (toRet -> sizeEntry), text);
 g_free (text);
 text = g_strdup_printf ("%ld", fill);
 gtk_entry_set_text (GTK_ENTRY (toRet -> fillEntry), text);
 g_free (text);
 if (GETGRIDMODE & MOORE_NEIGHBORS)
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toRet -> mooreRadio), TRUE);
 else if (GETGRIDMODE & VONNEUMANN_NEIGHBORS)
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toRet -> vonneRadio), TRUE);
 gtk_editable_set_editable (GTK_EDITABLE (toRet -> sizeEntry), editable);
 gtk_editable_set_editable (GTK_EDITABLE (toRet -> fillEntry), editable);
 gtk_widget_set_sensitive (toRet -> mooreRadio, editable);
 gtk_widget_set_sensitive (toRet -> vonneRadio, editable);
 return toRet;
}

void DestroyGridDialog (GRIDDIALOG* grid) {
 gtk_widget_destroy (grid -> dialog);
 g_slice_free1 (sizeof (GRIDDIALOG), grid);
}

ENTRYDIALOG* CounterEntryDialog (GtkWindow* parent, SCOUNTER count, gchar* name, gchar* caption) {
 ENTRYDIALOG* toRet;
 GtkWidget* separator;
 gchar* text;

 toRet = g_slice_alloc (sizeof (ENTRYDIALOG));
 toRet -> dialog = gtk_dialog_new_with_buttons (name, GTK_WINDOW (parent),
                       GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR,
                       GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, NULL);

 toRet -> entry = CreateEntry (caption, GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));

 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);

 text = g_strdup_printf ("%ld", count);
 gtk_entry_set_text (GTK_ENTRY (toRet -> entry), text);
 g_free (text);

 return toRet;
}

ENTRYDIALOG* RealEntryDialog (GtkWindow* parent, SREAL number, gchar* name, gchar* caption) {
 ENTRYDIALOG* toRet;
 GtkWidget* separator;
 gchar* text;

 toRet = g_slice_alloc (sizeof (ENTRYDIALOG));
 toRet -> dialog = gtk_dialog_new_with_buttons (name, GTK_WINDOW (parent),
                       GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR,
                       GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, NULL);

 toRet -> entry = CreateEntry (caption, GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));

 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);

 text = g_strdup_printf ("%.4f", number);
 CheckRealStr (text);
 gtk_entry_set_text (GTK_ENTRY (toRet -> entry), text);
 g_free (text);

 return toRet;
}

ENTRYDIALOG* TextEntryDialog (GtkWindow* parent, gchar* text, gchar* name, gchar* caption) {
 ENTRYDIALOG* toRet;
 GtkWidget* separator;

 toRet = g_slice_alloc (sizeof (ENTRYDIALOG));
 toRet -> dialog = gtk_dialog_new_with_buttons (name, GTK_WINDOW (parent),
                       GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR,
                       GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, NULL);

 toRet -> entry = CreateEntry (caption, GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));

 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);

 gtk_entry_set_text (GTK_ENTRY (toRet -> entry), text);
 return toRet;
}

void DestroyEntryDialog (ENTRYDIALOG* toDel) {
 gtk_widget_destroy (toDel -> dialog);
 g_slice_free1 (sizeof (ENTRYDIALOG), toDel);
}

CENTREDIALOG* CentreDialog () {
 CENTREDIALOG* toRet;
 GtkWidget* separator;

 toRet = g_slice_alloc (sizeof (CENTREDIALOG));
 toRet -> dialog = gtk_dialog_new_with_buttons ("Centre map", GTK_WINDOW (GETMAPWIN),
                       GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR,
                       GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, NULL);
 toRet -> latEntry = CreateEntry ("Latitude", GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));
 toRet -> lonEntry = CreateEntry ("Longitude", GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));

 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);

 gtk_entry_set_text (GTK_ENTRY (toRet -> latEntry), "0.0000");
 gtk_entry_set_text (GTK_ENTRY (toRet -> lonEntry), "0.0000");
 return toRet;
}

void DestroyCentreDialog (CENTREDIALOG* centre) {
 gtk_widget_destroy (centre -> dialog);
 g_slice_free1 (sizeof (CENTREDIALOG), centre);
}

MAPLIMITSDIALOG* MapLimitsDialog (SREAL n, SREAL s, SREAL e, SREAL w) {
 MAPLIMITSDIALOG* toRet;
 GtkWidget* separator;
 gchar* text;

 toRet = g_slice_alloc (sizeof (MAPLIMITSDIALOG));
 toRet -> dialog = gtk_dialog_new_with_buttons ("Set map limits", GTK_WINDOW (GETMAPWIN),
                       GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR,
                       GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, NULL);

 toRet -> northEntry = CreateEntry ("North (Latitude)", GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));
 toRet -> southEntry = CreateEntry ("South (Latitude)", GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));
 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);

 toRet -> eastEntry = CreateEntry ("East (Longitude)", GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));
 toRet -> westEntry = CreateEntry ("West (Longitude)", GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));
 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);

 text = g_strdup_printf ("%.4f", n);
 CheckRealStr (text);
 gtk_entry_set_text (GTK_ENTRY (toRet -> northEntry), text);
 g_free (text);
 text = g_strdup_printf ("%.4f", s);
 CheckRealStr (text);
 gtk_entry_set_text (GTK_ENTRY (toRet -> southEntry), text);
 g_free (text);
 text = g_strdup_printf ("%.4f", e);
 CheckRealStr (text);
 gtk_entry_set_text (GTK_ENTRY (toRet -> eastEntry), text);
 g_free (text);
 text = g_strdup_printf ("%.4f", w);
 CheckRealStr (text);
 gtk_entry_set_text (GTK_ENTRY (toRet -> westEntry), text);
 g_free (text);
 return toRet;
}

void DestroyMapLimitsDialog (MAPLIMITSDIALOG* mapLimits) {
 gtk_widget_destroy (mapLimits -> dialog);
 g_slice_free1 (sizeof (CENTREDIALOG), mapLimits);
}

SEARCHDIALOG* SearchDialog () {
 SEARCHDIALOG* toRet;
 GtkWidget* separator;
 gchar* text;

 toRet = g_slice_alloc (sizeof (SEARCHDIALOG));
 toRet -> dialog = gtk_dialog_new_with_buttons ("Search", GTK_WINDOW (GETMAINWIN),
                       GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR,
                       GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, NULL);

 toRet -> iterEntry = CreateEntry ("Iterations", GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));
 toRet -> equalEntry = CreateEntry ("Accept equals (percent)", GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));
 toRet -> pageCheck = gtk_check_button_new_with_mnemonic ("_Page\'s heuristic");
 toRet -> consCheck = gtk_check_button_new_with_mnemonic ("_OR constrained search");
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), toRet -> pageCheck, TRUE, TRUE, 2);
 gtk_widget_show (toRet -> pageCheck);
 toRet -> flipCheck = gtk_check_button_new_with_mnemonic ("_Flip nodes");
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), toRet -> flipCheck, TRUE, TRUE, 2);
 gtk_widget_show (toRet -> flipCheck);
 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);
 toRet -> noSecRadio = gtk_radio_button_new_with_mnemonic (NULL, "_No sectors");
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), toRet -> noSecRadio, TRUE, TRUE, 2);
 gtk_widget_show (toRet -> noSecRadio);
 toRet -> startSecRadio = gtk_radio_button_new_with_mnemonic_from_widget (GTK_RADIO_BUTTON (toRet -> noSecRadio), "_Start with a sector");
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), toRet -> startSecRadio, TRUE, TRUE, 2);
 gtk_widget_show (toRet -> startSecRadio);
 toRet -> secRadio = gtk_radio_button_new_with_mnemonic_from_widget (GTK_RADIO_BUTTON (toRet -> noSecRadio), "_Full sector search");
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), toRet -> secRadio, TRUE, TRUE, 2);
 gtk_widget_show (toRet -> secRadio);
 toRet -> secSizeEntry = CreateEntry ("Sector size", GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));
 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);
 toRet -> noAnnealRadio = gtk_radio_button_new_with_mnemonic (NULL, "N_o annealing");
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), toRet -> noAnnealRadio, TRUE, TRUE, 2);
 gtk_widget_show (toRet -> noAnnealRadio);
 toRet -> blindAnnealRadio = gtk_radio_button_new_with_mnemonic_from_widget
                             (GTK_RADIO_BUTTON (toRet -> noAnnealRadio), "_Blind annealing");
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), toRet -> blindAnnealRadio, TRUE, TRUE, 2);
 gtk_widget_show (toRet -> blindAnnealRadio);
 toRet -> gibbsAnnealRadio = gtk_radio_button_new_with_mnemonic_from_widget
                             (GTK_RADIO_BUTTON (toRet -> noAnnealRadio), "_Gibbs annealing");
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), toRet -> gibbsAnnealRadio, TRUE, TRUE, 2);
 gtk_widget_show (toRet -> gibbsAnnealRadio);
 toRet -> nixonAnnealRadio = gtk_radio_button_new_with_mnemonic_from_widget
                             (GTK_RADIO_BUTTON (toRet -> noAnnealRadio), "Ni_xon annealing");
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), toRet -> nixonAnnealRadio, TRUE, TRUE, 2);
 gtk_widget_show (toRet -> nixonAnnealRadio);
 toRet -> generationEntry = CreateEntry ("Generations", GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));
 toRet -> abandonEntry = CreateEntry ("Recheck each", GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));
 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);
 toRet -> keepCheck = gtk_check_button_new_with_mnemonic ("_Keep actual reconstructions");
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), toRet -> keepCheck, TRUE, TRUE, 2);
 gtk_widget_show (toRet -> keepCheck);
 toRet -> allCheck = gtk_check_button_new_with_mnemonic ("Keep _all reconstructions");
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), toRet -> allCheck, TRUE, TRUE, 2);
 gtk_widget_show (toRet -> allCheck);
 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);
 text = g_strdup_printf ("%ld", params . its);
 gtk_entry_set_text (GTK_ENTRY (toRet -> iterEntry), text);
 g_free (text);
 text = g_strdup_printf ("%ld", params . acceptEqual);
 gtk_entry_set_text (GTK_ENTRY (toRet -> equalEntry), text);
 g_free (text);
 text = g_strdup_printf ("%ld", params . sectSize);
 gtk_entry_set_text (GTK_ENTRY (toRet -> secSizeEntry), text);
 g_free (text);
 text = g_strdup_printf ("%ld", params . generations);
 gtk_entry_set_text (GTK_ENTRY (toRet -> generationEntry), text);
 g_free (text);
 text = g_strdup_printf ("%ld", params . abandon);
 gtk_entry_set_text (GTK_ENTRY (toRet -> abandonEntry), text);
 g_free (text);
 if (params . algorithm & PAGE_SEARCH)
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toRet -> pageCheck), TRUE);
 if (params . algorithm & FLIP_NODES)
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toRet -> flipCheck), TRUE);
 if (params . search & CONSTR_SEARCH)
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toRet -> consCheck), TRUE);
 if (params . algorithm & FIRST_SECTOR)
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toRet -> startSecRadio), TRUE);
 else if (params . algorithm & FULL_SECTOR)
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toRet -> secRadio), TRUE);
 else
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toRet -> noSecRadio), TRUE);
 if (params . algorithm & BLIND_ANNEAL)
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toRet -> blindAnnealRadio), TRUE);
 else if (params . algorithm & GIBBS_ANNEAL)
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toRet -> gibbsAnnealRadio), TRUE);
 else if (params . algorithm & NIXON_ANNEAL)
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toRet -> nixonAnnealRadio), TRUE);
 else
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toRet -> noAnnealRadio), TRUE);
 return toRet;
}

void DestroySearchDialog (SEARCHDIALOG* searchDialog) {
 gtk_widget_destroy (searchDialog -> dialog);
 g_slice_free1 (sizeof (SEARCHDIALOG), searchDialog);
}

static gint CanWindowClose (GtkWidget* widget, gpointer data) {
 SFLAG* progressUp;

 progressUp = (SFLAG*) data;
 return *progressUp;
}

static void OnStopSearch (GtkButton* stopButton, gpointer data) {
 PROGRESSSTRUCT* pBar;

 pBar = (PROGRESSSTRUCT*) data;
 pBar -> isActive = FALSE;
}

SEARCHPROGDLG* SearchProgress (PROGRESSSTRUCT* progress, SFLAG mode) {
 SEARCHPROGDLG* toRet;
 GtkWidget* vBox;
 GtkWidget* label;

 toRet = g_slice_alloc (sizeof (SEARCHPROGDLG));
 toRet -> dialog = gtk_window_new (GTK_WINDOW_TOPLEVEL);
 toRet -> progressUp = TRUE;
 gtk_window_set_title (GTK_WINDOW (toRet -> dialog), "Searching...");
 g_signal_connect (G_OBJECT (toRet -> dialog), "delete_event", G_CALLBACK (CanWindowClose), &toRet -> progressUp);
 gtk_container_border_width (GTK_CONTAINER (toRet -> dialog), 0);
 gtk_window_set_modal (GTK_WINDOW (toRet -> dialog), TRUE);

 vBox = gtk_vbox_new (FALSE, 0);
 gtk_container_add (GTK_CONTAINER (toRet -> dialog), vBox);
 gtk_widget_show (vBox);
 label = gtk_label_new ("Trees");
 gtk_box_pack_start (GTK_BOX (vBox), label, TRUE, TRUE, 2);
 gtk_widget_show (label);
 toRet -> treeProgress = gtk_progress_bar_new ();
 gtk_box_pack_start (GTK_BOX (vBox), toRet -> treeProgress, TRUE, TRUE, 2);
 gtk_widget_show (toRet -> treeProgress);
 label = gtk_label_new ("Iterations");
 gtk_box_pack_start (GTK_BOX (vBox), label, TRUE, TRUE, 2);
 gtk_widget_show (label);
 toRet -> iterProgress = gtk_progress_bar_new ();
 gtk_box_pack_start (GTK_BOX (vBox), toRet -> iterProgress, TRUE, TRUE, 2);
 gtk_widget_show (toRet -> iterProgress);
 toRet -> stopButton = gtk_button_new_from_stock (GTK_STOCK_STOP);
 g_signal_connect (G_OBJECT (toRet -> stopButton), "clicked", G_CALLBACK (OnStopSearch), progress);
 gtk_box_pack_start (GTK_BOX (vBox), toRet -> stopButton, FALSE, TRUE, 2);
 gtk_widget_show (toRet -> stopButton);
 gtk_widget_set_sensitive (toRet -> stopButton, mode);

 toRet -> progressData = progress;
 gtk_widget_show (toRet -> dialog);
 return toRet;
}

gboolean UpdateSearchProgress (gpointer data) {
 SEARCHPROGDLG* prog;
 gchar* text;

 prog = (SEARCHPROGDLG*) data;
 if ((prog -> progressData -> isActive == FALSE) || (prog -> progressData -> actBar1 >= prog -> progressData -> numBar1)) {
   prog -> progressUp = FALSE;
   gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (prog -> treeProgress), 1.0);
   gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (prog -> iterProgress), 1.0);
   gtk_widget_destroy (prog -> dialog);
   g_slice_free1 (sizeof (SEARCHPROGDLG), prog);
   return FALSE;
 }
 text = g_strdup_printf ("Tree %ld of %ld", prog -> progressData -> actBar1, prog -> progressData -> numBar1);
 gtk_progress_bar_set_text (GTK_PROGRESS_BAR (prog -> treeProgress), text);
 g_free (text);
 gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (prog -> treeProgress),
                                            (gdouble) prog -> progressData -> actBar1 / prog -> progressData -> numBar1);
 text = g_strdup_printf ("Iteration %ld of %ld", prog -> progressData -> actBar2, prog -> progressData -> numBar2);
 gtk_progress_bar_set_text (GTK_PROGRESS_BAR (prog -> iterProgress), text);
 g_free (text);
 gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (prog -> iterProgress),
                                            (gdouble) prog -> progressData -> actBar2 / prog -> progressData -> numBar2);
 return TRUE;
}

void DestroyParsingProgressDialog (PARSINGPROGDLG* toDel) {
 gtk_widget_destroy (toDel -> dialog);
 g_slice_free1 (sizeof (PARSINGPROGDLG), toDel);
}

PARSINGPROGDLG* ParsingProgress (PROGRESSSTRUCT* progress) {
 PARSINGPROGDLG* toRet;
 GtkWidget* vBox;
 GtkWidget* label;

 toRet = g_slice_alloc (sizeof (PARSINGPROGDLG));
 toRet -> dialog = gtk_window_new (GTK_WINDOW_TOPLEVEL);
 toRet -> progressUp = TRUE;
 gtk_window_set_title (GTK_WINDOW (toRet -> dialog), "Parsing file...");
 g_signal_connect (G_OBJECT (toRet -> dialog), "delete_event", G_CALLBACK (CanWindowClose), &toRet -> progressUp);
 gtk_container_border_width (GTK_CONTAINER (toRet -> dialog), 0);
 gtk_window_set_modal (GTK_WINDOW (toRet -> dialog), TRUE);

 vBox = gtk_vbox_new (FALSE, 0);
 gtk_container_add (GTK_CONTAINER (toRet -> dialog), vBox);
 gtk_widget_show (vBox);
 label = gtk_label_new ("Parsing...");
 gtk_box_pack_start (GTK_BOX (vBox), label, TRUE, TRUE, 2);
 gtk_widget_show (label);
 toRet -> fileProgress = gtk_progress_bar_new ();
 gtk_box_pack_start (GTK_BOX (vBox), toRet -> fileProgress, TRUE, TRUE, 2);
 gtk_widget_show (toRet -> fileProgress);

 toRet -> progressData = progress;
 gtk_widget_show (toRet -> dialog);
 return toRet;
}

gboolean UpdateParsingProgress (gpointer data) {
 PARSINGPROGDLG* prog;

 prog = (PARSINGPROGDLG*) data;
 if ((prog -> progressData -> isActive == FALSE) || (prog -> progressData -> actBar1 >= prog -> progressData -> numBar1)) {
   prog -> progressUp = FALSE;
   gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (prog -> fileProgress), 1.0);
   gtk_widget_destroy (prog -> dialog);
   g_slice_free1 (sizeof (PARSINGPROGDLG), prog);
   return FALSE;
 }
 gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (prog -> fileProgress),
                                            (gdouble) prog -> progressData -> actBar1 / prog -> progressData -> numBar1);
 return TRUE;
}

static void PutGeo (GEOREFBROWSERDLG* dlg) {
 TERMINAL* term;
 gchar* text;
 GEOREF* geo;

 geo = dlg -> actGeo;
 text = g_strdup_printf ("%.4f", geo -> lat);
 CheckRealStr (text);
 gtk_entry_set_text (GTK_ENTRY (dlg -> latEntry), text);
 g_free (text);
 text = g_strdup_printf ("%.4f", geo -> lon);
 CheckRealStr (text);
 gtk_entry_set_text (GTK_ENTRY (dlg -> lonEntry), text);
 g_free (text);
 term = (TERMINAL*) geo -> sp -> data;
 gtk_entry_set_text (GTK_ENTRY (dlg -> spEntry), term -> name);
 if (geo -> name != NULL)
   gtk_entry_set_text (GTK_ENTRY (dlg -> nameEntry), geo -> name);
 else
   gtk_entry_set_text (GTK_ENTRY (dlg -> nameEntry), "");
 if (geo -> collec != NULL)
   gtk_entry_set_text (GTK_ENTRY (dlg -> collecEntry), geo -> collec);
 else
   gtk_entry_set_text (GTK_ENTRY (dlg -> collecEntry), "");
 if (geo -> catalog != NULL)
   gtk_entry_set_text (GTK_ENTRY (dlg -> catalogEntry), geo -> catalog);
 else
   gtk_entry_set_text (GTK_ENTRY (dlg -> catalogEntry), "");
 if (geo -> url != NULL)
   gtk_entry_set_text (GTK_ENTRY (dlg -> urlEntry), geo -> url);
 else
   gtk_entry_set_text (GTK_ENTRY (dlg -> urlEntry), "");
 if (geo -> reference != NULL)
   gtk_entry_set_text (GTK_ENTRY (dlg -> referenceEntry), geo -> reference);
 else
   gtk_entry_set_text (GTK_ENTRY (dlg -> referenceEntry), "");
 if (geo -> identifier != NULL)
   gtk_entry_set_text (GTK_ENTRY (dlg -> identifierEntry), geo -> identifier);
 else
   gtk_entry_set_text (GTK_ENTRY (dlg -> identifierEntry), "");
 if (geo -> sampleDate != NULL)
   gtk_entry_set_text (GTK_ENTRY (dlg -> sampleEntry), geo -> sampleDate);
 else
   gtk_entry_set_text (GTK_ENTRY (dlg -> sampleEntry), "");
 if (geo -> collector != NULL)
   gtk_entry_set_text (GTK_ENTRY (dlg -> collectorEntry), geo -> collector);
 else
   gtk_entry_set_text (GTK_ENTRY (dlg -> collectorEntry), "");
 if (g_list_next (geo -> list) != NULL)
   gtk_widget_set_sensitive (dlg -> nextButton, TRUE);
 else
   gtk_widget_set_sensitive (dlg -> nextButton, FALSE);
 if (g_list_previous (geo -> list) != NULL)
   gtk_widget_set_sensitive (dlg -> prevButton, TRUE);
 else
   gtk_widget_set_sensitive (dlg -> prevButton, FALSE);
}

static void OnButtonPrevGeo (GtkButton* prevButton, gpointer data) {
 GEOREFBROWSERDLG* dlg;
 GList* geoList;

 dlg = (GEOREFBROWSERDLG*) data;
 geoList = g_list_previous (dlg -> actGeo -> list);
 dlg -> actGeo = (GEOREF*) geoList -> data;
 PutGeo (dlg);
}

static void OnButtonNextGeo (GtkButton* nextButton, gpointer data) {
 GEOREFBROWSERDLG* dlg;
 GList* geoList;

 dlg = (GEOREFBROWSERDLG*) data;
 geoList = g_list_next (dlg -> actGeo -> list);
 dlg -> actGeo = (GEOREF*) geoList -> data;
 PutGeo (dlg);
}

static void OnButtonApplyGeo (GtkButton* applyButton, gpointer data) {
 GEOREFBROWSERDLG* dlg;
 SFLAG change;
 SREAL nuLat, nuLon;
 gchar* entry;
 GEOREF* geo;
 TERMINAL* actTerm;

 dlg = (GEOREFBROWSERDLG*) data;
 change = FALSE;
 entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (dlg -> latEntry)));
 SetCommandLine (entry);
 g_free (entry);
 geo = dlg -> actGeo;
 if (GetCmdReal (&nuLat) != FALSE) {
   if ((nuLat > -90.0000) && (nuLat < 90.0000))
     change = TRUE;
   else
     nuLat = geo -> lat;
 }
 entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (dlg -> lonEntry)));
 SetCommandLine (entry);
 g_free (entry);
 if (GetCmdReal (&nuLon) != FALSE) {
   if ((nuLon > -180.0000) && (nuLon < 180.0000))
     change = TRUE;
  else
     nuLon = geo -> lon;
 }
 entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (dlg -> nameEntry)));
 if ((strlen (entry)) && (g_strcmp0 (entry, geo -> name))) {
   actTerm = (TERMINAL*) geo -> sp -> data;
   geo -> name = AddSynonim (actTerm, entry);
 }
 g_free (entry);
 entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (dlg -> collecEntry)));
 if ((strlen (entry)) && (g_strcmp0 (entry, geo -> collec))) {
   g_free (geo -> collec);
   geo -> collec = entry;
 }
 else
   g_free (entry);
 entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (dlg -> catalogEntry)));
 if ((strlen (entry)) && (g_strcmp0 (entry, geo -> catalog))) {
   g_free (geo -> catalog);
   geo -> catalog = entry;
 }
 else
   g_free (entry);
 entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (dlg -> urlEntry)));
 if ((strlen (entry)) && (g_strcmp0 (entry, geo -> url))) {
   g_free (geo -> url);
   geo -> url = entry;
 }
 else
   g_free (entry);
 entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (dlg -> referenceEntry)));
 if ((strlen (entry)) && (g_strcmp0 (entry, geo -> reference))) {
   g_free (geo -> reference);
   geo -> reference = entry;
 }
 else
   g_free (entry);
 entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (dlg -> identifierEntry)));
 if ((strlen (entry)) && (g_strcmp0 (entry, geo -> identifier))) {
   g_free (geo -> identifier);
   geo -> identifier = entry;
 }
 else
   g_free (entry);
 entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (dlg -> sampleEntry)));
 if ((strlen (entry)) && (g_strcmp0 (entry, geo -> sampleDate))) {
   g_free (geo -> sampleDate);
   geo -> sampleDate = entry;
 }
 else
   g_free (entry);
 entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (dlg -> collectorEntry)));
 if ((strlen (entry)) && (g_strcmp0 (entry, geo -> collector))) {
   g_free (geo -> collector);
   geo -> collector = entry;
 }
 else
   g_free (entry);
 if (change == TRUE) {
   geo -> lat = nuLat;
   geo -> lon = nuLon;
   geo -> inclination = FROMLATTOINC (nuLat);
   geo -> azimuth = FROMLONTOAZM (nuLon);
   UpdateLimits (geo);
   PaintNewGeoRef (geo);
 }
 gtk_widget_destroy (dlg -> dialog);
 g_slice_free1 (sizeof (GEOREFBROWSERDLG), dlg);
}

static void OnButtonDelGeo (GtkButton* delButton, gpointer data) {
 GEOREFBROWSERDLG* dlg;
 GEOREF* geo;

 dlg = (GEOREFBROWSERDLG*) data;
 geo = dlg -> actGeo;
 DeleteGeoRef (geo);
 ChangeGeo ();
 gtk_widget_destroy (dlg -> dialog);
 g_slice_free1 (sizeof (GEOREFBROWSERDLG), dlg);
}

static void OnButtonClose (GtkButton* closeButton, gpointer data) {
 GEOREFBROWSERDLG* dlg;

 dlg = (GEOREFBROWSERDLG*) data;
 gtk_widget_destroy (dlg -> dialog);
 g_slice_free1 (sizeof (GEOREFBROWSERDLG), dlg);
}

static gint OnCloseBrowser (GtkWidget* widget, gpointer data) {
 return TRUE;
}

GEOREFBROWSERDLG* GeoRefBrowser (GEOREF* geo, SFLAG mode) {
 GEOREFBROWSERDLG* toRet;
 GtkWidget* vBox;
 GtkWidget* hBox;
 GtkWidget* separator;
 GtkWidget* button;

 toRet = g_slice_alloc (sizeof (GEOREFBROWSERDLG));
 toRet -> dialog = gtk_window_new (GTK_WINDOW_TOPLEVEL);
 gtk_window_set_title (GTK_WINDOW (toRet -> dialog), "Record browser");
 g_signal_connect (G_OBJECT (toRet -> dialog), "delete_event", G_CALLBACK (OnCloseBrowser), NULL);
 gtk_container_border_width (GTK_CONTAINER (toRet -> dialog), 0);
 gtk_window_set_modal (GTK_WINDOW (toRet -> dialog), TRUE);
 vBox = gtk_vbox_new (FALSE, 0);
 gtk_container_add (GTK_CONTAINER (toRet -> dialog), vBox);
 gtk_widget_show (vBox);

 hBox = gtk_hbox_new (TRUE, 2);
 gtk_box_pack_start (GTK_BOX (vBox), hBox, FALSE, TRUE, 2);
 gtk_widget_show (hBox);
 toRet -> prevButton = gtk_button_new_from_stock (GTK_STOCK_GO_BACK);
 g_signal_connect (G_OBJECT (toRet -> prevButton), "clicked", G_CALLBACK (OnButtonPrevGeo), toRet);
 gtk_box_pack_start (GTK_BOX (hBox), toRet -> prevButton, FALSE, TRUE, 2);
 gtk_widget_show (toRet -> prevButton);
 toRet -> nextButton = gtk_button_new_from_stock (GTK_STOCK_GO_FORWARD);
 g_signal_connect (G_OBJECT (toRet -> nextButton), "clicked", G_CALLBACK (OnButtonNextGeo), toRet);
 gtk_box_pack_start (GTK_BOX (hBox), toRet -> nextButton, FALSE, TRUE, 2);
 gtk_widget_show (toRet -> nextButton);

 toRet -> spEntry = CreateEntry ("Terminal", GTK_BOX (vBox));
 separator = gtk_hseparator_new ();
 toRet -> nameEntry = CreateEntry ("Record name", GTK_BOX (vBox));
 toRet -> latEntry = CreateEntry ("Latitude", GTK_BOX (vBox));
 toRet -> lonEntry = CreateEntry ("Longitude", GTK_BOX (vBox));
 separator = gtk_hseparator_new ();
 toRet -> collecEntry = CreateEntry ("Collection", GTK_BOX (vBox));
 toRet -> catalogEntry = CreateEntry ("Catalogue code", GTK_BOX (vBox));
 toRet -> urlEntry = CreateEntry ("Url", GTK_BOX (vBox));
 toRet -> referenceEntry = CreateEntry ("Reference", GTK_BOX (vBox));
 toRet -> identifierEntry = CreateEntry ("Identified by", GTK_BOX (vBox));
 toRet -> sampleEntry = CreateEntry ("Sample date", GTK_BOX (vBox));
 toRet -> collectorEntry = CreateEntry ("Collector", GTK_BOX (vBox));
 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (vBox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);

 hBox = gtk_hbox_new (TRUE, 2);
 gtk_box_pack_start (GTK_BOX (vBox), hBox, FALSE, TRUE, 2);
 gtk_widget_show (hBox);
 toRet -> applyButton = gtk_button_new_from_stock (GTK_STOCK_APPLY);
 g_signal_connect (G_OBJECT (toRet -> applyButton), "clicked", G_CALLBACK (OnButtonApplyGeo), toRet);
 gtk_box_pack_start (GTK_BOX (hBox), toRet -> applyButton, FALSE, TRUE, 2);
 gtk_widget_show (toRet -> applyButton);
 toRet -> delButton = gtk_button_new_from_stock (GTK_STOCK_DELETE);
 g_signal_connect (G_OBJECT (toRet -> delButton), "clicked", G_CALLBACK (OnButtonDelGeo), toRet);
 gtk_box_pack_start (GTK_BOX (hBox), toRet -> delButton, FALSE, TRUE, 2);
 gtk_widget_show (toRet -> delButton);
 gtk_widget_set_sensitive (toRet -> applyButton, mode);
 gtk_widget_set_sensitive (toRet -> delButton, mode);
 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (vBox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);

 button = gtk_button_new_from_stock (GTK_STOCK_CLOSE);
 g_signal_connect (G_OBJECT (button), "clicked", G_CALLBACK (OnButtonClose), toRet);
 gtk_box_pack_start (GTK_BOX (vBox), button, FALSE, TRUE, 2);
 gtk_widget_show (button);
 toRet -> actGeo = geo;

 gtk_editable_set_editable (GTK_EDITABLE (toRet -> spEntry), FALSE);
 gtk_editable_set_editable (GTK_EDITABLE (toRet -> nameEntry), mode);
 gtk_editable_set_editable (GTK_EDITABLE (toRet -> latEntry), mode);
 gtk_editable_set_editable (GTK_EDITABLE (toRet -> lonEntry), mode);
 gtk_editable_set_editable (GTK_EDITABLE (toRet -> collecEntry), mode);
 gtk_editable_set_editable (GTK_EDITABLE (toRet -> catalogEntry), mode);
 gtk_editable_set_editable (GTK_EDITABLE (toRet -> referenceEntry), mode);
 gtk_editable_set_editable (GTK_EDITABLE (toRet -> identifierEntry), mode);
 gtk_editable_set_editable (GTK_EDITABLE (toRet -> sampleEntry), mode);
 gtk_editable_set_editable (GTK_EDITABLE (toRet -> collectorEntry), mode);

 PutGeo (toRet);
 gtk_widget_show (toRet -> dialog);

 return toRet;
}

NODEDATADIALOG* NodeDataDialog (GNode* node, SFLAG mode) {
 NODEDATADIALOG* toRet;
 GtkWidget* separator;
 PHYLONODE* actNode;
 gchar* text;

 toRet = g_slice_alloc (sizeof (NODEDATADIALOG));
 toRet -> dialog = gtk_dialog_new_with_buttons ("Node", GTK_WINDOW (GETTREEWIN),
                       GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR,
                       GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, NULL);

 toRet -> nameEntry = CreateEntry ("Node name", GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));

 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);

 toRet -> ageEntry = CreateEntry ("Node age", GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));

 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);

 actNode = (PHYLONODE*) node -> data;
 if (actNode -> name != NULL)
   text = g_strdup_printf (actNode -> name);
 else
   text = g_strdup_printf ("unnamed");
 gtk_entry_set_text (GTK_ENTRY (toRet -> nameEntry), text);
 g_free (text);
  text = g_strdup_printf ("%ld", actNode -> minAge);
 gtk_entry_set_text (GTK_ENTRY (toRet -> ageEntry), text);
 g_free (text);
 gtk_editable_set_editable (GTK_EDITABLE (toRet -> nameEntry), mode);
 gtk_editable_set_editable (GTK_EDITABLE (toRet -> ageEntry), mode);
 return toRet;
}

void DestroyNodeDataDialog (NODEDATADIALOG* toDel) {
 gtk_widget_destroy (toDel -> dialog);
 g_slice_free1 (sizeof (NODEDATADIALOG), toDel);
}
/* grid.c is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */

#include "grid.h"

static GNode* qTree = NULL;
static GPtrArray* cellArray = NULL;

static GNode* CreateQTree (GNode* anc, GNode* sibling, SINTEGER iniRow, SINTEGER endRow, SINTEGER iniCol, SINTEGER endCol,
                           INTARRAY* matrix);
static void MakeQTree ();

static void MakeCellObs (SCOUNTER destCell);

static void MapQNodeDown (GNode* qNode, BITFIELD* setA, BITFIELD* setB);
static void MapQNodeUp (GNode* qNode, BITFIELD* setA, BITFIELD* setB);

/* Phylogenetic diversity */
static SCOUNTER PhyloDivNode (GNode* node, PHYLONODE* anc, SCOUNTER cell);

/* Here I use a recursive function to build the qTree, becuase it is more difficult to build a non-recursive option! */
GNode* CreateQTree (GNode* anc, GNode* sibling, SINTEGER iniRow, SINTEGER endRow, SINTEGER iniCol, SINTEGER endCol,
                    INTARRAY* matrix) {
 QUADTREE* qNode;
 GNode* actNode;
 GNode* nuSibling;
 SCOUNTER r, c, actArea;
 SCOUNTER halfRow, halfCol;

 qNode = g_slice_alloc (sizeof (QUADTREE));
 qNode -> firstRow = iniRow;
 qNode -> lastRow = endRow;
 qNode -> firstCol = iniCol;
 qNode -> lastCol = endCol;
 qNode -> size = (endRow - iniRow + 1) * (endCol - iniCol + 1);
 qNode -> state = FALSE;
 qNode -> obsCell = FALSE;
 qNode -> cell = -1;
 actNode = g_node_new (qNode);
 AddChildren (actNode, anc, sibling);

 for (r = iniRow; r <= endRow; ++ r) {
   for (c = iniCol; c <= endCol; ++ c) {
     actArea = r * numCols + c;
     if (GetFromArray (matrix, actArea) > 0) {
       qNode -> state = TRUE;
       break;
     }
   }
   if (qNode -> state == TRUE) break;
 }

 if (qNode -> size == 1) {
   if (qNode -> state == TRUE) {
     qNode -> cell = numCells;
     g_ptr_array_add (cellArray, qNode);
     ++ numCells;
   }
   return actNode;
 }
 if (qNode -> state == FALSE) return actNode;
 nuSibling = NULL;
 if ((endCol > iniCol) && (endRow > iniRow)) {
   halfCol = iniCol + ((endCol - iniCol) / 2);
   halfRow = iniRow + ((endRow - iniRow) / 2);
   nuSibling = CreateQTree (actNode, nuSibling, iniRow, halfRow, iniCol, halfCol, matrix);
   nuSibling = CreateQTree (actNode, nuSibling, iniRow, halfRow, halfCol + 1, endCol, matrix);
   nuSibling = CreateQTree (actNode, nuSibling, halfRow + 1, endRow, iniCol, halfCol, matrix);
   CreateQTree (actNode, nuSibling, halfRow + 1, endRow, halfCol + 1, endCol, matrix);
 }
 else if (endCol > iniCol) {
   halfCol = iniCol + ((endCol - iniCol) / 2);
   nuSibling = CreateQTree (actNode, nuSibling, iniRow, endRow, iniCol, halfCol, matrix);
   CreateQTree (actNode, nuSibling, iniRow, endRow, halfCol + 1, endCol, matrix);
 }
 else if (endRow > iniRow) {
   halfRow = iniRow + ((endRow - iniRow) / 2);
   nuSibling = CreateQTree (actNode, nuSibling, iniRow, halfRow, iniCol, endCol, matrix);
   CreateQTree (actNode, nuSibling, halfRow + 1, endRow, iniCol, endCol, matrix);
 }
 return actNode;
}

SCOUNTER GetCell (SINTEGER row, SINTEGER col) {
 GNode* actNode;
 QUADTREE* qNode;

 qNode = NULL;
 for (actNode = qTree; actNode != NULL;) {
   qNode = (QUADTREE*) actNode -> data;
   if ((col >= qNode -> firstCol) && (col <= qNode -> lastCol) && (row >= qNode -> firstRow) && (row <= qNode -> lastRow)) {
     if (qNode -> size == 1) break;
     actNode = g_node_first_child (actNode);
     qNode = NULL;
     continue;
   }
   if (g_node_next_sibling (actNode) == NULL) {
     qNode = NULL;
     break;
   }
   actNode = g_node_next_sibling (actNode);
 }
 if (qNode == NULL) return -1;
 return qNode -> cell;
}

void MakeQTree () {
 SCOUNTER maxAreas, actArea;
 SCOUNTER row, col, r, c, num;
 GEOREF* actRec;
 GList* navGeo;
 INTARRAY* matrix;
 SREAL x, y;

 maxAreas = numCols *  numRows;
 matrix = CreateIntArray (maxAreas);
 num = 0;
 for (actArea = 0; actArea < maxAreas; ++ actArea)
   AddToArray (matrix, num);
 for (navGeo = GETGEOREF; navGeo != NULL; navGeo = g_list_next (navGeo)) {
   actRec = (GEOREF*) navGeo -> data;
   x = actRec -> azimuth - gridStartX;
   y = actRec -> inclination - gridStartY;
   col = (SINTEGER) (x / actualGridSize);
   row = (SINTEGER) (y / actualGridSize);
   for (r = row - actualFill; r <= (row + actualFill); ++ r) {
     if (r < 0) continue;
     if (r >= numRows) break;
     for (c = col - actualFill; c <= (col + actualFill); ++ c) {
       if (c < 0) continue;
       if (c >= numCols) break;
       actArea = (r * numCols) + c;
       num = GetFromArray (matrix, actArea);
       ++ num;
       InsertValToArray (matrix, num, actArea);
     }
   }
 }
 numCells = 0;
 cellArray = g_ptr_array_new ();
 qTree = CreateQTree (NULL, NULL, 0, numRows - 1, 0, numCols - 1, matrix);
 DestroyIntArray (matrix);
}

void CreateGrid () {
 GList* navTerm;
 GList* navGeo;
 TERMINAL* actTerm;
 PHYLONODE* node;
 GEOREF* actGeo;
 SCOUNTER actCell;
 SCOUNTER row, col;
 SREAL x, y;
 SCOUNTER termFill;

 numCols = ((GETMAXAZM - GETMINAZM) / actualGridSize) + 2;
 numRows = ((GETMAXINC - GETMININC) / actualGridSize) + 2;
 gridStartX = ((SINTEGER) (GETMINAZM / actualGridSize)) * actualGridSize;
 gridStartY = ((SINTEGER) (GETMININC / actualGridSize)) * actualGridSize;

 if (qTree == NULL) {
   MakeQTree ();
   StartBitFields (numCells);
   for (navTerm = GETSPP; navTerm != NULL; navTerm = g_list_next (navTerm)) {
     actTerm = (TERMINAL*) navTerm -> data;
     actTerm -> obs = CreateBitField ();
     node = (PHYLONODE*) actTerm -> node -> data;
     termFill = (node -> fill > actualFill) ? actualFill : node -> fill;
     for (navGeo = actTerm -> firstGeo; navGeo != NULL; navGeo = g_list_next (navGeo)) {
       actGeo = (GEOREF*) navGeo -> data;
       if (actGeo -> sp != navTerm) break;
       x = actGeo -> azimuth - gridStartX;
       y = actGeo -> inclination - gridStartY;
       col = (SINTEGER) (x / actualGridSize);
       row = (SINTEGER) (y / actualGridSize);
       actCell = GetCell (row, col);
       actGeo -> actCell = actCell;
       if (IsBitOn (actTerm -> obs, actCell) == TRUE) continue;
       MakeCellObs (actCell);
       PutBitOn (actTerm -> obs, actCell);
     }
   }
 }
 MakeOptimization (FALSE);
}

void DestroyGrid () {
 GList* navTerm;
 GList* navGeo;
 GNode* actNode;
 GNode* nuAct;
 TERMINAL* actTerm;
 GEOREF* actGeo;

 if (qTree == NULL) return;
 DestroyRecs ();
 RemoveAlloSets ();
 for (navTerm = GETSPP; navTerm != NULL; navTerm = g_list_next (navTerm)) {
   actTerm = (TERMINAL*) navTerm -> data;
   if (actTerm -> obs != NULL)
     g_slice_free1 (GETBITFIELDSIZE, actTerm -> obs);
   actTerm -> obs = NULL;
 }
 for (navGeo = GETGEOREF; navGeo != NULL; navGeo = g_list_next (navGeo)) {
   actGeo = (GEOREF*) navGeo -> data;
   actGeo -> actCell = -1;
 }
 actNode = qTree;
 while (actNode != NULL) {
   if (g_node_first_child (actNode) != NULL) {
     actNode = g_node_last_child (actNode);
     continue;
   }
   if (g_node_prev_sibling (actNode) != NULL)
     nuAct = g_node_prev_sibling (actNode);
   else
     nuAct = actNode -> parent;
   g_node_unlink (actNode);
   g_slice_free1 (sizeof (QUADTREE), actNode -> data);
   g_node_destroy (actNode);
   actNode = nuAct;
 }
 qTree = NULL;
 g_ptr_array_free (cellArray, TRUE);
 cellArray = NULL;
}

SFLAG GetGridState () {
 return (cellArray != NULL) ? TRUE : FALSE;
}

void GetColRowFromCell (SCOUNTER cell, SINTEGER* row, SINTEGER* col) {
 QUADTREE* qNode;

 qNode = (QUADTREE*) g_ptr_array_index (cellArray, cell);
 *row = qNode -> firstRow;
 *col = qNode -> firstCol;
}

void MapIntoQTree (BITFIELD* setA, BITFIELD* setB) {
 MapQNodeDown (qTree, setA, setB);
 MapQNodeUp (qTree, setA, setB);
}

void MapQNodeDown (GNode* qNode, BITFIELD* setA, BITFIELD* setB) {
 QUADTREE* quad;
 QUADTREE* child;
 GNode* navChild;

 quad = (QUADTREE*) qNode -> data;
 quad -> state = 0;
 if (quad -> size == 1) {
   if (quad -> cell != -1) {
     if (IsBitOn (setA, quad -> cell))
       quad -> state |= 1;
     if (IsBitOn (setB, quad -> cell))
       quad -> state |= 2;
   }
   return;
 }
 for (navChild = g_node_first_child (qNode); navChild != NULL; navChild = g_node_next_sibling (navChild)) {
   MapQNodeDown (navChild, setA, setB);
   child = (QUADTREE*) navChild -> data;
   quad -> state |= child -> state;
 }
 return;
}

void MapQNodeUp (GNode* qNode, BITFIELD* setA, BITFIELD* setB) {
 QUADTREE* quad;
 QUADTREE* anc;
 GNode* navChild;

 quad = (QUADTREE*) qNode -> data;

 if ((quad -> state == 0) && (qNode -> parent != NULL)) {
   anc = (QUADTREE*) qNode -> parent -> data;
   quad -> state = anc -> state;
 }
 for (navChild = g_node_first_child (qNode); navChild != NULL; navChild = g_node_next_sibling (navChild))
   MapQNodeUp (navChild, setA, setB);
}

SFLAG GetRowColState (SINTEGER row, SINTEGER col) {
 GNode* actNode;
 QUADTREE* qNode;

 qNode = NULL;
 for (actNode = qTree; actNode != NULL;) {
   qNode = (QUADTREE*) actNode -> data;
   if ((col >= qNode -> firstCol) && (col <= qNode -> lastCol) && (row >= qNode -> firstRow) && (row <= qNode -> lastRow)) {
     if (qNode -> state != 3) break;
     if (g_node_first_child (actNode) == NULL) break;
     actNode = g_node_first_child (actNode);
     qNode = NULL;
     continue;
   }
   if (g_node_next_sibling (actNode) == NULL) {
     qNode = NULL;
     break;
   }
   actNode = g_node_next_sibling (actNode);
 }
 if (qNode == NULL) return 0;
 return qNode -> state;
}

void MakeCellObs (SCOUNTER destCell) {
 QUADTREE* qNode;

 qNode = (QUADTREE*) g_ptr_array_index (cellArray, destCell);
 qNode -> obsCell = TRUE;
}

SCOUNTER GetCellIfObs (SINTEGER row, SINTEGER col) {
 GNode* actNode;
 QUADTREE* qNode;

 qNode = NULL;
 for (actNode = qTree; actNode != NULL;) {
   qNode = (QUADTREE*) actNode -> data;
   if ((col >= qNode -> firstCol) && (col <= qNode -> lastCol) && (row >= qNode -> firstRow) && (row <= qNode -> lastRow)) {
     if (qNode -> size == 1) break;
     actNode = g_node_first_child (actNode);
     qNode = NULL;
     continue;
   }
   if (g_node_next_sibling (actNode) == NULL) {
     qNode = NULL;
     break;
   }
   actNode = g_node_next_sibling (actNode);
 }
 if (qNode == NULL) return -1;
 return (qNode -> obsCell == TRUE) ? qNode -> cell : -1;
}

void GetObsSet (BITFIELD* obs, GNode* refNode) {
 GNode* actNode;
 PHYLONODE* node;
 PHYLONODE* anc;
 TERMINAL* term;

 node = (PHYLONODE*) refNode -> data;
 ResetBitField (obs);
 for (actNode = refNode; actNode != NULL;) {
   if (IsDesc (actNode, refNode) != TRUE) break;
   node = (PHYLONODE*) actNode -> data;
   if (actNode -> parent != NULL)
     anc = (PHYLONODE*) actNode -> parent -> data;
   if ((actNode != refNode) && ((node -> rec -> numCells == 0) ||
                 (node -> rec -> nodeFlag & REMOV_NODE) || (node == anc -> rec -> removSet))) {
     /* A removed distribution, ignores all of its descendants */
     while (actNode != NULL) {
       if (g_node_next_sibling (actNode) != NULL) {
         actNode = g_node_next_sibling (actNode);
         break;
       }
       actNode = actNode -> parent;
     }
     continue;
   }
   if (g_node_first_child (actNode) != NULL) {
   /* Not a term, goes to the first descendant */
     actNode = g_node_first_child (actNode);
     continue;
   }
   term = node -> term;
   BitFieldORsum (obs, term -> obs);
   while (actNode != NULL) {
     if (g_node_next_sibling (actNode) != NULL) {
       actNode = g_node_next_sibling (actNode);
       break;
     }
     actNode = actNode -> parent;
   }
 }
}

void DoFill (BITFIELD* dest, BITFIELD* obs, SCOUNTER fillValue) {
 SCOUNTER navCell;
 SCOUNTER actCell;
 SCOUNTER row, col, fill;
 SINTEGER r, c;
 SINTEGER xBlocks, yBlocks, blocks;

 ResetBitField (dest);
 fill = (fillValue < actualFill) ? fillValue : actualFill;
 for (navCell = 0; navCell <= numCells; ++ navCell) {
   if (IsBitOn (obs, navCell) != TRUE) continue;
   GetColRowFromCell (navCell, &row, &col);
   for (r = row - fill; r <= (row + fill); ++ r) {
     if (r < 0) continue;
     if (r >= numRows) break;
     for (c = col - fill; c <= (col + fill); ++ c) {
       if (c < 0) continue;
       if (c >= numCols) break;
       if (gridMode & VONNEUMANN_NEIGHBORS) {
         xBlocks = (r < row) ? row - r : r - row;
         yBlocks = (c < col) ? col - c : c - col;
         blocks = xBlocks + yBlocks;
         if (blocks > fill) continue;
       }
       actCell = GetCell (r, c);
       PutBitOn (dest, actCell);
     }
   }
 }
}

SCOUNTER PhyloDivNode (GNode* node, PHYLONODE* anc, SCOUNTER cell) {
 PHYLONODE* actNode;
 GNode* navDesc;
 SCOUNTER toRet;

 actNode = (PHYLONODE*) node -> data;
 if (IsBitOn (actNode -> rec -> cells, cell) != TRUE) return 0;
 if (anc != NULL)
   toRet = anc -> level - actNode -> level;
 else
   toRet = 0;
 if (g_node_first_child (node) != NULL) {
   for (navDesc = g_node_first_child (node); navDesc != NULL; navDesc = g_node_next_sibling (navDesc))
     toRet += PhyloDivNode (navDesc, actNode, cell);
 }
 return toRet;
}

SCOUNTER CalculatePhyloDiv (SCOUNTER cell, SCOUNTER max) {
 SCOUNTER toRet, thisTree;
 SCOUNTER extra;
 GList* navTree;
 PHYLOTREE* actTree;
 PHYLONODE* root;

 if (cellArray == NULL) return 0;
 toRet = 0;
 for (navTree = GETTREES; navTree != NULL; navTree = g_list_next (navTree)) {
   actTree = (PHYLOTREE*) navTree -> data;
   thisTree = PhyloDivNode (actTree -> root, NULL, cell);
   if (thisTree > 0) {
     root = (PHYLONODE*) actTree -> root -> data;
     extra = max - root -> level;
     thisTree += extra;
   }
   toRet += thisTree;
 }
 return toRet;
}
/* in.c is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */

#include "in.h"

/* Create a input stream */
INSTREAM* CreateInStream () {
 INSTREAM* toRet;

 toRet = (INSTREAM*) g_malloc (sizeof (INSTREAM));
 toRet -> stream = NULL;
 toRet -> name = NULL;
 toRet -> line = 0;
 toRet -> len = 0;
 toRet -> state = EOF_REACHED;
 toRet -> nextChar = 0;
 toRet -> actPos = 0;
 return toRet;
}

/* Destroy a input stream */
void DestroyInStream (INSTREAM* in) {
 if (in == NULL) return;
 if (in -> stream != NULL)
   g_free (in -> stream);
 g_free (in -> name);
 g_free (in);
}

/* Set stream to a input file */
SFLAG SetInStreamToFile (INSTREAM* in, gchar* name) {
 ResetInStream (in);
 in -> name = g_strdup_printf (name);
 if (g_file_get_contents (in -> name, &in -> stream, &in -> len, NULL) == FALSE) {
   errorMessage = g_strdup_printf ("\nUnable to open file %s\n", in -> name);
   OnError ();
   return FALSE;
 }
 in -> line = 1;
 in -> state = OPEN_STREAM;
 SkipSpaces (in);
 return TRUE;
}

/* Set stream to a GString */
SFLAG SetInStreamToString (INSTREAM* in, GString* src) {
 ResetInStream (in);
 in -> name = g_strdup_printf ("string");
 in -> stream = g_strdup (src -> str);
 in -> len = src -> len + 1;
 in -> line = 1;
 in -> state = OPEN_STREAM;
 in -> nextChar = in -> stream [0];
 return in -> state;
}

/* Reset stream */
void ResetInStream (INSTREAM* in) {
 if (in == NULL) return;

 if (in -> stream != NULL)
   g_free (in -> stream);
 in -> stream = NULL;
 if (in -> name != NULL)
   g_free (in -> name);
 in -> name = NULL;
 in -> line = 0;
 in -> len = 0;
 in -> state = EOF_REACHED;
 in -> nextChar = 0;
 in -> actPos = 0;
}

/* Read a string, returns the actual file state */
SFLAG ReadString (INSTREAM* in, GString* dest) {
 gchar c;

 if (dest != NULL)
   g_string_erase (dest, 0, -1);
 if (in -> state == EOF_REACHED) return in -> state;
 for (; in -> actPos < in -> len; ++ in -> actPos) {
   c = in -> stream [in -> actPos];
   if (g_ascii_isspace (c)) break;
   if ((c == ',') || (c == ';') || (c == '=') || (c == ':') || (c == '(') || (c == ')')) break;
   if (dest != NULL) {
     if ((g_ascii_isupper (c)) && (readingMode & ALL_LOWER))
       c += 32;
     g_string_append_c (dest, c);
   }
 }
 /* Put a NULL character in the end */
 if (dest != NULL)
   g_string_append_c (dest, 0);
 if (in -> actPos == in -> len) {
   in -> nextChar = 0;
   in -> state = EOF_REACHED;
 }
 else
   SkipSpaces (in);
 return in -> state;
}

/* Look for the next string on the file, but keep the actual position */
SFLAG PeekString (INSTREAM* in, GString* dest) {
 SCOUNTER line, pos;
 SFLAG state;
 gchar next;

 if (dest == NULL) return in -> state;
 if (in -> state == EOF_REACHED) return in -> state;
 line = in -> line;
 pos = in -> actPos;
 state = in -> state;
 next = in -> nextChar;
 ReadString (in, dest);
 in -> line = line;
 in -> actPos = pos;
 in -> state = state;
 in -> nextChar = next;
 return state;
}

SFLAG ReadStringUntil (INSTREAM* in, GString* dest, gchar end) {
 gchar c;
 SCOUNTER i;
 SFLAG inside;

 if (dest != NULL)
   g_string_erase (dest, 0, -1);
 if (in -> state == EOF_REACHED) return BAD_IN_FORMAT;
 c = 0;
 inside = FALSE;
 for (i = 0; in -> actPos < in -> len; ++ in -> actPos, ++ i) {
   c = in -> stream [in -> actPos];
   if (c == 34)
     inside = !inside;
   if ((c == end) && (inside == FALSE)) break;
   if (c == '\n') break;
   if ((dest != NULL) && (c != '\r')) {
     if ((g_ascii_isupper (c)) && (readingMode & ALL_LOWER))
       c += 32;
     else if (g_ascii_isspace (c))
       c = 32;
     else if ((c == '<') || (c == '>'))
       c = '_';
     if (c == '&')
       g_string_append (dest, "and");
     else
       g_string_append_c (dest, c);
   }
 }
 if ((dest != NULL) && (dest -> len > 0))
   g_string_append_c (dest, 0);
 if (in -> actPos == in -> len) {
   in -> nextChar = 0;
   in -> state = EOF_REACHED;
   return in -> state;
 }
 else {
   ++ in -> actPos;
   in -> nextChar = in -> stream [in -> actPos];
 }
 if (c != end) {
   ++ in -> line;
   return END_OF_IN_LINE;
 }
 return END_CHAR_FOUND;
}

/* Read a bounded comment, returns the actual file state,
   if no string is provided it ignores the inside part of the comment */
SFLAG ReadBoundComment (INSTREAM* in, GString* dest, gchar start, gchar end) {
 gchar c;

 if (dest != NULL)
   g_string_erase (dest, 0, -1);
 if (in -> state == EOF_REACHED) return BAD_IN_FORMAT;
 if (in -> nextChar != start) return BAD_IN_FORMAT;
 ++ in -> actPos;
 SkipSpaces (in);
 for (; in -> actPos < in -> len; ++ in -> actPos) {
   c = in -> stream [in -> actPos];
   if (c == end) break;
   if (dest != NULL)
     g_string_append_c (dest, c);
 }
 /* Put a NULL character in the end */
 if ((dest != NULL) && (dest -> len > 0))
   g_string_append_c (dest, 0);
 if (in -> actPos == in -> len) {
   in -> nextChar = 0;
   in -> state = EOF_REACHED;
   return in -> state;
 }
 else {
   ++ in -> actPos;
   SkipSpaces (in);
 }
 return 1;
}

/* Read an integer value */
SFLAG ReadInt (INSTREAM* in, SINTEGER* number) {
 gchar c;
 SINTEGER acum;
 SINTEGER kind;
 SCOUNTER prevPos;

 if (in -> state == EOF_REACHED) return in -> state;
 acum = 0;
 c = in -> nextChar;
 if (c == '-') {
   ++ in -> actPos;
   kind =-1;
 }
 else
   kind = 1;
 prevPos = in -> actPos;
 for (; in -> actPos < in -> len; ++ in -> actPos) {
   c = in -> stream [in -> actPos];
   if (!g_ascii_isdigit (c)) break;
   acum *= 10;
   acum += c - '0';
 }
 if (prevPos == in -> actPos) return BAD_IN_FORMAT;
 if (in -> actPos == in -> len) {
   in -> nextChar = 0;
   in -> state = EOF_REACHED;
 }
 else
   SkipSpaces (in);
 if (number != NULL)
   *number = (kind * acum);
 return in -> state;
}

/* Read a float pointed value */
SFLAG ReadReal (INSTREAM* in, SREAL* number) {
 SREAL acum;
 gchar c;
 SCOUNTER dig;
 SINTEGER num;
 SFLAG state;
 SCOUNTER prevPos;

 num = 0;
 state = ReadInt (in, &num);        /* Reads the integer part */
 c = in -> nextChar;
 acum = 0.000000;
 dig = 1;
 prevPos = in -> actPos;
 if (c == '.') {            /* Reads the decimal part */
   ++ in -> actPos;
   for (; in -> actPos < in -> len; ++ in -> actPos) {
     c = in -> stream [in -> actPos];
     if (!g_ascii_isdigit (c)) break;
     acum *= 10;
     acum += (SREAL) (c - '0');
     dig *= 10;
   }
   if (in -> actPos == in -> len) {
     in -> nextChar = 0;
     in -> state = EOF_REACHED;
   }
   else
     SkipSpaces (in);
   acum /= dig;
 }
 if ((state == BAD_IN_FORMAT) && (prevPos == in -> actPos)) return BAD_IN_FORMAT;
 if (num >= 0)
   acum += num;
 else
   acum = num - acum;
 if (number != NULL)
   *number = acum;
 return in -> state;
}

/* Returns the next character and moves the stream to the next readable char */
SFLAG ReadChar (INSTREAM* in, gchar* dest) {
 if (dest != NULL)
   *dest = in -> nextChar;
 if (in -> state != EOF_REACHED) {
   ++ in -> actPos;
   SkipSpaces (in);
 }
 return in -> state;
}

/* Ignore space characters of the data set, and count the EOL found */
void SkipSpaces (INSTREAM* in) {
 gchar c;

 c = 0;
 if (in -> state == EOF_REACHED) return;
 for (; in -> actPos < in -> len; ++ in -> actPos) {
   c = in -> stream [in -> actPos];
   if (!g_ascii_isspace (c)) break;
   if (c == '\n')
     ++ in -> line;
 }
 if (in -> actPos == in -> len) {
   in -> nextChar = 0;
   in -> state = EOF_REACHED;
   return;
 }
 in -> nextChar = c;
}
/* inxml.c is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */

#include "inxml.h"

static void SkipCommnet (INSTREAM* in);
static void SkipCData (INSTREAM* in);

static void CleanTokenStack (GList* toClean);

static void SkipUnicodeBOM (INSTREAM* in);

/* Check XML file */
SFLAG IsXMLFile (INSTREAM* in) {
 gchar c;
 SFLAG state;
 SFLAG tokState;
 GString* tempString;

 SkipUnicodeBOM (in);
 SkipSpaces (in);
 if (in -> nextChar != '<') return FALSE;
 ++ in -> actPos;
 in -> nextChar = in -> stream [in -> actPos];
 if (in -> nextChar != '?') {
   tempString = g_string_new (VOIDSTRING);
   in -> actPos = 0;
   in -> line = 0;
   SkipSpaces (in);
   ReadToken (in, tempString, &tokState);
   state = FALSE;
   if ((tokState & XML_TOKEN_OPEN) &&
       ((!g_strcmp0 (tempString -> str, "document")) || (!g_strcmp0 (tempString -> str, "document"))))
     state = TRUE;
   in -> actPos = 0;
   in -> line = 0;
   SkipSpaces (in);
   g_string_free (tempString, TRUE);
   return state;
 }
 ++ in -> actPos;
 for (; in -> actPos < in -> len; ++ in -> actPos) {
   c = in -> stream [in -> actPos];
   if (c == '>') break;
 }
 if (in -> actPos == in -> len) {
   in -> line = 0;
   in -> actPos = 0;
   SkipSpaces (in);
   return FALSE;
 }
 ++ in -> actPos;
 SkipSpaces (in);
 return TRUE;
}

/* Read a XML token */
SFLAG ReadToken (INSTREAM* in, GString* dest, SFLAG* tokenState) {
 gchar c;

 *tokenState = XML_ERROR;
 if (dest != NULL)
   g_string_erase (dest, 0, -1);
 if (in -> state == EOF_REACHED) return in -> state;
 if (in -> nextChar != '<') return in -> state;
 ++ in -> actPos;
 in -> nextChar = in -> stream [in -> actPos];
 if (in -> nextChar == '!') {
   ++ in -> actPos;
   in -> nextChar = in -> stream [in -> actPos];
   SkipCommnet (in);
   *tokenState = XML_TOKEN_COMMENT;
   return in -> state;
 }
 else if (in -> nextChar == '/') {
   ++ in -> actPos;
   *tokenState = XML_TOKEN_CLOSE;
 }
 else
   *tokenState = XML_TOKEN_OPEN;
 for (; in -> actPos < in -> len; ++ in -> actPos) {
   c = in -> stream [in -> actPos];
   if (c == '>') break;
   if (c == '/') {
      SkipParams (in);
      *tokenState = XML_TOKEN_EMPTY;
      return in -> state;
   }
   if (g_ascii_isspace (c)) {
     if (*tokenState == XML_TOKEN_CLOSE) {
       SkipParams (in);
       return in -> state;
     }
     else
       *tokenState |= XML_WITH_PARAMS;
     break;
   }
   if (dest != NULL) {
     if (g_ascii_isupper (c))
       c += 32;
     g_string_append_c (dest, c);
   }
 }
 /* Put a NULL character in the end */
 if ((dest != NULL) && (dest -> len > 0))
   g_string_append_c (dest, 0);
 if (in -> actPos == in -> len) {
   in -> nextChar = 0;
   in -> state = EOF_REACHED;
   *tokenState = XML_ERROR;
 }
 else {
   ++ in -> actPos;
   SkipSpaces (in);
 }
 return in -> state;
}

/* Read an attribute and its value */
SFLAG ReadXMLTokenParam (INSTREAM* in, GString* paramName, GString* paramVal, SFLAG* tokenState) {
 gchar c;

 *tokenState = XML_ERROR;
 if (paramName != NULL)
   g_string_erase (paramName, 0, -1);
 if (in -> nextChar == '/') {
   ++ in -> actPos;
   in -> nextChar = in -> stream [in -> actPos];
   if (in -> nextChar == '>') {
     *tokenState = XML_AUTO_CLOSE;
     ++ in -> actPos;
     SkipSpaces (in);
   }
   return in -> state;
 }
 for (; in -> actPos < in -> len; ++ in -> actPos) {
   c = in -> stream [in -> actPos];
   if (g_ascii_isspace (c)) return in -> state;
   if (c == '=') break;
   if (paramName != NULL) {
     if (g_ascii_isupper (c))
       c += 32;
     g_string_append_c (paramName, c);
   }
 }
 if ((paramName != NULL) && (paramName -> len > 0))
   g_string_append_c (paramName, 0);
 ++ in -> actPos;
 in -> nextChar = in -> stream [in -> actPos];
 if (in -> nextChar == '\"') {
   ++ in -> actPos;
   in -> nextChar = in -> stream [in -> actPos];
 }
 else return in -> state;
 if (paramVal != NULL)
   g_string_erase (paramVal, 0, -1);
 for (; in -> actPos < in -> len; ++ in -> actPos) {
   c = in -> stream [in -> actPos];
   if (c == '\"') break;
   if (paramVal != NULL)
     g_string_append_c (paramVal, c);
 }
 if ((paramVal != NULL) && (paramVal -> len > 0))
   g_string_append_c (paramVal, 0);
 ++ in -> actPos;
 in -> nextChar = in -> stream [in -> actPos];
 if (g_ascii_isspace (in -> nextChar)) {
   *tokenState = XML_WITH_PARAMS;
   SkipSpaces (in);
 }
 else if (in -> nextChar == '>') {
   *tokenState = XML_NO_PARAMS;
   ++ in -> actPos;
   SkipSpaces (in);
 }
 return in -> state;
}

/* Read XML Content */
SFLAG ReadXMLContent (INSTREAM* in, GString* dest, SFLAG* textState) {
 gchar c;

 *textState = XML_ERROR;
 if (dest != NULL)
   g_string_erase (dest, 0, -1);
 if (in -> state == EOF_REACHED) return in -> state;
 if (in -> state == EOF_REACHED) return in -> state;
 if (in -> nextChar == '<') {
   *textState = XML_TOKEN_NOCONTENT;
   return in -> state;
 }
 if (in -> nextChar == '!') {
   ++ in -> actPos;
   in -> nextChar = in -> stream [in -> actPos];
   if (in -> nextChar == '[') {
     SkipCData (in);
     *textState = XML_TOKEN_CDATA;
     return in -> state;
   }
 }
 for (; in -> actPos < in -> len; ++ in -> actPos) {
   c = in -> stream [in -> actPos];
   if (c == '<') {
     in -> nextChar = '<';
     break;
   }
   if ((dest != NULL) && (c != '\r') && (c != '&'))
     g_string_append_c (dest, c);
   else if ((dest != NULL) && (c == '&'))
     g_string_append (dest, "and");
 }
 /* Put a NULL character in the end */
 if ((dest != NULL) && (dest -> len > 0))
   g_string_append_c (dest, 0);
 if (in -> actPos == in -> len) {
   in -> nextChar = 0;
   in -> state = EOF_REACHED;
 }
 else {
   SkipSpaces (in);
    *textState = XML_TOKEN_CONTENT;
 }
 return in -> state;
}

void SkipCommnet (INSTREAM* in) {
 gchar c, prevC, farC;
 SFLAG cdataFlag;

 prevC = '!';
 farC = '<';
 cdataFlag = FALSE;
 if (in -> stream [in -> actPos] == '[')
   cdataFlag = TRUE;
 for (; in -> actPos < in -> len; ++ in -> actPos) {
   c = in -> stream [in -> actPos];
   if (c == '>') {
     if ((cdataFlag == TRUE) && (prevC == ']') && (farC == ']')) break;
     if ((cdataFlag == FALSE) && (prevC == '-') && (farC == '-')) break;
   }
   farC = prevC;
   prevC = c;
 }
 if (in -> actPos == in -> len) {
   in -> nextChar = 0;
   in -> state = EOF_REACHED;
 }
 else {
   ++ in -> actPos;
  SkipSpaces (in);
 }
}

void SkipCData (INSTREAM* in) {
 gchar c, prevC, farC;

 farC = '!';
 prevC = '[';
 for (; in -> actPos < in -> len; ++ in -> actPos) {
   c = in -> stream [in -> actPos];
   if ((c == '<') && (prevC == ']') && (farC == ']')) break;
   farC = prevC;
   prevC = c;
 }
 if (in -> actPos == in -> len) {
   in -> nextChar = 0;
   in -> state = EOF_REACHED;
 }
 else {
   ++ in -> actPos;
  SkipSpaces (in);
 }
}

SFLAG SkipParams (INSTREAM* in) {
 gchar c, prevC;

 prevC = c = '<';
 for (; in -> actPos < in -> len; ++ in -> actPos) {
   prevC = c;
   c = in -> stream [in -> actPos];
   if (c == '>') break;
 }
 if (in -> actPos == in -> len) {
   in -> nextChar = 0;
   in -> state = EOF_REACHED;
   return XML_ERROR;
 }
 else {
   ++ in -> actPos;
  SkipSpaces (in);
 }
 if (prevC == '/')
   return XML_AUTO_CLOSE;
 return XML_NO_PARAMS;
}

SFLAG ReadXMLFile (XMLPARSER* xmlParser) {
 GString* actToken;
 SFLAG toRet;

 toRet = TRUE;
 for (;;) {
   ReadToken (xmlParser -> in, xmlParser -> tempParserString, &xmlParser -> tokenState);
   if (xmlParser -> tokenState == XML_ERROR) {
     errorMessage = g_strdup_printf ("\nUnexpected file format\nFile %s, line %ld\n", xmlParser -> in -> name, xmlParser -> in -> line);
     if (xmlParser -> OnError != NULL)
       toRet = xmlParser -> OnError ();
     else {
       OnError ();
       toRet = FALSE;
     }
     break;
   }
   if (xmlParser -> tokenState & XML_TOKEN_CLOSE) {
     if (xmlParser -> tokenStack == NULL) {
       errorMessage = g_strdup_printf ("\nUnexpected \'</%s>\' token\nFile %s, line %ld\n",
                                       xmlParser -> tempParserString -> str, xmlParser -> in -> name, xmlParser -> in -> line);
       if (xmlParser -> OnError != NULL)
         toRet = xmlParser -> OnError ();
       else {
         OnError ();
         toRet = FALSE;
       }
       break;
     }
     actToken = (GString*) xmlParser -> tokenStack -> data;
     if (g_strcmp0 (xmlParser -> tempParserString -> str, actToken -> str)) {
       errorMessage = g_strdup_printf ("\nReading \'</%s>\', expecting token: </%s>\nFile %s, line %ld\n",
                         xmlParser -> tempParserString -> str, actToken -> str, xmlParser -> in -> name, xmlParser -> in -> line);
       if (xmlParser -> OnError != NULL)
         toRet = xmlParser -> OnError ();
       else {
         OnError ();
         toRet = FALSE;
       }
       break;
     }
     if (!g_strcmp0 (xmlParser -> tempParserString -> str, xmlParser -> fileHeader)) break;
     if ((xmlParser -> EndTag != NULL) && (xmlParser -> EndTag () == FALSE)) {
       if (xmlParser -> OnError != NULL)
         toRet = xmlParser -> OnError ();
       else
         toRet = FALSE;
       break;
     }
     RemoveFromTokenList (xmlParser);
   }
   else if (xmlParser -> tokenState & XML_TOKEN_OPEN) {
     actToken = g_string_new (xmlParser -> tempParserString -> str);
     xmlParser -> tokenStack = g_list_prepend (xmlParser -> tokenStack, actToken);
     if (xmlParser -> OpenTag == NULL) {
       if (IgnoreOpenTag (xmlParser) == FALSE) {
         if (xmlParser -> OnError != NULL)
           toRet = xmlParser -> OnError ();
         else
           toRet = FALSE;
         break;
       }
     }
     else if (xmlParser -> OpenTag () == FALSE) {
       if (xmlParser -> OnError != NULL)
         toRet = xmlParser -> OnError ();
       else
         toRet = FALSE;
       break;
     }
   }
 }
 if (xmlParser -> fileHeader != NULL)
   g_free (xmlParser -> fileHeader);
 xmlParser -> fileHeader = NULL;
 CleanTokenStack (xmlParser -> tokenStack);
 xmlParser -> tokenStack = NULL;
 return toRet;
}

void RemoveFromTokenList (XMLPARSER* xmlParser) {
 GList* toPop;
 toPop = xmlParser -> tokenStack;
 xmlParser -> tokenStack = g_list_next (xmlParser -> tokenStack);
 g_string_free (toPop -> data, TRUE);
 DestroyLink (toPop);
}

SFLAG IgnoreOpenTag (XMLPARSER* xmlParser) {
 SFLAG state;

 if (xmlParser -> tokenState & XML_WITH_PARAMS) {
   state = SkipParams (xmlParser -> in);
   if (state == XML_AUTO_CLOSE) {
     RemoveFromTokenList (xmlParser);
     return TRUE;
   }
   else if (state == XML_ERROR) return FALSE;
 }
 state = ReadXMLContent (xmlParser -> in, NULL, &xmlParser -> tokenState);
 if (state == XML_ERROR) return FALSE;
 return TRUE;
}

GString* GetPrevToken (XMLPARSER* xmlParser) {
 GList* prevToken;

 if (xmlParser -> tokenStack == NULL) return NULL;
 prevToken = g_list_next (xmlParser -> tokenStack);
 if (prevToken == NULL) return NULL;
 return  prevToken -> data;
}

void CleanTokenStack (GList* toClean) {
 GList* nav;
 for (nav = toClean; nav != NULL; nav = g_list_next (nav))
   g_string_free (nav -> data, TRUE);
 g_list_free (toClean);
}

void SkipUnicodeBOM (INSTREAM* in) {
 gchar c;

 c = in -> stream [in -> actPos];
 if (c == -17) {
   in -> actPos += 3;
   in -> nextChar = in -> stream [in -> actPos];
 }
}
/* link.c is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */

#include "ghead.h"

GList* InsertListElement (GList* prev, gpointer data) {
 GList* nu;

 nu = g_list_append (NULL, data);
 if (prev != NULL) {
   nu -> prev = prev;
   nu -> next = prev -> next;
   if (prev -> next != NULL)
     prev -> next -> prev = nu;
   prev -> next = nu;
 }
 return nu;
}

void PutAtListEnd (GList* dest, GList* toPut) {
 GList* last;

 last = g_list_last (dest);
 toPut -> prev = last;
 if (last != NULL)
   last -> next = toPut;
}

void DestroyList (GList* toDel, gsize sz) {
 GList* nav;
 if (toDel == NULL) return;
 if (g_list_previous (toDel) != NULL)
   g_list_previous (toDel) -> next = NULL;
 for (nav = toDel; nav != NULL; nav = g_list_next (nav))
   g_slice_free1 (sz, nav -> data);
 g_list_free (toDel);
}

void DestroyLinkElement (GList* toDel, gsize sz) {
 if (toDel == NULL) return;
 if (g_list_previous (toDel) != NULL)
   g_list_previous (toDel) -> next = g_list_next (toDel);
 if (g_list_next (toDel) != NULL)
   g_list_next (toDel) -> prev = g_list_previous (toDel);
 toDel -> prev = toDel -> next = NULL;
 g_slice_free1 (sz, toDel -> data);
 g_list_free (toDel);
}

void DestroyLink (GList* toDel) {
 if (toDel == NULL) return;
 if (g_list_previous (toDel) != NULL)
   g_list_previous (toDel) -> next = g_list_next (toDel);
 if (g_list_next (toDel) != NULL)
   g_list_next (toDel) -> prev = g_list_previous (toDel);
 toDel -> prev = toDel -> next = NULL;
 g_list_free (toDel);
}

void AddChildren (GNode* toAdd, GNode* anc, GNode* lastSibling) {
 if (anc == NULL) return;
 toAdd -> parent = anc;
 if (lastSibling == NULL)
   toAdd -> parent -> children = toAdd;
 else
   lastSibling -> next = toAdd;
 toAdd -> prev = lastSibling;
}

GNode* GetNextNode (GNode* actNode) {
 GNode* navNode;

 if (actNode == NULL) return NULL;
 if (g_node_first_child (actNode) != NULL)
   return g_node_first_child (actNode);

 for (navNode = actNode;; navNode = navNode -> parent) {
   if (g_node_next_sibling (navNode) != NULL) return g_node_next_sibling (navNode);
   if (navNode -> parent == NULL) break;
 }
 return navNode;
}

GNode* GetNextTerm (GNode* actNode) {
 GNode* navNode;

 if (actNode == NULL) return NULL;
 navNode = actNode;
 /* Search the next available node, if there are no more nodes, returns NULL */
 if (g_node_first_child (actNode) == NULL) {
   for (; navNode != NULL; navNode = navNode -> parent) {
     if (g_node_next_sibling (navNode) != NULL) {
       navNode = g_node_next_sibling (navNode);
       break;
     }
   }
   if (navNode == NULL) return NULL;
 }
 for (; g_node_first_child (navNode) != NULL; navNode = g_node_first_child (navNode));
 return navNode;
}

GNode* GetPrevNode (GNode* actNode) {
 GNode* navNode;

 if (actNode == NULL) return NULL;
 if (g_node_prev_sibling (actNode) != NULL) return g_node_prev_sibling (actNode);
 if (actNode -> parent != NULL) return actNode -> parent;
 /* actNode is the root, so the "previous" node is the last children of the whole tree */
 for (navNode = actNode; ; navNode = g_node_last_child (navNode))
   if (g_node_last_child (navNode) == NULL) break;
 return navNode;
}

SFLAG IsDesc (GNode* probe, GNode* anc) {
 GNode* navNode;

 if ((probe == NULL) || (anc == NULL)) return FALSE;
 for (navNode = probe; navNode != NULL; navNode = navNode -> parent)
   if (navNode == anc) return TRUE;
 return FALSE;
}

/* Lists of names */

GList* NameOnList (gchar* name, GList* list) {
 GList* nav;
 GList* toRet;
 GString* probe;
 GString* toAdd;
 SINTEGER cmp;

 toRet = list;
 if (list == NULL) {
   toAdd = g_string_new (name);
   toRet = InsertListElement (NULL, toAdd);
   return toRet;
 }
 for (nav = list; nav != NULL; nav = g_list_next (nav)) {
   probe = (GString*) nav -> data;
   cmp = g_strcmp0 (name, probe -> str);
   if (cmp == 0) break;
   else if (cmp < 0) {
     toAdd = g_string_new (name);
     if (nav == list) {
       toRet = InsertListElement (NULL, toAdd);
       toRet -> next = list;
       list -> prev = toRet;
       break;
     }
     InsertListElement (g_list_previous (nav), toAdd);
     break;
   }
   else if (g_list_next (nav) == NULL) {
     toAdd = g_string_new (name);
     InsertListElement (nav, toAdd);
     break;
   }
 }
 return toRet;
}
/* mainwin.c is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */

#include "wins.h"
#include "dialogs.h"
#include "parser.h"
#include "out.h"

#define VERSION_STRING "VIP v16 22 oct 2010\nJ. Salvador Arias\nCONICET, INSUE, Tucuman, Argentina\njsalarias@csnat.unt.edu.ar\n"

typedef struct tagMAINWINDATA MAINWINDATA;
struct tagMAINWINDATA {
 GtkWidget* menuGrid;
 GtkWidget* menuRec;
 GtkWidget* menuEdit;

 GtkWidget* menuOpen;
 GtkWidget* menuFeed;
 GtkWidget* menuSaveAs;
 GtkWidget* menuOpenRes;
 GtkWidget* menuSaveRes;
 GtkWidget* menuClose;
 GtkWidget* menuSetFill;
 GtkWidget* menuCreateGrid;
 GtkWidget* menuReDoGrid;
 GtkWidget* menuDestroyGrid;
 GtkWidget* menuGridTable;
 GtkWidget* menuFilter;
 GtkWidget* menuClean;
 GtkWidget* menuConsensus;
 GtkWidget* menuCleanConsensus;
 GtkWidget* menuHotMouse;
 GtkWidget* menuGeoBrowser;
 GtkWidget* menuRecEdit;

 GtkWidget* removEntry;
 GtkWidget* overlapEntry;
 GtkWidget* partEntry;
 GtkWidget* maxTimeEntry;
 GtkWidget* timeLagEntry;
 GtkWidget* lagCostEntry;

 GtkWidget* lateOverlapBox;
 GtkWidget* costFractionBox;
 GtkWidget* timeLagBox;
 GtkWidget* timeFractionBox;

 GtkWidget* buttonSearch;
 GtkWidget* buttonPrevTree;
 GtkWidget* buttonNextTree;
 GtkWidget* buttonPrevNode;
 GtkWidget* buttonNextNode;
 GtkWidget* buttonNextRec;
 GtkWidget* buttonPrevRec;
 GtkWidget* buttonCopy;
 GtkWidget* radioORRec;
 GtkWidget* radioSearchRec;
 GtkWidget* radioConsRec;

 GtkWidget* treeLabelEntry;
 GtkWidget* treeNavEntry;
 GtkWidget* costLabelEntry;
 GtkWidget* vicsLabelEntry;
 GtkWidget* removsLabelEntry;
 GtkWidget* recNavEntry;
 GtkWidget* nodeLabelEntry;

 GtkTextBuffer* logBuffer;
};

static gint OnDelete (GtkWidget* mainWin, GdkEvent* event, gpointer data);

/* --- Menu events --- */
static void OnMenuOpen (GtkMenuItem* openMenu, gpointer data);
static void OnMenuFeed (GtkMenuItem* openMenu, gpointer data);
static void OnMenuSaveAs (GtkMenuItem* saveAsMenu, gpointer data);
static void OnMenuOpenRes (GtkMenuItem* openMenu, gpointer data);
static void OnMenuSaveRes (GtkMenuItem* saveResMenu, gpointer data);
static void OnMenuClose (GtkMenuItem* closeMenu, gpointer data);
static void OnMenuQuit (GtkMenuItem* quitMenu, gpointer data);

static void OnMenuSetGrid (GtkMenuItem* gridMenu, gpointer data);
static void OnMenuSetFill (GtkMenuItem* gridMenu, gpointer data);
static void OnMenuCreateGrid (GtkMenuItem* gridMenu, gpointer data);
static void OnMenuReDoGrid (GtkMenuItem* gridMenu, gpointer data);
static void OnMenuDestroyGrid (GtkMenuItem* gridMenu, gpointer data);
static void OnMenuGridTable (GtkMenuItem* gridMenu, gpointer data);

static void OnMenuPageHeuristic (GtkMenuItem* searchMenu, gpointer data);
static void OnMenuConsensus (GtkMenuItem* consensusMenu, gpointer data);
static void OnMenuFilter (GtkMenuItem* filterMenu, gpointer data);
static void OnMenuClean (GtkMenuItem* cleanMenu, gpointer data);
static void OnMenuCleanConsensus (GtkMenuItem* cleanMenu, gpointer data);

static void OnMenuHotCheck (GtkMenuItem* editMenu, gpointer data);
static void OnMenuRecEditCheck (GtkMenuItem* editMenu, gpointer data);
static void OnMenuGeoBrowser (GtkMenuItem* browserMenu, gpointer data);
static void OnMenuSaveLog (GtkMenuItem* saveAsMenu, gpointer data);
static void OnMenuClearLog (GtkMenuItem* saveAsMenu, gpointer data);

/* --- Other events --- */
static void OnButtonApply (GtkButton* applyButton, gpointer data);
static void OnButtonSearch (GtkButton* searchButton, gpointer data);
static void OnButtonCancel (GtkButton* cancelButton, gpointer data);
static void OnButtonPrevTree (GtkButton* prevButton, gpointer data);
static void OnButtonNextTree (GtkButton* nextButton, gpointer data);
static void OnButtonPrevNode (GtkButton* prevButton, gpointer data);
static void OnButtonNextNode (GtkButton* nextButton, gpointer data);
static void OnButtonPrevRec (GtkButton* nextButton, gpointer data);
static void OnButtonNextRec (GtkButton* prevButton, gpointer data);
static void OnButtonCopy (GtkButton* copyButton, gpointer data);
static void OnRadioRec (GtkToggleButton* button, gpointer data);
static void OnUseTime (GtkCheckButton* button, gpointer data);


/* --- Other functions --- */
static int CloseData ();
static SFLAG ApplyParams (SFLAG toPrint);
static SFLAG CheckRadioRec ();

static void DoSearch (SFLAG keep);

static MAINWINDATA mainData;
static PROGRESSSTRUCT progress;

void StartMainWindow () {
 GtkWidget* gBox;
 GtkWidget* menuBar;
 GtkWidget* menu;
 GtkAccelGroup* accelGroup;

 GtkWidget* notebook;
 GtkWidget* vBox;
 GtkWidget* hBox;
 GtkWidget* label;
 GtkWidget* separator;
 GtkWidget* button;
 gchar* text;

 GtkWidget* scrollWin;
 GtkWidget* logTab;

 gBox = gtk_vbox_new (FALSE, 0);
 mainWindow = CreateWindow ("VIP: Main Window", 30, 0, 200, 500, G_CALLBACK (OnDelete), gBox);

 /* --- Sets menues --- */
 accelGroup = gtk_accel_group_new ();
 gtk_window_add_accel_group (GTK_WINDOW (mainWindow), accelGroup);
 menuBar = gtk_menu_bar_new ();
 gtk_box_pack_start (GTK_BOX (gBox), menuBar, FALSE, TRUE, 0);
 gtk_widget_show (menuBar);
 /* File */
 menu = gtk_menu_new ();
 mainData . menuOpen = CreateMenuItem (menu, "_Open...", accelGroup, GDK_O, GDK_CONTROL_MASK, G_CALLBACK (OnMenuOpen), NULL);
 gtk_widget_set_sensitive (mainData . menuOpen, TRUE);
 mainData . menuFeed = CreateMenuItem (menu, "_Feed records...", accelGroup, GDK_X, GDK_CONTROL_MASK,
                                         G_CALLBACK (OnMenuFeed), NULL);
 gtk_widget_set_sensitive (mainData . menuFeed, FALSE);
 mainData . menuSaveAs = CreateMenuItem (menu, "Save data _as...", accelGroup, GDK_S, GDK_CONTROL_MASK | GDK_SHIFT_MASK,
                                         G_CALLBACK (OnMenuSaveAs), NULL);
 gtk_widget_set_sensitive (mainData . menuSaveAs, FALSE);
 CreateMenuSeparator (menu);
 mainData . menuOpenRes = CreateMenuItem (menu, "Open _results...", accelGroup, GDK_O, GDK_CONTROL_MASK | GDK_SHIFT_MASK,
                                          G_CALLBACK (OnMenuOpenRes), NULL);
 gtk_widget_set_sensitive (mainData . menuOpenRes, FALSE);
 mainData . menuSaveRes = CreateMenuItem (menu, "_Save results as...", accelGroup, GDK_S, GDK_CONTROL_MASK,
                                         G_CALLBACK (OnMenuSaveRes), NULL);
 gtk_widget_set_sensitive (mainData . menuSaveRes, FALSE);
 CreateMenuSeparator (menu);
 mainData . menuClose = CreateMenuItem (menu, "_Close", accelGroup, GDK_W, GDK_CONTROL_MASK, G_CALLBACK (OnMenuClose), NULL);
 gtk_widget_set_sensitive (mainData . menuClose, FALSE);
 CreateMenuSeparator (menu);
 CreateMenuItem (menu, "_Quit...", accelGroup, GDK_Q, GDK_CONTROL_MASK, G_CALLBACK (OnMenuQuit), NULL);
 AddMenuToBar (menu, "_File", menuBar);
 /* Grid */
 menu = gtk_menu_new ();
 CreateMenuItem (menu, "_Grid settings...", accelGroup, GDK_G, GDK_CONTROL_MASK, G_CALLBACK (OnMenuSetGrid), NULL);
 mainData . menuSetFill = CreateMenuItem (menu, "Set _fill in all nodes", accelGroup, GDK_I, GDK_CONTROL_MASK,
                                          G_CALLBACK (OnMenuSetFill), NULL);
 CreateMenuSeparator (menu);
 mainData . menuCreateGrid = CreateMenuItem (menu, "_Create grid", accelGroup, GDK_A, GDK_CONTROL_MASK,
                                             G_CALLBACK (OnMenuCreateGrid), NULL);
 mainData . menuReDoGrid = CreateMenuItem (menu, "_Redo grid", accelGroup, GDK_B, GDK_CONTROL_MASK,
                                             G_CALLBACK (OnMenuReDoGrid), NULL);
 CreateMenuSeparator (menu);
 mainData . menuDestroyGrid = CreateMenuItem (menu, "_Destroy grid", accelGroup, GDK_D, GDK_CONTROL_MASK,
                                              G_CALLBACK (OnMenuDestroyGrid), NULL);
 CreateMenuSeparator (menu);
 mainData . menuGridTable = CreateMenuItem (menu, "Save grid _table...", accelGroup, GDK_M, GDK_CONTROL_MASK,
                                         G_CALLBACK (OnMenuGridTable), NULL);
 gtk_widget_set_sensitive (mainData . menuReDoGrid, FALSE);
 gtk_widget_set_sensitive (mainData . menuDestroyGrid, FALSE);
 gtk_widget_set_sensitive (mainData . menuGridTable, FALSE);
 mainData . menuGrid = AddMenuToBar (menu, "_Grid", menuBar);
 gtk_widget_set_sensitive (mainData . menuGrid, FALSE);

 SETFILL (1);
 SETGRIDSIZE (2.00);
 SETGRIDMODE (VONNEUMANN_NEIGHBORS);

 /* Recons */
 menu = gtk_menu_new ();
 CreateMenuItem (menu, "_Heuristic search...", accelGroup, GDK_E, GDK_CONTROL_MASK, G_CALLBACK (OnMenuPageHeuristic), NULL);
 CreateMenuSeparator (menu);
 mainData . menuConsensus = CreateMenuItem (menu, "_Consensus", accelGroup, GDK_U, GDK_CONTROL_MASK, G_CALLBACK (OnMenuConsensus), NULL);
 gtk_widget_set_sensitive (mainData . menuConsensus, FALSE);
 CreateMenuSeparator (menu);
 mainData . menuFilter = CreateMenuItem (menu, "_Filter...", accelGroup, GDK_F, GDK_CONTROL_MASK, G_CALLBACK (OnMenuFilter), NULL);
 gtk_widget_set_sensitive (mainData . menuFilter, FALSE);
 mainData . menuClean = CreateMenuItem (menu, "Clea_n", accelGroup, GDK_N, GDK_CONTROL_MASK, G_CALLBACK (OnMenuClean), NULL);
 gtk_widget_set_sensitive (mainData . menuClean, FALSE);
 mainData . menuCleanConsensus = CreateMenuItem (menu, "Clean consens_us", accelGroup, GDK_U, GDK_CONTROL_MASK,
                           G_CALLBACK (OnMenuCleanConsensus), NULL);
 gtk_widget_set_sensitive (mainData . menuCleanConsensus, FALSE);
 mainData . menuRec = AddMenuToBar (menu, "_Search", menuBar);
 gtk_widget_set_sensitive (mainData . menuRec, FALSE);


 /* Edit */
 menu = gtk_menu_new ();
 mainData . menuHotMouse = CreateMenuItemCheck (menu, "_Hot mouse (map)", accelGroup, GDK_H, GDK_CONTROL_MASK,
                                                    G_CALLBACK (OnMenuHotCheck), NULL);
 gtk_widget_set_sensitive (mainData . menuHotMouse, FALSE);
 mainData . menuRecEdit = CreateMenuItemCheck (menu, "Edit _reconstruction (tree)", accelGroup, GDK_T, GDK_CONTROL_MASK,
                                                    G_CALLBACK (OnMenuRecEditCheck), NULL);
 gtk_widget_set_sensitive (mainData . menuRecEdit, FALSE);
 CreateMenuSeparator (menu);
 mainData . menuGeoBrowser = CreateMenuItem (menu, "_Georeference browser", accelGroup, GDK_R, GDK_CONTROL_MASK,
                                                    G_CALLBACK (OnMenuGeoBrowser), NULL);
 gtk_widget_set_sensitive (mainData . menuGeoBrowser, FALSE);
 CreateMenuSeparator (menu);
 CreateMenuItem (menu, "_Save log...", accelGroup, GDK_L, GDK_CONTROL_MASK, G_CALLBACK (OnMenuSaveLog), NULL);
 CreateMenuItem (menu, "_Clean log", accelGroup, GDK_K, GDK_CONTROL_MASK, G_CALLBACK (OnMenuClearLog), NULL);
 mainData . menuEdit = AddMenuToBar (menu, "_Edit", menuBar);
 gtk_widget_set_sensitive (mainData . menuEdit, FALSE);

 SETFILL (1);
 SETGRIDSIZE (2.00);

 /* --- Layout --- */
 notebook = gtk_notebook_new ();
 gtk_box_pack_start (GTK_BOX (gBox), notebook, FALSE, TRUE, 0);
 gtk_widget_show (notebook);

 /* Parameters flap */
 vBox = gtk_vbox_new (FALSE, 2);
 gtk_notebook_append_page (GTK_NOTEBOOK (notebook), vBox, gtk_label_new_with_mnemonic ("_Parameters"));
 gtk_widget_show (vBox);

 mainData . removEntry = CreateEntry ("Cost of distribution removal", GTK_BOX (vBox));
 mainData . overlapEntry = CreateEntry ("Max overlap (percent)", GTK_BOX (vBox));
 mainData . costFractionBox = gtk_check_button_new_with_mnemonic ("Use fractional _cost");
 gtk_box_pack_start (GTK_BOX (vBox), mainData . costFractionBox, TRUE, TRUE, 2);
 gtk_widget_show (mainData . costFractionBox);
 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (vBox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);

 mainData . lateOverlapBox = gtk_check_button_new_with_mnemonic ("Use _partial removal");
 gtk_widget_show (mainData . lateOverlapBox);
 gtk_box_pack_start (GTK_BOX (vBox), mainData . lateOverlapBox, TRUE, TRUE, 2);
 mainData . partEntry = CreateEntry ("Cost of partial removal", GTK_BOX (vBox));
 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (vBox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);

 mainData . timeLagBox = gtk_check_button_new_with_mnemonic ("Use _time");
 gtk_box_pack_start (GTK_BOX (vBox), mainData . timeLagBox, TRUE, TRUE, 2);
 gtk_widget_show (mainData . timeLagBox);
 mainData . maxTimeEntry = CreateEntry ("Max time", GTK_BOX (vBox));
 mainData . timeLagEntry = CreateEntry ("Max time lag", GTK_BOX (vBox));
 mainData . lagCostEntry = CreateEntry ("Cost after time lag", GTK_BOX (vBox));
 mainData . timeFractionBox = gtk_check_button_new_with_mnemonic ("Use _fractional time cost");
 gtk_box_pack_start (GTK_BOX (vBox), mainData . timeFractionBox, TRUE, TRUE, 2);
 gtk_widget_show (mainData . timeFractionBox);
 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (vBox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);
 g_signal_connect (G_OBJECT (mainData . timeLagBox), "toggled", G_CALLBACK (OnUseTime), NULL);

 text = g_strdup_printf ("%.4f", params . costRemov);
 CheckRealStr (text);
 gtk_entry_set_text (GTK_ENTRY (mainData . removEntry), text);
 g_free (text);
 text = g_strdup_printf ("%.4f", params . costPartRemov);
 CheckRealStr (text);
 gtk_entry_set_text (GTK_ENTRY (mainData . partEntry), text);
 g_free (text);
 text = g_strdup_printf ("%.4f", params . overlap * 100);
 CheckRealStr (text);
 gtk_entry_set_text (GTK_ENTRY (mainData . overlapEntry), text);
 g_free (text);
 text = g_strdup_printf ("%.4f", params . costTimeLag);
 CheckRealStr (text);
 gtk_entry_set_text (GTK_ENTRY (mainData . lagCostEntry), text);
 g_free (text);
 text = g_strdup_printf ("%ld", params . maxTime);
 gtk_entry_set_text (GTK_ENTRY (mainData . maxTimeEntry), text);
 g_free (text);
 text = g_strdup_printf ("%ld", params . timeLag);
 gtk_entry_set_text (GTK_ENTRY (mainData . timeLagEntry), text);
 g_free (text);

 hBox = gtk_hbox_new (TRUE, 2);
 gtk_box_pack_start (GTK_BOX (vBox), hBox, FALSE, TRUE, 2);
 gtk_widget_show (hBox);
 button = gtk_button_new_from_stock (GTK_STOCK_APPLY);
 g_signal_connect (G_OBJECT (button), "clicked", G_CALLBACK (OnButtonApply), NULL);
 gtk_box_pack_start (GTK_BOX (hBox), button, FALSE, TRUE, 2);
 gtk_widget_show (button);
 button = gtk_button_new_from_stock (GTK_STOCK_CANCEL);
 g_signal_connect (G_OBJECT (button), "clicked", G_CALLBACK (OnButtonCancel), NULL);
 gtk_box_pack_start (GTK_BOX (hBox), button, FALSE, TRUE, 2);
 gtk_widget_show (button);
 mainData . buttonSearch = gtk_button_new_from_stock (GTK_STOCK_EXECUTE);
 g_signal_connect (G_OBJECT (mainData . buttonSearch), "clicked", G_CALLBACK (OnButtonSearch), NULL);
 gtk_box_pack_start (GTK_BOX (vBox), mainData . buttonSearch, FALSE, TRUE, 2);
 gtk_widget_show (mainData . buttonSearch);
 gtk_widget_set_sensitive (mainData . buttonSearch, FALSE);

 /* Reconstruction flap */
 vBox = gtk_vbox_new (FALSE, 2);
 gtk_notebook_append_page (GTK_NOTEBOOK (notebook), vBox, gtk_label_new_with_mnemonic ("_Reconstruction"));
 gtk_widget_show (vBox);
 label = gtk_label_new ("Tree navigator");
 gtk_box_pack_start (GTK_BOX (vBox), label, TRUE, TRUE, 2);
 gtk_widget_show (label);
 mainData . treeNavEntry = CreateEntry ("Trees in memory", GTK_BOX (vBox));
 mainData . treeLabelEntry = CreateEntry ("Tree", GTK_BOX (vBox));
 hBox = gtk_hbox_new (TRUE, 2);
 gtk_box_pack_start (GTK_BOX (vBox), hBox, FALSE, TRUE, 2);
 gtk_widget_show (hBox);
 mainData . buttonPrevTree = gtk_button_new_from_stock (GTK_STOCK_GO_BACK);
 g_signal_connect (G_OBJECT (mainData . buttonPrevTree), "clicked", G_CALLBACK (OnButtonPrevTree), NULL);
 gtk_box_pack_start (GTK_BOX (hBox), mainData . buttonPrevTree, FALSE, TRUE, 2);
 gtk_widget_show (mainData . buttonPrevTree);
 mainData . buttonNextTree = gtk_button_new_from_stock (GTK_STOCK_GO_FORWARD);
 g_signal_connect (G_OBJECT (mainData . buttonNextTree), "clicked", G_CALLBACK (OnButtonNextTree), NULL);
 gtk_box_pack_start (GTK_BOX (hBox), mainData . buttonNextTree, FALSE, TRUE, 2);
 gtk_widget_show (mainData . buttonNextTree);
 gtk_widget_set_sensitive (mainData . buttonPrevTree, FALSE);
 gtk_widget_set_sensitive (mainData . buttonNextTree, FALSE);
 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (vBox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);

 label = gtk_label_new ("Reconstruction navigator");
 gtk_box_pack_start (GTK_BOX (vBox), label, TRUE, TRUE, 2);
 gtk_widget_show (label);
 mainData . costLabelEntry = CreateEntry ("Cost", GTK_BOX (vBox));
 mainData . vicsLabelEntry = CreateEntry ("Disjunct sister pairs", GTK_BOX (vBox));
 mainData . removsLabelEntry = CreateEntry ("Nodes with removal", GTK_BOX (vBox));
 mainData . recNavEntry = CreateEntry ("Reconstructions in memory", GTK_BOX (vBox));
 hBox = gtk_hbox_new (TRUE, 2);
 gtk_box_pack_start (GTK_BOX (vBox), hBox, FALSE, TRUE, 2);
 gtk_widget_show (hBox);
 mainData . buttonPrevRec = gtk_button_new_from_stock (GTK_STOCK_GO_BACK);
 g_signal_connect (G_OBJECT (mainData . buttonPrevRec), "clicked", G_CALLBACK (OnButtonPrevRec), NULL);
 gtk_box_pack_start (GTK_BOX (hBox), mainData . buttonPrevRec, FALSE, TRUE, 2);
 gtk_widget_show (mainData . buttonPrevRec);
 mainData . buttonNextRec = gtk_button_new_from_stock (GTK_STOCK_GO_FORWARD);
 g_signal_connect (G_OBJECT (mainData . buttonNextRec), "clicked", G_CALLBACK (OnButtonNextRec), NULL);
 gtk_box_pack_start (GTK_BOX (hBox), mainData . buttonNextRec, FALSE, TRUE, 2);
 gtk_widget_show (mainData . buttonNextRec);
 mainData . buttonCopy = gtk_button_new_from_stock (GTK_STOCK_COPY);
 g_signal_connect (G_OBJECT (mainData . buttonCopy), "clicked", G_CALLBACK (OnButtonCopy), NULL);
 gtk_box_pack_start (GTK_BOX (vBox), mainData . buttonCopy, FALSE, TRUE, 2);
 gtk_widget_show (mainData . buttonCopy);
 gtk_widget_set_sensitive (mainData . buttonPrevRec, FALSE);
 gtk_widget_set_sensitive (mainData . buttonNextRec, FALSE);
 gtk_widget_set_sensitive (mainData . buttonCopy, FALSE);
 mainData . radioORRec = gtk_radio_button_new_with_mnemonic (NULL, "_OR reconstruction");
 g_signal_connect (G_OBJECT (mainData . radioORRec), "toggled", G_CALLBACK (OnRadioRec), NULL);
 gtk_box_pack_start (GTK_BOX (vBox), mainData . radioORRec, FALSE, TRUE, 2);
 gtk_widget_show (mainData . radioORRec);
 mainData . radioSearchRec =
    gtk_radio_button_new_with_mnemonic_from_widget (GTK_RADIO_BUTTON (mainData . radioORRec), "Reconstruction _buffer");
 g_signal_connect (G_OBJECT (mainData . radioSearchRec), "toggled", G_CALLBACK (OnRadioRec), NULL);
 gtk_box_pack_start (GTK_BOX (vBox), mainData . radioSearchRec, FALSE, TRUE, 2);
 gtk_widget_show (mainData . radioSearchRec);
 mainData . radioConsRec =
    gtk_radio_button_new_with_mnemonic_from_widget (GTK_RADIO_BUTTON (mainData . radioORRec), "_Consensus reconstruction");
 g_signal_connect (G_OBJECT (mainData . radioConsRec), "toggled", G_CALLBACK (OnRadioRec), NULL);
 gtk_box_pack_start (GTK_BOX (vBox), mainData . radioConsRec, FALSE, TRUE, 2);
 gtk_widget_show (mainData . radioConsRec);
 gtk_widget_set_sensitive (mainData . radioORRec, FALSE);
 gtk_widget_set_sensitive (mainData . radioSearchRec, FALSE);
 gtk_widget_set_sensitive (mainData . radioConsRec, FALSE);
 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (vBox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);

 label = gtk_label_new ("Node navigator");
 gtk_box_pack_start (GTK_BOX (vBox), label, TRUE, TRUE, 2);
 gtk_widget_show (label);
 mainData . nodeLabelEntry = CreateEntry ("Node", GTK_BOX (vBox));
 hBox = gtk_hbox_new (TRUE, 2);
 gtk_box_pack_start (GTK_BOX (vBox), hBox, FALSE, TRUE, 2);
 gtk_widget_show (hBox);
 mainData . buttonPrevNode = gtk_button_new_from_stock (GTK_STOCK_GO_BACK);
 g_signal_connect (G_OBJECT (mainData . buttonPrevNode), "clicked", G_CALLBACK (OnButtonPrevNode), NULL);
 gtk_box_pack_start (GTK_BOX (hBox), mainData . buttonPrevNode, FALSE, TRUE, 2);
 gtk_widget_show (mainData . buttonPrevNode);
 mainData . buttonNextNode = gtk_button_new_from_stock (GTK_STOCK_GO_FORWARD);
 g_signal_connect (G_OBJECT (mainData . buttonNextNode), "clicked", G_CALLBACK (OnButtonNextNode), NULL);
 gtk_box_pack_start (GTK_BOX (hBox), mainData . buttonNextNode, FALSE, TRUE, 2);
 gtk_widget_show (mainData . buttonNextNode);
 gtk_widget_set_sensitive (mainData . buttonPrevNode, FALSE);
 gtk_widget_set_sensitive (mainData . buttonNextNode, FALSE);
 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (vBox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);

 gtk_entry_set_text (GTK_ENTRY (mainData . treeNavEntry), "No tree");
 gtk_entry_set_text (GTK_ENTRY (mainData . treeLabelEntry), "No tree");
 gtk_entry_set_text (GTK_ENTRY (mainData . costLabelEntry), "No grid");
 gtk_entry_set_text (GTK_ENTRY (mainData . vicsLabelEntry), "No grid");
 gtk_entry_set_text (GTK_ENTRY (mainData . removsLabelEntry), "No grid");
 gtk_entry_set_text (GTK_ENTRY (mainData . recNavEntry), "No grid");
 gtk_entry_set_text (GTK_ENTRY (mainData . nodeLabelEntry), "No tree");
 gtk_editable_set_editable (GTK_EDITABLE (mainData . treeLabelEntry), FALSE);
 gtk_editable_set_editable (GTK_EDITABLE (mainData . costLabelEntry), FALSE);
 gtk_editable_set_editable (GTK_EDITABLE (mainData . vicsLabelEntry), FALSE);
 gtk_editable_set_editable (GTK_EDITABLE (mainData . removsLabelEntry), FALSE);
 gtk_editable_set_editable (GTK_EDITABLE (mainData . nodeLabelEntry), FALSE);
 gtk_editable_set_editable (GTK_EDITABLE (mainData . treeNavEntry), FALSE);
 gtk_editable_set_editable (GTK_EDITABLE (mainData . recNavEntry), FALSE);

 /* Log flag */
 vBox = gtk_vbox_new (FALSE, 2);
 gtk_notebook_append_page (GTK_NOTEBOOK (notebook), vBox, gtk_label_new_with_mnemonic ("_Log"));
 gtk_widget_show (vBox);

 logTab = gtk_text_view_new ();
 mainData . logBuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (logTab));
 gtk_text_view_set_editable (GTK_TEXT_VIEW (logTab), FALSE);
 gtk_text_buffer_set_text (mainData . logBuffer, VERSION_STRING, -1);
 scrollWin = gtk_scrolled_window_new (GTK_TEXT_VIEW (logTab) -> hadjustment, GTK_TEXT_VIEW (logTab) -> vadjustment);
 gtk_container_add (GTK_CONTAINER (scrollWin), logTab);
 gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrollWin), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
 gtk_widget_show (logTab);
 gtk_box_pack_start (GTK_BOX (vBox), scrollWin, TRUE, TRUE, 0);
 gtk_widget_show (scrollWin);

 gtk_widget_show (mainWindow);
 activeNode = NULL;
 activeTree = NULL;

 gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mainData . menuHotMouse), FALSE);
 gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mainData . menuRecEdit), FALSE);

 progress . numBar1 = 0;
 progress . actBar1 = 0;
 progress . numBar2 = 0;
 progress . actBar2 = 0;
 progress . isActive = FALSE;
}

void OnChangeRec () {
 PHYLONODE* node;
 PHYLONODE* root;
 PHYLOTREE* tree;
 gchar* text;

 if (GETTREES == NULL) {
   gtk_entry_set_text (GTK_ENTRY (mainData . treeLabelEntry), "No tree");
   gtk_entry_set_text (GTK_ENTRY (mainData . nodeLabelEntry), "No tree");
   gtk_entry_set_text (GTK_ENTRY (mainData . treeNavEntry), "No tree");
   gtk_entry_set_text (GTK_ENTRY (mainData . costLabelEntry), "No grid");
   gtk_entry_set_text (GTK_ENTRY (mainData . vicsLabelEntry), "No grid");
   gtk_entry_set_text (GTK_ENTRY (mainData . removsLabelEntry), "No grid");
   gtk_entry_set_text (GTK_ENTRY (mainData . recNavEntry), "No grid");
   return;
 }
 tree = (PHYLOTREE*) activeTree -> data;
 node = (PHYLONODE*) activeNode -> data;
 gtk_entry_set_text (GTK_ENTRY (mainData . treeLabelEntry), tree -> name);
 gtk_entry_set_text (GTK_ENTRY (mainData . nodeLabelEntry), node -> name);
 text = g_strdup_printf ("%ld of %ld", (SINTEGER) g_list_position (GETTREES, tree -> list), GETNUMTREES);
 gtk_entry_set_text (GTK_ENTRY (mainData . treeNavEntry), text);
 g_free (text);
 if (GetGridState () == TRUE) {
   CheckRadioRec ();
   root = (PHYLONODE*) tree -> root -> data;
   text = g_strdup_printf ("%.4f", root -> rec -> cost);
   CheckRealStr (text);
   gtk_entry_set_text (GTK_ENTRY (mainData . costLabelEntry), text);
   g_free (text);
   text = g_strdup_printf ("%ld", root -> rec -> numVics);
   gtk_entry_set_text (GTK_ENTRY (mainData . vicsLabelEntry), text);
   g_free (text);
   text = g_strdup_printf ("%ld", root -> rec -> numRemovNodes);
   gtk_entry_set_text (GTK_ENTRY (mainData . removsLabelEntry), text);
   g_free (text);
   if (root -> tree -> actRec != NULL)
     text = g_strdup_printf ("%ld of %ld", (SINTEGER) g_list_position (root -> tree -> recs, root -> tree -> actRec),
                             root -> tree -> stored);
   else
     text = g_strdup_printf ("%ld", root -> tree -> stored);
   gtk_entry_set_text (GTK_ENTRY (mainData . recNavEntry), text);
   g_free (text);
 }
 else {
   gtk_entry_set_text (GTK_ENTRY (mainData . costLabelEntry), "No grid");
   gtk_entry_set_text (GTK_ENTRY (mainData . vicsLabelEntry), "No grid");
   gtk_entry_set_text (GTK_ENTRY (mainData . removsLabelEntry), "No grid");
   gtk_entry_set_text (GTK_ENTRY (mainData . recNavEntry), "No grid");
 }
}

gint OnDelete (GtkWidget* widget, GdkEvent* event, gpointer data) {
 if ((GETTREES != NULL) && (CloseData () != TRUE)) return TRUE;
 gtk_main_quit ();
 return FALSE;
}

void OnMenuOpen (GtkMenuItem* openMenu, gpointer data) {
 GtkWidget* dialog;
 gchar* fileName;
 gchar* text;
 guint timer;
 PARSINGPROGDLG* progDlg;
 SCOUNTER numOlRecs, numNuRecs, numOlSpp, numNuSpp, numOlGeoSpp, numNuGeoSpp;
 GtkTextIter endBuffer;

 fileName = NULL;
 dialog = gtk_file_chooser_dialog_new ("Open File", GTK_WINDOW (GETMAINWIN), GTK_FILE_CHOOSER_ACTION_OPEN,
                                       GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);
 CreateFileFilter ("Phylogeny in a XML file (*.xml)", "*.xml", dialog);
 CreateFileFilter ("Phylogeny in a KML (Google Earth) file (*.kml)", "*.kml", dialog);
 CreateFileFilter ("All files (*.*)", "*.*", dialog);
 if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
   fileName = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
 gtk_widget_destroy (dialog);
 if (fileName != NULL) {
   progress . isActive = TRUE;
   progDlg = ParsingProgress (&progress);
   timer = g_timeout_add (100, UpdateParsingProgress, progDlg);
   numOlRecs = GETNUMGEOREFS;
   numOlSpp = GETNUMSPP;
   numOlGeoSpp = GetNumTermsWithGeo ();
   if (ReadFile (fileName, &progress) == TRUE) {
     gtk_widget_set_sensitive (mainData . menuFeed, TRUE);
     gtk_widget_set_sensitive (mainData . menuSaveAs, TRUE);
     gtk_widget_set_sensitive (mainData . menuClose, TRUE);
     gtk_widget_set_sensitive (mainData . menuGrid, TRUE);
     gtk_widget_set_sensitive (mainData . menuEdit, TRUE);
     gtk_widget_set_sensitive (mainData . menuHotMouse, TRUE);
     gtk_widget_set_sensitive (mainData . menuGeoBrowser, TRUE);
     OpenNewTree ();
     ChangeGeo ();
     OnChangeRec ();

     text = g_strdup_printf ("%ld", params . maxTime);
     gtk_entry_set_text (GTK_ENTRY (mainData . maxTimeEntry), text);
     g_free (text);
     text = g_strdup_printf ("%ld", params . timeLag);
     gtk_entry_set_text (GTK_ENTRY (mainData . timeLagEntry), text);
     g_free (text);
     gtk_widget_set_sensitive (mainData . buttonPrevNode, TRUE);
     gtk_widget_set_sensitive (mainData . buttonNextNode, TRUE);
     if (GETNUMTREES > 1)
        gtk_widget_set_sensitive (mainData . buttonNextTree, TRUE);
     numNuRecs = GETNUMGEOREFS - numOlRecs;
     numNuSpp = GETNUMSPP - numOlSpp;
     numNuGeoSpp = GetNumTermsWithGeo () - numOlGeoSpp;
     g_string_append_printf (outLog, "\n\nFile %s read\n\n%ld new teminals added (%ld with records)\n%ld new records added",
                             fileName, numNuSpp, numNuGeoSpp, numNuRecs);
     gtk_text_buffer_get_end_iter (mainData . logBuffer, &endBuffer);
     gtk_text_buffer_insert (mainData . logBuffer, &endBuffer, outLog -> str, -1);
     g_string_erase (outLog, 0, -1);
   }
   progress . isActive = FALSE;
 }
 g_free (fileName);
}

void OnMenuFeed (GtkMenuItem* openMenu, gpointer data) {
 GtkWidget* dialog;
 gchar* fileName;
 SCOUNTER numOlRecs, numNuRecs, numOlGeoSpp, numNuGeoSpp;
 SFLAG state;
 guint timer;
 PARSINGPROGDLG* progDlg;
 GtkTextIter endBuffer;

 fileName = NULL;
 dialog = gtk_file_chooser_dialog_new ("Feed records from file", GTK_WINDOW (GETMAINWIN), GTK_FILE_CHOOSER_ACTION_OPEN,
                                       GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);
 CreateFileFilter ("Vip XML file (*.xml)", "*.xml", dialog);
 CreateFileFilter ("Google Earth KML file (*.kml)", "*.kml", dialog);
 CreateFileFilter ("vNDM XYD file (*.xyd)", "*.xyd", dialog);
 CreateFileFilter ("Tab delimited text (*.txt)", "*.txt", dialog);
 CreateFileFilter ("All files (*.*)", "*.*", dialog);
 if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
   fileName = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
 gtk_widget_destroy (dialog);
 if (fileName != NULL) {
   progress . isActive = TRUE;
   progDlg = ParsingProgress (&progress);
   timer = g_timeout_add (100, UpdateParsingProgress, progDlg);
   numOlRecs = GETNUMGEOREFS;
   numOlGeoSpp = GetNumTermsWithGeo ();
   state = FeedFile (fileName, &progress);
   numNuRecs = GETNUMGEOREFS - numOlRecs;
   if (numNuRecs > 0) {
     ChangeGeo ();
     OnChangeRec ();
     numNuGeoSpp = GetNumTermsWithGeo () - numOlGeoSpp;
     g_string_append_printf (outLog, "\n\nFile %s read\n\n%ld new records added (%ld new terms with data)",
                             fileName, numNuRecs, numNuGeoSpp);
     if (state == FALSE) {
       errorMessage = g_strdup_printf ("Unable to read all records from file: %s", fileName);
       OnError ();
     }
   }
   else if (state == FALSE) {
     errorMessage = g_strdup_printf ("Unable to read file: %s", fileName);
     OnError ();
   }
   else
     g_string_append_printf (outLog, "\n\nFile %s read\nFile without records", fileName);
   gtk_text_buffer_get_end_iter (mainData . logBuffer, &endBuffer);
   gtk_text_buffer_insert (mainData . logBuffer, &endBuffer, outLog -> str, -1);
   g_string_erase (outLog, 0, -1);
   progress . isActive = FALSE;
 }
 g_free (fileName);
}

void OnMenuSaveAs (GtkMenuItem* saveAsMenu, gpointer data) {
 GtkWidget* dialog;
 GtkFileFilter* xmlFilter;
 GtkFileFilter* withFilter;
 GtkFileFilter* kmlFilter;
 GtkFileFilter* xydFilter;
 GtkFileFilter* tabFilter;
 GtkFileFilter* lstFilter;
 GtkFileFilter* cllFilter;
 GtkFileFilter* phyFilter;
 GtkFileFilter* claFilter;
 GtkFileFilter* userFilter;
 gchar* fileName;

 fileName = NULL;
 dialog = gtk_file_chooser_dialog_new ("Save File", GTK_WINDOW (GETMAINWIN), GTK_FILE_CHOOSER_ACTION_SAVE,
                                       GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT, NULL);
 gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (dialog), TRUE);
 gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (dialog), "unitled");
 xmlFilter = CreateFileFilter ("Phylogeny in a XML file (*.xml)", "*.xml", dialog);
 kmlFilter = CreateFileFilter ("Google Earth KML file (*.kml)", "*.kml", dialog);
 withFilter = CreateFileFilter ("Only terms and nodes with data (*.xml)", "*.xml", dialog);
 xydFilter = CreateFileFilter ("vNDM file (*.xyd)", "*.xyd", dialog);
 tabFilter = CreateFileFilter ("Tad delimited text file (*.txt)", "*.txt", dialog);
 lstFilter = CreateFileFilter ("Alphabetic term list (*.txt)", "*.txt", dialog);
 cllFilter = CreateFileFilter ("Alphabetic list of collections (*.txt)", "*.txt", dialog);
 claFilter = CreateFileFilter ("Phylogeny (just cladograms) in XML (*,xml)", "*.xml", dialog);
 phyFilter = CreateFileFilter ("Phylogeny in phyloXML (*.xml)", "*.xml", dialog);
 if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT) {
   fileName = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
   userFilter = gtk_file_chooser_get_filter (GTK_FILE_CHOOSER (dialog));
   if (userFilter == xmlFilter) {
     fileName = CheckExtention (fileName, "xml");
     WriteXMLFile (fileName);
   }
   else if (userFilter == kmlFilter) {
     fileName = CheckExtention (fileName, "kml");
     WriteKMLFile (fileName);
   }
   else if (userFilter == xydFilter) {
     fileName = CheckExtention (fileName, "xyd");
     WriteXYDFile (fileName);
   }
   else if (userFilter == tabFilter) {
     fileName = CheckExtention (fileName, "txt");
     WriteTABTable (fileName);
   }
   else if (userFilter == lstFilter) {
     fileName = CheckExtention (fileName, "txt");
     WriteLstFile (fileName);
   }
   else if (userFilter == cllFilter) {
     fileName = CheckExtention (fileName, "txt");
     WriteCllFile (fileName);
   }
   if (userFilter == withFilter) {
     fileName = CheckExtention (fileName, "xml");
     WriteXMLWithData (fileName);
   }
   if (userFilter == phyFilter) {
     fileName = CheckExtention (fileName, "xml");
     WritePhyFile (fileName);
   }
   if (userFilter == claFilter) {
     fileName = CheckExtention (fileName, "xml");
     WriteCladoFile (fileName);
   }
   g_free (fileName);
 }
 gtk_widget_destroy (dialog);
}

void OnMenuOpenRes (GtkMenuItem* openMenu, gpointer data) {
 GtkWidget* dialog;
 gchar* fileName;
 guint timer;
 PARSINGPROGDLG* progDlg;
 GtkTextIter endBuffer;
 GList* navTree;
 PHYLOTREE* onTree;

 fileName = NULL;
 dialog = gtk_file_chooser_dialog_new ("Open File", GTK_WINDOW (GETMAINWIN), GTK_FILE_CHOOSER_ACTION_OPEN,
                                       GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);
 CreateFileFilter ("VIP in a XML file (*.vrx)", "*.vrx", dialog);
 CreateFileFilter ("All files (*.*)", "*.*", dialog);
 if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
   fileName = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
 gtk_widget_destroy (dialog);
 if (fileName != NULL) {
   progress . isActive = TRUE;
   progDlg = ParsingProgress (&progress);
   timer = g_timeout_add (100, UpdateParsingProgress, progDlg);
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (mainData . radioORRec), TRUE);
   if (ReadResults (fileName, &progress) == TRUE) {
     g_string_append_printf (outLog, "\n\nFile %s read", fileName);
     for (navTree = trees; navTree != NULL; navTree = g_list_next (navTree)) {
       onTree = (PHYLOTREE*) navTree -> data;
       onTree -> actRec = onTree -> recs;
       SetRec ((GPtrArray*) onTree -> actRec -> data);
       g_string_append_printf (outLog, "\n%ld reconstructions for tree %ld", onTree -> stored, onTree -> id);
     }
     gtk_text_buffer_get_end_iter (mainData . logBuffer, &endBuffer);
     gtk_text_buffer_insert (mainData . logBuffer, &endBuffer, outLog -> str, -1);
     g_string_erase (outLog, 0, -1);

     gtk_widget_set_sensitive (mainData . radioSearchRec, TRUE);
     gtk_widget_set_sensitive (mainData . radioConsRec, TRUE);
     if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (mainData . radioSearchRec)) == FALSE)
       gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (mainData . radioSearchRec), TRUE);
     else {
       CheckRadioRec ();
       ChangeGeo ();
       OnUpdateTree ();
       OnChangeRec ();
     }
     gtk_widget_set_sensitive (mainData . menuSaveRes, TRUE);
     gtk_widget_set_sensitive (mainData . menuConsensus, TRUE);
     gtk_widget_set_sensitive (mainData . menuFilter, TRUE);
     gtk_widget_set_sensitive (mainData . menuClean, TRUE);
   }
   progress . isActive = FALSE;
 }
 g_free (fileName);
}

void OnMenuSaveRes (GtkMenuItem* saveResMenu, gpointer data) {
 GtkWidget* dialog;
 GtkFileFilter* vrxFilter;
 GtkFileFilter* userFilter;
 gchar* fileName;

 fileName = NULL;
 dialog = gtk_file_chooser_dialog_new ("Save File", GTK_WINDOW (GETMAINWIN), GTK_FILE_CHOOSER_ACTION_SAVE,
                                       GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT, NULL);
 gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (dialog), TRUE);
 gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (dialog), "unitled.vrx");
 vrxFilter = CreateFileFilter ("VIP Results XML file (*.vrx)", "*.vrx", dialog);
 if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT) {
   fileName = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
   userFilter = gtk_file_chooser_get_filter (GTK_FILE_CHOOSER (dialog));
   if (userFilter == vrxFilter) {
     fileName = CheckExtention (fileName, "vrx");
     WriteVRXFile (fileName);
   }
   g_free (fileName);
 }
 gtk_widget_destroy (dialog);
}

void OnMenuClose (GtkMenuItem* closeMenu, gpointer data) {
 if (CloseData () == TRUE) {
   gtk_widget_set_sensitive (mainData . menuFeed, FALSE);
   gtk_widget_set_sensitive (mainData . menuSaveAs, FALSE);
   gtk_widget_set_sensitive (mainData . menuOpenRes, FALSE);
   gtk_widget_set_sensitive (mainData . menuSaveRes, FALSE);
   gtk_widget_set_sensitive (mainData . menuClose, FALSE);
   gtk_widget_set_sensitive (mainData . menuSetFill, TRUE);
   gtk_widget_set_sensitive (mainData . menuCreateGrid, TRUE);
   gtk_widget_set_sensitive (mainData . menuReDoGrid, FALSE);
   gtk_widget_set_sensitive (mainData . menuDestroyGrid, FALSE);
   gtk_widget_set_sensitive (mainData . menuGridTable, FALSE);
   gtk_widget_set_sensitive (mainData . menuGrid, FALSE);
   gtk_widget_set_sensitive (mainData . menuOpen, TRUE);
   gtk_widget_set_sensitive (mainData . menuConsensus, FALSE);
   gtk_widget_set_sensitive (mainData . menuCleanConsensus, FALSE);
   gtk_widget_set_sensitive (mainData . menuFilter, FALSE);
   gtk_widget_set_sensitive (mainData . menuClean, FALSE);
   gtk_widget_set_sensitive (mainData . menuRec, FALSE);

   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (mainData . timeLagBox), FALSE);

   gtk_widget_set_sensitive (mainData . buttonSearch, FALSE);

   gtk_widget_set_sensitive (mainData . buttonPrevRec, FALSE);
   gtk_widget_set_sensitive (mainData . buttonNextRec, FALSE);
   gtk_widget_set_sensitive (mainData . buttonCopy, FALSE);

   gtk_widget_set_sensitive (mainData . radioORRec, FALSE);
   gtk_widget_set_sensitive (mainData . radioSearchRec, FALSE);
   gtk_widget_set_sensitive (mainData . radioConsRec, FALSE);

   gtk_widget_set_sensitive (mainData . buttonPrevTree, FALSE);
   gtk_widget_set_sensitive (mainData . buttonNextTree, FALSE);
   gtk_widget_set_sensitive (mainData . buttonPrevNode, FALSE);
   gtk_widget_set_sensitive (mainData . buttonNextNode, FALSE);

   gtk_widget_set_sensitive (mainData . menuEdit, FALSE);
   gtk_widget_set_sensitive (mainData . menuHotMouse, FALSE);
   gtk_widget_set_sensitive (mainData . menuRecEdit, FALSE);
   gtk_widget_set_sensitive (mainData . menuGeoBrowser, FALSE);
   gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mainData . menuHotMouse), FALSE);
   gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mainData . menuRecEdit), FALSE);

   activeNode = NULL;
   ChangeGeo ();
   OnChangeTree (NULL);
   OnChangeRec ();

   gtk_text_buffer_set_text (mainData . logBuffer, VERSION_STRING, -1);
 }
}

void OnMenuQuit (GtkMenuItem* quitMenu, gpointer data) {
 if ((GETTREES != NULL) && (CloseData () != TRUE)) return;
 gtk_main_quit ();
}

void OnMenuSetGrid (GtkMenuItem* gridMenu, gpointer data) {
 GRIDDIALOG* gridDialog;
 gchar* entry;
 SINTEGER nuFill;
 SREAL nuSize;

 gridDialog = GridDialog (GETGRIDSIZE, GETFILL, !GetGridState ());
 if ((gtk_dialog_run (GTK_DIALOG (gridDialog -> dialog)) == GTK_RESPONSE_ACCEPT) && (GetGridState () == FALSE)) {
   entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (gridDialog -> sizeEntry)));
   SetCommandLine (entry);
   if (GetCmdReal (&nuSize) != FALSE) {
     if (nuSize > 0) {
       SETGRIDSIZE (nuSize);
       ChangeGeo ();
     }
   }
   g_free (entry);
   entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (gridDialog -> fillEntry)));
   SetCommandLine (entry);
   g_free (entry);
   if (GetCmdNumber (&nuFill) != FALSE) {
     if (nuFill >= 0)
       SETFILL (nuFill);
   }
   if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (gridDialog  -> mooreRadio)) == TRUE)
     SETGRIDMODE (MOORE_NEIGHBORS);
   else if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (gridDialog  -> vonneRadio)) == TRUE)
     SETGRIDMODE (VONNEUMANN_NEIGHBORS);
 }
 DestroyGridDialog (gridDialog);
}

void OnMenuSetFill (GtkMenuItem* gridMenu, gpointer data) {
 GList* navTree;
 GNode* navNode;
 PHYLOTREE* tree;
 PHYLONODE* node;
 SCOUNTER i;

 for (navTree = GETTREES; navTree != NULL; navTree = g_list_next (navTree)) {
   tree = (PHYLOTREE*) navTree -> data;
   for (i = 0; i < tree -> nodeArray -> len; ++ i) {
     navNode = (GNode*) g_ptr_array_index (tree -> nodeArray, i);
     node = (PHYLONODE*) navNode -> data;
     node -> fill = 1000 + GETFILL;
   }
 }
}

void OnMenuReDoGrid (GtkMenuItem* gridMenu, gpointer data) {
 GtkWidget* dialog;
 GtkResponseType response;
 GList* navTree;
 PHYLOTREE* tree;

 dialog = gtk_message_dialog_new (GTK_WINDOW (mainWindow), GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO,
          "\nDo you want to recalculate the gird?\n");
 response = gtk_dialog_run (GTK_DIALOG (dialog));
 gtk_widget_destroy (dialog);
 if ((response == GTK_RESPONSE_OK) || (response == GTK_RESPONSE_YES) || (response == GTK_RESPONSE_ACCEPT)) {
   /* Clean the consensus */
   for (navTree = trees; navTree != NULL; navTree = g_list_next (navTree)) {
     tree = (PHYLOTREE*) navTree -> data;
     if (tree -> recCons != NULL) {
       DestroySingleRec (tree -> recCons);
       tree -> recCons = NULL;
     }
   }
   ApplyParams (FALSE);
   CreateGrid ();
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (mainData . radioORRec), TRUE);
   gtk_widget_set_sensitive (mainData . menuCleanConsensus, FALSE);
   ChangeGeo ();
   OnUpdateTree ();
   OnChangeRec ();
 }
}

void OnMenuCreateGrid (GtkMenuItem* gridMenu, gpointer data) {
 GtkWidget* dialog;
 GtkResponseType response;
 GtkTextIter endBuffer;

 dialog = gtk_message_dialog_new (GTK_WINDOW (mainWindow), GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO,
          "\nDo you want to create a grid with %.4lf size and %ld maximum fill?\n", GETGRIDSIZE, GETFILL);
 response = gtk_dialog_run (GTK_DIALOG (dialog));
 gtk_widget_destroy (dialog);
 if ((response == GTK_RESPONSE_OK) || (response == GTK_RESPONSE_YES) || (response == GTK_RESPONSE_ACCEPT)) {
   ApplyParams (TRUE);
   CreateGrid ();
   gtk_widget_set_sensitive (mainData . menuOpen, FALSE);
   gtk_widget_set_sensitive (mainData . menuFeed, FALSE);
   gtk_widget_set_sensitive (mainData . menuOpenRes, TRUE);
   gtk_widget_set_sensitive (mainData . menuCreateGrid, FALSE);
   gtk_widget_set_sensitive (mainData . menuReDoGrid, TRUE);
   gtk_widget_set_sensitive (mainData . menuDestroyGrid, TRUE);
   gtk_widget_set_sensitive (mainData . menuGridTable, TRUE);
   gtk_widget_set_sensitive (mainData . buttonSearch, TRUE);

   gtk_widget_set_sensitive (mainData . menuHotMouse, FALSE);
   gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mainData . menuHotMouse), FALSE);

   gtk_widget_set_sensitive (mainData . menuRec, TRUE);

   gtk_widget_set_sensitive (mainData . radioORRec, TRUE);
   gtk_widget_set_sensitive (mainData . buttonCopy, TRUE);
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (mainData . radioORRec), TRUE);

   ActivateGrid (TRUE);
   ChangeGeo ();
   OnUpdateTree ();
   OnChangeRec ();

   g_string_append_printf (outLog, "\n\n\nGrid created, %.4f size, %ld maximum fill", GETGRIDSIZE, GETFILL);
   if (GETGRIDMODE == MOORE_NEIGHBORS)
     g_string_append (outLog, "\nMoore neighborhood (fill with Chebyshev distance)");
   else if (GETGRIDMODE == VONNEUMANN_NEIGHBORS)
     g_string_append (outLog, "\nVonNeumann neighborhood (fill with Manhattan distance)");
   g_string_append_printf (outLog, "\nOriginal records: %ld\nEffecive cells: %ld", GETNUMGEOREFS, GETNUMCELLS);
   g_string_append_printf (outLog, "\nTerms with records: %ld of %ld", GetNumTermsWithGeo (), GETNUMSPP);
   gtk_text_buffer_get_end_iter (mainData . logBuffer, &endBuffer);
   gtk_text_buffer_insert (mainData . logBuffer, &endBuffer, outLog -> str, -1);
   g_string_erase (outLog, 0, -1);
 }
}

void OnMenuDestroyGrid (GtkMenuItem* gridMenu, gpointer data) {
 GtkWidget* dialog;
 GtkResponseType response;
 GtkTextIter endBuffer;

 dialog = gtk_message_dialog_new (GTK_WINDOW (mainWindow), GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO,
          "\nDo you want to destroy actual grid (Results will be lost!)?\n");
 response = gtk_dialog_run (GTK_DIALOG (dialog));
 gtk_widget_destroy (dialog);
 if ((response == GTK_RESPONSE_OK) || (response == GTK_RESPONSE_YES) || (response == GTK_RESPONSE_ACCEPT)) {
   DestroyGrid ();
   ActivateGrid (FALSE);
   gtk_widget_set_sensitive (mainData . menuOpen, TRUE);
   gtk_widget_set_sensitive (mainData . menuFeed, TRUE);
   gtk_widget_set_sensitive (mainData . menuOpenRes, FALSE);
   gtk_widget_set_sensitive (mainData . menuSaveRes, FALSE);
   gtk_widget_set_sensitive (mainData . menuCreateGrid, TRUE);
   gtk_widget_set_sensitive (mainData . menuReDoGrid, FALSE);
   gtk_widget_set_sensitive (mainData . menuDestroyGrid, FALSE);
   gtk_widget_set_sensitive (mainData . menuGridTable, FALSE);
   gtk_widget_set_sensitive (mainData . buttonSearch, FALSE);

   gtk_widget_set_sensitive (mainData . menuHotMouse, TRUE);
   gtk_widget_set_sensitive (mainData . menuRecEdit, FALSE);
   gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mainData . menuRecEdit), FALSE);

   gtk_widget_set_sensitive (mainData . menuConsensus, FALSE);
   gtk_widget_set_sensitive (mainData . menuFilter, FALSE);
   gtk_widget_set_sensitive (mainData . menuClean, FALSE);
   gtk_widget_set_sensitive (mainData . menuCleanConsensus, FALSE);
   gtk_widget_set_sensitive (mainData . menuRec, FALSE);

   gtk_widget_set_sensitive (mainData . buttonPrevRec, FALSE);
   gtk_widget_set_sensitive (mainData . buttonNextRec, FALSE);
   gtk_widget_set_sensitive (mainData . buttonCopy, FALSE);

   gtk_widget_set_sensitive (mainData . radioORRec, FALSE);
   gtk_widget_set_sensitive (mainData . radioSearchRec, FALSE);
   gtk_widget_set_sensitive (mainData . radioConsRec, FALSE);

   OnUpdateTree ();
   OnChangeRec ();

   g_string_append_printf (outLog, "\n\nGrid destroyed, results destroyed");
   gtk_text_buffer_get_end_iter (mainData . logBuffer, &endBuffer);
   gtk_text_buffer_insert (mainData . logBuffer, &endBuffer, outLog -> str, -1);
   g_string_erase (outLog, 0, -1);
 }
}

void OnMenuGridTable (GtkMenuItem* gridMenu, gpointer data) {
 GtkWidget* dialog;
 GtkFileFilter* txtRichFilter;
 GtkFileFilter* kmlRichFilter;
 GtkFileFilter* txtPDFilter;
 GtkFileFilter* kmlPDFilter;
 GtkFileFilter* ndmDATFilter;
 GtkFileFilter* userFilter;
 gchar* fileName;

 fileName = NULL;
 dialog = gtk_file_chooser_dialog_new ("Save File", GTK_WINDOW (GETMAINWIN), GTK_FILE_CHOOSER_ACTION_SAVE,
                                       GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT, NULL);
 gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (dialog), TRUE);
 gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (dialog), "unitled.kml");
 ndmDATFilter = CreateFileFilter ("NDM matrix (*.dat)", "*.dat", dialog);
 kmlRichFilter = CreateFileFilter ("Species richness on KML (*.kml)", "*.kml", dialog);
 txtRichFilter = CreateFileFilter ("Species richness table (*.txt)", "*.txt", dialog);
 kmlPDFilter = CreateFileFilter ("Phylogenetic diversity on KML (*.kml)", "*.kml", dialog);
 txtPDFilter = CreateFileFilter ("Phylogenetic diversity table (*.txt)", "*.txt", dialog);
 if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT) {
   fileName = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
   userFilter = gtk_file_chooser_get_filter (GTK_FILE_CHOOSER (dialog));
   if (userFilter == ndmDATFilter) {
     fileName = CheckExtention (fileName, "dat");
     WriteNDMMatrix (fileName);
   }
   if (userFilter == kmlRichFilter) {
     fileName = CheckExtention (fileName, "kml");
     WriteRichKMLFile (fileName);
   }
   if (userFilter == txtRichFilter) {
     fileName = CheckExtention (fileName, "txt");
     WriteSpRichness (fileName);
   }
   if (userFilter == txtPDFilter) {
     fileName = CheckExtention (fileName, "txt");
     WritePhyloDiv (fileName);
   }
   if (userFilter == kmlPDFilter) {
     fileName = CheckExtention (fileName, "kml");
     WritePDKMLFile (fileName);
   }
   g_free (fileName);
 }
 gtk_widget_destroy (dialog);
 gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (mainData . radioORRec), TRUE);
 CheckRadioRec ();
 ChangeGeo ();
 OnUpdateTree ();
 OnChangeRec ();
}

void OnMenuPageHeuristic (GtkMenuItem* searchMenu, gpointer data) {
 SEARCHDIALOG* searchDialog;
 gchar* entry;
 SCOUNTER iters;
 SCOUNTER accept;
 SFLAG keep;
 SFLAG page;
 SFLAG cons;

 iters = 0;
 keep = FALSE;
 if (ApplyParams (TRUE) == TRUE)
   MakeOptimization (TRUE);
 searchDialog = SearchDialog ();
 if (gtk_dialog_run (GTK_DIALOG (searchDialog -> dialog)) == GTK_RESPONSE_ACCEPT) {
   entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (searchDialog -> iterEntry)));
   SetCommandLine (entry);
   if (GetCmdNumber (&iters) == FALSE)
     iters = 0;
   entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (searchDialog -> equalEntry)));
   SetCommandLine (entry);
   if (GetCmdNumber (&accept) == FALSE)
     accept = 50;
   keep = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (searchDialog -> keepCheck));
   params . keepAll = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (searchDialog -> allCheck));
   page = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (searchDialog -> pageCheck));
   cons = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (searchDialog -> consCheck));
   if (page == TRUE)
     params . algorithm = PAGE_SEARCH;
   else
     params . algorithm = FAST_SEARCH;
   if (cons == TRUE)
     params . search |= CONSTR_SEARCH;
   else
     params . search &= ~CONSTR_SEARCH;
   if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (searchDialog -> flipCheck)))
     params . algorithm |= FLIP_NODES;
   if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (searchDialog -> startSecRadio)))
     params . algorithm |= FIRST_SECTOR;
   else if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (searchDialog -> secRadio)))
     params . algorithm |= FULL_SECTOR;
   entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (searchDialog -> secSizeEntry)));
   SetCommandLine (entry);
   if (GetCmdNumber (&params . sectSize) == FALSE)
     params . sectSize = 20;
   if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (searchDialog -> blindAnnealRadio)))
     params . algorithm |= BLIND_ANNEAL;
   else if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (searchDialog -> gibbsAnnealRadio)))
     params . algorithm |= GIBBS_ANNEAL;
   else if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (searchDialog -> nixonAnnealRadio)))
     params . algorithm |= NIXON_ANNEAL;
   entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (searchDialog -> generationEntry)));
   SetCommandLine (entry);
   if (GetCmdNumber (&params . generations) == FALSE)
     params . generations = 1000;
   entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (searchDialog -> abandonEntry)));
   SetCommandLine (entry);
   if (GetCmdNumber (&params . abandon) == FALSE)
     params . abandon = 20;
 }
 DestroySearchDialog (searchDialog);
 if (iters <= 0) return;
 params . its = iters;
 params . acceptEqual = 50;
 if ((accept >= 0) && (accept <= 100))
   params . acceptEqual = accept;
 DoSearch (keep);
}

void OnMenuConsensus (GtkMenuItem* consensusMenu, gpointer data) {
 GList* navTree;
 PHYLOTREE* tree;

 RemoveAlloSets ();
 for (navTree = trees; navTree != NULL; navTree = g_list_next (navTree)) {
   tree = (PHYLOTREE*) navTree -> data;
   if (tree -> recCons != NULL) {
     DestroySingleRec (tree -> recCons);
     tree -> recCons = NULL;
   }
 }
 gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (mainData . radioConsRec), TRUE);
 gtk_widget_set_sensitive (mainData . menuCleanConsensus, TRUE);
}

void OnMenuClean (GtkMenuItem* cleanMenu, gpointer data) {
 GList* navTree;
 PHYLOTREE* tree;
 GtkTextIter endBuffer;

 RemoveAlloSets ();
 for (navTree = trees; navTree != NULL; navTree = g_list_next (navTree)) {
   tree = (PHYLOTREE*) navTree -> data;
   if (tree -> recs != NULL) {
     DestroyRecList (tree -> recs);
     tree -> recs = NULL;
   }
   tree -> actRec = NULL;
   tree -> stored = 0;
   if (tree -> recCons != NULL) {
     DestroySingleRec (tree -> recCons);
     tree -> recCons = NULL;
   }
 }
 gtk_widget_set_sensitive (mainData . menuSaveRes, FALSE);
 gtk_widget_set_sensitive (mainData . menuCleanConsensus, FALSE);
 gtk_widget_set_sensitive (mainData . radioSearchRec, FALSE);
 gtk_widget_set_sensitive (mainData . radioConsRec, FALSE);
 gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (mainData . radioORRec), TRUE);
 gtk_widget_set_sensitive (mainData . menuRecEdit, FALSE);
 gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mainData . menuRecEdit), FALSE);

 g_string_append_printf (outLog, "\n\nResults destroyed, only OR reconstruction keeped");
 gtk_text_buffer_get_end_iter (mainData . logBuffer, &endBuffer);
 gtk_text_buffer_insert (mainData . logBuffer, &endBuffer, outLog -> str, -1);
 g_string_erase (outLog, 0, -1);
}

void OnMenuCleanConsensus (GtkMenuItem* cleanMenu, gpointer data) {
 GList* navTree;
 PHYLOTREE* tree;

 for (navTree = trees; navTree != NULL; navTree = g_list_next (navTree)) {
   tree = (PHYLOTREE*) navTree -> data;
   if (tree -> recCons != NULL) {
     DestroySingleRec (tree -> recCons);
     tree -> recCons = NULL;
   }
   SetRec (tree -> recOR);
 }
 gtk_widget_set_sensitive (mainData . menuCleanConsensus, FALSE);
 CheckRadioRec ();
 ChangeGeo ();
 OnUpdateTree ();
 OnChangeRec ();
}

void OnMenuFilter (GtkMenuItem* filterMenu, gpointer data) {
 ENTRYDIALOG* radiusDialog;
 GtkTextIter endBuffer;
 gchar* entry;
 SREAL extra;

 radiusDialog = CounterEntryDialog (GTK_WINDOW (mapWindow), 0.00, "Filter", "Extra cost");
 if (gtk_dialog_run (GTK_DIALOG (radiusDialog -> dialog)) == GTK_RESPONSE_ACCEPT) {
   entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (radiusDialog -> entry)));
   SetCommandLine (entry);
   g_free (entry);
   if (GetCmdReal (&extra) != FALSE) {
     if (extra >= 0) {
       g_string_append_printf (outLog, "\n\nResults filtered");
       FilterRecs (extra);
       gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (mainData . radioORRec), TRUE);
       gtk_text_buffer_get_end_iter (mainData . logBuffer, &endBuffer);
       gtk_text_buffer_insert (mainData . logBuffer, &endBuffer, outLog -> str, -1);
       g_string_erase (outLog, 0, -1);
     }
   }
 }
 DestroyEntryDialog (radiusDialog);
}

void OnMenuHotCheck (GtkMenuItem* editMenu, gpointer data) {
 SetEdition (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mainData . menuHotMouse)));
}

void OnMenuRecEditCheck (GtkMenuItem* editMenu, gpointer data) {
 SetRecEdition (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mainData . menuRecEdit)));
}

void OnMenuGeoBrowser (GtkMenuItem* browserMenu, gpointer data) {
 GEOREFBROWSERDLG* browser;
 PHYLONODE* node;
 GList* geoList;

 node = (PHYLONODE*) activeNode -> data;
 geoList = NULL;
 if (node -> term != NULL)
  geoList = node -> term -> firstGeo;
 if (geoList == NULL)
   geoList = GETGEOREF;
 browser = GeoRefBrowser ((GEOREF*) geoList -> data, gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mainData . menuHotMouse)));
}

void OnMenuSaveLog (GtkMenuItem* saveAsMenu, gpointer data) {
 GtkWidget* dialog;
 GtkFileFilter* txtFilter;
 gchar* fileName;
 gchar* onBuffer;
 GString* toSave;
 GtkTextIter onStart;
 GtkTextIter onEnd;

 fileName = NULL;
 dialog = gtk_file_chooser_dialog_new ("Save File", GTK_WINDOW (GETMAINWIN), GTK_FILE_CHOOSER_ACTION_SAVE,
                                       GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT, NULL);
 gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (dialog), TRUE);
 gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (dialog), "unitled.txt");
 txtFilter = CreateFileFilter ("Text file (*.txt)", "*.txt", dialog);
 if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT) {
   fileName = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
   fileName = CheckExtention (fileName, "txt");
   gtk_text_buffer_get_bounds (mainData . logBuffer, &onStart, &onEnd);
   onBuffer = gtk_text_buffer_get_text (mainData .  logBuffer, &onStart, &onEnd, FALSE);
   toSave = g_string_new (onBuffer);
   WriteOutLog (fileName, toSave);
   g_free (fileName);
   g_free (onBuffer);
   g_string_free (toSave, TRUE);
 }
 gtk_widget_destroy (dialog);
}

void OnMenuClearLog (GtkMenuItem* saveAsMenu, gpointer data) {
 gtk_text_buffer_set_text (mainData . logBuffer, VERSION_STRING, -1);
}

void OnButtonApply (GtkButton* applyButton, gpointer data) {
 if ((ApplyParams (FALSE) == TRUE) && (GetGridState () == TRUE)) {
   MakeOptimization (TRUE);
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (mainData . radioORRec), TRUE);
   ChangeGeo ();
   OnUpdateTree ();
   OnChangeRec ();
 }
}

void OnButtonSearch (GtkButton* searchButton, gpointer data) {
 OnMenuPageHeuristic (NULL, NULL);
}

void OnButtonCancel (GtkButton* cancelButton, gpointer data) {
 gchar* text;

 text = g_strdup_printf ("%.4f", params . costPartRemov);
 CheckRealStr (text);
 gtk_entry_set_text (GTK_ENTRY (mainData . partEntry), text);
 g_free (text);
 text = g_strdup_printf ("%.4f", params . costRemov);
 CheckRealStr (text);
 gtk_entry_set_text (GTK_ENTRY (mainData . removEntry), text);
 g_free (text);
 text = g_strdup_printf ("%.4f", params . overlap);
 CheckRealStr (text);
 gtk_entry_set_text (GTK_ENTRY (mainData . overlapEntry), text);
 g_free (text);
 text = g_strdup_printf ("%.4f", params . costTimeLag);
 CheckRealStr (text);
 gtk_entry_set_text (GTK_ENTRY (mainData . lagCostEntry), text);
 g_free (text);
 text = g_strdup_printf ("%ld", params . maxTime);
 gtk_entry_set_text (GTK_ENTRY (mainData . maxTimeEntry), text);
 g_free (text);
 text = g_strdup_printf ("%ld", params . timeLag);
 gtk_entry_set_text (GTK_ENTRY (mainData . timeLagEntry), text);
 g_free (text);
}

void OnButtonPrevTree (GtkButton* prevButton, gpointer data) {
 GList* nuTree;

 nuTree = g_list_previous (activeTree);
 if (nuTree != NULL) {
   ChangeGeo ();
   OnChangeTree (nuTree);
   OnChangeRec ();
   gtk_widget_set_sensitive (mainData . buttonNextTree, TRUE);
   if (g_list_previous (nuTree) == NULL)
     gtk_widget_set_sensitive (mainData . buttonPrevTree, FALSE);
 }
}

void OnButtonNextTree (GtkButton* nextButton, gpointer data) {
 GList* nuTree;

 nuTree = g_list_next (activeTree);
 if (nuTree != NULL) {
   ChangeGeo ();
   OnChangeTree (nuTree);
   OnChangeRec ();
   gtk_widget_set_sensitive (mainData . buttonPrevTree, TRUE);
   if (g_list_next (nuTree) == NULL)
     gtk_widget_set_sensitive (mainData . buttonNextTree, FALSE);
 }
}

void OnButtonPrevNode (GtkButton* prevButton, gpointer data) {
 activeNode = GetPrevNode (activeNode);
 ChangeGeo ();
 OnUpdateTree ();
 OnChangeRec ();
}

void OnButtonNextNode (GtkButton* nextButton, gpointer data) {
 activeNode = GetNextNode (activeNode);
 ChangeGeo ();
 OnUpdateTree ();
 OnChangeRec ();
}

void OnButtonPrevRec (GtkButton* nextButton, gpointer data) {
 PHYLOTREE* tree;

 tree = (PHYLOTREE*) activeTree -> data;
 tree -> actRec = g_list_previous (tree -> actRec);
 SetRec ((GPtrArray*) tree -> actRec -> data);
 gtk_widget_set_sensitive (mainData . buttonNextRec, TRUE);
 if (g_list_previous (tree -> actRec) == NULL)
   gtk_widget_set_sensitive (mainData . buttonPrevRec, FALSE);
 ChangeGeo ();
 OnUpdateTree ();
 OnChangeRec ();
}

void OnButtonCopy (GtkButton* copyButton, gpointer data) {
 PHYLOTREE* tree;
 GList* nuRec;

 tree = (PHYLOTREE*) activeTree -> data;
 nuRec = CopyActRec (tree);
 tree -> actRec = nuRec;
 SetRec ((GPtrArray*) tree -> actRec -> data);
 gtk_widget_set_sensitive (mainData . radioSearchRec, TRUE);
 if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (mainData . radioSearchRec)) == FALSE)
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (mainData . radioSearchRec), TRUE);
 else {
   CheckRadioRec ();
   ChangeGeo ();
   OnUpdateTree ();
   OnChangeRec ();
 }
 gtk_widget_set_sensitive (mainData . menuSaveRes, TRUE);
 gtk_widget_set_sensitive (mainData . menuFilter, TRUE);
 gtk_widget_set_sensitive (mainData . menuClean, TRUE);
}

void OnButtonNextRec (GtkButton* prevButton, gpointer data) {
 PHYLOTREE* tree;

 tree = (PHYLOTREE*) activeTree -> data;
 tree -> actRec = g_list_next (tree -> actRec);
 SetRec ((GPtrArray*) tree -> actRec -> data);
 gtk_widget_set_sensitive (mainData . buttonPrevRec, TRUE);
 if (g_list_next (tree -> actRec) == NULL)
   gtk_widget_set_sensitive (mainData . buttonNextRec, FALSE);
 ChangeGeo ();
 OnUpdateTree ();
 OnChangeRec ();
 gtk_widget_set_sensitive (mainData . menuSaveRes, TRUE);
}

void OnRadioRec (GtkToggleButton* button, gpointer data) {
 SFLAG change;

 change = CheckRadioRec ();
 if (change == TRUE) {
   ChangeGeo ();
   OnUpdateTree ();
   OnChangeRec ();
 }
}

void OnUseTime (GtkCheckButton* button, gpointer data) {
 if ((gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (mainData . timeLagBox)) == TRUE) && (params . maxTime == 0))
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (mainData . timeLagBox), FALSE);
}

int CloseData () {
 GtkWidget* dialog;
 gint ret;

 dialog = gtk_message_dialog_new (GTK_WINDOW (GETMAINWIN), GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO,
                                  "\nClose actual data?\n");
 ret = gtk_dialog_run (GTK_DIALOG (dialog));
 gtk_widget_destroy (dialog);
 if (ret != GTK_RESPONSE_YES) return FALSE;
 DestroyGrid ();
 DestroyBioGeo ();

 return TRUE;
}

SFLAG ApplyParams (SFLAG toPrint) {
 gchar* entry;
 gchar* text;
 SREAL nu;
 SFLAG change, prevMode;
 SINTEGER nuInt;
 GtkTextIter endBuffer;

 change = FALSE;
 prevMode = params . mode;
 params . mode = USE_INTS;
 params . search = NORMAL_SEARCH;

 entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (mainData . partEntry)));
 SetCommandLine (entry);
 g_free (entry);
 g_string_append_printf (outLog, "\n");
 if (GetCmdReal (&nu) != FALSE) {
   if ((nu >= 0) && (nu != params . costPartRemov)) {
     params . costPartRemov = nu;
     change = TRUE;
  }
 }
 g_string_append_printf (outLog, "\nPartial elimination set to %.4f", params . costPartRemov);
 text = g_strdup_printf ("%.4f", params . costPartRemov);
 CheckRealStr (text);
 gtk_entry_set_text (GTK_ENTRY (mainData . partEntry), text);
 g_free (text);

 entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (mainData . removEntry)));
 SetCommandLine (entry);
 g_free (entry);
 if (GetCmdReal (&nu) != FALSE) {
   if (nu < 1.00000000)
     nu = 0;
   if (nu != params . costRemov) {
     params . costRemov = nu;
     change = TRUE;
   }
 }
 g_string_append_printf (outLog, "\nDistribution eliminiation set to %.4f", params . costRemov);
 text = g_strdup_printf ("%.4f", params . costRemov);
 CheckRealStr (text);
 gtk_entry_set_text (GTK_ENTRY (mainData . removEntry), text);
 g_free (text);

 entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (mainData . overlapEntry)));
 SetCommandLine (entry);
 g_free (entry);
 if (GetCmdReal (&nu) != FALSE) {
   if ((nu >= 0) && (nu < 100) && ((nu / 100) != params . overlap)) {
     params . overlap = nu / 100;
     change = TRUE;
   }
 }
 g_string_append_printf (outLog, "\nOverlap set to %.4f", params . overlap * 100);
 text = g_strdup_printf ("%.4f", params . overlap * 100);
 CheckRealStr (text);
 gtk_entry_set_text (GTK_ENTRY (mainData . overlapEntry), text);
 g_free (text);

 if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (mainData . costFractionBox)) == TRUE) {
   params . mode |= OVER_REAL;
   g_string_append_printf (outLog, "\nUsing fractional cost for overlap");
 }
 if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (mainData . lateOverlapBox)) == TRUE) {
   params . search |= USE_LATEOVER;
   g_string_append_printf (outLog, "\nUsing partial removal");
 }

 if ((gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (mainData . timeLagBox)) == TRUE) && (params . maxTime > 0)) {
   params . mode |= USE_TIME;
   entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (mainData . lagCostEntry)));
   SetCommandLine (entry);
   g_free (entry);
   g_string_append_printf (outLog, "\nUsing time");
   if (GetCmdReal (&nu) != FALSE) {
     if ((nu >= 0) && (nu != params . costTimeLag) && (nu < params . costNoVic)) {
       params . costTimeLag = nu;
       change = TRUE;
     }
   }
   text = g_strdup_printf ("%.4f", params . costTimeLag);
   CheckRealStr (text);
   gtk_entry_set_text (GTK_ENTRY (mainData . lagCostEntry), text);
   g_free (text);

   entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (mainData . maxTimeEntry)));
   SetCommandLine (entry);
   g_free (entry);
   if (GetCmdNumber (&nuInt) != FALSE) {
     if ((nuInt >= 0) && (nuInt != params . maxTime)) {
       params . maxTime = nuInt;
       change = TRUE;
     }
   }
   g_string_append_printf (outLog, "\nMaximum time set to %ld", params . maxTime);
   text = g_strdup_printf ("%ld", params . maxTime);
   gtk_entry_set_text (GTK_ENTRY (mainData . maxTimeEntry), text);
   g_free (text);

   entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (mainData . timeLagEntry)));
   SetCommandLine (entry);
   g_free (entry);
   if (GetCmdNumber (&nuInt) != FALSE) {
     if ((nuInt >= 0) && (nuInt <= params . maxTime) && (nuInt != params . timeLag)) {
       params . timeLag = nuInt;
       change = TRUE;
     }
   }
   if (params . timeLag > params . maxTime) {
     params . timeLag = params . maxTime;
     change = TRUE;
   }
   g_string_append_printf (outLog, "\nTime lag set to %ld", params . timeLag);
   params . timeDiff = params . maxTime - params . timeLag;
   text = g_strdup_printf ("%ld", params . timeLag);
   gtk_entry_set_text (GTK_ENTRY (mainData . timeLagEntry), text);
   g_free (text);
   if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (mainData . timeFractionBox)) == TRUE) {
     params . mode |= TIME_REAL;
     g_string_append_printf (outLog, "\nUsing fractionals cost on time");
   }
 }
 if (params . mode != prevMode)
   change = TRUE;
 if ((change == TRUE) || (toPrint == TRUE)) {
   gtk_text_buffer_get_end_iter (mainData . logBuffer, &endBuffer);
   gtk_text_buffer_insert (mainData . logBuffer, &endBuffer, outLog -> str, -1);
 }
 g_string_erase (outLog, 0, -1);
 return change;
}

SFLAG CheckRadioRec () {
 SFLAG change;
 PHYLOTREE* tree;
 GtkTextIter endBuffer;

 change = TRUE;
 tree = (PHYLOTREE*) activeTree -> data;
 if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (mainData . radioORRec)) == TRUE) {
   tree -> actRec = NULL;
   SetRec (tree -> recOR);
   gtk_widget_set_sensitive (mainData . buttonPrevRec, FALSE);
   gtk_widget_set_sensitive (mainData . buttonNextRec, FALSE);
   change = TRUE;
   gtk_widget_set_sensitive (mainData . menuRecEdit, FALSE);
   gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mainData . menuRecEdit), FALSE);
 }
 if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (mainData . radioConsRec)) == TRUE) {
   if (tree -> recs == NULL) {
     SetRec (tree -> recOR);
     gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (mainData . radioORRec), TRUE);
     return TRUE;
   }
   tree -> actRec = NULL;
   if (tree -> recCons == NULL) {
     MakeCommonBarriers ();
     gtk_text_buffer_get_end_iter (mainData . logBuffer, &endBuffer);
     gtk_text_buffer_insert (mainData . logBuffer, &endBuffer, outLog -> str, -1);
     g_string_erase (outLog, 0, -1);
   }
   SetRec (tree -> recCons);
   gtk_widget_set_sensitive (mainData . buttonPrevRec, FALSE);
   gtk_widget_set_sensitive (mainData . buttonNextRec, FALSE);
   gtk_widget_set_sensitive (mainData . menuRecEdit, FALSE);
   gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mainData . menuRecEdit), FALSE);
   change = TRUE;
   gtk_widget_set_sensitive (mainData . menuCleanConsensus, TRUE);
 }
 if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (mainData . radioSearchRec)) == TRUE) {
   if (tree -> actRec == NULL) {
     if (tree -> recs == NULL) {
       SetRec (tree -> recOR);
       gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (mainData . radioORRec), TRUE);
       return TRUE;
     }
     tree -> actRec = tree -> recs;
     SetRec ((GPtrArray*) tree -> actRec -> data);
     change = TRUE;
   }
   if (g_list_previous (tree -> actRec) != NULL)
     gtk_widget_set_sensitive (mainData . buttonPrevRec, TRUE);
   else
     gtk_widget_set_sensitive (mainData . buttonPrevRec, FALSE);
   if (g_list_next (tree -> actRec) != NULL)
     gtk_widget_set_sensitive (mainData . buttonNextRec, TRUE);
   else
     gtk_widget_set_sensitive (mainData . buttonNextRec, FALSE);
   gtk_widget_set_sensitive (mainData . menuRecEdit, TRUE);
 }
 return change;
}

void DoSearch (SFLAG keep) {
 GList* navTree;
 PHYLOTREE* tree;
 SEARCHPROGDLG* progDlg;
 guint timer;
 GtkTextIter endBuffer;

 /* Clean previous recontructions */
 gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (mainData . radioORRec), TRUE);
 for (navTree = trees; navTree != NULL; navTree = g_list_next (navTree)) {
   tree = (PHYLOTREE*) navTree -> data;
   if (keep == FALSE) {
     if (tree -> recs != NULL) {
       DestroyRecList (tree -> recs);
       tree -> recs = NULL;
     }
     tree -> actRec = NULL;
     tree -> stored = 0;
   }
   if (tree -> recCons != NULL) {
     DestroySingleRec (tree -> recCons);
     tree -> recCons = NULL;
   }
 }
 RemoveAlloSets ();
 progress . numBar1 = GETNUMTREES;
 progress . numBar2 = params . its;
 progress . isActive = TRUE;
 progDlg = SearchProgress (&progress, TRUE);
 timer = g_timeout_add (100, UpdateSearchProgress, progDlg);
 g_string_append (outLog,"\n\n------ > Searching < ------\n\n");
 if (params . algorithm & PAGE_SEARCH)
   g_string_append (outLog,"Using Page\'s heuristics");
 else if (params . algorithm & FAST_SEARCH)
   g_string_append (outLog, "Using simple Hill climbing heuristic");
 if (params . algorithm & FLIP_NODES)
   g_string_append (outLog,"\nFlipping nodes");
 else
   g_string_append (outLog, "\nRemoving nodes");
 if (params . algorithm & FIRST_SECTOR)
   g_string_append_printf (outLog,"\nStarting with a sector of %ld nodes", params . sectSize);
 else if (params . algorithm & FULL_SECTOR)
   g_string_append_printf (outLog,"\nUsing sectors of %ld nodes", params . sectSize);
 if (params . algorithm & BLIND_ANNEAL)
   g_string_append_printf (outLog,"\nUsing bling annelaing (%ld generations)", params . generations);
 else if (params . algorithm & GIBBS_ANNEAL)
   g_string_append_printf (outLog,"\nUsing Gibb\'s annealing (%ld generations)", params . generations);
 else if (params . algorithm & NIXON_ANNEAL)
   g_string_append_printf (outLog,"\nUsing Nixon\'s annealing (%ld generations)", params . generations);
 if (params . search & CONSTR_SEARCH)
   g_string_append_printf (outLog,"\nConstraining to OR reconstruction");
 g_string_append_printf (outLog, "\n%ld iterations per tree\nPercentage of accepting equals: %ld", params . its, params . acceptEqual);
 SetConstrains (params . search & CONSTR_SEARCH);
 SearchMain (&progress);
 SetConstrains (FALSE);
 progress . isActive = FALSE;
 gtk_widget_set_sensitive (mainData . radioSearchRec, TRUE);
 gtk_widget_set_sensitive (mainData . radioConsRec, TRUE);
 if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (mainData . radioSearchRec)) == FALSE)
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (mainData . radioSearchRec), TRUE);
 else {
   CheckRadioRec ();
   ChangeGeo ();
   OnUpdateTree ();
   OnChangeRec ();
 }
 gtk_widget_set_sensitive (mainData . menuSaveRes, TRUE);
 gtk_widget_set_sensitive (mainData . menuConsensus, TRUE);
 gtk_widget_set_sensitive (mainData . menuFilter, TRUE);
 gtk_widget_set_sensitive (mainData . menuClean, TRUE);
 gtk_widget_set_sensitive (mainData . menuCleanConsensus, FALSE);

 gtk_text_buffer_get_end_iter (mainData . logBuffer, &endBuffer);
 gtk_text_buffer_insert (mainData . logBuffer, &endBuffer, outLog -> str, -1);
 g_string_erase (outLog, 0, -1);
}
/* map.c is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */

#include "wins.h"
#include "dialogs.h"
#include "parser.h"
#include "grid.h"
#include "tessellation.h"
#include "out.h"

typedef struct tagMAPWINDOWDATA MAPWINDATA;
struct tagMAPWINDOWDATA {
 GdkPixbuf* mapImg;
 GdkPixmap* pixMap;
 GtkWidget* drawable;

 GdkGC* blackPen;
 GdkGC* greenPen;
 GdkGC* bluePen;
 GdkGC* redPen;
 GdkGC* greyPen;
 GdkGC* whitePen;

 GtkWidget* menuDrawGrid;
 GtkWidget* menuDrawBarrier;
 GtkWidget* menuDrawRecords;
 GtkWidget* menuDrawCells;
 GtkWidget* menuDrawAll;
 GtkWidget* menuDrawRemoved;
 GtkWidget* menuBarrier;
 GtkWidget* menuBarrierVoronoi;
 GtkWidget* menuBarrierDelaunay;
 GtkWidget* menuBarrierMidTri;
 GtkWidget* menuBarrierAncestor;
 GtkWidget* menuBarrierCommon;

 GtkWidget* labelMousePos;

 gint clientX;
 gint clientY;
 SCOUNTER width;
 SCOUNTER height;
 SINTEGER originX;
 SINTEGER originY;
 SINTEGER startX;
 SINTEGER startY;
 SREAL north;
 SREAL south;
 SREAL west;
 SREAL east;
 gint prevX;
 gint prevY;
 gint radius;

 SREAL xScale;
 SREAL yScale;

 GNode* actNode;

 SFLAG mode;
 SFLAG barrierMode;
 SFLAG viewMode;
};

#define DRAW_MAP_MODE 0
#define EDIT_MAP_MODE 1

#define VIEW_RECORDS 0
#define VIEW_CELLS   1

#define DRAW_AS_VORONOI  0
#define DRAW_AS_DELAUNAY 1
#define DRAW_AS_MIDTRI   2

#define MENU_VORONOI  "voronoi_barrier"
#define MENU_DELAUNAY "delaunay_barrier"
#define MENU_MIDTRI "midtri_barrier"


static gint OnDelete (GtkWidget* mapWin, GdkEvent* event, gpointer data);
static gint OnExpose (GtkWidget* drawable, GdkEventExpose* event, gpointer data);
static gint OnConfigure (GtkWidget* drawable, GdkEvent* event, gpointer data);

static gint OnMotion (GtkWidget* drawable, GdkEventMotion* event, gpointer data);
static gint OnMouseButton (GtkWidget* drawable, GdkEventButton* event, gpointer data);

static void OnMenuLoad (GtkMenuItem* loadMenu, gpointer data);
static void OnMenuSave (GtkMenuItem* saveMenu, gpointer data);
static void OnMenuLimits (GtkMenuItem* limitsMenu, gpointer data);
static void OnMenuCentre (GtkMenuItem* centreMenu, gpointer data);
static void OnMenuRadius (GtkMenuItem* radiusMenu, gpointer data);
static void OnMenuViewMode (GtkMenuItem* drawCheckMenu, gpointer data);
static void OnMenuDrawCheck (GtkMenuItem* drawCheckMenu, gpointer data);
static void OnMenuBarrierMode (GtkMenuItem* modeMenu, gpointer data);
static void OnMenuBarrierOptions (GtkMenuItem* barrierMenu, gpointer data);

static void PrepareGeoRefs (SFLAG updateAll);
static void DrawMap ();
static void DrawPixMap (GdkPixmap* pixmap, SCOUNTER clientX, SCOUNTER clientY, SCOUNTER originX, SCOUNTER originY);
static void DrawActRec (GdkPixmap* pixmap, SCOUNTER clientX, SCOUNTER clientY, SCOUNTER originX, SCOUNTER originY,
                        GNode* refNode, GdkGC* pen, BITFIELD* outCells);
static GEOREF* SearchNearestPoint (gint x, gint y);

static void DrawBarrier (GdkPixmap* pixmap, SCOUNTER clientX, SCOUNTER clientY, SCOUNTER originX, SCOUNTER originY,
                         GList* triangles);
static void DrawBarrierFromRec (GdkPixmap* pixmap, SCOUNTER clientX, SCOUNTER clientY, SCOUNTER originX, SCOUNTER originY,
                                RECONS* rec);
static void DrawBarrierFromAlloSet (GdkPixmap* pixmap, SCOUNTER clientX, SCOUNTER clientY, SCOUNTER originX, SCOUNTER originY,
                                    PHYLONODE* node);

static MAPWINDATA mWinData;

void StartMapWindow () {
 GtkWidget* gBox;
 GtkWidget* menuBar;
 GtkWidget* menu;
 GtkAccelGroup* accelGroup;
 GSList* group;

 GdkColor blueColor = {0, 0x0000, 0x0000, 0xFFFF};
 GdkColor greenColor = {0, 0x0000, 0xFFFF, 0x0000};
 GdkColor redColor = {0, 0xFFFF, 0x0000, 0x0000};
 GdkColor greyColor = {0, 0x6000, 0x6000, 0x6000};

 mWinData . pixMap = NULL;
 mWinData . mapImg = NULL;
 mWinData . width = 720;
 mWinData . height = 360;
 mWinData . originX = mWinData . originY = 0;
 mWinData . startX = mWinData . startY = 0;
 mWinData . north = mWinData . west = 0;
 mWinData . south = 180.0000;
 mWinData . east = 360.0000;
 mWinData . radius = 1;

 mWinData . mode = DRAW_MAP_MODE;

 gBox = gtk_vbox_new (FALSE, 0);
 mapWindow = CreateWindow ("VIP: MapWindow", 100, 200, 800, 500, G_CALLBACK (OnDelete), gBox);

 /* --- Set menues --- */
 accelGroup = gtk_accel_group_new ();
 gtk_window_add_accel_group (GTK_WINDOW (mapWindow), accelGroup);
 menuBar = gtk_menu_bar_new ();
 gtk_box_pack_start (GTK_BOX (gBox), menuBar, FALSE, TRUE, 0);
 gtk_widget_show (menuBar);
 /* Map */
 menu = gtk_menu_new ();
 CreateMenuItem (menu, "_Open...", accelGroup, GDK_O, GDK_CONTROL_MASK, G_CALLBACK (OnMenuLoad), NULL);
 CreateMenuSeparator (menu);
 CreateMenuItem (menu, "Map _limits...", accelGroup, GDK_L, GDK_CONTROL_MASK, G_CALLBACK (OnMenuLimits), NULL);
 CreateMenuItem (menu, "_Centre on...", accelGroup, GDK_C, GDK_CONTROL_MASK, G_CALLBACK (OnMenuCentre), NULL);
 CreateMenuSeparator (menu);
 CreateMenuItem (menu, "_Save map...", accelGroup, GDK_S, GDK_CONTROL_MASK | GDK_SHIFT_MASK, G_CALLBACK (OnMenuSave), NULL);
 AddMenuToBar (menu, "_Map", menuBar);
 /* Drawing */
 menu = gtk_menu_new ();
 group = NULL;
 mWinData . menuDrawRecords = CreateMenuItemRadio (menu, "View _records", accelGroup, GDK_R, GDK_CONTROL_MASK,
                                             G_CALLBACK (OnMenuViewMode), NULL, group);
 group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (mWinData . menuDrawRecords));
 gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawRecords), TRUE);
 mWinData . menuDrawCells = CreateMenuItemRadio (menu, "View c_ells", accelGroup, GDK_E, GDK_CONTROL_MASK,
                                             G_CALLBACK (OnMenuViewMode), NULL, group);
 gtk_widget_set_sensitive (mWinData . menuDrawCells, FALSE);
 CreateMenuSeparator (menu);
 CreateMenuItem (menu, "Record radi_us...", accelGroup, GDK_U, GDK_CONTROL_MASK, G_CALLBACK (OnMenuRadius), NULL);
 CreateMenuSeparator (menu);
 mWinData . menuDrawAll = CreateMenuItemCheck (menu, "_All records", accelGroup, GDK_A, GDK_CONTROL_MASK,
                                                    G_CALLBACK (OnMenuDrawCheck), NULL);
 mWinData . menuDrawRemoved = CreateMenuItemCheck (menu, "Re_moved records", accelGroup, GDK_M, GDK_CONTROL_MASK,
                                                    G_CALLBACK (OnMenuDrawCheck), NULL);
 mWinData . menuDrawBarrier = CreateMenuItemCheck (menu, "Barrier", accelGroup, GDK_B, GDK_CONTROL_MASK,
                                                    G_CALLBACK (OnMenuDrawCheck), NULL);
 CreateMenuSeparator (menu);
 mWinData . menuDrawGrid = CreateMenuItemCheck (menu, "_Grid", accelGroup, GDK_G, GDK_CONTROL_MASK,
                                                    G_CALLBACK (OnMenuDrawCheck), NULL);
 gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawGrid), FALSE);
 gtk_widget_set_sensitive (mWinData . menuDrawBarrier, FALSE);
 gtk_widget_set_sensitive (mWinData . menuDrawRemoved, FALSE);
 AddMenuToBar (menu, "_Drawing", menuBar);
 /* Barrier */
 menu = gtk_menu_new ();
 group = NULL;
 mWinData . menuBarrierVoronoi = CreateMenuItemRadio (menu, "_Voronoi lines", accelGroup, GDK_V, GDK_CONTROL_MASK,
                                             G_CALLBACK (OnMenuBarrierMode), MENU_VORONOI, group);
 group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (mWinData . menuBarrierVoronoi));
 gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mWinData . menuBarrierVoronoi), TRUE);
 mWinData . menuBarrierDelaunay = CreateMenuItemRadio (menu, "_Delaunay triangles", accelGroup, GDK_D, GDK_CONTROL_MASK,
                                             G_CALLBACK (OnMenuBarrierMode), MENU_DELAUNAY, group);
 group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (mWinData . menuBarrierDelaunay));
 mWinData . menuBarrierMidTri = CreateMenuItemRadio (menu, "Mid tr_iangles", accelGroup, GDK_I, GDK_CONTROL_MASK,
                                             G_CALLBACK (OnMenuBarrierMode), MENU_MIDTRI, group);
 CreateMenuSeparator (menu);
 mWinData . menuBarrierAncestor = CreateMenuItemCheck (menu, "Barriers on a_ncestors", accelGroup, GDK_N, GDK_CONTROL_MASK,
                                                    G_CALLBACK (OnMenuBarrierOptions), NULL);
 mWinData . menuBarrierCommon = CreateMenuItemCheck (menu, "Common barr_iers (on consensus)", accelGroup, GDK_I, GDK_CONTROL_MASK,
                                                    G_CALLBACK (OnMenuBarrierOptions), NULL);
 mWinData . menuBarrier = AddMenuToBar (menu, "_Barriers", menuBar);
 gtk_widget_set_sensitive (mWinData . menuBarrier, FALSE);

 /* --- Drawing area --- */
 mWinData . drawable = gtk_drawing_area_new ();
 gtk_box_pack_start (GTK_BOX (gBox), mWinData . drawable, TRUE, TRUE, 0);
 gtk_widget_show (mWinData . drawable);
 g_signal_connect (G_OBJECT (mWinData . drawable), "expose_event", G_CALLBACK (OnExpose), NULL);
 g_signal_connect (G_OBJECT (mWinData . drawable), "configure_event", G_CALLBACK (OnConfigure), NULL);
 gtk_widget_add_events (mWinData . drawable, GDK_POINTER_MOTION_MASK | GDK_BUTTON_PRESS_MASK);
 g_signal_connect (G_OBJECT (mWinData . drawable), "motion_notify_event", G_CALLBACK (OnMotion), NULL);
 g_signal_connect (G_OBJECT (mWinData . drawable), "button_press_event", G_CALLBACK (OnMouseButton), NULL);

 mWinData . labelMousePos = gtk_label_new ("mouse off map");
 gtk_label_set_justify (GTK_LABEL (mWinData . labelMousePos), GTK_JUSTIFY_RIGHT);
 gtk_box_pack_start (GTK_BOX (gBox), mWinData . labelMousePos, FALSE, TRUE, 0);
 gtk_widget_show (mWinData . labelMousePos);

 gtk_widget_show (mapWindow);

 mWinData . blackPen = gdk_gc_new (mWinData . drawable -> window);
 gdk_gc_copy (mWinData . blackPen, mWinData . drawable -> style -> black_gc);

 mWinData . greenPen = gdk_gc_new (mWinData . drawable -> window);
 gdk_gc_copy (mWinData . greenPen, mWinData . drawable -> style -> black_gc);
 gdk_gc_set_rgb_fg_color (mWinData . greenPen, &greenColor);

 mWinData . bluePen = gdk_gc_new (mWinData . drawable -> window);
 gdk_gc_copy (mWinData . bluePen, mWinData . drawable -> style -> black_gc);
 gdk_gc_set_rgb_fg_color (mWinData . bluePen, &blueColor);

 mWinData . redPen = gdk_gc_new (mWinData . drawable -> window);
 gdk_gc_copy (mWinData . redPen, mWinData . drawable -> style -> black_gc);
 gdk_gc_set_rgb_fg_color (mWinData . redPen, &redColor);

 mWinData . greyPen = gdk_gc_new (mWinData . drawable -> window);
 gdk_gc_copy (mWinData . greyPen, mWinData . drawable -> style -> black_gc);
 gdk_gc_set_rgb_fg_color (mWinData . greyPen, &greyColor);

 mWinData . whitePen = gdk_gc_new (mWinData . drawable -> window);
 gdk_gc_copy (mWinData . whitePen, mWinData . drawable -> style -> white_gc);

 gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawAll), TRUE);
 gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawRemoved), FALSE);
 gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawRecords), TRUE);

 mWinData . actNode = NULL;
 mWinData . barrierMode = DRAW_AS_VORONOI;
 mWinData . viewMode = VIEW_RECORDS;
}

void ChangeGeo () {
 PrepareGeoRefs (FALSE);
 if (activeNode != mWinData . actNode)
   mWinData . actNode = activeNode;
 DrawMap ();
}

void ActivateGrid (SFLAG mode) {
 gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawGrid), mode);
 gtk_widget_set_sensitive (mWinData . menuDrawBarrier, mode);
 gtk_widget_set_sensitive (mWinData . menuDrawRemoved, mode);
 gtk_widget_set_sensitive (mWinData . menuDrawCells, mode);
 if (mode == FALSE) {
   mWinData . viewMode = VIEW_RECORDS;
   gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawRecords), TRUE);
 }
 DrawMap ();
}

void SetEdition (SFLAG mode) {
 if (mode == TRUE)
   mWinData . mode = EDIT_MAP_MODE;
 else
   mWinData . mode = DRAW_MAP_MODE;
}

void PaintNewGeoRef (GEOREF* geo) {
 geo -> actX = mWinData . xScale * geo -> azimuth;
 geo -> actY = mWinData . yScale * geo -> inclination;
 DrawMap ();
}

gint OnDelete (GtkWidget* treeWin, GdkEvent* event, gpointer data) {
 gtk_window_iconify (GTK_WINDOW (mapWindow));
 return TRUE;
}

gint OnExpose (GtkWidget* drawable, GdkEventExpose* event, gpointer data) {
 gdk_draw_pixmap (mWinData . drawable -> window, mWinData . drawable -> style -> fg_gc [GTK_WIDGET_STATE (mWinData . drawable)],
     mWinData . pixMap, event -> area . x, event -> area . y, event -> area . x, event -> area . y,
     event -> area . width, event -> area . height);
 return FALSE;
}

gint OnConfigure (GtkWidget* drawable, GdkEvent* event, gpointer data) {
 mWinData . clientX = mWinData . drawable -> allocation . width;
 mWinData . clientY = mWinData . drawable -> allocation . height;
 if (mWinData . pixMap)
   g_object_unref (mWinData . pixMap);
 mWinData . pixMap = gdk_pixmap_new (mWinData . drawable -> window, mWinData . clientX, mWinData . clientY, -1);
 DrawMap ();
 return TRUE;
}

gint OnMotion (GtkWidget* drawable, GdkEventMotion* event, gpointer data) {
 SINTEGER nu;
 SREAL nuLon, nuLat;
 gchar* text;

 if (event -> state & GDK_BUTTON1_MASK) {
   if ((mWinData . prevX != 0) || (mWinData . prevY != 0)) {
     nu = event -> x - mWinData . prevX;
     mWinData . originX += nu;
     if (mWinData . originX < - (mWinData . width - 10))
       mWinData . originX = - (mWinData . width - 10);
     else if (mWinData . originX > (mWinData . clientX - 10))
       mWinData . originX = mWinData . clientX - 10;
     nu = event -> y - mWinData . prevY;
     mWinData . originY += nu;
     if (mWinData . originY < - (mWinData . height - 10))
       mWinData . originY = - (mWinData . height - 10);
     else if (mWinData . originY > (mWinData . clientY - 10))
       mWinData . originY = mWinData . clientY - 10;
     DrawMap ();
   }
   mWinData . prevX = event -> x;
   mWinData . prevY = event -> y;
 }
 else {
   mWinData . prevX = 0;
   mWinData . prevY = 0;
   nuLon = (SREAL) (event -> x - mWinData . originX) / mWinData . xScale;
   nuLat = (SREAL) (event -> y - mWinData . originY) / mWinData . yScale;
   if ((nuLon < 0) || (nuLon > 360) || (nuLat < 0) || (nuLat > 180))
     gtk_label_set_text (GTK_LABEL (mWinData . labelMousePos), "mouse off map");
   else {
     nuLat = FROMINCTOLAT (nuLat);
     nuLon = FROMAZMTOLON (nuLon);
     text = g_strdup_printf ("lat: %.4f\tlon: %.4f", nuLat, nuLon);
     gtk_label_set_text (GTK_LABEL (mWinData . labelMousePos), text);
     g_free (text);
   }
 }
 return TRUE;
}

gint OnMouseButton (GtkWidget* drawable, GdkEventButton* event, gpointer data) {
 GList* navGeo;
 GEOREF* actGeo;
 TERMINAL* term;
 PHYLONODE* actNode;
 GEOREFBROWSERDLG* browser;
 ENTRYDIALOG* radiusDialog;
 SREAL nuLon, nuLat;
 gchar* entry;
 SCOUNTER nuRadius;
 SFLAG paint;

 if (activeNode == NULL) return FALSE;
 if ((event -> type == GDK_BUTTON_PRESS) && (event -> button == 1) &&
     (!(event -> state & GDK_SHIFT_MASK)) && (!(event -> state & GDK_CONTROL_MASK))) {
   actGeo = SearchNearestPoint (event -> x - mWinData . originX, event -> y - mWinData . originY);
   if (actGeo != NULL)
     browser = GeoRefBrowser (actGeo, FALSE);
 }
 else if ((event -> type == GDK_BUTTON_PRESS) && (event -> button == 3) &&
          (event -> state & GDK_CONTROL_MASK) && (!(event -> state & GDK_SHIFT_MASK))) {
   actGeo = SearchNearestPoint (event -> x - mWinData . originX, event -> y - mWinData . originY);
   if (actGeo != NULL) {
     term = (TERMINAL*) actGeo -> sp -> data;
     if ((mWinData . mode == EDIT_MAP_MODE) && (term -> node == activeNode))
       browser = GeoRefBrowser (actGeo, TRUE);
     else
       browser = GeoRefBrowser (actGeo, FALSE);
   }
 }
 else if ((mWinData . mode == EDIT_MAP_MODE) && (event -> button == 3) &&
          (event -> state & GDK_SHIFT_MASK) && (!(event -> state & GDK_CONTROL_MASK)) && (g_node_first_child (activeNode) == NULL)) {
   nuLon = (SREAL) (event -> x - mWinData . originX) / mWinData . xScale;
   nuLat = (SREAL) (event -> y - mWinData . originY) / mWinData . yScale;
   nuLat = FROMINCTOLAT (nuLat);
   nuLon = FROMAZMTOLON (nuLon);
   if ((nuLon > -180.00001) && (nuLon < 180.00001) && (nuLat > -90.00001) && (nuLat < 90.00001)) {
     actGeo = AddGeoRef (nuLon, nuLat, activeNode);
     PaintNewGeoRef (actGeo);
   }
 }
 else if ((mWinData . mode == EDIT_MAP_MODE) && (event -> button == 1) &&
          (event -> state & GDK_SHIFT_MASK) && (event -> state & GDK_CONTROL_MASK) && (g_node_first_child (activeNode) == NULL)) {
   radiusDialog = CounterEntryDialog (GTK_WINDOW (mapWindow), GETGRIDSIZE, "Delete around", "Radius");
   if (gtk_dialog_run (GTK_DIALOG (radiusDialog -> dialog)) == GTK_RESPONSE_ACCEPT) {
     entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (radiusDialog -> entry)));
     SetCommandLine (entry);
     g_free (entry);
     if (GetCmdNumber (&nuRadius) != FALSE) {
       actNode = (PHYLONODE*) activeNode -> data;
       if (actNode -> term == NULL) return FALSE;
       term = actNode -> term;
       nuLon = (SREAL) (event -> x - mWinData . originX) / mWinData . xScale;
       nuLat = (SREAL) (event -> y - mWinData . originY) / mWinData . yScale;
       navGeo = term -> firstGeo;
       paint = FALSE;
       while (navGeo != NULL) {
         actGeo = (GEOREF*) navGeo -> data;
         navGeo = g_list_next (navGeo);
         if (term != (TERMINAL*) actGeo -> sp -> data) break;
         if ((actGeo -> inclination > (nuLat - nuRadius)) && (actGeo -> inclination < (nuLat + nuRadius)) &&
             (actGeo -> azimuth > (nuLon - nuRadius)) && (actGeo -> azimuth < (nuLon + nuRadius))) {
           paint = TRUE;
           DeleteGeoRef (actGeo);
         }
       }
       if (paint == TRUE)
         DrawMap ();
     }
   }
   DestroyEntryDialog (radiusDialog);
 }
 return FALSE;
}

void OnMenuLoad (GtkMenuItem* openMenu, gpointer data) {
 GtkWidget* dialog;
 GtkFileFilter* filter;
 GdkPixbuf* nuMap;
 gchar* fileName;

 fileName = NULL;
 dialog = gtk_file_chooser_dialog_new ("Open Map", GTK_WINDOW (GETMAINWIN), GTK_FILE_CHOOSER_ACTION_OPEN,
                                       GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);
 filter = gtk_file_filter_new ();
 gtk_file_filter_add_pixbuf_formats (filter);
 gtk_file_chooser_set_filter (GTK_FILE_CHOOSER (dialog), filter);
 if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
   fileName = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
 gtk_widget_destroy (dialog);
 if (fileName != NULL) {
   nuMap = gdk_pixbuf_new_from_file (fileName, NULL);
   if (nuMap != NULL) {
     if (mWinData . mapImg != NULL)
       g_object_unref (mWinData . mapImg);
     mWinData . mapImg = nuMap;
     mWinData . width = gdk_pixbuf_get_width (nuMap);
     mWinData . height = gdk_pixbuf_get_height (nuMap);
     mWinData . originX = mWinData . originY = 0;
     mWinData . startX = mWinData . startY = 0;
     mWinData . north = mWinData . west = 0;
     mWinData . south = 180.0000;
     mWinData . east = 360.0000;
     PrepareGeoRefs (TRUE);
     DrawMap ();
   }
 }
 g_free (fileName);
}

void OnMenuSave (GtkMenuItem* saveMenu, gpointer data) {
 GtkWidget* dialog;
 GtkFileFilter* fullJpgFilter;
 GtkFileFilter* actualJpgFilter;
 GtkFileFilter* actualKMLFilter;
 GtkFileFilter* treeKMLFilter;
 GtkFileFilter* userFilter;
 GdkPixmap* tmpPix;
 GdkPixbuf* tmpBuf;
 gchar* fileName;
 SFLAG modeRecords;
 PHYLONODE* node;

 fileName = NULL;
 dialog = gtk_file_chooser_dialog_new ("Save File", GTK_WINDOW (GETMAINWIN), GTK_FILE_CHOOSER_ACTION_SAVE,
                                       GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT, NULL);
 gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (dialog), TRUE);
 gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (dialog), "unitled");
 fullJpgFilter = CreateFileFilter ("Image of full map (*.jpg)", "*.jpg", dialog);
 actualJpgFilter = CreateFileFilter ("Image of actual screen (*.jpg)", "*.jpg", dialog);
 if (activeNode != NULL)
   actualKMLFilter = CreateFileFilter ("KML file of actual node reconstruction (*.kml)", "*.kml", dialog);
 else
   actualKMLFilter = NULL;
 if (GetGridState () == TRUE)
   treeKMLFilter = CreateFileFilter ("KML of vicariant nodes of actual tree reconstruction (*.kml)", "*.kml", dialog);
 else
   treeKMLFilter = NULL;
 if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT) {
   fileName = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
   userFilter = gtk_file_chooser_get_filter (GTK_FILE_CHOOSER (dialog));
   if (userFilter == fullJpgFilter) {
     fileName = CheckExtention (fileName, "jpg");
     tmpPix = gdk_pixmap_new (mWinData . drawable -> window, mWinData . width, mWinData . height, -1);
     DrawPixMap (tmpPix, mWinData . width, mWinData . height, 0, 0);
     tmpBuf = gdk_pixbuf_get_from_drawable (NULL, tmpPix, NULL, 0, 0, 0, 0, mWinData . width, mWinData . height);
     gdk_pixbuf_save (tmpBuf, fileName, "jpeg", NULL, "quality", "100", NULL);
     g_object_unref (tmpPix);
     g_object_unref (tmpBuf);
   }
   else if (userFilter == actualJpgFilter) {
     fileName = CheckExtention (fileName, "jpg");
     tmpPix = gdk_pixmap_new (mWinData . drawable -> window, mWinData . clientX, mWinData . clientY, -1);
     DrawPixMap (tmpPix, mWinData . clientX, mWinData . clientY, mWinData . originX, mWinData . originY);
     tmpBuf = gdk_pixbuf_get_from_drawable (NULL, tmpPix, NULL, 0, 0, 0, 0, mWinData . clientX, mWinData . clientY);
     gdk_pixbuf_save (tmpBuf, fileName, "jpeg", NULL, "quality", "100", NULL);
     g_object_unref (tmpPix);
     g_object_unref (tmpBuf);
   }
   else if ((userFilter == actualKMLFilter) && (actualKMLFilter != NULL)) {
     fileName = CheckExtention (fileName, "kml");
     modeRecords = (mWinData . viewMode == VIEW_RECORDS);
     WriteKMLRecons (fileName, activeNode, modeRecords, gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawRemoved)));
   }
   else if ((userFilter == treeKMLFilter) && (treeKMLFilter != NULL)) {
     fileName = CheckExtention (fileName, "kml");
     node = (PHYLONODE*) activeNode -> data;
     modeRecords = (mWinData . viewMode == VIEW_RECORDS);
     WriteKMLTree (fileName, node -> tree, modeRecords,
                               gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawRemoved)));
   }
   g_free (fileName);
 }
 gtk_widget_destroy (dialog);
}

void OnMenuLimits (GtkMenuItem* limitsMenu, gpointer data) {
 MAPLIMITSDIALOG* mapLimitsDlg;
 gchar* entry;
 SREAL tmp, nuN, nuS, nuW, nuE;

 mapLimitsDlg = MapLimitsDialog (FROMINCTOLAT (mWinData . north), FROMINCTOLAT (mWinData . south), FROMAZMTOLON (mWinData . east),
                                 FROMAZMTOLON (mWinData . west));
 nuN = mWinData . north;
 nuS = mWinData . south;
 nuE = mWinData . east;
 nuW = mWinData . west;
 if (gtk_dialog_run (GTK_DIALOG (mapLimitsDlg -> dialog)) == GTK_RESPONSE_ACCEPT) {
   entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (mapLimitsDlg -> northEntry)));
   SetCommandLine (entry);
   g_free (entry);
   if (GetCmdReal (&tmp) != FALSE) {
     if ((tmp > -90.0000) && (tmp < 90.0000))
       nuN = FROMLATTOINC (tmp);
     else
       nuN = mWinData . north;
   }
   entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (mapLimitsDlg -> southEntry)));
   SetCommandLine (entry);
   g_free (entry);
   if (GetCmdReal (&tmp) != FALSE) {
     if ((tmp > -90.0000) && (tmp < 90.0000))
       nuS = FROMLATTOINC (tmp);
     else
       nuS = mWinData . south;
   }
   if (nuS < nuN) {
     nuN = mWinData . north;
     nuS = mWinData . south;
   }
   entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (mapLimitsDlg -> westEntry)));
   SetCommandLine (entry);
   g_free (entry);
   if (GetCmdReal (&tmp) != FALSE) {
     if ((tmp > -180.0000) && (tmp < 180.0000))
       nuW = FROMLONTOAZM (tmp);
     else
       nuW = mWinData . west;
   }
   entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (mapLimitsDlg -> eastEntry)));
   SetCommandLine (entry);
   g_free (entry);
   if (GetCmdReal (&tmp) != FALSE) {
     if ((tmp > -180.0000) && (tmp < 180.0000))
       nuE = FROMLONTOAZM (tmp);
     else
       nuE = mWinData . east;
   }
   if (nuE < nuW) {
     nuW = mWinData . west;
     nuE = mWinData . east;
   }
 }
 DestroyMapLimitsDialog (mapLimitsDlg);
 if ((nuN != mWinData . north) || (nuS != mWinData . south) || (nuE != mWinData . east) || (nuW != mWinData . west)) {
  mWinData . north = nuN;
  mWinData . south = nuS;
  mWinData . east = nuE;
  mWinData . west = nuW;
  PrepareGeoRefs (TRUE);
  mWinData . startX = mWinData . xScale * mWinData . west;
  mWinData . startY = mWinData . yScale * mWinData . north;
  DrawMap ();
 }
}

void OnMenuCentre (GtkMenuItem* centreMenu, gpointer data) {
 CENTREDIALOG* centreDialog;
 gchar* entry;
 SREAL nuX, nuY, cLat, cLon;
 SFLAG change;

 centreDialog = CentreDialog ();
 change = FALSE;
 if (gtk_dialog_run (GTK_DIALOG (centreDialog -> dialog)) == GTK_RESPONSE_ACCEPT) {
   entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (centreDialog -> latEntry)));
   SetCommandLine (entry);
   g_free (entry);
   if (GetCmdReal (&cLat) != FALSE) {
     if ((cLat > -90.0000) && (cLat < 90.0000)) {
       cLat = FROMLATTOINC (cLat);
       change = TRUE;
     }
   }
   entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (centreDialog -> lonEntry)));
   SetCommandLine (entry);
   g_free (entry);
   if ((GetCmdReal (&cLon) != FALSE) && (change == TRUE)) {
     if ((cLon > -180.0000) && (cLon < 180.0000))
       cLon = FROMLONTOAZM (cLon);
     else
       change = FALSE;
   }
   else
     change = FALSE;
 }
 DestroyCentreDialog (centreDialog);
 if (change == TRUE) {
   nuX = mWinData . xScale * cLon;
   nuY = mWinData . yScale * cLat;
   mWinData . originX = (mWinData . clientX / 2) - nuX;
   mWinData . originY = (mWinData . clientY / 2) - nuY;
   DrawMap ();
 }
}

void OnMenuRadius (GtkMenuItem* radiusMenu, gpointer data) {
 ENTRYDIALOG* radiusDialog;
 gchar* entry;
 SCOUNTER nuRadius;
 SFLAG paint;

 radiusDialog = CounterEntryDialog (GTK_WINDOW (mapWindow), mWinData . radius, "Set drawing radius", "Radius");
 paint = FALSE;
 if (gtk_dialog_run (GTK_DIALOG (radiusDialog -> dialog)) == GTK_RESPONSE_ACCEPT) {
   entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (radiusDialog -> entry)));
   SetCommandLine (entry);
   g_free (entry);
   if (GetCmdNumber (&nuRadius) != FALSE) {
     if ((nuRadius >= 0) && (nuRadius != mWinData . radius)) {
       mWinData . radius = nuRadius;
       paint = TRUE;
     }
   }
 }
 DestroyEntryDialog (radiusDialog);
 if (paint == TRUE)
   DrawMap ();
}

void OnMenuViewMode (GtkMenuItem* modeMenu, gpointer data) {
 SFLAG prevMode;

 prevMode = mWinData . viewMode;
 if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawRecords)) == TRUE)
   mWinData . viewMode = VIEW_RECORDS;
 if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawCells)) == TRUE)
   mWinData . viewMode = VIEW_CELLS;
 if (prevMode != mWinData . viewMode)
   DrawMap ();
}

void OnMenuDrawCheck (GtkMenuItem* drawCheckMenu, gpointer data) {
 if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawBarrier)) == TRUE)
   gtk_widget_set_sensitive (mWinData . menuBarrier, TRUE);
 else
   gtk_widget_set_sensitive (mWinData . menuBarrier, FALSE);
 DrawMap ();
}

void OnMenuBarrierMode (GtkMenuItem* modeMenu, gpointer data) {
 SFLAG prevMode;

 prevMode = mWinData . barrierMode;
 if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuBarrierVoronoi)) == TRUE)
   mWinData . barrierMode = DRAW_AS_VORONOI;
 else if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuBarrierDelaunay)) == TRUE)
   mWinData . barrierMode = DRAW_AS_DELAUNAY;
 else if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuBarrierMidTri)) == TRUE)
   mWinData . barrierMode = DRAW_AS_MIDTRI;
 if (prevMode != mWinData . barrierMode)
   DrawMap ();
}

void OnMenuBarrierOptions (GtkMenuItem* barrierMenu, gpointer data) {
 DrawMap ();
}

void PrepareGeoRefs (SFLAG updateAll) {
 GList* navGeo;
 GEOREF* ref;

 mWinData . xScale = (SREAL) mWinData . width / (mWinData . east - mWinData . west);
 mWinData . yScale = (SREAL) mWinData . height / (mWinData . south - mWinData . north);
 for (navGeo = GETLASTGEOREF; navGeo != NULL; navGeo = g_list_previous (navGeo)) {
   ref = (GEOREF*) navGeo -> data;
   if ((updateAll == FALSE) && (ref -> actX != -1)) continue;
   ref -> actX = mWinData . xScale * ref -> azimuth;
   ref -> actY = mWinData . yScale * ref -> inclination;
 }
}

void DrawMap () {
 GdkRectangle updateRec;

 if (mWinData . pixMap == NULL)
   mWinData . pixMap = gdk_pixmap_new (mWinData . drawable -> window, mWinData . clientX, mWinData . clientY, -1);
 DrawPixMap (mWinData . pixMap, mWinData . clientX, mWinData . clientY, mWinData . originX, mWinData . originY);
 /* Invalidate the window */
 updateRec . x = 0;
 updateRec . y = 0;
 updateRec . width = mWinData . clientX;
 updateRec . height = mWinData . clientY;
 gdk_window_invalidate_rect (mWinData . drawable -> window, &updateRec, TRUE);
}

void DrawPixMap (GdkPixmap* pixmap, SCOUNTER clientX, SCOUNTER clientY, SCOUNTER originX, SCOUNTER originY) {
 GList* navGeo;
 GEOREF* ref;
 PHYLONODE* node;
 GNode* navAnc;
 PHYLONODE* ancNode;
 SINTEGER x, y;
 SCOUNTER r, c, sX, sY, eX, eY;
 GdkGCValues gcVals;
 SREAL tmpNum;
 GdkGC* pen;

 gdk_draw_rectangle (pixmap, mWinData . drawable -> style -> white_gc, TRUE, 0, 0, clientX, clientY);
 if (mWinData . mapImg != NULL) {
   gdk_draw_pixbuf (pixmap, mWinData . drawable -> style -> fg_gc [GTK_WIDGET_STATE (mWinData . drawable)], mWinData . mapImg,
                    0, 0, originX + mWinData . startX, originY + mWinData . startY,
                    -1, -1, GDK_RGB_DITHER_NONE, 0, 0);
 }
 /* Draw the grid */
 if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawGrid)) == TRUE) {
   sX = originX;
   eX = (360 * mWinData . xScale) + originX;
   if (((sX >= 0) && (sX <= clientX)) || ((eX >= 0) && (eX <= clientX)) || ((sX <= 0) && (eX >= clientX))) {
     if (sX < 0)
       sX = 0;
     if (eX > clientX)
       eX = clientX;
     for (r = 0; ; ++ r) {
       tmpNum = GETGRIDSIZE * r;
       if (tmpNum > 180) break;
       sY = (tmpNum * mWinData . yScale) + originY;
       if (sY < 0) continue;
       if (sY > clientY) break;
       gdk_draw_line (pixmap, mWinData . greyPen, sX, sY, eX, sY);
     }
   }
   sY = mWinData . originY;
   eY = (180 * mWinData . yScale) + originY;
   if (((sY >= 0) && (sY <= clientY)) || ((eY >= 0) && (eY <= clientY)) || ((sY <= 0) && (eY >= clientY))) {
     if (sY < 0)
       sY = 0;
     if (eY > clientY)
       eY = clientY;
     for (c = 0; ; ++ c) {
       tmpNum = GETGRIDSIZE * c;
       if (tmpNum > 360) break;
       sX = (tmpNum * mWinData . xScale) + originX;
       if (sX < 0) continue;
       if (sX > clientX) break;
       gdk_draw_line (pixmap, mWinData . greyPen, sX, sY, sX, eY);
     }
   }
 }
 /* First draw all records */
 if ((gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawAll)) == TRUE) &&
     (mWinData . viewMode == VIEW_RECORDS)) {
   for (navGeo = GETGEOREF; navGeo != NULL; navGeo = g_list_next (navGeo)) {
     ref = (GEOREF*) navGeo -> data;
     x = ref -> actX + originX;
     if ((x < 0) || (x > clientX)) continue;
     y = ref -> actY + originY;
     if ((y < 0) || (y > clientY)) continue;
     gdk_draw_rectangle (pixmap, mWinData . blackPen, FALSE, x - mWinData . radius, y - mWinData . radius,
                 (2 * mWinData . radius) + 1, (2 * mWinData . radius) + 1);
   }
 }
 /* Then draws the active descendants of the actual node */
 if (activeNode != NULL) {
   node = (PHYLONODE*) activeNode -> data;
   if (node -> rec == NULL)
     DrawActRec (pixmap, clientX, clientY, originX, originY, activeNode, mWinData . greenPen, NULL);
   else if (node -> rec -> nodeFlag & VIC_NODE) {
     if ((node -> rec -> removSet != NULL) && (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawRemoved)))
         && (node -> rec -> removSet == node -> rec -> setB))
       pen = mWinData . whitePen;
     else
       pen = mWinData . bluePen;
     DrawActRec (pixmap, clientX, clientY, originX, originY,
                 node -> rec -> setA -> node, pen, node -> rec -> setB -> rec -> cells);
     if ((node -> rec -> removSet != NULL) && (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawRemoved)))
         && (node -> rec -> removSet == node -> rec -> setA))
       pen = mWinData . whitePen;
     else
       pen = mWinData . redPen;
     DrawActRec (pixmap, clientX, clientY, originX, originY,
                 node -> rec -> setB -> node, pen, node -> rec -> setA -> rec -> cells);
     if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawBarrier)) == TRUE) {
       gdk_gc_get_values (mWinData . greenPen, &gcVals);
       if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuBarrierAncestor)) == TRUE) {
         gdk_gc_set_line_attributes (mWinData . greenPen, gcVals . line_width, GDK_LINE_ON_OFF_DASH, gcVals . cap_style,
                                     gcVals . join_style);
         for (navAnc = node -> node -> parent; navAnc != NULL; navAnc = navAnc -> parent) {
           ancNode = (PHYLONODE*) navAnc -> data;
           if (ancNode -> rec -> nodeFlag & VIC_NODE)
             DrawBarrierFromRec (pixmap, clientX, clientY, originX, originY, ancNode -> rec);
         }
       }
       if ((gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuBarrierCommon)) == TRUE) &&
           (node -> alloSet != NULL)) {
         gdk_gc_set_line_attributes (mWinData . greenPen, gcVals . line_width, gcVals . line_style, gcVals . cap_style,
                                     gcVals . join_style);
         DrawBarrierFromAlloSet (pixmap, clientX, clientY, originX, originY, node);
       }
       gdk_gc_set_line_attributes (mWinData . greenPen, gcVals . line_width + 2, gcVals . line_style, gcVals . cap_style,
                                   gcVals . join_style);
       DrawBarrierFromRec (pixmap, clientX, clientY, originX, originY, node -> rec);
       gdk_gc_set_line_attributes (mWinData . greenPen, gcVals . line_width, gcVals . line_style, gcVals . cap_style,
                                   gcVals . join_style);
     }
   }
   else
     DrawActRec (pixmap, clientX, clientY, originX, originY, activeNode, mWinData . greenPen, NULL);
 }
}

void DrawActRec (GdkPixmap* pixmap, SCOUNTER clientX, SCOUNTER clientY, SCOUNTER originX, SCOUNTER originY,
                 GNode* refNode, GdkGC* pen, BITFIELD* outCells) {
 GNode* actNode;
 GList* navGeo;
 PHYLONODE* node;
 PHYLONODE* anc;
 TERMINAL* term;
 GEOREF* ref;
 SCOUNTER c, sX, sY;
 SINTEGER x, y, x1, y1;
 BITFIELD* inCells;
 SFLAG in, out;

 node = (PHYLONODE*) refNode -> data;
 if (mWinData . viewMode == VIEW_RECORDS) {
   if (node -> rec != NULL)
     inCells = node -> rec -> cells;
   else
     inCells = NULL;
   for (actNode = refNode; actNode != NULL;) {
     if (IsDesc (actNode, refNode) != TRUE) break;
     node = (PHYLONODE*) actNode -> data;
     if (actNode -> parent != NULL)
       anc = (PHYLONODE*) actNode -> parent -> data;
     if ((inCells != NULL) && (actNode != refNode) && ((node -> rec -> numCells == 0) ||
                           (node -> rec -> nodeFlag & REMOV_NODE) || (node == anc -> rec -> removSet))) {
       /* A removed distribution, ignores all of its descendants */
       while (actNode != NULL) {
         if (g_node_next_sibling (actNode) != NULL) {
           actNode = g_node_next_sibling (actNode);
           break;
         }
         actNode = actNode -> parent;
       }
       continue;
     }
     if (g_node_first_child (actNode) != NULL) {
     /* Not a term, goes to the first descendant */
       actNode = g_node_first_child (actNode);
       continue;
     }
     term = node -> term;
     for (navGeo = term -> firstGeo; navGeo != NULL; navGeo = g_list_next (navGeo)) {
       ref = (GEOREF*) navGeo -> data;
       if (term != (TERMINAL*) ref -> sp -> data) break;
       x = ref -> actX + originX;
       if ((x < 0) || (x > clientX)) continue;
       y = ref -> actY + originY;
       if ((y < 0) || (y > clientY)) continue;
       if (inCells != NULL) {
         in = IsBitOn (inCells, ref -> actCell);
         if (outCells != NULL)
           out = IsBitOn (outCells, ref -> actCell);
         else
           out = FALSE;
       }
       else {
         in = TRUE;
         out = FALSE;
       }
       if (in != TRUE) {
         if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawRemoved)) == TRUE) {
           gdk_draw_rectangle (pixmap, mWinData . whitePen, TRUE, x - mWinData . radius, y - mWinData . radius,
                 (2 * mWinData . radius) + 1, (2 * mWinData . radius) + 1);
           gdk_draw_rectangle (pixmap, mWinData . blackPen, FALSE, x - mWinData . radius, y - mWinData . radius,
                 (2 * mWinData . radius) + 1, (2 * mWinData . radius) + 1);
         }
       }
       else if ((in == TRUE) && (out != TRUE)) {
         gdk_draw_rectangle (pixmap, pen, TRUE, x - mWinData . radius, y - mWinData . radius,
                 (2 * mWinData . radius) + 1, (2 * mWinData . radius) + 1);
         gdk_draw_rectangle (pixmap, mWinData . blackPen, FALSE, x - mWinData . radius, y - mWinData . radius,
                 (2 * mWinData . radius) + 1, (2 * mWinData . radius) + 1);
       }
       else {
         gdk_draw_rectangle (pixmap, mWinData . greenPen, TRUE, x - mWinData . radius, y - mWinData . radius,
                 (2 * mWinData . radius) + 1, (2 * mWinData . radius) + 1);
         gdk_draw_rectangle (pixmap, mWinData . blackPen, FALSE, x - mWinData . radius, y - mWinData . radius,
                 (2 * mWinData . radius) + 1, (2 * mWinData . radius) + 1);
       }
     }
     while (actNode != NULL) {
       if (g_node_next_sibling (actNode) != NULL) {
         actNode = g_node_next_sibling (actNode);
         break;
       }
       actNode = actNode -> parent;
     }
   }
 }
 else {
   for (c = 0; c < GETNUMCELLS; ++ c) {
     if (IsBitOn (node -> rec -> cells, c)) {
       GetColRowFromCell (c, &sY, &sX);
       x = (((GETGRIDSIZE * sX) + GETGRIDSTARTX) * mWinData . xScale) + originX;
       x1 = (((GETGRIDSIZE * (sX + 1)) + GETGRIDSTARTX) * mWinData . xScale) + originX;
       if ((x1 < 0) || (x > clientX)) continue;
       y = (((GETGRIDSIZE * sY) + GETGRIDSTARTY) * mWinData . yScale) + originY;
       y1 = (((GETGRIDSIZE * (sY + 1)) + GETGRIDSTARTY) * mWinData . yScale) + originY;
       if ((y1 < 0) || (y > clientY)) continue;
       if (outCells != NULL)
         out = IsBitOn (outCells, c);
       else
         out = FALSE;
       if (out != TRUE)
         gdk_draw_rectangle (pixmap, pen, TRUE, x, y, x1 - x, y1 - y);
       else
         gdk_draw_rectangle (pixmap, mWinData . greenPen, TRUE, x, y, x1 - x, y1 - y);
     }
   }
 }
}

GEOREF* SearchNearestPoint (gint x, gint y) {
 GList* navGeo;
 GEOREF* actGeo;
 GEOREF* toRet;
 SINTEGER difX, difY, bestDif, actDif;
 PHYLONODE* node;

 toRet = NULL;
 bestDif = (2 * mWinData . radius) + 1;
 /* First look for points of the active terminal, if any */
 if (activeNode != NULL) {
   node = (PHYLONODE*) activeNode -> data;
   if (node -> term != NULL) {
     for (navGeo = node -> term -> firstGeo; navGeo != NULL; navGeo = g_list_next (navGeo)) {
       actGeo = (GEOREF*) navGeo -> data;
       if (actGeo -> sp != node -> term -> list) break;
       difX = x - actGeo -> actX;
       difX = (difX >= 0) ? difX : - difX;
       if (difX > (mWinData . radius + 1)) continue;
       difY = y - actGeo -> actY;
       difY = (difY >= 0) ? difY : - difY;
       if (difY > (mWinData . radius + 1)) continue;
       actDif = difX + difY;
       if (actDif < bestDif) {
         bestDif = actDif;
         toRet = actGeo;
       }
       if (actDif == 0) break;
     }
   }
 }
 if (toRet != NULL) return toRet;
 /* Then, check all points */
 for (navGeo = GETGEOREF; navGeo != NULL; navGeo = g_list_next (navGeo)) {
   actGeo = (GEOREF*) navGeo -> data;
   difX = x - actGeo -> actX;
   difX = (difX >= 0) ? difX : - difX;
   if (difX > (mWinData . radius + 1)) continue;
   difY = y - actGeo -> actY;
   difY = (difY >= 0) ? difY : - difY;
   if (difY > (mWinData . radius + 1)) continue;
   actDif = difX + difY;
   if (actDif < bestDif) {
     bestDif = actDif;
     toRet = actGeo;
   }
   if (actDif == 0) break;
 }
 return toRet;
}

void DrawBarrierFromRec (GdkPixmap* pixmap, SCOUNTER clientX, SCOUNTER clientY, SCOUNTER originX, SCOUNTER originY, RECONS* rec) {
 BITFIELD* setA;
 BITFIELD* setB;

 if (rec -> triangles == NULL) {
   setA = CreateBitField ();
   setB = CreateBitField ();
   CopyBitField (setA, rec -> setA -> rec -> obs);
   CopyBitField (setB, rec -> setB -> rec -> obs);
   BitFieldRemove (setA, rec -> setB -> rec -> cells);
   BitFieldRemove (setB, rec -> setA -> rec -> cells);
   rec -> triangles = MakeVoronoi (setA, setB);
   g_slice_free1 (GETBITFIELDSIZE, setA);
   g_slice_free1 (GETBITFIELDSIZE, setB);
 }
 DrawBarrier (pixmap, clientX, clientY, originX, originY, rec -> triangles);
}

void DrawBarrierFromAlloSet (GdkPixmap* pixmap, SCOUNTER clientX, SCOUNTER clientY, SCOUNTER originX, SCOUNTER originY, PHYLONODE* node) {
 if (node -> alloSet -> triangles == NULL)
   node -> alloSet -> triangles = MakeVoronoi (node -> alloSet -> setA, node -> alloSet -> setB);
 DrawBarrier (pixmap, clientX, clientY, originX, originY, node -> alloSet -> triangles);
}

void DrawBarrier (GdkPixmap* pixmap, SCOUNTER clientX, SCOUNTER clientY, SCOUNTER originX, SCOUNTER originY, GList* triangles) {
 GList* nav;
 SCOUNTER aR, aC, bR, bC, cR, cC;
 SCOUNTER sX, sY, eX, eY, mid;
 SCOUNTER cX, cY, nX, nY;
 DELAUNAYTRIANGLE* actTri;
 DELAUNAYTRIANGLE* neiTri;

 for (nav = triangles; nav != NULL; nav = g_list_next (nav)) {
   actTri = (DELAUNAYTRIANGLE*) nav -> data;
   aR = actTri -> row [0];
   aC = actTri -> col [0];
   bR = actTri -> row [1];
   bC = actTri -> col [1];
   cR = actTri -> row [2];
   cC = actTri -> col [2];
   aC = (GETGRIDSTARTX + (GETGRIDSIZE * aC) + (GETGRIDSIZE / 2)) * mWinData . xScale;
   bC = (GETGRIDSTARTX + (GETGRIDSIZE * bC) + (GETGRIDSIZE / 2)) * mWinData . xScale;
   cC = (GETGRIDSTARTX + (GETGRIDSIZE * cC) + (GETGRIDSIZE / 2)) * mWinData . xScale;
   aR = (GETGRIDSTARTY + (GETGRIDSIZE * aR) + (GETGRIDSIZE / 2)) * mWinData . yScale;
   bR = (GETGRIDSTARTY + (GETGRIDSIZE * bR) + (GETGRIDSIZE / 2)) * mWinData . yScale;
   cR = (GETGRIDSTARTY + (GETGRIDSIZE * cR) + (GETGRIDSIZE / 2)) * mWinData . yScale;
   sX = sY = eX = eY =-1;

   if (mWinData . barrierMode == DRAW_AS_DELAUNAY) {
     /* Draws the triangles */
     sX = aC + originX;
     sY = aR + originY;
     eX = bC + originX;
     eY = bR + originY;
     gdk_draw_line (pixmap, mWinData . greenPen, sX, sY, eX, eY);
     eX = cC + originX;
     eY = cR + originY;
     gdk_draw_line (pixmap, mWinData . greenPen, sX, sY, eX, eY);
     sX = bC + originX;
     sY = bR + originY;
     gdk_draw_line (pixmap, mWinData . greenPen, sX, sY, eX, eY);
   }
   else if (mWinData . barrierMode == DRAW_AS_MIDTRI) {
     /* Draw middle triangles, "pseudovoronoi" */
     if ((actTri -> state [0] != actTri -> state [1]) && ((actTri -> state [0] != 0) && (actTri -> state [1] != 0))) {
       mid = (aC > bC) ? ((aC - bC) / 2) + bC : ((bC - aC) / 2) + aC;
       sX = mid + originX;
       mid = (aR > bR) ? ((aR - bR) / 2) + bR : ((bR - aR) / 2) + aR;
       sY = mid + originY;
     }
     if ((actTri -> state [0] != actTri -> state [2]) && ((actTri -> state [0] != 0) && (actTri -> state [2] != 0))) {
       mid = (aC > cC) ? ((aC - cC) / 2) + cC : ((cC - aC) / 2) + aC;
       eX = mid + originX;
       mid = (aR > cR) ? ((aR - cR) / 2) + cR : ((cR - aR) / 2) + aR;
       eY = mid + originY;
     }
     if ((actTri -> state [1] != actTri -> state [2]) && ((actTri -> state [1] != 0) && (actTri -> state [2] != 0))) {
       if (sX != -1) {
         mid = (bC > cC) ? ((bC - cC) / 2) + cC : ((cC - bC) / 2) + bC;
         eX = mid + originX;
         mid = (bR > cR) ? ((bR - cR) / 2) + cR : ((cR - bR) / 2) + bR;
         eY = mid + originY;
       }
       else if (sX == -1) {
         mid = (bC > cC) ? ((bC - cC) / 2) + cC : ((cC - bC) / 2) + bC;
         sX = mid + originX;
         mid = (bR > cR) ? ((bR - cR) / 2) + cR : ((cR - bR) / 2) + bR;
         sY = mid + originY;
       }
     }
     if (((sX < 0) && (eX < 0)) || ((sY < 0) && (eY < 0)) ||
        ((sX > clientX) && (eX > clientX)) || ((sY > clientY) && (eY > clientY)))
       continue;
     gdk_draw_line (pixmap, mWinData . greenPen, sX, sY, eX, eY);
   }
   else {
     /* Draw Voronoi lines */
     cX = ((GETGRIDSTARTX + (GETGRIDSIZE * actTri -> cX) + (GETGRIDSIZE / 2)) * mWinData . xScale) + originX;
     cY = ((GETGRIDSTARTY + (GETGRIDSIZE * actTri -> cY) + (GETGRIDSIZE / 2)) * mWinData . yScale) + originY;
     if ((actTri -> neighbor [2] != NULL) &&
         ((actTri -> state [0] != actTri -> state [1]) && ((actTri -> state [0] != 0) && (actTri -> state [1] != 0)))) {
       neiTri = (DELAUNAYTRIANGLE*) actTri -> neighbor [2];
       nX = ((GETGRIDSTARTX + (GETGRIDSIZE * neiTri -> cX) + (GETGRIDSIZE / 2)) * mWinData . xScale) + originX;
       nY = ((GETGRIDSTARTY + (GETGRIDSIZE * neiTri -> cY) + (GETGRIDSIZE / 2)) * mWinData . yScale) + originY;
       if (!(((cX < 0) && (nX < 0)) || ((cY < 0) && (nY < 0)) ||
        ((cX > mWinData . clientX) && (nX > clientX)) || ((cY > clientY) && (nY > clientY))))
          gdk_draw_line (pixmap, mWinData . greenPen, cX, cY, nX, nY);
     }
     if ((actTri -> neighbor [1]) &&
         ((actTri -> state [0] != actTri -> state [2]) && ((actTri -> state [0] != 0) && (actTri -> state [2] != 0)))) {
       neiTri = (DELAUNAYTRIANGLE*) actTri -> neighbor [1];
       nX = ((GETGRIDSTARTX + (GETGRIDSIZE * neiTri -> cX) + (GETGRIDSIZE / 2)) * mWinData . xScale) + originX;
       nY = ((GETGRIDSTARTY + (GETGRIDSIZE * neiTri -> cY) + (GETGRIDSIZE / 2)) * mWinData . yScale) + originY;
       if (!(((cX < 0) && (nX < 0)) || ((cY < 0) && (nY < 0)) ||
        ((cX > clientX) && (nX > clientX)) || ((cY > clientY) && (nY > clientY))))
         gdk_draw_line (pixmap, mWinData . greenPen, cX, cY, nX, nY);
     }
     if ((actTri -> neighbor [0] != NULL) &&
         ((actTri -> state [2] != actTri -> state [1]) && ((actTri -> state [2] != 0) && (actTri -> state [1] != 0)))){
       neiTri = (DELAUNAYTRIANGLE*) actTri -> neighbor [0];
       nX = ((GETGRIDSTARTX + (GETGRIDSIZE * neiTri -> cX) + (GETGRIDSIZE / 2)) * mWinData . xScale) + originX;
       nY = ((GETGRIDSTARTY + (GETGRIDSIZE * neiTri -> cY) + (GETGRIDSIZE / 2)) * mWinData . yScale) + originY;
       if (!(((cX < 0) && (nX < 0)) || ((cY < 0) && (nY < 0)) ||
        ((cX > clientX) && (nX > clientX)) || ((cY > clientY) && (nY > clientY))))
         gdk_draw_line (pixmap, mWinData . greenPen, cX, cY, nX, nY);
     }
   }
 }
}
/* out.c is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */

#include "out.h"
#include "grid.h"
#include <stdio.h>

static GString* DateString (gchar* date);
static void WriteNodeWithData (GNode* actNode, GString* outBuffer);

static void WritePhyloXMLTree (PHYLOTREE* tree, GString* outBuffer);
static void WriteCladoXMLTree (PHYLOTREE* tree, GString* outBuffer);

static void WriteKMLSingle (GNode* refNode, BITFIELD* outCells, SFLAG modeRecords, SFLAG modeRemoved, gchar* style,
                            GString* outBuffer);

static SCOUNTER CalculateRichness (INTARRAY* sppPerCell, INTARRAY* obsPerCell);
static SCOUNTER CalculatePD (INTARRAY* pdPerCell, INTARRAY* obsPerCell);

static gchar GetHexChar (SCOUNTER num);
static void WriteInHex (gchar* str, SCOUNTER num);

void WriteXMLFile (gchar* fileName) {
 gsize iniLen;
 GString* outBuffer;
 GList* navTree;
 GNode* navNode;
 GList* navGeo;
 GList* navName;
 PHYLONODE* node;
 PHYLOTREE* tree;
 GEOREF* geo;
 FILE* out;
 gchar* numStr;
 SCOUNTER c;

 iniLen = (85 * GETNUMSPP) + (65 * GETNUMGEOREFS);
 outBuffer = g_string_sized_new (iniLen);
 g_string_append (outBuffer, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
 g_string_append (outBuffer, "\n\n<Vip>");
 for (navTree = GETTREES; navTree != NULL; navTree = g_list_next (navTree)) {
   tree = (PHYLOTREE*) navTree -> data;
   g_string_append (outBuffer, "\n<Tree>");
   if (tree -> name != NULL)
     g_string_append_printf (outBuffer, "\n\t<name>%s</name>", tree -> name);
   for (navNode = tree -> root; navNode != NULL; ) {
     node = (PHYLONODE*) navNode -> data;
     g_string_append_printf (outBuffer, "\n\t<Node id=\"%ld\">", node -> id);
     if (node -> name != NULL)
       g_string_append_printf (outBuffer, "\n\t\t<name>%s</name>", node -> name);
     if ((node -> minAge > 0) && (node -> minAge != NO_AGE))
       g_string_append_printf (outBuffer, "\n\t\t<age>%ld</age>", node -> minAge);
     if (g_node_first_child (navNode) != NULL) {
       navNode = g_node_first_child (navNode);
       continue;
     }
     for (c = 0; ; ++ c) {
       navName = g_list_nth (node -> term -> synonyms, c);
       if (navName == NULL) break;
       if (navName -> data == node -> term -> name) continue;
       g_string_append_printf (outBuffer, "\n\t\t<synonym>%s</synonym>", (gchar*) navName -> data);
     }
     for (navGeo = node -> term -> firstGeo; navGeo != NULL; navGeo = g_list_next (navGeo)) {
       geo = (GEOREF*) navGeo -> data;
       if (geo -> sp != node -> term -> list) break;
       g_string_append (outBuffer, "\n\t\t<SpecimenInfo>");
       if (geo -> name != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\t<name>%s</name>", geo -> name);
       if (geo -> collec != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\t<collection>%s</collection>", geo -> collec);
       if (geo -> catalog != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\t<catalogCode>%s</catalogCode>", geo -> catalog);
       if (geo -> url != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\t<url>%s</url>", geo -> url);
       if (geo -> reference != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\t<reference>%s</reference>", geo -> reference);
       if (geo -> identifier != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\t<identifiedBy>%s</identifiedBy>", geo -> identifier);
       if (geo -> sampleDate != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\t<collectionDate>%s</collectionDate>", geo -> sampleDate);
       if (geo -> collector != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\t<collector>%s</collector>", geo -> collector);
       g_string_append (outBuffer, "\n\t\t\t<Point>");
       g_string_append (outBuffer, "\n\t\t\t\t<coordinates>");
       numStr = g_strdup_printf ("%.6f", geo -> lon);
       CheckRealStr (numStr);
       g_string_append_printf (outBuffer, "%s,", numStr);
       g_free (numStr);
       numStr = g_strdup_printf ("%.6f", geo -> lat);
       CheckRealStr (numStr);
       g_string_append (outBuffer, numStr);
       g_free (numStr);
       g_string_append (outBuffer, "</coordinates>");
       g_string_append (outBuffer, "\n\t\t\t</Point>\n\t\t</SpecimenInfo>");
     }
     while (navNode != NULL) {
       g_string_append (outBuffer, "\n\t</Node>");
       if (g_node_next_sibling (navNode) != NULL) {
         navNode = g_node_next_sibling (navNode);
         break;
       }
       navNode = navNode -> parent;
     }
   }
   g_string_append (outBuffer, "\n</Tree>");
 }
 g_string_append (outBuffer, "\n\n</Vip>\n");
 out = fopen (fileName, "w");
 if (out == NULL) {
   errorMessage = g_strdup_printf ("\nUnable to open %s\n", fileName);
   OnError ();
   g_string_free (outBuffer, TRUE);
   return;
 }
 if (fwrite (outBuffer -> str, sizeof (gchar), outBuffer -> len, out) != outBuffer -> len) {
   errorMessage = g_strdup_printf ("\nUnable to save whole data\nFile%s\n", fileName);
   OnError ();
 }
 fclose (out);
 g_string_free (outBuffer, TRUE);
}

void WriteKMLFile (gchar* fileName) {
 gsize iniLen;
 GString* outBuffer;
 GList* navTree;
 GNode* navNode;
 GList* navGeo;
 PHYLONODE* node;
 PHYLOTREE* tree;
 GEOREF* geo;
 FILE* out;
 gchar* numStr;
 GString* date;

 iniLen = (85 * GETNUMSPP) + (65 * GETNUMGEOREFS);
 outBuffer = g_string_sized_new (iniLen);
 g_string_append (outBuffer, "<?xml version=\"1.0\"?><kml xmlns=\"http://earth.google.com/kml/2.1\">");
 g_string_append (outBuffer, "\n\n<Document>");
 g_string_append_printf (outBuffer, "\n\t<name>%s</name>", fileName);
 g_string_append (outBuffer, "\n\t<description><![CDATA[<p>KML file generated by VIP (Arias, 2009)</p>");
 g_string_append (outBuffer, "\n\t\t<p>Useful links:<br>");
 g_string_append (outBuffer, "<a href=\"http://www.zmuc.dk/public/phylogeny/VIP\">");
 g_string_append (outBuffer, "www.zmuc.dk/public/phylogeny/VIP</a>: VIP download<br />");
 g_string_append (outBuffer, "\n\t\t<a href=\"mailto:sal.arias@gmail.com\">sal.arias@gmail.com</a>: ");
 g_string_append (outBuffer, "Salva e-mail</p>\n\t\t<p><b>Acknowledgment</b><br />Pablo, Claudia, ");
 g_string_append (outBuffer, "and Lolo, for suggestions and interest!</p>]]></description>\n");
 g_string_append (outBuffer, "\n");
 for (navTree = GETTREES; navTree != NULL; navTree = g_list_next (navTree)) {
   tree = (PHYLOTREE*) navTree -> data;
   g_string_append (outBuffer, "\n");
   for (navNode = tree -> root; navNode != NULL; ) {
     node = (PHYLONODE*) navNode -> data;
     g_string_append_printf (outBuffer, "\n<Folder id=\"%ld\">", node -> id);
     if (node -> name != NULL)
       g_string_append_printf (outBuffer, "\n\t<name>%s</name>", node -> name);
     if (g_node_first_child (navNode) != NULL) {
       navNode = g_node_first_child (navNode);
       continue;
     }
     for (navGeo = node -> term -> firstGeo; navGeo != NULL; navGeo = g_list_next (navGeo)) {
       geo = (GEOREF*) navGeo -> data;
       if (geo -> sp != node -> term -> list) break;
       g_string_append (outBuffer, "\n\t<Placemark>");
       if (geo -> name != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<name>%s</name>", geo -> name);
       g_string_append (outBuffer, "\n\t\t<description><![CDATA[<p>");
       if (geo -> name != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\tTerminal: %s<br />", geo -> name);
       if (geo -> collec != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\tCollection: %s<br />", geo -> collec);
       if (geo -> catalog != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\tCatalogue code: %s <br />", geo -> catalog);
       if (geo -> url != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\tUrl: <a hfer=\"%s\">%s</a><br />", geo -> url, geo -> url);
       if (geo -> reference != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\tReference: %s<br />", geo -> reference);
       if (geo -> identifier != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\tIdentified by: %s<br />", geo -> identifier);
       if (geo -> sampleDate != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\tCollection date: %s<br />", geo -> sampleDate);
       if (geo -> collector != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\tCollector: %s<br />", geo -> collector);
       g_string_append (outBuffer, "\n\t\t</p>]]></description>");
       if (geo -> sampleDate != NULL) {
         date = DateString (geo -> sampleDate);
         g_string_append (outBuffer, "\n\t\t<TimeStap>");
         g_string_append_printf (outBuffer, "\n\t\t\t<when>%s</when>", date -> str);
         g_string_append (outBuffer, "\n\t\t</TimeStap>");
         g_string_free (date, TRUE);
       }
       g_string_append (outBuffer, "\n\t\t<Point>");
       g_string_append (outBuffer, "\n\t\t\t<extrude>0</extrude>");
       g_string_append (outBuffer, "\n\t\t\t<altitudeMode>absolute</altitudeMode>");
       g_string_append (outBuffer, "\n\t\t\t<coordinates>");
       numStr = g_strdup_printf ("%.6f", geo -> lon);
       CheckRealStr (numStr);
       g_string_append_printf (outBuffer, "%s,", numStr);
       g_free (numStr);
       numStr = g_strdup_printf ("%.6f", geo -> lat);
       CheckRealStr (numStr);
       g_string_append (outBuffer, numStr);
       g_free (numStr);
       g_string_append (outBuffer, "</coordinates>");
       g_string_append (outBuffer, "\n\t\t</Point>\n\t</Placemark>");
     }
     while (navNode != NULL) {
       g_string_append (outBuffer, "\n</Folder>");
       if (g_node_next_sibling (navNode) != NULL) {
         navNode = g_node_next_sibling (navNode);
         break;
       }
       navNode = navNode -> parent;
     }
   }
 }
 g_string_append (outBuffer, "\n\n</Document>\n</kml>\n");
 out = fopen (fileName, "w");
 if (out == NULL) {
   errorMessage = g_strdup_printf ("\nUnable to open %s\n", fileName);
   OnError ();
   g_string_free (outBuffer, TRUE);
   return;
 }
 if (fwrite (outBuffer -> str, sizeof (gchar), outBuffer -> len, out) != outBuffer -> len) {
   errorMessage = g_strdup_printf ("\nUnable to save whole data\nFile%s\n", fileName);
   OnError ();
 }
 fclose (out);
 g_string_free (outBuffer, TRUE);
}

void WriteVRXFile (gchar* fileName) {
 gsize iniLen;
 GString* outBuffer;
 GList* navTree;
 PHYLOTREE* tree;
 GPtrArray* recArray;
 FILE* out;
 SCOUNTER i;
 GList* navRec;
 RECONS* actRec;
 PHYLONODE* anc;

 iniLen = (85 * GETNUMSPP) + (65 * GETNUMGEOREFS);
 outBuffer = g_string_sized_new (iniLen);
 g_string_append (outBuffer, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
 g_string_append (outBuffer, "\n\n<VipResults>");
 for (navTree = GETTREES; navTree != NULL; navTree = g_list_next (navTree)) {
   tree = (PHYLOTREE*) navTree -> data;
   if (tree -> recs == NULL) continue;
   for (navRec = tree -> recs; navRec != NULL; navRec = g_list_next (navRec)) {
     g_string_append_printf (outBuffer,"\n\n<Reconstruction>\n\t<treeID>%ld</treeID>", tree -> id);
     recArray = (GPtrArray*) navRec -> data;
     for (i = 0; i < recArray -> len; ++ i) {
       actRec = (RECONS*) g_ptr_array_index (recArray, i);
       if (actRec -> nodeFlag & REMOV_NODE) {
         g_string_append_printf (outBuffer, "\n\t<DistroElimination>\n\t\t<nodeID>%ld</nodeID>\n\t</DistroElimination>",
                                 actRec -> node -> id);
       }
       if (actRec -> node -> node -> parent != NULL) {
         anc = (PHYLONODE*) actRec -> node -> node -> parent -> data;
         if (anc -> rec -> removSet == actRec -> node)
           g_string_append_printf (outBuffer, "\n\t<PartialElimination>\n\t\t<nodeID>%ld</nodeID>\n\t</PartialElimination>",
                                 actRec -> node -> id);
       }
     }
     g_string_append (outBuffer,"\n</Reconstruction>");
   }
 }
 g_string_append (outBuffer, "\n\n</VipResults>\n");
 out = fopen (fileName, "w");
 if (out == NULL) {
   errorMessage = g_strdup_printf ("\nUnable to open %s\n", fileName);
   OnError ();
   g_string_free (outBuffer, TRUE);
   return;
 }
 if (fwrite (outBuffer -> str, sizeof (gchar), outBuffer -> len, out) != outBuffer -> len) {
   errorMessage = g_strdup_printf ("\nUnable to save whole data\nFile%s\n", fileName);
   OnError ();
 }
 fclose (out);
 g_string_free (outBuffer, TRUE);
}

void WriteXYDFile (gchar* fileName) {
 gsize iniLen;
 GString* outBuffer;
 GList* navTerm;
 GList* navGeo;
 TERMINAL* actTerm;
 GEOREF* geo;
 FILE* out;
 SCOUNTER i;
 gchar* numStr;

 iniLen = (20 * GETNUMGEOREFS);
 outBuffer = g_string_sized_new (iniLen);
 g_string_append_printf (outBuffer, "\nspp %ld", GETNUMSPP);
 g_string_append (outBuffer, "\nynegative");
 g_string_append (outBuffer, "\nxydata");
 i = 0;
 for (navTerm = GETSPP; navTerm != NULL; navTerm = g_list_next (navTerm)) {
   actTerm = (TERMINAL*) navTerm -> data;
   if (actTerm -> name != NULL)
     g_string_append_printf (outBuffer, "\n\nsp %ld [%s]", i, actTerm -> name);
   else
     g_string_append_printf (outBuffer, "\n\nsp %ld [species_%ld]", i, i);
   ++ i;
   for (navGeo = actTerm -> firstGeo; navGeo != NULL; navGeo = g_list_next (navGeo)) {
     geo = (GEOREF*) navGeo -> data;
     if (geo -> sp != actTerm -> list) break;
     numStr = g_strdup_printf ("%.6f", geo -> lon);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer, "\n%s ", numStr);
     g_free (numStr);
     numStr = g_strdup_printf ("%.6f", geo -> lat);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer, numStr);
     g_free (numStr);
   }
 }
 g_string_append (outBuffer, "\n\n");
 out = fopen (fileName, "w");
 if (out == NULL) {
   errorMessage = g_strdup_printf ("\nUnable to open %s\n", fileName);
   OnError ();
   g_string_free (outBuffer, TRUE);
   return;
 }
 if (fwrite (outBuffer -> str, sizeof (gchar), outBuffer -> len, out) != outBuffer -> len) {
   errorMessage = g_strdup_printf ("\nUnable to save whole data\nFile%s\n", fileName);
   OnError ();
 }
 fclose (out);
 g_string_free (outBuffer, TRUE);
}

void WriteTABTable (gchar* fileName) {
 gsize iniLen;
 GString* outBuffer;
 GList* navGeo;
 GEOREF* geo;
 TERMINAL* term;
 FILE* out;
 gchar* numStr;

 iniLen = (20 * GETNUMGEOREFS);
 outBuffer = g_string_sized_new (iniLen);
 g_string_append_printf (outBuffer,
    "Dataset\tDate collected\tCollector\tCatalogue No\tIdentifier\tScientific name\tLatitude\tLongitude\tURL\tPrimary reference\n");
 for (navGeo = GETGEOREF; navGeo != NULL; navGeo = g_list_next (navGeo)) {
   geo = (GEOREF*) navGeo -> data;
   if (geo -> collec != NULL)
     g_string_append_printf (outBuffer, "%s\t", geo -> collec);
   else
     g_string_append (outBuffer, "\t");
   if (geo -> sampleDate != NULL)
     g_string_append_printf (outBuffer, "%s\t", geo -> sampleDate);
   else
     g_string_append (outBuffer, "\t");
   if (geo -> collector != NULL)
     g_string_append_printf (outBuffer, "%s\t", geo -> collector);
   else
     g_string_append (outBuffer, "\t");
   if (geo -> catalog != NULL)
     g_string_append_printf (outBuffer, "%s\t", geo -> catalog);
   else
     g_string_append (outBuffer, "\t");
   if (geo -> identifier != NULL)
     g_string_append_printf (outBuffer, "%s\t", geo -> identifier);
   else
     g_string_append (outBuffer, "\t");
   if (geo -> name != NULL)
     g_string_append_printf (outBuffer, "%s\t", geo -> name);
   else {
     term = (TERMINAL*) geo -> sp -> data;
     if (term -> name != NULL)
       g_string_append_printf (outBuffer, "%s\t", term -> name);
     else
       g_string_append (outBuffer, "\t");
   }
   numStr = g_strdup_printf ("%.6f", geo -> lat);
   CheckRealStr (numStr);
   g_string_append_printf (outBuffer, "%s\t", numStr);
   g_free (numStr);
   numStr = g_strdup_printf ("%.6f", geo -> lon);
   CheckRealStr (numStr);
   g_string_append_printf (outBuffer, "%s\t", numStr);
   g_free (numStr);
   if (geo -> url != NULL)
     g_string_append_printf (outBuffer, "%s\t", geo -> url);
   else
     g_string_append (outBuffer, "\t");
   if (geo -> reference != NULL)
     g_string_append_printf (outBuffer, "%s\n", geo -> reference);
   else
     g_string_append (outBuffer, "\n");
 }
 out = fopen (fileName, "w");
 if (out == NULL) {
   errorMessage = g_strdup_printf ("\nUnable to open %s\n", fileName);
   OnError ();
   g_string_free (outBuffer, TRUE);
   return;
 }
 if (fwrite (outBuffer -> str, sizeof (gchar), outBuffer -> len, out) != outBuffer -> len) {
   errorMessage = g_strdup_printf ("\nUnable to save whole data\nFile%s\n", fileName);
   OnError ();
 }
 fclose (out);
 g_string_free (outBuffer, TRUE);
}

void WriteLstFile (gchar* fileName) {
 gsize iniLen;
 GString* outBuffer;
 GList* navSpp;
 TERMINAL* actSp;
 FILE* out;
 GList* termList;
 GList* navName;
 GString* name;

 iniLen = (50 * GETNUMSPP);
 outBuffer = g_string_sized_new (iniLen);
 termList = NULL;
 /* Sort the list alphabetically */
 for (navSpp = GETSPP; navSpp != NULL; navSpp = g_list_next (navSpp)) {
   actSp = (TERMINAL*) navSpp -> data;
   termList = NameOnList (actSp -> name, termList);
 }
 for (navName = termList; navName != NULL; navName = g_list_next (navName)) {
   name = (GString*) navName -> data;
   g_string_append_printf (outBuffer, "%s\n", name -> str);
 }
 for (navName = termList; navName != NULL; navName = g_list_next (navName))
   g_string_free (navName -> data, TRUE);
 g_list_free (termList);
 out = fopen (fileName, "w");
 if (out == NULL) {
   errorMessage = g_strdup_printf ("\nUnable to open %s\n", fileName);
   OnError ();
   g_string_free (outBuffer, TRUE);
   return;
 }
 if (fwrite (outBuffer -> str, sizeof (gchar), outBuffer -> len, out) != outBuffer -> len) {
   errorMessage = g_strdup_printf ("\nUnable to save whole data\nFile%s\n", fileName);
   OnError ();
 }
 fclose (out);
 g_string_free (outBuffer, TRUE);
}

void WriteCllFile (gchar* fileName) {
 gsize iniLen;
 GString* outBuffer;
 GList* navRec;
 GEOREF* actRec;
 FILE* out;
 GList* collecList;
 GList* navName;
 GString* name;

 iniLen = (50 * GETNUMSPP);
 outBuffer = g_string_sized_new (iniLen);
 collecList = NULL;
 /* Sort the list alphabetically */
 for (navRec = GETGEOREF; navRec != NULL; navRec = g_list_next (navRec)) {
   actRec = (GEOREF*) navRec -> data;
   if (actRec -> collec != NULL)
     collecList = NameOnList (actRec -> collec, collecList);
 }
 for (navName = collecList; navName != NULL; navName = g_list_next (navName)) {
   name = (GString*) navName -> data;
   g_string_append_printf (outBuffer, "%s\n", name -> str);
 }
 for (navName = collecList; navName != NULL; navName = g_list_next (navName))
   g_string_free (navName -> data, TRUE);
 g_list_free (collecList);
 out = fopen (fileName, "w");
 if (out == NULL) {
   errorMessage = g_strdup_printf ("\nUnable to open %s\n", fileName);
   OnError ();
   g_string_free (outBuffer, TRUE);
   return;
 }
 if (fwrite (outBuffer -> str, sizeof (gchar), outBuffer -> len, out) != outBuffer -> len) {
   errorMessage = g_strdup_printf ("\nUnable to save whole data\nFile%s\n", fileName);
   OnError ();
 }
 fclose (out);
 g_string_free (outBuffer, TRUE);
}

void WriteOutLog (gchar* fileName, GString* outBuffer) {
 FILE* out;

 out = fopen (fileName, "w");
 if (out == NULL) {
   errorMessage = g_strdup_printf ("\nUnable to open %s\n", fileName);
   OnError ();
   return;
 }
 if (fwrite (outBuffer -> str, sizeof (gchar), outBuffer -> len, out) != outBuffer -> len) {
   errorMessage = g_strdup_printf ("\nUnable to save whole data\nFile%s\n", fileName);
   OnError ();
 }
 fclose (out);
}

/* This functions ensure that the out uses points and not commas, This is because my system uses spanish system (i.e. commas) !!" */
void CheckRealStr (gchar* num) {
 SCOUNTER c;

 for (c = 0; ; ++ c) {
   if (num [c] == '.') break;
   if (num [c] == ',') {
     num [c] = '.';
     break;
   }
   if (num [c] == 0) break;
 }
}

GString* DateString (gchar* date) {
 GString* toRet;
 SCOUNTER c;

 toRet = g_string_sized_new (16);
 for (c = 0; ; ++ c) {
   if (date [c] == 0) break;
   if (g_ascii_isspace (date [c])) break;
   g_string_append_c (toRet, date [c]);
 }
 g_string_append_c (toRet, 0);
 return toRet;
}

gchar* CheckExtention (gchar* name, gchar* extention) {
 SCOUNTER c;
 SFLAG noExt;
 gchar* toRet;

 noExt = TRUE;
 for (c = 0; ; ++c) {
   if (name [c] == '\0') {
     noExt = TRUE;
     break;
   }
   if (name [c] == '.') {
     noExt = FALSE;
     continue;
   }
   if ((noExt == FALSE) && (!(g_ascii_isspace (name [c])))) break;
 }
 toRet = name;
 if (noExt == TRUE) {
   toRet = g_strdup_printf ("%s.%s", name, extention);
   g_free (name);
 }
 return toRet;
}

void WriteNodeWithData (GNode* actNode, GString* outBuffer) {
 GNode* navNode;
 GList* navGeo;
 GList* navName;
 PHYLONODE* node;
 PHYLONODE* desc;
 GEOREF* geo;
 SFLAG isOpen;
 gchar* numStr;
 SCOUNTER c;

 node = (PHYLONODE*) actNode -> data;
 if (node -> withData == 0) return;
 /* If the node is a term, write it and returns */
 if (node -> term != NULL) {
   g_string_append_printf (outBuffer, "\n<Node id=\"%ld\">", node -> id);
   if (node -> name != NULL)
     g_string_append_printf (outBuffer, "\n\t<name>%s</name>", node -> name);
   for (c = 0; ; ++ c) {
     navName = g_list_nth (node -> term -> synonyms, c);
     if (navName == NULL) break;
     if (navName -> data == node -> term -> name) continue;
     g_string_append_printf (outBuffer, "\n\t<synonym>%s</synonym>", (gchar*) navName -> data);
   }
   for (navGeo = node -> term -> firstGeo; navGeo != NULL; navGeo = g_list_next (navGeo)) {
     geo = (GEOREF*) navGeo -> data;
     if (geo -> sp != node -> term -> list) break;
     g_string_append (outBuffer, "\n\t<SpecimenInfo>");
     if (geo -> name != NULL)
       g_string_append_printf (outBuffer, "\n\t\t<name>%s</name>", geo -> name);
     if (geo -> collec != NULL)
       g_string_append_printf (outBuffer, "\n\t\t<collection>%s</collection>", geo -> collec);
     if (geo -> catalog != NULL)
       g_string_append_printf (outBuffer, "\n\t\t<catalogCode>%s</catalogCode>", geo -> catalog);
     if (geo -> url != NULL)
       g_string_append_printf (outBuffer, "\n\t\t<url>%s</url>", geo -> url);
     if (geo -> reference != NULL)
       g_string_append_printf (outBuffer, "\n\t\t<reference>%s</reference>", geo -> reference);
     if (geo -> identifier != NULL)
       g_string_append_printf (outBuffer, "\n\t\t<identifiedBy>%s</identifiedBy>", geo -> identifier);
     if (geo -> sampleDate != NULL)
       g_string_append_printf (outBuffer, "\n\t\t<collectionDate>%s</collectionDate>", geo -> sampleDate);
     if (geo -> collector != NULL)
       g_string_append_printf (outBuffer, "\n\t\t<collector>%s</collector>", geo -> collector);
     g_string_append (outBuffer, "\n\t\t<Point>");
     g_string_append (outBuffer, "\n\t\t\t<coordinates>");
     numStr = g_strdup_printf ("%.6f", geo -> lon);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer, "%s,", numStr);
     g_free (numStr);
     numStr = g_strdup_printf ("%.6f", geo -> lat);
     CheckRealStr (numStr);
     g_string_append (outBuffer, numStr);
     g_free (numStr);
     g_string_append (outBuffer, "</coordinates>");
     g_string_append (outBuffer, "\n\t\t</Point>\n\t</SpecimenInfo>");
   }
   g_string_append (outBuffer, "\n</Node>");
   return;
 }
 isOpen = FALSE;
 /* Checks if the number of nodes with data is diferent in any descendant */
 for (navNode = g_node_first_child (actNode); navNode != NULL; navNode = g_node_next_sibling (navNode)) {
   desc = (PHYLONODE*) navNode -> data;
   if ((desc -> withData > 0) && (desc -> withData != node -> withData)) {
     isOpen = TRUE;
     break;
   }
 }
 if (isOpen == TRUE) {
   g_string_append_printf (outBuffer, "\n<Node id=\"%ld\">", node -> id);
   if (node -> name != NULL)
     g_string_append_printf (outBuffer, "\n\t<name>%s</name>", node -> name);
 }
 /* Write descendants */
 for (navNode = g_node_first_child (actNode); navNode != NULL; navNode = g_node_next_sibling (navNode))
   WriteNodeWithData (navNode, outBuffer);
 if (isOpen == TRUE)
   g_string_append (outBuffer, "\n</Node>");
 return;
}

void WriteXMLWithData (gchar* fileName) {
 gsize iniLen;
 GString* outBuffer;
 GList* navTree;
 PHYLOTREE* tree;
 FILE* out;

 iniLen = (85 * GETNUMSPP) + (65 * GETNUMGEOREFS);
 outBuffer = g_string_sized_new (iniLen);
 g_string_append (outBuffer, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
 g_string_append (outBuffer, "\n\n<Vip>");
 for (navTree = GETTREES; navTree != NULL; navTree = g_list_next (navTree)) {
   tree = (PHYLOTREE*) navTree -> data;
   g_string_append (outBuffer, "\n");
   WriteNodeWithData (tree -> root, outBuffer);
 }
 g_string_append (outBuffer, "\n\n</Vip>\n");
 out = fopen (fileName, "w");
 if (out == NULL) {
   errorMessage = g_strdup_printf ("\nUnable to open %s\n", fileName);
   OnError ();
   g_string_free (outBuffer, TRUE);
   return;
 }
 if (fwrite (outBuffer -> str, sizeof (gchar), outBuffer -> len, out) != outBuffer -> len) {
   errorMessage = g_strdup_printf ("\nUnable to save whole data\nFile%s\n", fileName);
   OnError ();
 }
 fclose (out);
 g_string_free (outBuffer, TRUE);
}

void WriteSVGFile (gchar* fileName, PHYLOTREE* tree, SFLAG withRec) {
 gsize iniLen;
 GString* outBuffer;
 FILE* out;
 GNode* navNode;
 PHYLONODE* node;
 PHYLONODE* ancNode;
 SINTEGER endX, startX, startY;
 gchar* text;

 node = (PHYLONODE*) tree -> root -> data;
 iniLen = (80 * node -> numTerms);
 outBuffer = g_string_sized_new (iniLen);
 g_string_append (outBuffer, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
 g_string_append_printf (outBuffer, "\n<svg width=\"%ld\" height=\"%ld\" xmlns=\"http://www.w3.org/2000/svg\">",
                                    tree -> x + 100, tree -> y);
 g_string_append_printf (outBuffer, "\n\t<desc>Tree %s</desc>", tree -> name);
 g_string_append (outBuffer,
                  "\n\t<g stroke-width=\"1\" stroke=\"black\" stroke-linecap=\"round\" font-family=\"Verdana\" font-size=\"10\">");
 for (navNode = tree -> root; navNode != NULL; ) {
   node = (PHYLONODE*) navNode -> data;
   if (node -> name != NULL)
     g_string_append_printf (outBuffer, "\n\t\t<!-- Node: %s -->", node -> name);
   else
     g_string_append_printf (outBuffer, "\n\t\t<!-- Node -->");
   /* draws the horizontal line */
   if (navNode -> parent != NULL) {
     ancNode = (PHYLONODE*) navNode -> parent -> data;
     endX = ancNode -> actX;
   }
   else {
     ancNode = NULL;
     endX = node -> actX - 2;
   }
   g_string_append_printf (outBuffer,"\n\t\t<line x1=\"%ld\" y1=\"%ld\" x2=\"%ld\" y2=\"%ld\" />",
                              node -> actX, node -> actY, endX, node -> actY);
   if ((g_node_first_child (navNode) == NULL) && (node -> name != NULL)) {
     /* write the term name */
     g_string_append_printf (outBuffer,"\n\t\t<text x=\"%ld\" y=\"%ld\" stroke-width=\"0\">%s</text>",
                                node -> actX + 5, node -> actY + 5, node -> name);
     if ((withRec & SVG_WITH_REC) && (node -> rec -> nodeFlag & REMOV_NODE))
          g_string_append_printf (outBuffer, "\n\t\t<circle cx=\"%ld\" cy=\"%ld\" r=\"2\" fill=\"white\" stroke=\"black\" />",
               node -> actX, node -> actY);
   }
   else if (g_node_first_child (navNode) != NULL) {
     /* draws the vertical line */
     g_string_append_printf (outBuffer,"\n\t\t<line x1=\"%ld\" y1=\"%ld\" x2=\"%ld\" y2=\"%ld\" />",
                                        node -> actX, node -> topY, node -> actX, node -> downY);
     if (withRec & SVG_WITH_REC) {
       if (node -> rec -> nodeFlag & VIC_NODE) {
         startX = node -> actX - 2;
         startY = node -> actY - 2;
         if (node -> rec -> nodeFlag & REMOV_NODE)
           g_string_append_printf (outBuffer,
              "\n\t\t<rect x=\"%ld\" y=\"%ld\" width=\"5\" height=\"5\" fill=\"white\" stroke=\"black\" />", startX, startY);
         else if (node -> rec -> removSet != NULL) {
           g_string_append_printf (outBuffer,
              "\n\t\t<rect x=\"%ld\" y=\"%ld\" width=\"5\" height=\"5\" fill=\"white\" stroke=\"black\" />", startX, startY);
           g_string_append_printf (outBuffer,"\n\t\t<line x1=\"%ld\" y1=\"%ld\" x2=\"%ld\" y2=\"%ld\" />",
                                        startX, startY, startX + 5, startY + 5);
         }
         else if (node -> rec -> nodeFlag & TIME_LAG)
           g_string_append_printf (outBuffer,
              "\n\t\t<rect x=\"%ld\" y=\"%ld\" width=\"5\" height=\"5\" fill=\"darkgrey\" stroke=\"black\" />", startX, startY);
         else
           g_string_append_printf (outBuffer,
              "\n\t\t<rect x=\"%ld\" y=\"%ld\" width=\"5\" height=\"5\" fill=\"black\" stroke=\"black\" />", startX, startY);
         if (withRec & SVG_WITH_SUPP) {
           if (node -> bremer > node -> tree -> worst)
             text = g_strdup_printf ("?> %.2f", node -> tree -> worst - node -> tree -> best);
           else
             text = g_strdup_printf ("%.2f", node -> bremer - node -> tree -> best);
           CheckRealStr (text);
           g_string_append_printf (outBuffer,"\n\t\t<text x=\"%ld\" y=\"%ld\" stroke-width=\"0\">%s</text>",
                                   node -> actX + 5, node -> actY + 5, text);
           g_free (text);
         }
       }
       else if (node -> rec -> nodeFlag & REMOV_NODE)
          g_string_append_printf (outBuffer, "\n\t\t<circle cx=\"%ld\" cy=\"%ld\" r=\"2\" fill=\"white\" stroke=\"black\" />",
               node -> actX, node -> actY);
     }
     navNode = g_node_first_child (navNode);
     continue;
   }
   while (navNode != NULL) {
     if (g_node_next_sibling (navNode) != NULL) {
       navNode = g_node_next_sibling (navNode);
       break;
     }
     navNode = navNode -> parent;
   }
 }
 g_string_append (outBuffer,"\n\t</g>\n</svg>\n");
 out = fopen (fileName, "w");
 if (out == NULL) {
   errorMessage = g_strdup_printf ("\nUnable to open %s\n", fileName);
   OnError ();
   g_string_free (outBuffer, TRUE);
   return;
 }
 if (fwrite (outBuffer -> str, sizeof (gchar), outBuffer -> len, out) != outBuffer -> len) {
   errorMessage = g_strdup_printf ("\nUnable to save whole data\nFile%s\n", fileName);
   OnError ();
 }
 fclose (out);
 g_string_free (outBuffer, TRUE);
}

void WritePhyTree (gchar* fileName, PHYLOTREE* tree) {
 gsize iniLen;
 GString* outBuffer;
 FILE* out;

 iniLen = (85 * GETNUMSPP);
 outBuffer = g_string_sized_new (iniLen);
 g_string_append (outBuffer, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
 g_string_append (outBuffer, "\n\n<phyloxml>");
 WritePhyloXMLTree (tree, outBuffer);
 g_string_append (outBuffer, "\n\n</phyloxml>\n");
 out = fopen (fileName, "w");
 if (out == NULL) {
   errorMessage = g_strdup_printf ("\nUnable to open %s\n", fileName);
   OnError ();
   g_string_free (outBuffer, TRUE);
   return;
 }
 if (fwrite (outBuffer -> str, sizeof (gchar), outBuffer -> len, out) != outBuffer -> len) {
   errorMessage = g_strdup_printf ("\nUnable to save whole data\nFile%s\n", fileName);
   OnError ();
 }
 fclose (out);
 g_string_free (outBuffer, TRUE);
}

void WritePhyFile (gchar* fileName) {
 gsize iniLen;
 GString* outBuffer;
 GList* navTree;
 PHYLOTREE* tree;
 FILE* out;

 iniLen = (85 * GETNUMSPP);
 outBuffer = g_string_sized_new (iniLen);
 g_string_append (outBuffer, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
 g_string_append (outBuffer, "\n\n<phyloxml>");
 for (navTree = GETTREES; navTree != NULL; navTree = g_list_next (navTree)) {
   tree = (PHYLOTREE*) navTree -> data;
   WritePhyloXMLTree (tree, outBuffer);
 }
 g_string_append (outBuffer, "\n\n</phyloxml>\n");
 out = fopen (fileName, "w");
 if (out == NULL) {
   errorMessage = g_strdup_printf ("\nUnable to open %s\n", fileName);
   OnError ();
   g_string_free (outBuffer, TRUE);
   return;
 }
 if (fwrite (outBuffer -> str, sizeof (gchar), outBuffer -> len, out) != outBuffer -> len) {
   errorMessage = g_strdup_printf ("\nUnable to save whole data\nFile%s\n", fileName);
   OnError ();
 }
 fclose (out);
 g_string_free (outBuffer, TRUE);
}

void WritePhyloXMLTree (PHYLOTREE* tree, GString* outBuffer) {
 GNode* navNode;
 PHYLONODE* node;

 g_string_append (outBuffer, "\n<phylogeny rooted=\"true\">");
 if (tree -> name != NULL)
   g_string_append_printf (outBuffer, "\n\t<name>%s</name>", tree -> name);
 for (navNode = tree -> root; navNode != NULL; ) {
   node = (PHYLONODE*) navNode -> data;
   g_string_append (outBuffer, "\n\t<clade>");
   if (node -> name != NULL)
     g_string_append_printf (outBuffer, "\n\t\t<name>%s</name>", node -> name);
   if (node -> minAge > 0)
     g_string_append_printf (outBuffer, "\n\t\t<date>\n\t\t\t<minimum>%ld</minimum>\n\t\t</date>", node -> minAge);
   if (g_node_first_child (navNode) != NULL) {
     navNode = g_node_first_child (navNode);
     continue;
   }
   while (navNode != NULL) {
     g_string_append (outBuffer, "\n\t</clade>");
     if (g_node_next_sibling (navNode) != NULL) {
       navNode = g_node_next_sibling (navNode);
       break;
     }
     navNode = navNode -> parent;
   }
 }
 g_string_append (outBuffer, "\n</phylogeny>");
}

void WriteCladoTree (gchar* fileName, PHYLOTREE* tree) {
 gsize iniLen;
 GString* outBuffer;
 FILE* out;

 iniLen = (85 * GETNUMSPP);
 outBuffer = g_string_sized_new (iniLen);
 g_string_append (outBuffer, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
 g_string_append (outBuffer, "\n\n<Vip>");
 WriteCladoXMLTree (tree, outBuffer);
 g_string_append (outBuffer, "\n\n</Vip>\n");
 out = fopen (fileName, "w");
 if (out == NULL) {
   errorMessage = g_strdup_printf ("\nUnable to open %s\n", fileName);
   OnError ();
   g_string_free (outBuffer, TRUE);
   return;
 }
 if (fwrite (outBuffer -> str, sizeof (gchar), outBuffer -> len, out) != outBuffer -> len) {
   errorMessage = g_strdup_printf ("\nUnable to save whole data\nFile%s\n", fileName);
   OnError ();
 }
 fclose (out);
 g_string_free (outBuffer, TRUE);
}

void WriteCladoFile (gchar* fileName) {
 gsize iniLen;
 GString* outBuffer;
 GList* navTree;
 PHYLOTREE* tree;
 FILE* out;

 iniLen = (85 * GETNUMSPP);
 outBuffer = g_string_sized_new (iniLen);
 g_string_append (outBuffer, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
 g_string_append (outBuffer, "\n\n<Vip>");
 for (navTree = GETTREES; navTree != NULL; navTree = g_list_next (navTree)) {
   tree = (PHYLOTREE*) navTree -> data;
   WriteCladoXMLTree (tree, outBuffer);
 }
 g_string_append (outBuffer, "\n\n</Vip>\n");
 out = fopen (fileName, "w");
 if (out == NULL) {
   errorMessage = g_strdup_printf ("\nUnable to open %s\n", fileName);
   OnError ();
   g_string_free (outBuffer, TRUE);
   return;
 }
 if (fwrite (outBuffer -> str, sizeof (gchar), outBuffer -> len, out) != outBuffer -> len) {
   errorMessage = g_strdup_printf ("\nUnable to save whole data\nFile%s\n", fileName);
   OnError ();
 }
 fclose (out);
 g_string_free (outBuffer, TRUE);
}

void WriteCladoXMLTree (PHYLOTREE* tree, GString* outBuffer) {
 GNode* navNode;
 PHYLONODE* node;

 g_string_append (outBuffer, "\n<tree>");
 if (tree -> name != NULL)
   g_string_append_printf (outBuffer, "\n\t<name>%s</name>", tree -> name);
 for (navNode = tree -> root; navNode != NULL; ) {
   node = (PHYLONODE*) navNode -> data;
   g_string_append_printf (outBuffer, "\n\t<node id=\"%ld\">", node -> id);
   if (node -> name != NULL)
     g_string_append_printf (outBuffer, "\n\t\t<name>%s</name>", node -> name);
   if (node -> minAge > 0)
     g_string_append_printf (outBuffer, "\n\t\t<age>%ld</age>", node -> minAge);
   if (g_node_first_child (navNode) != NULL) {
     navNode = g_node_first_child (navNode);
     continue;
   }
   while (navNode != NULL) {
     g_string_append (outBuffer, "\n\t</node>");
     if (g_node_next_sibling (navNode) != NULL) {
       navNode = g_node_next_sibling (navNode);
       break;
     }
     navNode = navNode -> parent;
   }
 }
 g_string_append (outBuffer, "\n</tree>");
}

void WriteKMLRecons (gchar* fileName, GNode* refNode, SFLAG modeRecords, SFLAG modeRemoved) {
 gsize iniLen;
 GString* outBuffer;
 PHYLONODE* node;
 FILE* out;

 node = (PHYLONODE*) refNode -> data;
 iniLen = (85 * node -> numTerms) + (65 * GETNUMGEOREFS);
 outBuffer = g_string_sized_new (iniLen);
 g_string_append (outBuffer, "<?xml version=\"1.0\"?><kml xmlns=\"http://earth.google.com/kml/2.1\">");
 g_string_append (outBuffer, "\n\n<Document>");
 g_string_append_printf (outBuffer, "\n\t<name>%s</name>", fileName);
 g_string_append (outBuffer, "\n\t<description><![CDATA[<p>KML file generated by VIP (Arias, 2009)</p>");
 g_string_append (outBuffer, "\n\t\t<p>Useful links:<br>");
 g_string_append (outBuffer, "<a href=\"http://www.zmuc.dk/public/phylogeny/VIP\">");
 g_string_append (outBuffer, "www.zmuc.dk/public/phylogeny/VIP</a>: VIP download<br />");
 g_string_append (outBuffer, "\n\t\t<a href=\"mailto:sal.arias@gmail.com\">sal.arias@gmail.com</a>: ");
 g_string_append (outBuffer, "Salva e-mail</p>\n\t\t<p><b>Acknowledgment</b><br />Pablo, Claudia, ");
 g_string_append (outBuffer, "and Lolo, for suggestions and interest!</p>]]></description>\n");
 g_string_append (outBuffer, "\n");
 if (modeRecords == TRUE) {
   g_string_append (outBuffer,
                          "\n<Style id=\"overlapStyle\">\n\t<IconStyle>\n\t\t<color>ff00ff00</color>");
   g_string_append (outBuffer, "\n\t\t<Icon>\n\t\t\t<href>http://maps.google.com/mapfiles/kml/paddle/wht-blank.png</href>\n\t\t</Icon>");
   g_string_append (outBuffer, "\n\t</IconStyle>\n</Style>");
   g_string_append (outBuffer,
                           "\n<Style id=\"setAStyle\">\n\t<IconStyle>\n\t\t<color>ffff0000</color>\n\t</IconStyle>\n</Style>");
   g_string_append (outBuffer, "\n\t\t<Icon>\n\t\t\t<href>http://maps.google.com/mapfiles/kml/paddle/wht-blank.png</href>\n\t\t</Icon>");
   g_string_append (outBuffer, "\n\t</IconStyle>\n</Style>");
   g_string_append (outBuffer,
                           "\n<Style id=\"setBStyle\">\n\t<IconStyle>\n\t\t<color>ff0000ff</color>\n\t</IconStyle>\n</Style>");
   g_string_append (outBuffer, "\n\t\t<Icon>\n\t\t\t<href>http://maps.google.com/mapfiles/kml/paddle/wht-blank.png</href>\n\t\t</Icon>");
   g_string_append (outBuffer, "\n\t</IconStyle>\n</Style>");
   g_string_append (outBuffer,
                           "\n<Style id=\"removStyle\">\n\t<IconStyle>\n\t\t<color>ff000000</color>\n\t</IconStyle>\n</Style>");
   g_string_append (outBuffer, "\n\t\t<Icon>\n\t\t\t<href>http://maps.google.com/mapfiles/kml/paddle/wht-blank.png</href>\n\t\t</Icon>");
   g_string_append (outBuffer, "\n\t</IconStyle>\n</Style>");
 }
 else {
   g_string_append (outBuffer, "<Style id=\"overlapStyle\">\n\t<PolyStyle>\n\t\t<color>9900ff00</color>\n\t</PolyStyle>\n</Style>");
   g_string_append (outBuffer, "<Style id=\"setAStyle\">\n\t<PolyStyle>\n\t\t<color>99ff0000</color>\n\t</PolyStyle>\n</Style>");
   g_string_append (outBuffer, "<Style id=\"setBStyle\">\n\t<PolyStyle>\n\t\t<color>990000ff</color>\n\t</PolyStyle>\n</Style>");
 }
 g_string_append (outBuffer, "\n");
 if ((node -> rec != NULL) && (node -> rec -> nodeFlag & VIC_NODE)) {
   g_string_append_printf (outBuffer, "\n<Folder>");
   if (node -> name != NULL)
     g_string_append_printf (outBuffer, "\n\t<name>%s</name>", node -> name);
   g_string_append (outBuffer, "\n\t<description><![CDATA[<p>Vicariant distribution</p>]]></description>\n");
   WriteKMLSingle (node -> rec -> setA -> node, node -> rec -> setB -> rec -> cells, modeRecords, modeRemoved, "setAStyle",
                   outBuffer);
   WriteKMLSingle (node -> rec -> setB -> node, node -> rec -> setA -> rec -> cells, modeRecords, modeRemoved, "setBStyle",
                   outBuffer);
   g_string_append_printf (outBuffer, "\n</Folder>");
 }
 else
   WriteKMLSingle (refNode, NULL, modeRecords, modeRemoved, "overlapStyle", outBuffer);
 g_string_append (outBuffer, "\n\n</Document>\n</kml>\n");
 out = fopen (fileName, "w");
 if (out == NULL) {
   errorMessage = g_strdup_printf ("\nUnable to open %s\n", fileName);
   OnError ();
   g_string_free (outBuffer, TRUE);
   return;
 }
 if (fwrite (outBuffer -> str, sizeof (gchar), outBuffer -> len, out) != outBuffer -> len) {
   errorMessage = g_strdup_printf ("\nUnable to save whole data\nFile%s\n", fileName);
   OnError ();
 }
 fclose (out);
 g_string_free (outBuffer, TRUE);
}

void WriteKMLTree (gchar* fileName, PHYLOTREE* tree, SFLAG modeRecords, SFLAG modeRemoved) {
 gsize iniLen;
 GString* outBuffer;
 PHYLONODE* node;
 GNode* refNode;
 FILE* out;
 SCOUNTER i, c;

 node = (PHYLONODE*) tree -> root -> data;
 iniLen = (85 * node -> numTerms) + (65 * GETNUMGEOREFS);
 outBuffer = g_string_sized_new (iniLen);
 g_string_append (outBuffer, "<?xml version=\"1.0\"?><kml xmlns=\"http://earth.google.com/kml/2.1\">");
 g_string_append (outBuffer, "\n\n<Document>");
 g_string_append_printf (outBuffer, "\n\t<name>%s</name>", fileName);
 g_string_append (outBuffer, "\n\t<description><![CDATA[<p>KML file generated by VIP (Arias, 2009)</p>");
 g_string_append (outBuffer, "\n\t\t<p>Useful links:<br>");
 g_string_append (outBuffer, "<a href=\"http://www.zmuc.dk/public/phylogeny/VIP\">");
 g_string_append (outBuffer, "www.zmuc.dk/public/phylogeny/VIP</a>: VIP download<br />");
 g_string_append (outBuffer, "\n\t\t<a href=\"mailto:sal.arias@gmail.com\">sal.arias@gmail.com</a>: ");
 g_string_append (outBuffer, "Salva e-mail</p>\n\t\t<p><b>Acknowledgment</b><br />Pablo, Claudia, ");
 g_string_append (outBuffer, "and Lolo, for suggestions and interest!</p>]]></description>\n");
 g_string_append (outBuffer, "\n");
 if (modeRecords == TRUE) {
   g_string_append (outBuffer,
                          "\n<Style id=\"overlapStyle\">\n\t<IconStyle>\n\t\t<color>ff00ff00</color>");
   g_string_append (outBuffer, "\n\t\t<Icon>\n\t\t\t<href>http://maps.google.com/mapfiles/kml/paddle/wht-blank.png</href>\n\t\t</Icon>");
   g_string_append (outBuffer, "\n\t</IconStyle>\n</Style>");
   g_string_append (outBuffer,
                           "\n<Style id=\"setAStyle\">\n\t<IconStyle>\n\t\t<color>ffff0000</color>\n\t</IconStyle>\n</Style>");
   g_string_append (outBuffer, "\n\t\t<Icon>\n\t\t\t<href>http://maps.google.com/mapfiles/kml/paddle/wht-blank.png</href>\n\t\t</Icon>");
   g_string_append (outBuffer, "\n\t</IconStyle>\n</Style>");
   g_string_append (outBuffer,
                           "\n<Style id=\"setBStyle\">\n\t<IconStyle>\n\t\t<color>ff0000ff</color>\n\t</IconStyle>\n</Style>");
   g_string_append (outBuffer, "\n\t\t<Icon>\n\t\t\t<href>http://maps.google.com/mapfiles/kml/paddle/wht-blank.png</href>\n\t\t</Icon>");
   g_string_append (outBuffer, "\n\t</IconStyle>\n</Style>");
   g_string_append (outBuffer,
                           "\n<Style id=\"removStyle\">\n\t<IconStyle>\n\t\t<color>ff000000</color>\n\t</IconStyle>\n</Style>");
   g_string_append (outBuffer, "\n\t\t<Icon>\n\t\t\t<href>http://maps.google.com/mapfiles/kml/paddle/wht-blank.png</href>\n\t\t</Icon>");
   g_string_append (outBuffer, "\n\t</IconStyle>\n</Style>");
 }
 else {
   g_string_append (outBuffer, "<Style id=\"overlapStyle\">\n\t<PolyStyle>\n\t\t<color>9900ff00</color>\n\t</PolyStyle>\n</Style>");
   g_string_append (outBuffer, "<Style id=\"setAStyle\">\n\t<PolyStyle>\n\t\t<color>99ff0000</color>\n\t</PolyStyle>\n</Style>");
   g_string_append (outBuffer, "<Style id=\"setBStyle\">\n\t<PolyStyle>\n\t\t<color>990000ff</color>\n\t</PolyStyle>\n</Style>");
 }
 g_string_append (outBuffer, "\n");
 for (i = 0, c = 0; i < tree -> nodeArray -> len; ++ i) {
   refNode = (GNode*) g_ptr_array_index (tree -> nodeArray, i);
   node = (PHYLONODE*) refNode -> data;
   if (node -> rec -> nodeFlag & VIC_NODE) {
     ++ c;
     g_string_append_printf (outBuffer, "\n<Folder>");
     if (node -> name != NULL)
       g_string_append_printf (outBuffer, "\n\t<name>%s</name>", node -> name);
     g_string_append_printf (outBuffer, "\n\t<description><![CDATA[<p>Vicariant distribution %ld</p>]]></description>\n", c);
     WriteKMLSingle (node -> rec -> setA -> node, node -> rec -> setB -> rec -> cells, modeRecords, modeRemoved, "setAStyle",
                   outBuffer);
     WriteKMLSingle (node -> rec -> setB -> node, node -> rec -> setA -> rec -> cells, modeRecords, modeRemoved, "setBStyle",
                   outBuffer);
     g_string_append_printf (outBuffer, "\n</Folder>");
   }
 }
 g_string_append (outBuffer, "\n\n</Document>\n</kml>\n");
 out = fopen (fileName, "w");
 if (out == NULL) {
   errorMessage = g_strdup_printf ("\nUnable to open %s\n", fileName);
   OnError ();
   g_string_free (outBuffer, TRUE);
   return;
 }
 if (fwrite (outBuffer -> str, sizeof (gchar), outBuffer -> len, out) != outBuffer -> len) {
   errorMessage = g_strdup_printf ("\nUnable to save whole data\nFile%s\n", fileName);
   OnError ();
 }
 fclose (out);
 g_string_free (outBuffer, TRUE);
}

void WriteKMLSingle (GNode* refNode, BITFIELD* outCells, SFLAG modeRecords, SFLAG modeRemoved, gchar* style,
                     GString* outBuffer) {
 GNode* actNode;
 GList* navGeo;
 PHYLONODE* node;
 PHYLONODE* anc;
 TERMINAL* term;
 GEOREF* geo;
 BITFIELD* inCells;
 SFLAG in, out;
 SREAL x, x1, y, y1;
 SCOUNTER c, sX, sY;
 gchar* numStr;
 GString* date;

 node = (PHYLONODE*) refNode -> data;
 g_string_append_printf (outBuffer, "\n<Folder>");
 if (node -> name != NULL)
   g_string_append_printf (outBuffer, "\n\t<name>%s</name>", node -> name);
 if (modeRecords == TRUE) {
   if (node -> rec != NULL)
     inCells = node -> rec -> cells;
   else
     inCells = NULL;
   for (actNode = refNode; actNode != NULL;) {
     if (IsDesc (actNode, refNode) != TRUE) break;
     node = (PHYLONODE*) actNode -> data;
     if (actNode -> parent != NULL)
       anc = (PHYLONODE*) actNode -> parent -> data;
     if ((inCells != NULL) && (actNode != refNode) && ((node -> rec -> numCells == 0) ||
                 (node -> rec -> nodeFlag & REMOV_NODE) || (node == anc -> rec -> removSet))) {
       /* A removed distribution, ignores all of its descendants */
       while (actNode != NULL) {
         if (g_node_next_sibling (actNode) != NULL) {
           actNode = g_node_next_sibling (actNode);
           break;
         }
         actNode = actNode -> parent;
       }
       continue;
     }
     if (g_node_first_child (actNode) != NULL) {
     /* Not a term, goes to the first descendant */
       actNode = g_node_first_child (actNode);
       continue;
     }
     term = node -> term;
     for (navGeo = term -> firstGeo; navGeo != NULL; navGeo = g_list_next (navGeo)) {
       geo = (GEOREF*) navGeo -> data;
       if (term != (TERMINAL*) geo -> sp -> data) break;
       if (inCells != NULL) {
         in = IsBitOn (inCells, geo -> actCell);
         if (outCells != NULL)
           out = IsBitOn (outCells, geo -> actCell);
         else
           out = FALSE;
       }
       else {
         in = TRUE;
         out = FALSE;
       }
       if (in != TRUE) {
         if (modeRemoved == TRUE) {
           g_string_append (outBuffer, "\n\t<Placemark>");
           if (geo -> name != NULL)
             g_string_append_printf (outBuffer, "\n\t\t<name>%s</name>", geo -> name);
           g_string_append (outBuffer, "\n\t\t<styleUrl>#removStyle</styleUrl>");
         }
         else continue;
       }
       else if ((in == TRUE) && (out != TRUE)) {
         g_string_append (outBuffer, "\n\t<Placemark>");
         if (geo -> name != NULL)
           g_string_append_printf (outBuffer, "\n\t\t<name>%s</name>", geo -> name);
         g_string_append_printf (outBuffer, "\n\t\t<styleUrl>#%s</styleUrl>", style);
       }
       else {
         g_string_append (outBuffer, "\n\t<Placemark>");
         if (geo -> name != NULL)
           g_string_append_printf (outBuffer, "\n\t\t<name>%s</name>", geo -> name);
         g_string_append (outBuffer, "\n\t\t<styleUrl>#overlapStyle</styleUrl>");
       }
       g_string_append (outBuffer, "\n\t\t<description><![CDATA[<p>");
       if (geo -> name != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\tTerminal: %s<br />", geo -> name);
       if (geo -> collec != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\tCollection: %s<br />", geo -> collec);
       if (geo -> catalog != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\tCatalogue code: %s <br />", geo -> catalog);
       if (geo -> url != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\tUrl: <a hfer=\"%s\">%s</a><br />", geo -> url, geo -> url);
       if (geo -> reference != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\tReference: %s<br />", geo -> reference);
       if (geo -> identifier != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\tIdentified by: %s<br />", geo -> identifier);
       if (geo -> sampleDate != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\tCollection date: %s<br />", geo -> sampleDate);
       if (geo -> collector != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\tCollector: %s<br />", geo -> collector);
       g_string_append (outBuffer, "\n\t\t</p>]]></description>");
       if (geo -> sampleDate != NULL) {
         date = DateString (geo -> sampleDate);
         g_string_append (outBuffer, "\n\t\t<TimeStap>");
         g_string_append_printf (outBuffer, "\n\t\t\t<when>%s</when>", date -> str);
         g_string_append (outBuffer, "\n\t\t</TimeStap>");
         g_string_free (date, TRUE);
       }
       g_string_append (outBuffer, "\n\t\t<Point>");
       g_string_append (outBuffer, "\n\t\t\t<extrude>0</extrude>");
       g_string_append (outBuffer, "\n\t\t\t<altitudeMode>absolute</altitudeMode>");
       g_string_append (outBuffer, "\n\t\t\t<coordinates>");
       numStr = g_strdup_printf ("%.6f", geo -> lon);
       CheckRealStr (numStr);
       g_string_append_printf (outBuffer, "%s,", numStr);
       g_free (numStr);
       numStr = g_strdup_printf ("%.6f", geo -> lat);
       CheckRealStr (numStr);
       g_string_append (outBuffer, numStr);
       g_free (numStr);
       g_string_append (outBuffer, "</coordinates>");
       g_string_append (outBuffer, "\n\t\t</Point>\n\t</Placemark>");
     }
     while (actNode != NULL) {
       if (g_node_next_sibling (actNode) != NULL) {
         actNode = g_node_next_sibling (actNode);
         break;
       }
       actNode = actNode -> parent;
     }
   }
 }
 else {
   for (c = 0; c < GETNUMCELLS; ++ c) {
     if (IsBitOn (node -> rec -> cells, c)) {
       GetColRowFromCell (c, &sY, &sX);
       x = ((GETGRIDSIZE * sX) + GETGRIDSTARTX);
       x = FROMAZMTOLON (x);
       x1 = ((GETGRIDSIZE * (sX + 1)) + GETGRIDSTARTX);
       x1 = FROMAZMTOLON (x1);
       y = ((GETGRIDSIZE * sY) + GETGRIDSTARTY);
       y = FROMINCTOLAT (y);
       y1 = ((GETGRIDSIZE * (sY + 1)) + GETGRIDSTARTY);
       y1 = FROMINCTOLAT (y1);
       g_string_append (outBuffer, "\n\t<Placemark>");
       if (node -> name != NULL)
          g_string_append_printf (outBuffer, "\n\t\t<name>%s</name>", node -> name);
       if (outCells != NULL)
         out = IsBitOn (outCells, c);
       else
         out = FALSE;
       if (out != TRUE)
         g_string_append_printf (outBuffer, "\n\t\t<styleUrl>#%s</styleUrl>", style);
       else
         g_string_append (outBuffer, "\n\t\t<styleUrl>#overlapStyle</styleUrl>");
       g_string_append (outBuffer, "\n\t\t<Polygon>");
       g_string_append (outBuffer, "\n\t\t\t<outerBoundaryIs>\n\t\t\t\t<LinearRing>\n\t\t\t\t\t<coordinates>");
       numStr = g_strdup_printf ("%.6f", x);
       CheckRealStr (numStr);
       g_string_append_printf (outBuffer, "\n\t\t\t\t\t\t%s,", numStr);
       g_free (numStr);
       numStr = g_strdup_printf ("%.6f", y);
       CheckRealStr (numStr);
       g_string_append_printf (outBuffer, "%s", numStr);
       g_free (numStr);
       numStr = g_strdup_printf ("%.6f", x1);
       CheckRealStr (numStr);
       g_string_append_printf (outBuffer, "\n\t\t\t\t\t\t%s,", numStr);
       g_free (numStr);
       numStr = g_strdup_printf ("%.6f", y);
       CheckRealStr (numStr);
       g_string_append_printf (outBuffer, "%s", numStr);
       g_free (numStr);
       numStr = g_strdup_printf ("%.6f", x1);
       CheckRealStr (numStr);
       g_string_append_printf (outBuffer, "\n\t\t\t\t\t\t%s,", numStr);
       g_free (numStr);
       numStr = g_strdup_printf ("%.6f", y1);
       CheckRealStr (numStr);
       g_string_append_printf (outBuffer, "%s", numStr);
       g_free (numStr);
       numStr = g_strdup_printf ("%.6f", x);
       CheckRealStr (numStr);
       g_string_append_printf (outBuffer, "\n\t\t\t\t\t\t%s,", numStr);
       g_free (numStr);
       numStr = g_strdup_printf ("%.6f", y1);
       CheckRealStr (numStr);
       g_string_append_printf (outBuffer, "%s", numStr);
       g_free (numStr);
       numStr = g_strdup_printf ("%.6f", x);
       CheckRealStr (numStr);
       g_string_append_printf (outBuffer, "\n\t\t\t\t\t\t%s,", numStr);
       g_free (numStr);
       numStr = g_strdup_printf ("%.6f", y);
       CheckRealStr (numStr);
       g_string_append_printf (outBuffer, "%s", numStr);
       g_free (numStr);
       g_string_append (outBuffer, "\n\t\t\t\t\t</coordinates>\n\t\t\t\t</LinearRing>\n\t\t\t</outerBoundaryIs>");
       g_string_append (outBuffer, "\n\t\t</Polygon>\n\t</Placemark>");
     }
   }
 }
 g_string_append_printf (outBuffer, "\n</Folder>");
}

void WriteNewickFile (gchar* fileName, PHYLOTREE* tree, SFLAG mode) {
 gsize iniLen;
 GString* outBuffer;
 FILE* out;
 GNode* navNode;
 PHYLONODE* node;
 PHYLONODE* anc;
 SCOUNTER i, id, terms;
 INTARRAY* termIds;

 if (tree == NULL) return;
 iniLen = (85 * GETNUMSPP);
 outBuffer = g_string_sized_new (iniLen);
 node = (PHYLONODE*) tree -> root -> data;

 if (mode & USE_NEXUS) {
   g_string_append (outBuffer, "#NEXUS\n");
   if (tree -> name != NULL)
     g_string_append_printf (outBuffer, "\n[ %s ]", tree -> name);
   g_string_append (outBuffer, "\nbegin trees;");
   if (!(mode & USE_NAMES))
     g_string_append (outBuffer, "\ntranslate");
 }
 if (mode == USE_TNT) {
   g_string_append (outBuffer, "xread");
   if (tree -> name != NULL)
     g_string_append_printf (outBuffer, "\n\' %s \'", tree -> name);
   g_string_append_printf (outBuffer, "1 %ld", node -> numTerms);
 }
 terms = node -> numTerms;
 termIds = CreateIntArray (terms);
 for (i = 0; i < tree -> nodeArray -> len; ++ i) {
   navNode = (GNode*) g_ptr_array_index (tree -> nodeArray, i);
   if (g_node_first_child (navNode) != NULL) continue;
   node = (PHYLONODE*) navNode -> data;
   AddToArray (termIds, node -> id);
 }
 if (!(mode & USE_NAMES)) {
   for (i = 0; i < terms; ++ i) {
     id = GetFromArray (termIds, i);
     node = GetNodeFromID (tree, id);
     if (mode == USE_NEXUS) {
       if (node -> name != NULL)
         g_string_append_printf (outBuffer, "\n\t%ld\t%s", i + 1, node -> name);
       else
         g_string_append_printf (outBuffer, "\n\t%ld\tnode_%ld", i + 1, id);
       if ((i + 1) < terms)
         g_string_append (outBuffer, ",");
     }
     if (mode == USE_TNT) {
       if (node -> name != NULL)
         g_string_append_printf (outBuffer, "\n%s          ?", node -> name);
       else
         g_string_append_printf (outBuffer, "\nnode_%ld          ?", id);
     }
   }
   g_string_append (outBuffer, ";");
 }
 if (mode & USE_NEXUS)
   g_string_append_printf (outBuffer, "\n\n\ttree %s = ", (tree -> name != NULL) ? tree -> name : "unnamedtree");
 if (mode & USE_TNT)
   g_string_append (outBuffer, "\n\ntread\n");
 for (navNode = tree -> root; navNode != NULL; ) {
   if (g_node_first_child (navNode) != NULL) {
     g_string_append (outBuffer, "(");
     navNode = g_node_first_child (navNode);
     continue;
   }
   node = (PHYLONODE*) navNode -> data;
   i = SearchIndexValue (termIds, node -> id);
   if (mode == USE_NEXUS)
     g_string_append_printf (outBuffer, "%ld", i + 1);
   if (mode == USE_TNT)
     g_string_append_printf (outBuffer, " %ld", i);
   if (mode & USE_NAMES) {
     if (mode & USE_TNT)
       g_string_append (outBuffer, " ");
     if (node -> name != NULL)
       g_string_append_printf (outBuffer, "%s", node -> name);
     else
       g_string_append_printf (outBuffer, "node_%ld", id);
   }
   while (navNode != NULL) {
     if  ((mode & USE_NEXUS) && (node -> minAge > 0) && (navNode -> parent != NULL) && (g_node_first_child (navNode) == NULL)) {
       anc = (PHYLONODE*) navNode -> parent -> data;
       g_string_append_printf (outBuffer, ":%ld", anc -> minAge - node -> minAge);
     }
     if (g_node_first_child (navNode) != NULL)
       g_string_append (outBuffer, ")");
     if  ((mode & USE_NEXUS) && (node -> minAge > 0) && (navNode -> parent != NULL) && (g_node_first_child (navNode) != NULL)) {
       anc = (PHYLONODE*) navNode -> parent -> data;
       g_string_append_printf (outBuffer, ":%ld", anc -> minAge - node -> minAge);
     }
     if (g_node_next_sibling (navNode) != NULL) {
       if (mode & USE_NEXUS)
         g_string_append (outBuffer, ",");
       navNode = g_node_next_sibling (navNode);
       break;
     }
     navNode = navNode -> parent;
   }
 }
 g_string_append (outBuffer, ";\n");
 if (mode & USE_NEXUS)
   g_string_append_printf (outBuffer, "end;\n");
 if (mode & USE_TNT)
   g_string_append (outBuffer, "p/;\n");
 DestroyIntArray (termIds);
 out = fopen (fileName, "w");
 if (out == NULL) {
   errorMessage = g_strdup_printf ("\nUnable to open %s\n", fileName);
   OnError ();
   g_string_free (outBuffer, TRUE);
   return;
 }
 if (fwrite (outBuffer -> str, sizeof (gchar), outBuffer -> len, out) != outBuffer -> len) {
   errorMessage = g_strdup_printf ("\nUnable to save whole data\nFile%s\n", fileName);
   OnError ();
 }
 fclose (out);
 g_string_free (outBuffer, TRUE);
}

SCOUNTER CalculateRichness (INTARRAY* sppPerCell, INTARRAY* obsPerCell) {
 SCOUNTER cell, num, max;
 GList* spNav;
 TERMINAL* actSp;
 PHYLONODE* actNode;

 for (cell = 0; cell < GETNUMCELLS; ++ cell) {
   if (sppPerCell != NULL)
     AddToArray (sppPerCell, 0);
   AddToArray (obsPerCell, 0);
 }
 max = 0;
 for (spNav = GETSPP; spNav != NULL; spNav = g_list_next (spNav)) {
   actSp = (TERMINAL*) spNav -> data;
   actNode = (PHYLONODE*) actSp -> node -> data;
   for (cell = 0; cell < GETNUMCELLS; ++ cell) {
     if (!IsBitOn (actNode -> rec -> cells, cell)) continue;
     if (IsBitOn (actSp -> obs, cell))
       InsertValToArray (obsPerCell, 1, cell);
     if (sppPerCell != NULL) {
       num = GetFromArray (sppPerCell, cell) + 1;
       InsertValToArray (sppPerCell, num, cell);
       if (num > max)
         max = num;
     }
     else {
       if (IsBitOn (actNode -> rec -> cells, cell))
         InsertValToArray (obsPerCell, 1, cell);
     }
   }
 }
 return max;
}

void WriteNDMMatrix (gchar* fileName) {
 gsize iniLen;
 GString* outBuffer;
 FILE* out;
 SCOUNTER c, r, cell;
 INTARRAY* obsPerCell;
 GList* spNav;
 GList* navTree;
 PHYLOTREE* actTree;
 TERMINAL* actSp;
 PHYLONODE* actNode;
 gchar* text;

 obsPerCell = CreateIntArray (GETNUMCELLS);
 for (navTree = GETTREES; navTree != NULL; navTree = g_list_next (navTree)) {
   actTree = (PHYLOTREE*) navTree -> data;
   SetRec (actTree -> recOR);
 }
 CalculateRichness (NULL, obsPerCell);
 iniLen = (GETNUMCELLS * GETNUMSPP);
 outBuffer = g_string_sized_new (iniLen);

 text = g_strdup_printf ("%.4f", FROMAZMTOLON (GETGRIDSTARTX));
 CheckRealStr (text);
 g_string_append_printf (outBuffer, "gridx %s ", text);
 g_free (text);
 text = g_strdup_printf ("%.4f", GETGRIDSIZE);
 CheckRealStr (text);
 g_string_append_printf (outBuffer, "%s\n", text);
 g_free (text);
 text = g_strdup_printf ("%.4f", FROMINCTOLAT (GETGRIDSTARTY));
 CheckRealStr (text);
 g_string_append_printf (outBuffer, "gridy %s ", text);
 g_free (text);
 text = g_strdup_printf ("%.4f", GETGRIDSIZE);
 CheckRealStr (text);
 g_string_append_printf (outBuffer, "%s\n", text);
 g_free (text);
 g_string_append_printf (outBuffer, "rows %ld\ncols %ld\nspp%ld\ndata", GETNUMROWS, GETNUMCOLS, GETNUMSPP);
 for (r = 0; r < GETNUMROWS; ++ r) {
   for (c = 0; c < GETNUMCOLS; ++ c) {
     cell = GetCell (r, c);
     if ((cell == -1) || (GetFromArray (obsPerCell, cell) == 0)) continue;
     g_string_append_printf (outBuffer, "\nA%ld-%ld\t", r, c);
     for (spNav = GETSPP; spNav != NULL; spNav = g_list_next (spNav)) {
       actSp = (TERMINAL*) spNav -> data;
       actNode = (PHYLONODE*) actSp -> node -> data;
       if (IsBitOn (actNode -> rec -> cells, cell))
         g_string_append (outBuffer, "1");
       else
         g_string_append (outBuffer, "0");
     }
   }
 }
 g_string_append (outBuffer, "\n;\n");
 DestroyIntArray (obsPerCell);
 out = fopen (fileName, "w");
 if (out == NULL) {
   errorMessage = g_strdup_printf ("\nUnable to open %s\n", fileName);
   OnError ();
   g_string_free (outBuffer, TRUE);
   return;
 }
 if (fwrite (outBuffer -> str, sizeof (gchar), outBuffer -> len, out) != outBuffer -> len) {
   errorMessage = g_strdup_printf ("\nUnable to save whole data\nFile%s\n", fileName);
   OnError ();
 }
 fclose (out);
 g_string_free (outBuffer, TRUE);
}

void WriteSpRichness (gchar* fileName) {
 gsize iniLen;
 GString* outBuffer;
 FILE* out;
 SCOUNTER c, r;
 SCOUNTER cell, max;
 INTARRAY* sppPerCell;
 INTARRAY* obsPerCell;
 GList* navTree;
 PHYLOTREE* actTree;
 SREAL lat, lon;
 gchar* numStr;

 sppPerCell = CreateIntArray (GETNUMCELLS);
 obsPerCell = CreateIntArray (GETNUMCELLS);
 for (navTree = GETTREES; navTree != NULL; navTree = g_list_next (navTree)) {
   actTree = (PHYLOTREE*) navTree -> data;
   SetRec (actTree -> recOR);
 }
 max = CalculateRichness (sppPerCell, obsPerCell);
 iniLen = (3 * GETNUMROWS) * (3 * GETNUMCOLS);
 outBuffer = g_string_sized_new (iniLen);
 g_string_append (outBuffer, "R-C\t");
 for (c = 0; c < GETNUMCOLS; ++ c) {
   lon = GETGRIDSTARTX + (GETGRIDSIZE * c);
   numStr = g_strdup_printf ("%.3f", FROMAZMTOLON (lon));
   CheckRealStr (numStr);
   g_string_append_printf (outBuffer, "%s\t", numStr);
   g_free (numStr);
 }
 for (r = 0; r < GETNUMROWS; ++ r) {
   lat = GETGRIDSTARTY + (GETGRIDSIZE * r);
   numStr = g_strdup_printf ("%.3f", FROMINCTOLAT (lat));
   CheckRealStr (numStr);
   g_string_append_printf (outBuffer, "%s\t", numStr);
   g_free (numStr);
   for (c = 0; c < GETNUMCOLS; ++ c) {
     cell = GetCell (r, c);
     if ((cell == -1) || (GetFromArray (obsPerCell, cell) == 0)) {
       g_string_append (outBuffer, "0\t");
       continue;
     }
     numStr = g_strdup_printf ("%.4f", (SREAL) GetFromArray (sppPerCell, cell) / max);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer, "%s\t", numStr);
     g_free (numStr);
   }
 }
 DestroyIntArray (sppPerCell);
 DestroyIntArray (obsPerCell);
 out = fopen (fileName, "w");
 if (out == NULL) {
   errorMessage = g_strdup_printf ("\nUnable to open %s\n", fileName);
   OnError ();
   g_string_free (outBuffer, TRUE);
   return;
 }
 if (fwrite (outBuffer -> str, sizeof (gchar), outBuffer -> len, out) != outBuffer -> len) {
   errorMessage = g_strdup_printf ("\nUnable to save whole data\nFile%s\n", fileName);
   OnError ();
 }
 fclose (out);
 g_string_free (outBuffer, TRUE);
}

SCOUNTER CalculatePD (INTARRAY* pdPerCell, INTARRAY* obsPerCell) {
 SCOUNTER cell, max, toRet, num;
 GList* spNav;
 TERMINAL* actSp;
 GList* navTree;
 PHYLOTREE* actTree;
 PHYLONODE* root;

 for (cell = 0; cell < GETNUMCELLS; ++ cell)
   AddToArray (obsPerCell, 0);
 for (spNav = GETSPP; spNav != NULL; spNav = g_list_next (spNav)) {
   actSp = (TERMINAL*) spNav -> data;
   for (cell = 0; cell < GETNUMCELLS; ++ cell) {
     if (!IsBitOn (actSp -> obs, cell)) continue;
     InsertValToArray (obsPerCell, 1, cell);
   }
 }
 max = 0;
 for (navTree = GETTREES; navTree != NULL; navTree = g_list_next (navTree)) {
   actTree = (PHYLOTREE*) navTree -> data;
   root = (PHYLONODE*) actTree -> root -> data;
   if (root -> level > max)
     max = root -> level;
   SetRec (actTree -> recOR);
 }
 toRet = 0;
 for (cell = 0; cell < GETNUMCELLS; ++ cell) {
   num = CalculatePhyloDiv (cell, max);
   AddToArray (pdPerCell, num);
   InsertValToArray (pdPerCell, num, cell);
   if (num > toRet)
     toRet = num;
 }
 return toRet;
}

void WritePhyloDiv (gchar* fileName) {
 gsize iniLen;
 GString* outBuffer;
 FILE* out;
 SCOUNTER c, r;
 SCOUNTER cell, max;
 INTARRAY* obsPerCell;
 INTARRAY* pdPerCell;
 SREAL lat, lon;
 gchar* numStr;

 pdPerCell = CreateIntArray (GETNUMCELLS);
 obsPerCell = CreateIntArray (GETNUMCELLS);
 max = CalculatePD (pdPerCell, obsPerCell);
 iniLen = (3 * GETNUMROWS) * (3 * GETNUMCOLS);
 outBuffer = g_string_sized_new (iniLen);
 g_string_append (outBuffer, "R-C\t");
 for (c = 0; c < GETNUMCOLS; ++ c) {
   lon = GETGRIDSTARTX + (GETGRIDSIZE * c);
   numStr = g_strdup_printf ("%.3f", FROMAZMTOLON (lon));
   CheckRealStr (numStr);
   g_string_append_printf (outBuffer, "%s\t", numStr);
   g_free (numStr);
 }
 for (r = 0; r < GETNUMROWS; ++ r) {
   lat = GETGRIDSTARTY + (GETGRIDSIZE * r);
   numStr = g_strdup_printf ("%.3f", FROMINCTOLAT (lat));
   CheckRealStr (numStr);
   g_string_append_printf (outBuffer, "%s\t", numStr);
   g_free (numStr);
   for (c = 0; c < GETNUMCOLS; ++ c) {
     cell = GetCell (r, c);
     if ((cell == -1) || (GetFromArray (obsPerCell, cell) == 0)) {
       g_string_append (outBuffer, "0\t");
       continue;
     }
     numStr = g_strdup_printf ("%.4f", (SREAL) GetFromArray (pdPerCell, cell) / max);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer, "%s\t", numStr);
     g_free (numStr);
   }
 }
 DestroyIntArray (pdPerCell);
 DestroyIntArray (obsPerCell);
 out = fopen (fileName, "w");
 if (out == NULL) {
   errorMessage = g_strdup_printf ("\nUnable to open %s\n", fileName);
   OnError ();
   g_string_free (outBuffer, TRUE);
   return;
 }
 if (fwrite (outBuffer -> str, sizeof (gchar), outBuffer -> len, out) != outBuffer -> len) {
   errorMessage = g_strdup_printf ("\nUnable to save whole data\nFile%s\n", fileName);
   OnError ();
 }
 fclose (out);
 g_string_free (outBuffer, TRUE);
}

void WriteRichKMLFile (gchar* fileName) {
 gsize iniLen;
 GString* outBuffer;
 FILE* out;
 gchar hexStr [3];
 INTARRAY* sppPerCell;
 INTARRAY* obsPerCell;
 SCOUNTER i, cell, max, num, r, c;
 SREAL x, x1, y, y1;
 gchar* numStr;
 GList* navTree;
 PHYLOTREE* actTree;

 sppPerCell = CreateIntArray (GETNUMCELLS);
 obsPerCell = CreateIntArray (GETNUMCELLS);
 for (navTree = GETTREES; navTree != NULL; navTree = g_list_next (navTree)) {
   actTree = (PHYLOTREE*) navTree -> data;
   SetRec (actTree -> recOR);
 }
 max = CalculateRichness (sppPerCell, obsPerCell);
 iniLen = (80 * GETNUMROWS * GETNUMCOLS);
 outBuffer = g_string_sized_new (iniLen);
 g_string_append (outBuffer, "<?xml version=\"1.0\"?><kml xmlns=\"http://earth.google.com/kml/2.1\">");
 g_string_append (outBuffer, "\n\n<Document>");
 g_string_append_printf (outBuffer, "\n\t<name>%s</name>", fileName);
 g_string_append (outBuffer, "\n\t<description><![CDATA[<p>KML file generated by VIP (Arias, 2009)</p>");
 g_string_append (outBuffer, "\n\t\t<p>Useful links:<br>");
 g_string_append (outBuffer, "<a href=\"http://www.zmuc.dk/public/phylogeny/VIP\">");
 g_string_append (outBuffer, "www.zmuc.dk/public/phylogeny/VIP</a>: VIP download<br />");
 g_string_append (outBuffer, "\n\t\t<a href=\"mailto:sal.arias@gmail.com\">sal.arias@gmail.com</a>: ");
 g_string_append (outBuffer, "Salva e-mail</p>\n\t\t<p><b>Acknowledgment</b><br />Pablo, Claudia, ");
 g_string_append (outBuffer, "and Lolo, for suggestions and interest!</p>]]></description>\n");
 g_string_append (outBuffer, "\n");
 for (i = 0; i < 100; ++ i) {
   g_string_append_printf (outBuffer, "<Style id=\"val%ld\">\n\t<PolyStyle>", i + 1);
   if (i < 50) {
     WriteInHex (hexStr, 250 - (5 * i));
     g_string_append_printf (outBuffer, "\n\t\t<color>99%s", hexStr);
     if (i >= 25) {
       WriteInHex (hexStr, 10 * (i - 25));
       g_string_append_printf (outBuffer, "%s00</color>\n\t</PolyStyle>\n</Style>", hexStr);
     }
     else
       g_string_append (outBuffer, "0000</color>\n\t</PolyStyle>\n</Style>");
   }
   else {
     if (i < 75) {
       WriteInHex (hexStr, 250 - (10 * (i - 50)));
       g_string_append_printf (outBuffer, "\n\t\t<color>9900%s", hexStr);
     }
     else
       g_string_append (outBuffer, "\n\t\t<color>990000");
     WriteInHex (hexStr, 5 * (i - 50));
     g_string_append_printf (outBuffer, "%s</color>\n\t</PolyStyle>\n</Style>", hexStr);
   }
 }
 g_string_append (outBuffer, "\n\n<Folder>");
 for (r = 0; r < GETNUMROWS; ++ r) {
   for (c = 0; c < GETNUMCOLS; ++ c) {
     cell = GetCell (r, c);
     if ((cell == -1) || (GetFromArray (obsPerCell, cell) == 0)) continue;
     x = ((GETGRIDSIZE * c) + GETGRIDSTARTX);
     x = FROMAZMTOLON (x);
     x1 = ((GETGRIDSIZE * (c + 1)) + GETGRIDSTARTX);
     x1 = FROMAZMTOLON (x1);
     y = ((GETGRIDSIZE * r) + GETGRIDSTARTY);
     y = FROMINCTOLAT (y);
     y1 = ((GETGRIDSIZE * (r + 1)) + GETGRIDSTARTY);
     y1 = FROMINCTOLAT (y1);
     g_string_append (outBuffer, "\n\t<Placemark>");
     g_string_append_printf (outBuffer, "\n\t\t<name>cell%ld</name>", cell);
     num = GetFromArray (sppPerCell, cell);
     numStr = g_strdup_printf ("%.3f", (SREAL) num / max);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer,
            "\n\t\t<description><![CDATA[<p>Richness = %ld (scaled = %s)</p>]]></description>", num, numStr);
     g_free (numStr);
     num = (num * 100) / max;
     g_string_append_printf (outBuffer, "\n\t\t<styleUrl>#val%ld</styleUrl>", num);
     g_string_append (outBuffer, "\n\t\t<Polygon>");
     g_string_append (outBuffer, "\n\t\t\t<outerBoundaryIs>\n\t\t\t\t<LinearRing>\n\t\t\t\t\t<coordinates>");
     numStr = g_strdup_printf ("%.6f", x);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer, "\n\t\t\t\t\t\t%s,", numStr);
     g_free (numStr);
     numStr = g_strdup_printf ("%.6f", y);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer, "%s", numStr);
     g_free (numStr);
     numStr = g_strdup_printf ("%.6f", x1);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer, "\n\t\t\t\t\t\t%s,", numStr);
     g_free (numStr);
     numStr = g_strdup_printf ("%.6f", y);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer, "%s", numStr);
     g_free (numStr);
     numStr = g_strdup_printf ("%.6f", x1);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer, "\n\t\t\t\t\t\t%s,", numStr);
     g_free (numStr);
     numStr = g_strdup_printf ("%.6f", y1);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer, "%s", numStr);
     g_free (numStr);
     numStr = g_strdup_printf ("%.6f", x);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer, "\n\t\t\t\t\t\t%s,", numStr);
     g_free (numStr);
     numStr = g_strdup_printf ("%.6f", y1);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer, "%s", numStr);
     g_free (numStr);
     numStr = g_strdup_printf ("%.6f", x);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer, "\n\t\t\t\t\t\t%s,", numStr);
     g_free (numStr);
     numStr = g_strdup_printf ("%.6f", y);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer, "%s", numStr);
     g_free (numStr);
     g_string_append (outBuffer, "\n\t\t\t\t\t</coordinates>\n\t\t\t\t</LinearRing>\n\t\t\t</outerBoundaryIs>");
     g_string_append (outBuffer, "\n\t\t</Polygon>\n\t</Placemark>");
   }
 }
 g_string_append_printf (outBuffer, "\n</Folder>");
 g_string_append (outBuffer, "\n\n</Document>\n</kml>\n");
 DestroyIntArray (sppPerCell);
 DestroyIntArray (obsPerCell);
 out = fopen (fileName, "w");
 if (out == NULL) {
   errorMessage = g_strdup_printf ("\nUnable to open %s\n", fileName);
   OnError ();
   g_string_free (outBuffer, TRUE);
   return;
 }
 if (fwrite (outBuffer -> str, sizeof (gchar), outBuffer -> len, out) != outBuffer -> len) {
   errorMessage = g_strdup_printf ("\nUnable to save whole data\nFile%s\n", fileName);
   OnError ();
 }
 fclose (out);
 g_string_free (outBuffer, TRUE);
}

void WritePDKMLFile (gchar* fileName) {
 gsize iniLen;
 GString* outBuffer;
 FILE* out;
 gchar hexStr [3];
 INTARRAY* pdPerCell;
 INTARRAY* obsPerCell;
 SCOUNTER i, cell, max, num, r, c;
 SREAL x, x1, y, y1;
 gchar* numStr;

 pdPerCell = CreateIntArray (GETNUMCELLS);
 obsPerCell = CreateIntArray (GETNUMCELLS);
 max = CalculatePD (pdPerCell, obsPerCell);
 iniLen = (80 * GETNUMROWS * GETNUMCOLS);
 outBuffer = g_string_sized_new (iniLen);
 g_string_append (outBuffer, "<?xml version=\"1.0\"?><kml xmlns=\"http://earth.google.com/kml/2.1\">");
 g_string_append (outBuffer, "\n\n<Document>");
 g_string_append_printf (outBuffer, "\n\t<name>%s</name>", fileName);
 g_string_append (outBuffer, "\n\t<description><![CDATA[<p>KML file generated by VIP (Arias, 2009)</p>");
 g_string_append (outBuffer, "\n\t\t<p>Useful links:<br>");
 g_string_append (outBuffer, "<a href=\"http://www.zmuc.dk/public/phylogeny/VIP\">");
 g_string_append (outBuffer, "www.zmuc.dk/public/phylogeny/VIP</a>: VIP download<br />");
 g_string_append (outBuffer, "\n\t\t<a href=\"mailto:sal.arias@gmail.com\">sal.arias@gmail.com</a>: ");
 g_string_append (outBuffer, "Salva e-mail</p>\n\t\t<p><b>Acknowledgment</b><br />Pablo, Claudia, ");
 g_string_append (outBuffer, "and Lolo, for suggestions and interest!</p>]]></description>\n");
 g_string_append (outBuffer, "\n");
 for (i = 0; i < 100; ++ i) {
   g_string_append_printf (outBuffer, "<Style id=\"val%ld\">\n\t<PolyStyle>", i + 1);
   if (i < 50) {
     WriteInHex (hexStr, 250 - (5 * i));
     g_string_append_printf (outBuffer, "\n\t\t<color>99%s", hexStr);
     if (i >= 25) {
       WriteInHex (hexStr, 10 * (i - 25));
       g_string_append_printf (outBuffer, "%s00</color>\n\t</PolyStyle>\n</Style>", hexStr);
     }
     else
       g_string_append (outBuffer, "0000</color>\n\t</PolyStyle>\n</Style>");
   }
   else {
     if (i < 75) {
       WriteInHex (hexStr, 250 - (10 * (i - 50)));
       g_string_append_printf (outBuffer, "\n\t\t<color>9900%s", hexStr);
     }
     else
       g_string_append (outBuffer, "\n\t\t<color>990000");
     WriteInHex (hexStr, 5 * (i - 50));
     g_string_append_printf (outBuffer, "%s</color>\n\t</PolyStyle>\n</Style>", hexStr);
   }
 }
 g_string_append (outBuffer, "\n\n<Folder>");
 for (r = 0; r < GETNUMROWS; ++ r) {
   for (c = 0; c < GETNUMCOLS; ++ c) {
     cell = GetCell (r, c);
     if ((cell == -1) || (GetFromArray (obsPerCell, cell) == 0)) continue;
     x = ((GETGRIDSIZE * c) + GETGRIDSTARTX);
     x = FROMAZMTOLON (x);
     x1 = ((GETGRIDSIZE * (c + 1)) + GETGRIDSTARTX);
     x1 = FROMAZMTOLON (x1);
     y = ((GETGRIDSIZE * r) + GETGRIDSTARTY);
     y = FROMINCTOLAT (y);
     y1 = ((GETGRIDSIZE * (r + 1)) + GETGRIDSTARTY);
     y1 = FROMINCTOLAT (y1);
     g_string_append (outBuffer, "\n\t<Placemark>");
     g_string_append_printf (outBuffer, "\n\t\t<name>cell%ld</name>", cell);
     num = GetFromArray (pdPerCell, cell);
     numStr = g_strdup_printf ("%.3f", (SREAL) num / max);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer,
            "\n\t\t<description><![CDATA[<p>PD = %ld (scaled = %s)</p>]]></description>", num, numStr);
     g_free (numStr);
     num = (num * 100) / max;
     g_string_append_printf (outBuffer, "\n\t\t<styleUrl>#val%ld</styleUrl>", num);
     g_string_append (outBuffer, "\n\t\t<Polygon>");
     g_string_append (outBuffer, "\n\t\t\t<outerBoundaryIs>\n\t\t\t\t<LinearRing>\n\t\t\t\t\t<coordinates>");
     numStr = g_strdup_printf ("%.6f", x);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer, "\n\t\t\t\t\t\t%s,", numStr);
     g_free (numStr);
     numStr = g_strdup_printf ("%.6f", y);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer, "%s", numStr);
     g_free (numStr);
     numStr = g_strdup_printf ("%.6f", x1);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer, "\n\t\t\t\t\t\t%s,", numStr);
     g_free (numStr);
     numStr = g_strdup_printf ("%.6f", y);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer, "%s", numStr);
     g_free (numStr);
     numStr = g_strdup_printf ("%.6f", x1);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer, "\n\t\t\t\t\t\t%s,", numStr);
     g_free (numStr);
     numStr = g_strdup_printf ("%.6f", y1);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer, "%s", numStr);
     g_free (numStr);
     numStr = g_strdup_printf ("%.6f", x);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer, "\n\t\t\t\t\t\t%s,", numStr);
     g_free (numStr);
     numStr = g_strdup_printf ("%.6f", y1);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer, "%s", numStr);
     g_free (numStr);
     numStr = g_strdup_printf ("%.6f", x);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer, "\n\t\t\t\t\t\t%s,", numStr);
     g_free (numStr);
     numStr = g_strdup_printf ("%.6f", y);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer, "%s", numStr);
     g_free (numStr);
     g_string_append (outBuffer, "\n\t\t\t\t\t</coordinates>\n\t\t\t\t</LinearRing>\n\t\t\t</outerBoundaryIs>");
     g_string_append (outBuffer, "\n\t\t</Polygon>\n\t</Placemark>");
   }
 }
 g_string_append_printf (outBuffer, "\n</Folder>");
 g_string_append (outBuffer, "\n\n</Document>\n</kml>\n");
 DestroyIntArray (pdPerCell);
 DestroyIntArray (obsPerCell);
 out = fopen (fileName, "w");
 if (out == NULL) {
   errorMessage = g_strdup_printf ("\nUnable to open %s\n", fileName);
   OnError ();
   g_string_free (outBuffer, TRUE);
   return;
 }
 if (fwrite (outBuffer -> str, sizeof (gchar), outBuffer -> len, out) != outBuffer -> len) {
   errorMessage = g_strdup_printf ("\nUnable to save whole data\nFile%s\n", fileName);
   OnError ();
 }
 fclose (out);
 g_string_free (outBuffer, TRUE);
}

gchar GetHexChar (SCOUNTER num) {
 if (num < 10) return (num + '0');
 return ((num -10) + 'a');
}

void WriteInHex (gchar* str, SCOUNTER num) {

 str [0] = str [1] = str [2] = '\0';
 if (num > 255) return;
 str [0] = GetHexChar (num / 16);
 str [1] = GetHexChar (num % 16);
}
/* parcer.c is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */

#include "parser.h"
#include "in.h"
#include "inxml.h"
#include "biogeo.h"
#include "out.h"

static SFLAG ReadRealTuple (INSTREAM* from, SREAL* r1, SREAL* r2);

static GString* tempParserString = NULL;
static INSTREAM* in = NULL;
static INSTREAM* cmdLine = NULL;

static PROGRESSSTRUCT* progBar = NULL;

static XMLPARSER xmlParser = { NULL, NULL, NULL, NULL, NULL, NULL, NULL, XML_ERROR };

static SFLAG BasicOpen ();
static SFLAG FeedOpen ();
static SFLAG BasicEnd ();
static SFLAG BasicError ();

static SFLAG OpenVIPTag ();
static SFLAG EndVIPTag ();
static SFLAG OnVIPError ();

static SFLAG OpenFeedTag ();
static SFLAG EndFeedTag ();
static SFLAG OnFeedError ();

static SFLAG FeedDarwinTag ();
static SFLAG EndDarwinTag ();

static SFLAG FeedXYDFile ();
static SFLAG FeedTABTable ();

static void StartNuFeed ();
static void FinishUnRead ();

static SFLAG OpenVRXTag ();
static SFLAG EndVRXTag ();
static SFLAG OnVRXError ();
static void FinishVRXFile ();

static void CleanTokenStack (GList* toClean);

static SFLAG IsXYDFile ();

SFLAG ReadFile (gchar* name, PROGRESSSTRUCT* progress) {
 SFLAG toRet;

 toRet = FALSE;
 if (in == NULL)
   in = CreateInStream ();
 if (cmdLine == NULL)
   cmdLine = CreateInStream ();
 if (tempParserString == NULL)
   tempParserString = g_string_new (VOIDSTRING);
 SETREADINGMODE (ALL_LOWER);
 if (SetInStreamToFile (in, name) != TRUE) return FALSE;
 if (IsXMLFile (in) == FALSE) {
   errorMessage = g_strdup_printf ("\nFile %s format is unknown\n", name);
   OnError ();
   ResetInStream (in);
   return FALSE;
 }
 progBar = progress;
 progBar -> numBar1 = in -> len;
 progBar -> actBar1 = 0;
 xmlParser . OpenTag = BasicOpen;
 xmlParser . EndTag = BasicEnd;
 xmlParser . OnError = BasicError;
 xmlParser . tempParserString = tempParserString;
 xmlParser . in = in;
 toRet = ReadXMLFile (&xmlParser);
 ResetInStream (in);
 ResetInStream (cmdLine);
 xmlParser . OpenTag = NULL;
 xmlParser . EndTag = NULL;
 xmlParser . OnError = NULL;
 xmlParser . tempParserString = NULL;
 xmlParser . in = NULL;
 return toRet;
}

SFLAG FeedFile (gchar* name, PROGRESSSTRUCT* progress) {
 SFLAG toRet;

 toRet = FALSE;
 if (in == NULL)
   in = CreateInStream ();
 if (cmdLine == NULL)
   cmdLine = CreateInStream ();
 if (tempParserString == NULL)
   tempParserString = g_string_new (VOIDSTRING);
 SETREADINGMODE (ALL_LOWER);
 if (SetInStreamToFile (in, name) != TRUE) return FALSE;
 progBar = progress;
 progBar -> numBar1 = in -> len;
 progBar -> actBar1 = 0;
 if (IsXMLFile (in) == TRUE) {
   xmlParser . OpenTag = FeedOpen;
   xmlParser . EndTag = BasicEnd;
   xmlParser . OnError = BasicError;
   xmlParser . tempParserString = tempParserString;
   xmlParser . in = in;
   toRet = ReadXMLFile (&xmlParser);
   xmlParser . OpenTag = NULL;
   xmlParser . EndTag = NULL;
   xmlParser . OnError = NULL;
   xmlParser . tempParserString = NULL;
   xmlParser . in = NULL;
 }
 else if (IsXYDFile () == TRUE)
   toRet = FeedXYDFile ();
 else
   toRet = FeedTABTable ();
 if (toRet != FALSE)
   FinishUnRead ();
 else {
   ResetInStream (in);
   return FALSE;
 }
 ResetInStream (in);
 ResetInStream (cmdLine);
 return toRet;
}

SFLAG ReadResults (gchar* name, PROGRESSSTRUCT* progress) {
 SFLAG toRet;

 toRet = FALSE;
 if (in == NULL)
   in = CreateInStream ();
 if (cmdLine == NULL)
   cmdLine = CreateInStream ();
 if (tempParserString == NULL)
   tempParserString = g_string_new (VOIDSTRING);
 SETREADINGMODE (ALL_LOWER);
 if (SetInStreamToFile (in, name) != TRUE) return FALSE;
 if (IsXMLFile (in) == FALSE) {
   errorMessage = g_strdup_printf ("\nFile %s format is unknown\n", name);
   OnError ();
   ResetInStream (in);
   return FALSE;
 }
 progBar = progress;
 progBar -> numBar1 = in -> len;
 progBar -> actBar1 = 0;
 xmlParser . OpenTag = OpenVRXTag;
 xmlParser . EndTag = EndVRXTag;
 xmlParser . OnError = OnVRXError;
 xmlParser . tempParserString = tempParserString;
 xmlParser . in = in;
 toRet = ReadXMLFile (&xmlParser);
 if (toRet != FALSE)
   FinishVRXFile ();
 ResetInStream (in);
 ResetInStream (cmdLine);
 xmlParser . OpenTag = NULL;
 xmlParser . EndTag = NULL;
 xmlParser . OnError = NULL;
 xmlParser . tempParserString = NULL;
 xmlParser . in = NULL;
 return toRet;
}

SFLAG BasicOpen () {
 progBar -> actBar1 = in -> actPos;
 UpdateOnCalculus ();
 if ((!g_strcmp0 (tempParserString -> str, "vip")) || (!g_strcmp0 (tempParserString -> str, "document")) ||
     (!g_strcmp0 (tempParserString -> str, "phyloxml")) || (!g_strcmp0 (tempParserString -> str, "kml"))) {
   if (xmlParser . tokenState & XML_WITH_PARAMS) {
     if (SkipParams (in) == XML_AUTO_CLOSE) {
       RemoveFromTokenList (&xmlParser);
       return TRUE;
     }
   }
   xmlParser . OpenTag = OpenVIPTag;
   xmlParser . EndTag = EndVIPTag;
   xmlParser . OnError = OnVIPError;
   xmlParser . fileHeader = g_strdup (tempParserString -> str);
 }
 else
   return IgnoreOpenTag (&xmlParser);
 return TRUE;
}

SFLAG FeedOpen () {
 progBar -> actBar1 = in -> actPos;
 UpdateOnCalculus ();
 if ((!g_strcmp0 (tempParserString -> str, "vip")) || (!g_strcmp0 (tempParserString -> str, "document")) ||
     (!g_strcmp0 (tempParserString -> str, "kml"))) {
   if (xmlParser . tokenState & XML_WITH_PARAMS) {
     if (SkipParams (in) == XML_AUTO_CLOSE) {
       RemoveFromTokenList (&xmlParser);
       return TRUE;
     }
   }
   StartNuFeed ();
   xmlParser . OpenTag = OpenFeedTag;
   xmlParser . EndTag = EndFeedTag;
   xmlParser . OnError = OnFeedError;
   xmlParser . fileHeader = g_strdup (tempParserString -> str);
 }
 else if (!g_strcmp0 (tempParserString -> str, "gbifresponse")) {
   if (xmlParser . tokenState & XML_WITH_PARAMS) {
     if (SkipParams (in) == XML_AUTO_CLOSE) {
       RemoveFromTokenList (&xmlParser);
       return TRUE;
     }
   }
   StartNuFeed ();
   xmlParser . OpenTag = FeedDarwinTag;
   xmlParser . EndTag = EndDarwinTag;
   xmlParser . OnError = OnFeedError;
   xmlParser . fileHeader = g_strdup (tempParserString -> str);
 }
 else
   return IgnoreOpenTag (&xmlParser);
 return TRUE;
}

SFLAG BasicEnd () {
 return TRUE;
}

SFLAG BasicError () {
 return FALSE;
}

/* Data reading variables */

static GNode* actNode = NULL;
static GNode* anc = NULL;
static GNode* sibling = NULL;
static GList* firstTerm = NULL;
static GList* lastTerm = NULL;
static GList* actTerm = NULL;
static GList* firstGeo = NULL;
static GList* lastGeo = NULL;
static GEOREF* actGeo = NULL;
static PHYLOTREE* nuTree = NULL;
static PHYLONODE* node = NULL;
static TERMINAL* term = NULL;

static GList* unReadList = NULL;
static GString* termName = NULL;
static GString* outBuffer = NULL;

static SCOUNTER numGeo = 0;
static SCOUNTER nodeId = 1000000;
static SREAL incMin = 360.00;
static SREAL incMax = 0;
static SREAL azmMin = 360.00;
static SREAL azmMax = 0;

static SREAL addLon = 360.00;
static SREAL addLat = 360.00;

static gchar* recordName = NULL;
static gchar* collecName = NULL;
static gchar* catalogName = NULL;
static gchar* urlName = NULL;
static gchar* referenceName = NULL;
static gchar* identifierName = NULL;
static gchar* sampleName = NULL;
static gchar* collectorName = NULL;

/* Function reading */
static void StartNuTree ();
static void ResetTreeVars ();
static void FreeNames ();
static void CloseTree ();

SFLAG OpenVIPTag () {
 GString* paramName;
 GString* actToken;
 PHYLONODE* parent;
 SINTEGER tmpNum;
 SFLAG onMarkup;
 SREAL lon, lat;
 SINTEGER age;

 progBar -> actBar1 = in -> actPos;
 UpdateOnCalculus ();
 if ((!g_strcmp0 (tempParserString -> str, "tree")) || (!g_strcmp0 (tempParserString -> str, "phylogeny"))) {
   if (xmlParser . tokenState & XML_WITH_PARAMS) {
     if (SkipParams (in) == XML_AUTO_CLOSE) {
       RemoveFromTokenList (&xmlParser);
       return TRUE;
     }
   }
   if (nuTree == NULL)
     StartNuTree ();
   else {
     errorMessage = g_strdup_printf ("\nBad \'<%s>\' token\nFile %s, line %ld\n", tempParserString -> str, in -> name, in -> line);
     return FALSE;
   }
 }
 else if ((!g_strcmp0 (tempParserString -> str, "node")) || (!g_strcmp0 (tempParserString -> str, "folder")) ||
     (!g_strcmp0 (tempParserString -> str, "clade"))) {
   actToken = (GString*) xmlParser . tokenStack -> data;
   if (nuTree == NULL)
     StartNuTree ();
   paramName = g_string_new (VOIDSTRING);
   tmpNum = nodeId;
   while (xmlParser . tokenState & XML_WITH_PARAMS) {
     ReadXMLTokenParam (in, paramName, tempParserString, &xmlParser . tokenState);
     if (xmlParser . tokenState == XML_ERROR) {
       g_string_free (paramName, TRUE);
       errorMessage = g_strdup_printf ("\nBad \'<%s>\' token\nFile %s, line %ld\n", actToken -> str, in -> name, in -> line);
       return FALSE;
     }
     else if (xmlParser . tokenState == XML_AUTO_CLOSE) {
       g_string_free (paramName, TRUE);
       RemoveFromTokenList (&xmlParser);
       return TRUE;
     }
     if (!g_strcmp0 (paramName -> str, "id")) {
       SetInStreamToString (cmdLine, tempParserString);
       if (ReadInt (cmdLine, &tmpNum) == FALSE) {
         errorMessage = g_strdup_printf ("\nBad \'id\' parameter on \'<Mode>\' token\nFile %s, line %ld\n", in -> name, in -> line);
         OnError ();
         tmpNum = nodeId;
       }
       if (xmlParser . tokenState & XML_WITH_PARAMS) {
         if (SkipParams (in) == XML_AUTO_CLOSE) {
           RemoveFromTokenList (&xmlParser);
           return TRUE;
         }
       }
       break;
     }
   }
   actNode = CreateNode (nuTree, anc, sibling, tmpNum);
   node = (PHYLONODE*) actNode -> data;
   g_ptr_array_add (nuTree -> nodeArray, actNode);
   if (nuTree -> root == NULL)
     nuTree -> root = actNode;
   else {
     parent = (PHYLONODE*) anc -> data;
     node -> nested = parent -> nested + 1;
   }
   anc = actNode;
   sibling = NULL;
   term = NULL;
   g_string_free (paramName, TRUE);
   ++ nodeId;
 }
 else if ((!g_strcmp0 (tempParserString -> str, "placemark")) || (!g_strcmp0 (tempParserString -> str, "point"))
                || (!g_strcmp0 (tempParserString -> str, "specimeninfo"))) {
   if (xmlParser . tokenState & XML_WITH_PARAMS) {
     if (SkipParams (in) == XML_AUTO_CLOSE) {
       RemoveFromTokenList (&xmlParser);
       return TRUE;
     }
   }
   if (term == NULL) {
     term = CreateTerminal (actNode);
     actTerm = InsertListElement (lastTerm, term);
     term -> list = actTerm;
     if (firstTerm == NULL)
       firstTerm = actTerm;
     lastTerm = actTerm;
     node -> term = term;
     FreeNames ();
   }
   actGeo = NULL;
   if ((!g_strcmp0 (tempParserString -> str, "placemark")) || (!g_strcmp0 (tempParserString -> str, "specimeninfo")))
     FreeNames ();
 }
 else if (!g_strcmp0 (tempParserString -> str, "coordinates")) {
   actToken = GetPrevToken (&xmlParser);
   if (actToken == NULL) {
     errorMessage = g_strdup_printf ("\nUnexpected \'<coordinates>\' token\nFile %s, line %ld\n", in -> name, in -> line);
     return FALSE;
   }
   if ((g_strcmp0 (actToken -> str, "point")) && (g_strcmp0 (actToken -> str, "placemark")) &&
       (g_strcmp0 (tempParserString -> str, "specimeninfo"))) {
     errorMessage = g_strdup_printf ("\nReading \'<coordinates>\', without <point>\nFile %s, line %ld\n", in -> name, in -> line);
     return FALSE;
   }
   if (actGeo != NULL) {
     errorMessage = g_strdup_printf ("\nUnexpected \'<coordinates>\' token\nFile %s, line %ld\n", in -> name, in -> line);
     return FALSE;
   }
   if (xmlParser . tokenState & XML_WITH_PARAMS) {
     if (SkipParams (in) == XML_AUTO_CLOSE) {
       RemoveFromTokenList (&xmlParser);
       return TRUE;
     }
   }
   ReadXMLContent (in, tempParserString, &xmlParser . tokenState);
   if (xmlParser . tokenState != XML_TOKEN_CONTENT) {
     errorMessage = g_strdup_printf ("\nBad \'<coordinates>\' token\nFile %s, line %ld\n", in -> name, in -> line);
     return FALSE;
   }
   SetInStreamToString (cmdLine, tempParserString);
   if (ReadRealTuple (cmdLine, &lon, &lat) == TRUE) {
     if ((lon > -180.00001) && (lon < 180.00001) && (lat > -90.00001) && (lat < 90.00001)) {
       lastGeo = CreateGeoRef (lastGeo, actTerm,  lon, lat);
       if (firstGeo == NULL)
         firstGeo = lastGeo;
       if (term -> firstGeo == NULL)
         term -> firstGeo = lastGeo;
       ++ term -> numGeo;
       ++ numGeo;
       actGeo = (GEOREF*) lastGeo -> data;
       if (actGeo -> inclination > incMax)
         incMax = actGeo -> inclination;
       if (actGeo -> inclination < incMin)
         incMin = actGeo -> inclination;
       if (actGeo -> azimuth > azmMax)
         azmMax = actGeo -> azimuth;
       if (actGeo -> azimuth < azmMin)
         azmMin = actGeo -> azimuth;
     }
     else {
       errorMessage = g_strdup_printf ("\nBad \'<coordinates>\' token\nFile %s, line %ld\n", in -> name, in -> line);
       OnError ();
       actGeo = NULL;
     }
   }
   else {
     errorMessage = g_strdup_printf ("\nBad \'<coordinates>\' token\nFile %s, line %ld\n", in -> name, in -> line);
     OnError ();
     actGeo = NULL;
   }
 }
 else if (!g_strcmp0 (tempParserString -> str, "name")) {
   actToken = GetPrevToken (&xmlParser);
   if (actToken == NULL) {
     errorMessage = g_strdup_printf ("\nUnexpected \'<name>\' token\nFile %s, line %ld\n", in -> name, in -> line);
     return FALSE;
   }
   onMarkup = 0;
   if ((!g_strcmp0 (actToken -> str, "tree")) || (!g_strcmp0 (actToken -> str, "phylogeny")))
     onMarkup = 1;
   else if ((!g_strcmp0 (actToken -> str, "node")) || (!g_strcmp0 (actToken -> str, "folder")) || (!g_strcmp0 (actToken -> str, "clade")))
     onMarkup = 2;
   else if ((!g_strcmp0 (actToken -> str, "placemark")) || (!g_strcmp0 (actToken -> str, "specimeninfo")))
     onMarkup = 4;
   else {
     IgnoreOpenTag (&xmlParser);
     return TRUE;
   }
   if (xmlParser . tokenState & XML_WITH_PARAMS) {
     if (SkipParams (in) == XML_AUTO_CLOSE) {
       RemoveFromTokenList (&xmlParser);
       return TRUE;
     }
   }
   ReadXMLContent (in, tempParserString, &xmlParser . tokenState);
   if (xmlParser . tokenState != XML_TOKEN_CONTENT) {
     errorMessage = g_strdup_printf ("\nBad \'<name>\' token: %s\nFile %s, line %ld\n", tempParserString -> str, in -> name, in -> line);
     OnError ();
     return TRUE;
   }
   if (onMarkup == 1) {
     if (nuTree -> name != NULL)
       g_free (nuTree -> name);
     nuTree -> name = g_strdup (tempParserString -> str);
   }
   else if (onMarkup == 2) {
     if (node -> name != NULL)
       g_free (node -> name);
     node -> name = g_strdup (tempParserString -> str);
   }
   else if (onMarkup == 4)
     recordName = AddSynonim (term, tempParserString -> str);
 }
 else if (!g_strcmp0 (tempParserString -> str, "synonym")) {
   actToken = GetPrevToken (&xmlParser);
   if (actToken == NULL) {
     errorMessage = g_strdup_printf ("\nUnexpected \'<synonym>\' token\nFile %s, line %ld\n", in -> name, in -> line);
     return FALSE;
   }
   if ((g_strcmp0 (actToken -> str, "node")) && (g_strcmp0 (actToken -> str, "folder")) && (g_strcmp0 (actToken -> str, "clade"))) {
     IgnoreOpenTag (&xmlParser);
     return TRUE;
   }
   if (xmlParser . tokenState & XML_WITH_PARAMS) {
     if (SkipParams (in) == XML_AUTO_CLOSE) {
       RemoveFromTokenList (&xmlParser);
       return TRUE;
     }
   }
   ReadXMLContent (in, tempParserString, &xmlParser . tokenState);
   if (xmlParser . tokenState != XML_TOKEN_CONTENT) {
     errorMessage = g_strdup_printf ("\nBad \'<synonym>\' token: %s\nFile %s, line %ld\n", tempParserString -> str, in -> name, in -> line);
     OnError ();
     return TRUE;
   }
   if (term == NULL) {
     term = CreateTerminal (actNode);
     actTerm = InsertListElement (lastTerm, term);
     term -> list = actTerm;
     if (firstTerm == NULL)
       firstTerm = actTerm;
     lastTerm = actTerm;
     node -> term = term;
     FreeNames ();
   }
   AddSynonim (term, tempParserString -> str);
 }
 else if (!g_strcmp0 (tempParserString -> str, "taxonomy")) {
   actToken = GetPrevToken (&xmlParser);
   if (actToken == NULL) {
     errorMessage = g_strdup_printf ("\nUnexpected \'<%s>\' token\nFile %s, line %ld\n", tempParserString -> str, in -> name, in -> line);
     return FALSE;
   }
   if ((g_strcmp0 (actToken -> str, "node")) && (g_strcmp0 (actToken -> str, "folder")) && (g_strcmp0 (actToken -> str, "clade"))) {
     IgnoreOpenTag (&xmlParser);
     return TRUE;
   }
   if (xmlParser . tokenState & XML_WITH_PARAMS) {
     if (SkipParams (in) == XML_AUTO_CLOSE) {
       RemoveFromTokenList (&xmlParser);
       return TRUE;
     }
   }
 }
 else if (!g_strcmp0 (tempParserString -> str, "scientific_name")) {
   actToken = GetPrevToken (&xmlParser);
   if (actToken == NULL) {
     errorMessage = g_strdup_printf ("\nUnexpected \'<%s>\' token\nFile %s, line %ld\n", tempParserString -> str, in -> name, in -> line);
     return FALSE;
   }
   if (g_strcmp0 (actToken -> str, "taxonomy")) {
     IgnoreOpenTag (&xmlParser);
     return TRUE;
   }
   if (xmlParser . tokenState & XML_WITH_PARAMS) {
     if (SkipParams (in) == XML_AUTO_CLOSE) {
       RemoveFromTokenList (&xmlParser);
       return TRUE;
     }
   }
   ReadXMLContent (in, tempParserString, &xmlParser . tokenState);
   if (xmlParser . tokenState != XML_TOKEN_CONTENT) {
     errorMessage = g_strdup_printf ("\nBad \'<scientific_name>\' token: %s\nFile %s, line %ld\n", tempParserString -> str,
                                     in -> name, in -> line);
     OnError ();
     return TRUE;
   }
   if (node -> name != NULL)
     g_free (node -> name);
   node -> name = g_strdup (tempParserString -> str);
 }
 else if (!g_strcmp0 (tempParserString -> str, "date")) {
   actToken = GetPrevToken (&xmlParser);
   if (actToken == NULL) {
     errorMessage = g_strdup_printf ("\nUnexpected \'<%s>\' token\nFile %s, line %ld\n", tempParserString -> str, in -> name, in -> line);
     return FALSE;
   }
   if ((g_strcmp0 (actToken -> str, "node")) && (g_strcmp0 (actToken -> str, "folder")) && (g_strcmp0 (actToken -> str, "clade"))) {
     IgnoreOpenTag (&xmlParser);
     return TRUE;
   }
   if (xmlParser . tokenState & XML_WITH_PARAMS) {
     if (SkipParams (in) == XML_AUTO_CLOSE) {
       RemoveFromTokenList (&xmlParser);
       return TRUE;
     }
   }
 }
 else if ((!g_strcmp0 (tempParserString -> str, "age")) || (!g_strcmp0 (tempParserString -> str, "date")) ||
          (!g_strcmp0 (tempParserString -> str, "minimum")) || (!g_strcmp0 (tempParserString -> str, "maximum"))) {
   actToken = GetPrevToken (&xmlParser);
   if (actToken == NULL) {
     errorMessage = g_strdup_printf ("\nUnexpected \'<%s>\' token\nFile %s, line %ld\n", tempParserString -> str, in -> name, in -> line);
     return FALSE;
   }
   if ((!g_strcmp0 (tempParserString -> str, "age")) || (g_strcmp0 (tempParserString -> str, "date"))) {
     if ((g_strcmp0 (actToken -> str, "node")) && (g_strcmp0 (actToken -> str, "folder")) && (g_strcmp0 (actToken -> str, "clade"))) {
       IgnoreOpenTag (&xmlParser);
       return TRUE;
     }
   }
   else if (!g_strcmp0 (tempParserString -> str, "minimum")) {
     if ((!g_strcmp0 (tempParserString -> str, "age")) || (g_strcmp0 (actToken -> str, "date"))) {
       IgnoreOpenTag (&xmlParser);
       return TRUE;
     }
   }
   if (xmlParser . tokenState & XML_WITH_PARAMS) {
     if (SkipParams (in) == XML_AUTO_CLOSE) {
       RemoveFromTokenList (&xmlParser);
       return TRUE;
     }
   }
   ReadXMLContent (in, tempParserString, &xmlParser . tokenState);
   if (xmlParser . tokenState != XML_TOKEN_CONTENT) {
     errorMessage = g_strdup_printf ("\nBad \'<age>\' token: %s\nFile %s, line %ld\n", tempParserString -> str, in -> name, in -> line);
     OnError ();
     return TRUE;
   }
   SetInStreamToString (cmdLine, tempParserString);
   ReadInt (cmdLine, &age);
   if (age >= 0)
     SetNodeAge (actNode, (SCOUNTER) age);
 }

 else if ((!g_strcmp0 (tempParserString -> str, "lat")) || (!g_strcmp0 (tempParserString -> str, "long"))) {
   actToken = GetPrevToken (&xmlParser);
   if (actToken == NULL) {
     errorMessage = g_strdup_printf ("\nUnexpected \'<%s>\' token\nFile %s, line %ld\n", tempParserString -> str, in -> name, in -> line);
     return FALSE;
   }
   if (g_strcmp0 (actToken -> str, "point")) {
     IgnoreOpenTag (&xmlParser);
     return TRUE;
   }
   if (actGeo != NULL) {
     errorMessage = g_strdup_printf ("\nUnexpected \'<%s>\' token\nFile %s, line %ld\n", tempParserString -> str, in -> name, in -> line);
     return FALSE;
   }
   if (xmlParser . tokenState & XML_WITH_PARAMS) {
     if (SkipParams (in) == XML_AUTO_CLOSE) {
       RemoveFromTokenList (&xmlParser);
       return TRUE;
     }
   }
   onMarkup = 1;
   if (!g_strcmp0 (tempParserString -> str, "lat"))
     onMarkup = 2;
   ReadXMLContent (in, tempParserString, &xmlParser . tokenState);
   if (xmlParser . tokenState != XML_TOKEN_CONTENT) {
     errorMessage = g_strdup_printf ("\nBad \'<age>\' token: %s\nFile %s, line %ld\n", tempParserString -> str, in -> name, in -> line);
     OnError ();
     return TRUE;
   }
   SetInStreamToString (cmdLine, tempParserString);
   if (onMarkup == 1)
     ReadReal (cmdLine, &addLon);
   else
     ReadReal (cmdLine, &addLat);
 }
 else if ((!g_strcmp0 (tempParserString -> str, "collection")) || (!g_strcmp0 (tempParserString -> str, "catalogcode")) ||
          (!g_strcmp0 (tempParserString -> str, "url")) || (!g_strcmp0 (tempParserString -> str, "reference")) ||
          (!g_strcmp0 (tempParserString -> str, "identifiedby")) || (!g_strcmp0 (tempParserString -> str, "collectiondate")) ||
          (!g_strcmp0 (tempParserString -> str, "collector"))) {
   actToken = GetPrevToken (&xmlParser);
   if (actToken == NULL) {
     errorMessage = g_strdup_printf ("\nUnexpected \'<%s>\' token\nFile %s, line %ld\n", tempParserString -> str,
                                     in -> name, in -> line);
     return FALSE;
   }
   if ((g_strcmp0 (actToken -> str, "placemark")) && (g_strcmp0 (actToken -> str, "specimeninfo"))) {
     errorMessage = g_strdup_printf ("\nReading \'<%s>\', without <SpecimenInfo>\nFile %s, line %ld\n",
                                     tempParserString -> str, in -> name, in -> line);
     return FALSE;
   }
   if (xmlParser . tokenState & XML_WITH_PARAMS) {
     if (SkipParams (in) == XML_AUTO_CLOSE) {
       RemoveFromTokenList (&xmlParser);
       return TRUE;
     }
   }
   actToken = xmlParser . tokenStack -> data;
   ReadXMLContent (in, tempParserString, &xmlParser . tokenState);
   if (xmlParser . tokenState == XML_TOKEN_NOCONTENT) return TRUE;
   if (xmlParser . tokenState != XML_TOKEN_CONTENT) {
     errorMessage = g_strdup_printf ("\nBad \'<%s>\' token\nFile %s, line %ld\n", actToken -> str, in -> name, in -> line);
     return FALSE;
   }
   if (!g_strcmp0 (actToken -> str, "collection"))
     collecName = g_strdup (tempParserString -> str);
   else if (!g_strcmp0 (actToken -> str, "catalogcode"))
     catalogName = g_strdup (tempParserString -> str);
   else if (!g_strcmp0 (actToken -> str, "url"))
     urlName = g_strdup (tempParserString -> str);
   else if (!g_strcmp0 (actToken -> str, "reference"))
     referenceName = g_strdup (tempParserString -> str);
   else if (!g_strcmp0 (actToken -> str, "identifiedby"))
     identifierName = g_strdup (tempParserString -> str);
   else if (!g_strcmp0 (actToken -> str, "collectiondate"))
     sampleName = g_strdup (tempParserString -> str);
   else if (!g_strcmp0 (actToken -> str, "collector"))
     collectorName = g_strdup (tempParserString -> str);
 }
 else
   return IgnoreOpenTag (&xmlParser);
 return TRUE;
}

SFLAG EndVIPTag () {
 GString* actToken;
 PHYLONODE* parent;
 SINTEGER tmpNum;

 actToken = (GString*) xmlParser . tokenStack -> data;
 if (!g_strcmp0 (tempParserString -> str, "tree"))
   nuTree = NULL;
 else if ((!g_strcmp0 (tempParserString -> str, "node")) || (!g_strcmp0 (tempParserString -> str, "folder"))
           || (!g_strcmp0 (tempParserString -> str, "clade"))) {
   tmpNum = g_node_n_children (actNode);
   if (tmpNum == 1) {
     errorMessage = g_strdup_printf ("\nFewer than 2 descendants in a node\nFile %s, line %ld\n", in -> name, in -> line);
     return FALSE;
   }
   else if ((tmpNum != 0) && (node -> term != NULL)) {
     errorMessage = g_strdup_printf ("\nTerminals must be without descendants\nFile %s, line %ld\n", in -> name, in -> line);
     return FALSE;
   }
   else if (tmpNum == 0) {
     node -> level = 1;
     node -> numTerms = 1;
     node -> size = 1;
     if (term == NULL) {
       term = CreateTerminal (actNode);
       actTerm = InsertListElement (lastTerm, term);
       term -> list = actTerm;
       if (firstTerm == NULL)
         firstTerm = actTerm;
       lastTerm = actTerm;
       node -> term = term;
     }
     if (node -> name != NULL)
       term -> name = AddSynonim (term, node -> name);
     UpdateWithData (term);
   }
   anc = actNode -> parent;
   if (anc != NULL) {
     parent = (PHYLONODE*) anc -> data;
     parent -> numTerms += node -> numTerms;
     parent -> size += node -> size;
     if (node -> level >= parent -> level)
       parent -> level = node -> level + 1;
     /* Update the actual node */
     sibling = actNode;
     actNode = anc;
     node = (PHYLONODE*) actNode -> data;
   }
   else
     CloseTree ();
 }
 else if (!g_strcmp0 (tempParserString -> str, "point")) {
   if (actGeo == NULL) {
     if ((addLon > -180.00001) && (addLon < 180.00001) && (addLat > -90.00001) && (addLat < 90.00001)) {
       if (term != NULL) {
         lastGeo = CreateGeoRef (lastGeo, actTerm,  addLon, addLat);
         if (firstGeo == NULL)
           firstGeo = lastGeo;
         if (term -> firstGeo == NULL)
           term -> firstGeo = lastGeo;
         ++ term -> numGeo;
         ++ numGeo;
         actGeo = (GEOREF*) lastGeo -> data;
         if (actGeo -> inclination > incMax)
           incMax = actGeo -> inclination;
         if (actGeo -> inclination < incMin)
           incMin = actGeo -> inclination;
         if (actGeo -> azimuth > azmMax)
           azmMax = actGeo -> azimuth;
         if (actGeo -> azimuth < azmMin)
           azmMin = actGeo -> azimuth;
       }
       actGeo = NULL;
       addLon = 360.00;
       addLat = 360.00;
     }
   }
 }
 else if ((!g_strcmp0 (tempParserString -> str, "placemark")) || (!g_strcmp0 (tempParserString -> str, "specimeninfo"))) {
   if (actGeo != NULL) {
     actGeo -> name = recordName;
     actGeo -> collec = collecName;
     actGeo -> catalog = catalogName;
     actGeo -> url = urlName;
     actGeo -> reference = referenceName;
     actGeo -> identifier = identifierName;
     actGeo -> sampleDate = sampleName;
     actGeo -> collector = collectorName;
     recordName = NULL;
     collecName = NULL;
     catalogName = NULL;
     urlName = NULL;
     referenceName = NULL;
     identifierName = NULL;
     sampleName = NULL;
     collectorName = NULL;
   }
   else
     FreeNames ();
   actGeo = NULL;
 }
 return TRUE;
}

SFLAG OnVIPError () {
 OnError ();
 DestroyTree (nuTree);
 DestroyTermList (firstTerm);
 DestroyGeoList (firstGeo);
 if (GETLASTSPP == NULL)
   SETSPP (NULL);
 if (GETLASTGEOREF == NULL)
   SETGEOREF (NULL);
 FreeNames ();
 nuTree = NULL;
 return FALSE;
}

void StartNuTree () {
 ResetTreeVars ();
 lastTerm = GETLASTSPP;
 lastGeo = GETLASTGEOREF;
 actGeo = NULL;
 incMin = GETMININC;
 incMax = GETMAXINC;
 azmMin = GETMINAZM;
 azmMax = GETMAXAZM;
 nuTree = CreateTree (NULL, GETNUMTREES);
}

void ResetTreeVars () {
 actNode = anc = sibling = NULL;
 firstTerm = NULL;
 firstGeo = NULL;
 actGeo = NULL;
 numGeo = 0;
 node = NULL;
 term = NULL;
 actTerm = NULL;
 nodeId = g_random_int_range (1000000, 9999999);
 FreeNames ();
}

void CloseTree () {
 GString* actToken;

 if (nuTree == NULL) return;
 if (nuTree -> root == NULL) {
   DestroyTree (nuTree);
   return;
 }
 PrepareTree (nuTree);
 nuTree -> list = InsertListElement (GETLASTTREE, nuTree);
 if (GETTREES == NULL)
   SETTREES (nuTree -> list);
 SETLASTTREE (nuTree -> list);
 INCREASENUMTREES (1);
 if (GETSPP == NULL)
   SETSPP (firstTerm);
 SETLASTSPP (lastTerm);
 node = (PHYLONODE*) nuTree -> root -> data;
 INCREASENUMSPP (node -> numTerms);
 if (GETGEOREF == NULL)
   SETGEOREF (firstGeo);
 SETLASTGEOREF (lastGeo);
 INCREASENUMGEOREFS (numGeo);
 if (GETMAXAZM < azmMax)
   SETMAXAZM (azmMax);
 if (GETMINAZM > azmMin)
   SETMINAZM (azmMin);
 if (GETMAXINC < incMax)
   SETMAXINC (incMax);
 if (GETMININC > incMin)
   SETMININC (incMin);
 node = (PHYLONODE*) nuTree -> root -> data;
 ResetTreeVars ();
 actToken = GetPrevToken (&xmlParser);
 if (g_strcmp0 (tempParserString -> str, "tree"))
   nuTree = NULL;
}

SFLAG OpenFeedTag () {
 SFLAG onMarkup;
 GString* actToken;

 progBar -> actBar1 = in -> actPos;
 UpdateOnCalculus ();
 if (!g_strcmp0 (tempParserString -> str, "name")) {
   actToken = GetPrevToken (&xmlParser);
   if (actToken == NULL) {
     errorMessage = g_strdup_printf ("\nUnexpected \'<name>\' token\nFile %s, line %ld\n", in -> name, in -> line);
     return FALSE;
   }
   onMarkup = 0;
   if ((!g_strcmp0 (actToken -> str, "node")) || (!g_strcmp0 (actToken -> str, "folder")) || (!g_strcmp0 (actToken -> str, "clade")))
     onMarkup = 2;
   else if ((!g_strcmp0 (actToken -> str, "placemark")) || (!g_strcmp0 (actToken -> str, "point"))
             || (!g_strcmp0 (actToken -> str, "specimeninfo")))
     onMarkup = 4;
   else {
     IgnoreOpenTag (&xmlParser);
     return TRUE;
   }
   if (xmlParser . tokenState & XML_WITH_PARAMS) {
     if (SkipParams (in) == XML_AUTO_CLOSE) {
       RemoveFromTokenList (&xmlParser);
       return TRUE;
     }
   }
   ReadXMLContent (in, tempParserString, &xmlParser . tokenState);
   if (xmlParser . tokenState != XML_TOKEN_CONTENT) {
     errorMessage = g_strdup_printf ("\nBad \'<name>\' token: %s\nFile %s, line %ld\n", tempParserString -> str, in -> name, in -> line);
     OnError ();
     return TRUE;
   }
   if (onMarkup == 2) {
     term = GetTermFromName (tempParserString -> str);
     if (termName != NULL)
       g_string_free (termName, TRUE);
     termName = g_string_new (tempParserString -> str);
   }
   else if (onMarkup == 4) {
     if (term != NULL)
       recordName = AddSynonim (term, tempParserString -> str);
     else {
       if (termName != NULL)
         g_string_free (termName, TRUE);
       termName = g_string_new (tempParserString -> str);
     }
   }
 }
 else if (!g_strcmp0 (tempParserString -> str, "synonym")) {
   actToken = GetPrevToken (&xmlParser);
   if (actToken == NULL) {
     errorMessage = g_strdup_printf ("\nUnexpected \'<synonym>\' token\nFile %s, line %ld\n", in -> name, in -> line);
     return FALSE;
   }
   if ((g_strcmp0 (actToken -> str, "node")) && (g_strcmp0 (actToken -> str, "folder")) && (g_strcmp0 (actToken -> str, "clade"))) {
     IgnoreOpenTag (&xmlParser);
     return TRUE;
   }
   if (xmlParser . tokenState & XML_WITH_PARAMS) {
     if (SkipParams (in) == XML_AUTO_CLOSE) {
       RemoveFromTokenList (&xmlParser);
       return TRUE;
     }
   }
   ReadXMLContent (in, tempParserString, &xmlParser . tokenState);
   if (xmlParser . tokenState != XML_TOKEN_CONTENT) {
     errorMessage = g_strdup_printf ("\nBad \'<synonym>\' token\nFile %s, line %ld\n", in -> name, in -> line);
     OnError ();
     return TRUE;
   }
   if (term != NULL)
     AddSynonim (term, tempParserString -> str);
 }
 else if (!g_strcmp0 (tempParserString -> str, "coordinates")) {
   actToken = GetPrevToken (&xmlParser);
   if (actToken == NULL) {
     errorMessage = g_strdup_printf ("\nUnexpected \'<coordinates>\' token\nFile %s, line %ld\n", in -> name, in -> line);
     return FALSE;
   }
   if ((g_strcmp0 (actToken -> str, "point")) && (g_strcmp0 (actToken -> str, "placemark")) &&
       (g_strcmp0 (tempParserString -> str, "specimeninfo"))) {
     errorMessage = g_strdup_printf ("\nReading \'<coordinates>\', without <point>\nFile %s, line %ld\n", in -> name, in -> line);
     return FALSE;
   }
   if (actGeo != NULL) {
     errorMessage = g_strdup_printf ("\nUnexpected \'<coordinates>\' token\nFile %s, line %ld\n", in -> name, in -> line);
     return FALSE;
   }
   if (xmlParser . tokenState & XML_WITH_PARAMS) {
     if (SkipParams (in) == XML_AUTO_CLOSE) {
       RemoveFromTokenList (&xmlParser);
       return TRUE;
     }
   }
   ReadXMLContent (in, tempParserString, &xmlParser . tokenState);
   if (xmlParser . tokenState != XML_TOKEN_CONTENT) {
     errorMessage = g_strdup_printf ("\nBad \'<coordinates>\' token\nFile %s, line %ld\n", in -> name, in -> line);
     return FALSE;
   }
   SetInStreamToString (cmdLine, tempParserString);
   if (ReadRealTuple (cmdLine, &addLon, &addLat) == TRUE) {
     if ((addLon > -180.00001) && (addLon < 180.00001) && (addLat > -90.00001) && (addLat < 90.00001)) {
       if (term != NULL)
         actGeo = AddGeoRef (addLon, addLat, term -> node);
       else
         actGeo = NULL;
     }
     else {
       errorMessage = g_strdup_printf ("\nBad \'<coordinates>\' token\nFile %s, line %ld\n", in -> name, in -> line);
       OnError ();
       actGeo = NULL;
     }
   }
   else {
     errorMessage = g_strdup_printf ("\nBad \'<coordinates>\' token\nFile %s, line %ld\n", in -> name, in -> line);
     OnError ();
     actGeo = NULL;
   }
 }
 else if ((!g_strcmp0 (tempParserString -> str, "collection")) || (!g_strcmp0 (tempParserString -> str, "catalogcode")) ||
          (!g_strcmp0 (tempParserString -> str, "url")) || (!g_strcmp0 (tempParserString -> str, "reference")) ||
          (!g_strcmp0 (tempParserString -> str, "identifiedby")) || (!g_strcmp0 (tempParserString -> str, "collectiondate")) ||
          (!g_strcmp0 (tempParserString -> str, "collector"))) {
   actToken = GetPrevToken (&xmlParser);
   if (actToken == NULL) {
     errorMessage = g_strdup_printf ("\nUnexpected \'<%s>\' token\nFile %s, line %ld\n", tempParserString -> str,
                                     in -> name, in -> line);
     return FALSE;
   }
   if ((g_strcmp0 (actToken -> str, "placemark")) && (g_strcmp0 (actToken -> str, "specimeninfo"))) {
     errorMessage = g_strdup_printf ("\nReading \'<%s>\', without <SpecimenInfo>\nFile %s, line %ld\n",
                                     tempParserString -> str, in -> name, in -> line);
     return FALSE;
   }
   if (xmlParser . tokenState & XML_WITH_PARAMS) {
     if (SkipParams (in) == XML_AUTO_CLOSE) {
       RemoveFromTokenList (&xmlParser);
       return TRUE;
     }
   }
   actToken = xmlParser . tokenStack -> data;
   ReadXMLContent (in, tempParserString, &xmlParser . tokenState);
   if (xmlParser . tokenState == XML_TOKEN_NOCONTENT) return TRUE;
   if (xmlParser . tokenState != XML_TOKEN_CONTENT) {
     errorMessage = g_strdup_printf ("\nBad \'<%s>\' token\nFile %s, line %ld\n", actToken -> str, in -> name, in -> line);
     return FALSE;
   }
   if (!g_strcmp0 (actToken -> str, "collection"))
     collecName = g_strdup (tempParserString -> str);
   else if (!g_strcmp0 (actToken -> str, "catalogcode"))
     catalogName = g_strdup (tempParserString -> str);
   else if (!g_strcmp0 (actToken -> str, "url"))
     urlName = g_strdup (tempParserString -> str);
   else if (!g_strcmp0 (actToken -> str, "reference"))
     referenceName = g_strdup (tempParserString -> str);
   else if (!g_strcmp0 (actToken -> str, "identifiedby"))
     identifierName = g_strdup (tempParserString -> str);
   else if (!g_strcmp0 (actToken -> str, "collectiondate"))
     sampleName = g_strdup (tempParserString -> str);
   else if (!g_strcmp0 (actToken -> str, "collector"))
     collectorName = g_strdup (tempParserString -> str);
 }
 else
   return IgnoreOpenTag (&xmlParser);
 return TRUE;
}

SFLAG EndFeedTag () {
 GString* actToken;
 gchar* numStr;

 actToken = (GString*) xmlParser . tokenStack -> data;
 if ((!g_strcmp0 (tempParserString -> str, "placemark")) || (!g_strcmp0 (tempParserString -> str, "specimeninfo"))) {
   if (actGeo != NULL) {
     actGeo -> name = recordName;
     actGeo -> collec = collecName;
     actGeo -> catalog = catalogName;
     actGeo -> url = urlName;
     actGeo -> reference = referenceName;
     actGeo -> identifier = identifierName;
     actGeo -> sampleDate = sampleName;
     actGeo -> collector = collectorName;
     recordName = NULL;
     collecName = NULL;
     catalogName = NULL;
     urlName = NULL;
     referenceName = NULL;
     identifierName = NULL;
     sampleName = NULL;
     collectorName = NULL;
   }
   else {
     if ((addLon > -180.00001) && (addLon < 180.00001) && (addLat > -90.00001) && (addLat < 90.00001) && (termName != NULL)) {
       g_string_append_printf (outBuffer, "\n<Folder>");
       g_string_append_printf (outBuffer, "\n\t<name>%s</name>", termName -> str);
       g_string_append (outBuffer, "\n\t<SpecimenInfo>");
       if (recordName != NULL) {
         g_string_append_printf (outBuffer,"\n\t\t<name>%s</name>", recordName);
         g_free (recordName);
       }
       if (collecName != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<collection>%s</collection>", collecName);
       if (catalogName != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<catalogCode>%s</catalogCode>", catalogName);
       if (urlName != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<url>%s</url>", urlName);
       if (referenceName != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<reference>%s</reference>", referenceName);
       if (identifierName != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<identifiedBy>%s</identifiedBy>", identifierName);
       if (sampleName != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<collectionDate>%s</collectionDate>", sampleName);
       if (collectorName != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<collector>%s</collector>", collectorName);
       g_string_append (outBuffer, "\n\t\t<Point>");
       g_string_append (outBuffer, "\n\t\t\t<coordinates>");
       numStr = g_strdup_printf ("%.6f", addLon);
       CheckRealStr (numStr);
       g_string_append_printf (outBuffer, "%s,", numStr);
       g_free (numStr);
       numStr = g_strdup_printf ("%.6f", addLat);
       CheckRealStr (numStr);
       g_string_append (outBuffer, numStr);
       g_free (numStr);
       g_string_append (outBuffer, "</coordinates>");
       g_string_append (outBuffer, "\n\t\t</Point>\n\t</SpecimenInfo>");
       g_string_append (outBuffer, "\n</Folder>");
       unReadList = NameOnList (termName -> str, unReadList);
     }
     FreeNames ();
   }
   actGeo = NULL;
   addLat = 360.00;
   addLon = 360.00;
 }
 else if ((!g_strcmp0 (tempParserString -> str, "node")) || (!g_strcmp0 (tempParserString -> str, "folder"))
           || (!g_strcmp0 (tempParserString -> str, "clade"))) {
   term = NULL;
   if (termName != NULL)
     g_string_free (termName, TRUE);
   termName = NULL;
   actGeo = NULL;
 }
 return TRUE;
}

SFLAG OnFeedError () {
 OnError ();
 FinishUnRead ();
 FreeNames ();
 return FALSE;
}

SFLAG FeedDarwinTag () {
 GString* actToken;
 GString* paramName;
 SREAL tmp;

 progBar -> actBar1 = in -> actPos;
 UpdateOnCalculus ();
 if (!g_strcmp0 (tempParserString -> str, "tax:taxonoccurrence")) {
   actToken = (GString*) xmlParser . tokenStack -> data;
   paramName = g_string_new (VOIDSTRING);
   while (xmlParser . tokenState & XML_WITH_PARAMS) {
     ReadXMLTokenParam (in, paramName, tempParserString, &xmlParser . tokenState);
     if (xmlParser . tokenState == XML_ERROR) {
       g_string_free (paramName, TRUE);
       errorMessage = g_strdup_printf ("\nBad \'<%s>\' token\nFile %s, line %ld\n", actToken -> str, in -> name, in -> line);
       return FALSE;
     }
     else if (xmlParser . tokenState == XML_AUTO_CLOSE) {
       g_string_free (paramName, TRUE);
       RemoveFromTokenList (&xmlParser);
       return TRUE;
     }
     if (!g_strcmp0 (paramName -> str, "rdf:about")) {
       urlName = g_strdup (tempParserString -> str);
       if (xmlParser . tokenState & XML_WITH_PARAMS) {
         if (SkipParams (in) == XML_AUTO_CLOSE) {
           RemoveFromTokenList (&xmlParser);
           return TRUE;
         }
       }
       break;
     }
   }
   term = NULL;
   if (termName != NULL)
     g_string_free (termName, TRUE);
   termName = NULL;
   actGeo = NULL;
   addLon = 360.00;
   addLat = 360.00;
 }
 else if (!g_strcmp0 (tempParserString -> str, "name")) {
   actToken = GetPrevToken (&xmlParser);
   if (actToken == NULL) {
     errorMessage = g_strdup_printf ("\nUnexpected \'<name>\' token\nFile %s, line %ld\n", in -> name, in -> line);
     return FALSE;
   }
   if (g_strcmp0 (actToken -> str, "dataresource")) {
     IgnoreOpenTag (&xmlParser);
     return TRUE;
   }
   if (xmlParser . tokenState & XML_WITH_PARAMS) {
     if (SkipParams (in) == XML_AUTO_CLOSE) {
       RemoveFromTokenList (&xmlParser);
       return TRUE;
     }
   }
   ReadXMLContent (in, tempParserString, &xmlParser . tokenState);
   if (xmlParser . tokenState != XML_TOKEN_CONTENT) {
     errorMessage = g_strdup_printf ("\nBad \'<name>\' token: %s\nFile %s, line %ld\n", tempParserString -> str, in -> name, in -> line);
     OnError ();
     return TRUE;
   }
   collecName = g_strdup (tempParserString -> str);
 }
 else if (!g_strcmp0 (tempParserString -> str, "tax:catalognumber")) {
   actToken = GetPrevToken (&xmlParser);
   if (actToken == NULL) {
     errorMessage = g_strdup_printf ("\nUnexpected \'<%s>\' token\nFile %s, line %ld\n", tempParserString -> str, in -> name, in -> line);
     return FALSE;
   }
   if (g_strcmp0 (actToken -> str, "tax:taxonoccurrence")) {
     errorMessage = g_strdup_printf ("\nUnexpected \'<%s>\' token\nFile %s, line %ld\n", tempParserString -> str, in -> name, in -> line);
     return FALSE;
   }
   if (xmlParser . tokenState & XML_WITH_PARAMS) {
     if (SkipParams (in) == XML_AUTO_CLOSE) {
       RemoveFromTokenList (&xmlParser);
       return TRUE;
     }
   }
   ReadXMLContent (in, tempParserString, &xmlParser . tokenState);
   if (xmlParser . tokenState != XML_TOKEN_CONTENT) {
     errorMessage = g_strdup_printf ("\nBad \'<tax:catalogNumber>\' token\nFile %s, line %ld\n", in -> name, in -> line);
     OnError ();
     return TRUE;
   }
   catalogName = g_strdup (tempParserString -> str);
 }
 else if ((!g_strcmp0 (tempParserString -> str, "tax:decimallatitude")) || (!g_strcmp0 (tempParserString -> str, "tax:decimallongitude"))) {
   actToken = GetPrevToken (&xmlParser);
   if (actToken == NULL) {
     errorMessage = g_strdup_printf ("\nUnexpected \'<%s>\' token\nFile %s, line %ld\n", tempParserString -> str, in -> name, in -> line);
     return FALSE;
   }
   if (g_strcmp0 (actToken -> str, "tax:taxonoccurrence")) {
     errorMessage = g_strdup_printf ("\nUnexpected \'<%s>\' token\nFile %s, line %ld\n", tempParserString -> str, in -> name, in -> line);
     return FALSE;
   }
   if (xmlParser . tokenState & XML_WITH_PARAMS) {
     if (SkipParams (in) == XML_AUTO_CLOSE) {
       RemoveFromTokenList (&xmlParser);
       return TRUE;
     }
   }
   actToken = (GString*) xmlParser . tokenStack -> data;
   ReadXMLContent (in, tempParserString, &xmlParser . tokenState);
   if (xmlParser . tokenState != XML_TOKEN_CONTENT) {
     errorMessage = g_strdup_printf ("\nBad \'<%s>\' token\nFile %s, line %ld\n", actToken -> str, in -> name, in -> line);
     return FALSE;
   }
   SetInStreamToString (cmdLine, tempParserString);
   if (GetCmdReal (&tmp) == TRUE) {
     if (!g_strcmp0 (actToken -> str, "tax:decimallatitude"))
       addLat = tmp;
     else if (!g_strcmp0 (actToken -> str, "tax:decimallongitude"))
       addLon = tmp;
   }
 }
 else if (!g_strcmp0 (tempParserString -> str, "tax2:namecomplete")) {
   if (xmlParser . tokenState & XML_WITH_PARAMS) {
     if (SkipParams (in) == XML_AUTO_CLOSE) {
       RemoveFromTokenList (&xmlParser);
       return TRUE;
     }
   }
   actToken = (GString*) xmlParser . tokenStack -> data;
   ReadXMLContent (in, tempParserString, &xmlParser . tokenState);
   if (xmlParser . tokenState != XML_TOKEN_CONTENT) {
     errorMessage = g_strdup_printf ("\nBad \'<%s>\' token\nFile %s, line %ld\n", actToken -> str, in -> name, in -> line);
     return FALSE;
   }
   termName = g_string_new (tempParserString -> str);
 }
 else if (!g_strcmp0 (tempParserString -> str, "tax:earliestdatecollected")) {
   actToken = GetPrevToken (&xmlParser);
   if (actToken == NULL) {
     errorMessage = g_strdup_printf ("\nUnexpected \'<%s>\' token\nFile %s, line %ld\n", tempParserString -> str, in -> name, in -> line);
     return FALSE;
   }
   if (g_strcmp0 (actToken -> str, "tax:taxonoccurrence")) {
     errorMessage = g_strdup_printf ("\nUnexpected \'<%s>\' token\nFile %s, line %ld\n", tempParserString -> str, in -> name, in -> line);
     return FALSE;
   }
   if (xmlParser . tokenState & XML_WITH_PARAMS) {
     if (SkipParams (in) == XML_AUTO_CLOSE) {
       RemoveFromTokenList (&xmlParser);
       return TRUE;
     }
   }
   ReadXMLContent (in, tempParserString, &xmlParser . tokenState);
   if (xmlParser . tokenState == XML_TOKEN_NOCONTENT) return TRUE;
   if (xmlParser . tokenState != XML_TOKEN_CONTENT) {
     errorMessage = g_strdup_printf ("\nBad \'<tax:catalogNumber>\' token\nFile %s, line %ld\n", in -> name, in -> line);
     OnError ();
     return TRUE;
   }
   sampleName = g_strdup (tempParserString -> str);
 }
 else
   return IgnoreOpenTag (&xmlParser);
 return TRUE;
}

SFLAG EndDarwinTag () {
 gchar* numStr;
 gchar* tmp;

 if (!g_strcmp0 (tempParserString -> str, "tax:taxonoccurrence")) {
   if ((addLon > -180.00001) && (addLon < 180.00001) && (addLat > -90.00001) && (addLat < 90.00001) && (termName != NULL)) {
     term = GetTermFromName (termName -> str);
     if (term != NULL) {
       actGeo = AddGeoRef (addLon, addLat, term -> node);
       actGeo -> name = AddSynonim (term, termName -> str);
       actGeo -> collec = g_strdup (collecName);
       actGeo -> catalog = catalogName;
       actGeo -> url = urlName;
       actGeo -> reference = referenceName;
       actGeo -> identifier = identifierName;
       actGeo -> sampleDate = sampleName;
       recordName = NULL;
       catalogName = NULL;
       urlName = NULL;
       referenceName = NULL;
       identifierName = NULL;
       sampleName = NULL;
     }
     else {
       g_string_append_printf (outBuffer, "\n<Folder>");
       g_string_append_printf (outBuffer, "\n\t<name>%s</name>", termName -> str);
       g_string_append (outBuffer, "\n\t<SpecimenInfo>");
       g_string_append_printf (outBuffer,"\n\t\t<name>%s</name>", termName -> str);
       if (collecName != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<collection>%s</collection>", collecName);
       if (catalogName != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<catalogCode>%s</catalogCode>", catalogName);
       if (urlName != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<url>%s</url>", urlName);
       if (referenceName != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<reference>%s</reference>", referenceName);
       if (identifierName != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<identifiedBy>%s</identifiedBy>", identifierName);
       if (sampleName != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<collectionDate>%s</collectionDate>", sampleName);
       if (collectorName != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<collector>%s</collector>", collectorName);
       g_string_append (outBuffer, "\n\t\t<Point>");
       g_string_append (outBuffer, "\n\t\t\t<coordinates>");
       numStr = g_strdup_printf ("%.6f", addLon);
       CheckRealStr (numStr);
       g_string_append_printf (outBuffer, "%s,", numStr);
       g_free (numStr);
       numStr = g_strdup_printf ("%.6f", addLat);
       CheckRealStr (numStr);
       g_string_append (outBuffer, numStr);
       g_free (numStr);
       g_string_append (outBuffer, "</coordinates>");
       g_string_append (outBuffer, "\n\t\t</Point>\n\t</SpecimenInfo>");
       g_string_append (outBuffer, "\n</Folder>");
       unReadList = NameOnList (termName -> str, unReadList);
     }
   }
   tmp = g_strdup (collecName);
   FreeNames ();
   collecName = tmp;
   term = NULL;
   if (termName != NULL)
     g_string_free (termName, TRUE);
   termName = NULL;
   actGeo = NULL;
   addLon = 360.00;
   addLat = 360.00;
 }
 else if (!g_strcmp0 (tempParserString -> str, "dataresource")) {
   if (collecName != NULL)
     g_free (collecName);
   collecName = NULL;
 }
 return TRUE;
}

static GPtrArray* recArray = NULL;
static SFLAG validFile = FALSE;

SFLAG OpenVRXTag () {
 GString* actToken;
 SCOUNTER idVal;
 GList* nuRec;
 SFLAG onMarkup;

 progBar -> actBar1 = in -> actPos;
 UpdateOnCalculus ();
 if (!g_strcmp0 (tempParserString -> str, "vipresults")) {
   validFile = TRUE;
   recArray = NULL;
   actNode = NULL;
   nuTree = NULL;
   xmlParser . fileHeader = g_strdup (tempParserString -> str);
 }
 if (!g_strcmp0 (tempParserString -> str, "treeid")) {
   if (validFile == FALSE) {
     errorMessage = g_strdup_printf ("\nUnknown file format\nFile %s, line %ld\n", in -> name, in -> line);
     return FALSE;
   }
   actToken = GetPrevToken (&xmlParser);
   if (actToken == NULL) {
     errorMessage = g_strdup_printf ("\nUnexpected \'<treeID>\' token\nFile %s, line %ld\n", in -> name, in -> line);
     return FALSE;
   }
   if (g_strcmp0 (actToken -> str, "reconstruction")) {
     errorMessage = g_strdup_printf ("\nUnexpected \'<treeID>\' token\nFile %s, line %ld\n", in -> name, in -> line);
     IgnoreOpenTag (&xmlParser);
     return TRUE;
   }
   if (xmlParser . tokenState & XML_WITH_PARAMS) {
     if (SkipParams (in) == XML_AUTO_CLOSE) {
       RemoveFromTokenList (&xmlParser);
       return TRUE;
     }
   }
   ReadXMLContent (in, tempParserString, &xmlParser . tokenState);
   if (xmlParser . tokenState != XML_TOKEN_CONTENT) {
     errorMessage = g_strdup_printf ("\nBad \'<treeID>\' token: %s\nFile %s, line %ld\n", tempParserString -> str,
                                      in -> name, in -> line);
     OnError ();
     return TRUE;
   }
   SetInStreamToString (cmdLine, tempParserString);
   if (GetCmdNumber (&idVal) == TRUE)
     nuTree = GetTreeFromID (idVal);
   if (nuTree != NULL) {
     SetRec (nuTree -> recOR);
     nuRec = CopyActRec (nuTree);
     recArray = (GPtrArray*) nuRec -> data;
     SetRec (recArray);
     actNode = NULL;
     node = NULL;
   }
 }
 else if (!g_strcmp0 (tempParserString -> str, "nodeid")) {
   if (validFile == FALSE) {
     errorMessage = g_strdup_printf ("\nUnknown file format\nFile %s, line %ld\n", in -> name, in -> line);
     return FALSE;
   }
   actToken = GetPrevToken (&xmlParser);
   if (actToken == NULL) {
     errorMessage = g_strdup_printf ("\nUnexpected \'<nodeID>\' token\nFile %s, line %ld\n", in -> name, in -> line);
     return FALSE;
   }
   onMarkup = 0;
   if (!g_strcmp0 (actToken -> str, "distroelimination"))
     onMarkup = 1;
   else if (!g_strcmp0 (actToken -> str, "partialelimination"))
     onMarkup = 2;
   else {
     IgnoreOpenTag (&xmlParser);
     return TRUE;
   }
   if (xmlParser . tokenState & XML_WITH_PARAMS) {
     if (SkipParams (in) == XML_AUTO_CLOSE) {
       RemoveFromTokenList (&xmlParser);
       return TRUE;
     }
   }
   ReadXMLContent (in, tempParserString, &xmlParser . tokenState);
   if (xmlParser . tokenState != XML_TOKEN_CONTENT) {
     errorMessage = g_strdup_printf ("\nBad \'<nodeID>\' token: %s\nFile %s, line %ld\n", tempParserString -> str,
                                      in -> name, in -> line);
     OnError ();
     return TRUE;
   }
   SetInStreamToString (cmdLine, tempParserString);
   if (GetCmdNumber (&idVal) == TRUE) {
     node = GetNodeFromID (nuTree, idVal);
     if (node != NULL)
       actNode = node -> node;
   }
   if (actNode != NULL) {
     if (onMarkup == 1)
       SetNodeAsRemoved (actNode);
     else if (onMarkup == 2)
       SetNodeAsPartialRemov (actNode);
     actNode = NULL;
     node = NULL;
   }
   else {
     errorMessage = g_strdup_printf ("\nBad \'<nodeID>\' token: %s\nFile %s, line %ld\n", tempParserString -> str,
                                      in -> name, in -> line);
     OnError ();
   }
 }
 else {
   if (validFile == FALSE) {
     errorMessage = g_strdup_printf ("\nUnknown file format\nFile %s, line %ld\n", in -> name, in -> line);
     return FALSE;
   }
   return IgnoreOpenTag (&xmlParser);
 }
 return TRUE;
}

SFLAG EndVRXTag () {
 if (!g_strcmp0 (tempParserString -> str, "reconstruction")) {
   recArray = NULL;
   actNode = NULL;
   nuTree = NULL;
 }
 return TRUE;
}

SFLAG OnVRXError () {
 OnError ();
 FinishVRXFile ();
 return FALSE;
}

void FinishVRXFile () {
 recArray = NULL;
 actNode = NULL;
 nuTree = NULL;
 validFile = FALSE;
}

SFLAG FeedXYDFile () {
 SFLAG state;
 gchar* numStr;
 SCOUNTER numRecs;

 StartNuFeed ();
 for (;;) {
   progBar -> actBar1 = in -> actPos;
   UpdateOnCalculus ();
   if (ReadString (in, tempParserString) == EOF_REACHED) {
     errorMessage = g_strdup_printf ("Unexpected End of File\nExpecting \'xydata\' keyword\nFile: %s", in -> name);
     return OnFeedError ();
   }
   if (!g_strcmp0 (tempParserString -> str, "xydata")) break;
 }
 term = NULL;
 if (termName != NULL)
   g_string_free (termName, TRUE);
 termName = NULL;
 for (;;) {
   progBar -> actBar1 = in -> actPos;
   UpdateOnCalculus ();
   if (ReadString (in, tempParserString) == EOF_REACHED) break;
   if (g_strcmp0 (tempParserString -> str, "sp")) {
     errorMessage = g_strdup_printf ("Unknown expression %s\nExpecting \'sp\' keyword\nFile: %s, line: %ld",
                    tempParserString -> str, in -> name, in -> line);
     return OnFeedError ();
   }
   if (ReadInt (in, NULL) == BAD_IN_FORMAT) {
     errorMessage = g_strdup_printf ("Expecting number after \'sp\' keyword\nFile: %s, line: %ld", in -> name, in -> line);
     return OnFeedError ();
   }
   state = ReadBoundComment (in, tempParserString, '[', ']');
   if (state == EOF_REACHED) {
     errorMessage = g_strdup_printf ("Unclosed species name\nExpecting \']\'\nFile: %s", in -> name);
     return OnFeedError ();
   }
   if (state != BAD_IN_FORMAT) {
     termName = g_string_new (tempParserString -> str);
     term = GetTermFromName (termName -> str);
     if (term == NULL)
       unReadList = NameOnList (termName -> str, unReadList);
   }
   numRecs = 0;
   for (;;) {
     progBar -> actBar1 = in -> actPos;
     UpdateOnCalculus ();
     if (in -> state == EOF_REACHED) break;
     if ((in -> nextChar == 's') || (in -> nextChar == 'S')) break;
     addLat = addLon = 360.00;
     if (ReadReal (in, &addLon) == BAD_IN_FORMAT) {
       errorMessage = g_strdup_printf ("Expecting number\nFile: %s, line: %ld", in -> name, in -> line);
       return OnFeedError ();
     }
     if (in -> state == EOF_REACHED) {
       errorMessage = g_strdup_printf ("Unexpected End of File\nFile: %s", in -> name);
       return OnFeedError ();
     }
     if (ReadReal (in, &addLat) == BAD_IN_FORMAT) {
       errorMessage = g_strdup_printf ("Expecting number\nFile: %s, line: %ld", in -> name, in -> line);
       return OnFeedError ();
     }
     if ((addLon > -180.00001) && (addLon < 180.00001) && (addLat > -90.00001) && (addLat < 90.00001) && (termName != NULL)) {
       if (term != NULL) {
         actGeo = AddGeoRef (addLon, addLat, term -> node);
         actGeo -> name = AddSynonim (term, termName -> str);
       }
       else {
         if (numRecs == 0) {
           g_string_append_printf (outBuffer, "\n<Folder>");
           g_string_append_printf (outBuffer, "\n\t<name>%s</name>", termName -> str);
         }
         g_string_append (outBuffer, "\n\t<SpecimenInfo>");
         g_string_append_printf (outBuffer,"\n\t\t<name>%s</name>", termName -> str);
         g_string_append (outBuffer, "\n\t\t<Point>");
         g_string_append (outBuffer, "\n\t\t\t<coordinates>");
         numStr = g_strdup_printf ("%.6f", addLon);
         CheckRealStr (numStr);
         g_string_append_printf (outBuffer, "%s,", numStr);
         g_free (numStr);
         numStr = g_strdup_printf ("%.6f", addLat);
         CheckRealStr (numStr);
         g_string_append (outBuffer, numStr);
         g_free (numStr);
         g_string_append (outBuffer, "</coordinates>");
         g_string_append (outBuffer, "\n\t\t</Point>\n\t</SpecimenInfo>");
         ++ numRecs;
       }
     }
   }
   if ((numRecs > 0) && (term == NULL))
     g_string_append (outBuffer, "\n</Folder>");
   if (termName != NULL)
     g_string_free (termName, TRUE);
   term = NULL;
   termName = NULL;
 }
 return TRUE;
}

SFLAG FeedTABTable () {
 SINTEGER nameCount;
 SINTEGER collecCount;
 SINTEGER catalogCount;
 SINTEGER urlCount;
 SINTEGER referenceCount;
 SINTEGER identifierCount;
 SINTEGER sampleCount;
 SINTEGER collectorCount;
 SINTEGER latCount;
 SINTEGER lonCount;
 SCOUNTER i;
 SFLAG state;
 SREAL tmpNum;
 gchar* numStr;

 nameCount = collecCount = catalogCount = referenceCount = collectorCount = -1;
 urlCount = identifierCount = sampleCount = -1;
 latCount = lonCount = -1;

 /* Read headers */
 for (i = 0;; ++ i) {
   progBar -> actBar1 = in -> actPos;
   UpdateOnCalculus ();
   state = ReadStringUntil (in, tempParserString, '\t');
   if (state == EOF_REACHED) {
     errorMessage = g_strdup_printf ("Unknown file format\nFile: %s", in -> name);
     OnError ();
     return FALSE;
   }
   if (!g_strcmp0 (tempParserString -> str, "scientific name"))
     nameCount = i;
   else if (!g_strcmp0 (tempParserString -> str, "latitude"))
     latCount = i;
   else if (!g_strcmp0 (tempParserString -> str, "longitude"))
     lonCount = i;
   else if (!g_strcmp0 (tempParserString -> str, "dataset"))
     collecCount = i;
   else if ((!g_strcmp0 (tempParserString -> str, "catalogue no")) || (!g_strcmp0 (tempParserString -> str, "catalogue number")) ||
            (!g_strcmp0 (tempParserString -> str, "catalogue code")))
     catalogCount = i;
   else if ((!g_strcmp0 (tempParserString -> str, "gbif webservice url")) || (!g_strcmp0 (tempParserString -> str, "url")) ||
           (!g_strcmp0 (tempParserString -> str, "link")))
     urlCount = i;
   else if ((!g_strcmp0 (tempParserString -> str, "reference")) || (!g_strcmp0 (tempParserString -> str, "primary reference")))
     referenceCount = i;
   else if ((!g_strcmp0 (tempParserString -> str, "identifier")) || (!g_strcmp0 (tempParserString -> str, "authority")))
     identifierCount = i;
   else if (!g_strcmp0 (tempParserString -> str, "date collected"))
     sampleCount = i;
   else if ((!g_strcmp0 (tempParserString -> str, "collector")) || (!g_strcmp0 (tempParserString -> str, "collector name")) ||
           (!g_strcmp0 (tempParserString -> str, "collected by")))
     collectorCount = i;
   if (state == END_OF_IN_LINE) break;
 }
 if ((nameCount == -1) || (latCount == -1) || (lonCount == -1)) {
   errorMessage = g_strdup_printf ("Unknown file format\nFile: %s", in -> name);
   OnError ();
   return FALSE;
 }
 SETREADINGMODE (READ_AS_IS);
 StartNuFeed ();
 for (;;) {
   progBar -> actBar1 = in -> actPos;
   UpdateOnCalculus ();
   for (i = 0;; ++ i) {
     state = ReadStringUntil (in, tempParserString, '\t');
     if (state == EOF_REACHED) break;
     if (tempParserString -> len == 0) {
       if (state == END_OF_IN_LINE) break;
       continue;
     }
     if (i == nameCount) {
       termName = g_string_new (tempParserString -> str);
       term = GetTermFromName (termName -> str);
       if (term == NULL)
         unReadList = NameOnList (termName -> str, unReadList);
     }
     else if ((i == latCount) || (i == lonCount)) {
       SetInStreamToString (cmdLine, tempParserString);
       if (GetCmdReal (&tmpNum) != TRUE) {
         errorMessage = g_strdup_printf ("Expecting real value\nFile: %s, line: %ld, column: %ld", in -> name, in -> line, i);
         return OnFeedError ();
       }
       if (i == latCount)
         addLat = tmpNum;
       else if (i == lonCount)
         addLon = tmpNum;
     }
     else if (i == collecCount)
       collecName = g_strdup (tempParserString -> str);
     else if (i == catalogCount)
       catalogName = g_strdup (tempParserString -> str);
     else if (i == urlCount)
       urlName = g_strdup (tempParserString -> str);
     else if (i == referenceCount)
       referenceName = g_strdup (tempParserString -> str);
     else if (i == identifierCount)
       identifierName = g_strdup (tempParserString -> str);
     else if (i == sampleCount)
       sampleName = g_strdup (tempParserString -> str);
     else if (i == collectorCount)
       collectorName = g_strdup (tempParserString -> str);
     if (state == END_OF_IN_LINE) break;
   }
   if (state == EOF_REACHED) break;
   if ((addLon > -180.00001) && (addLon < 180.00001) && (addLat > -90.00001) && (addLat < 90.00001) && (termName != NULL)) {
     if (term != NULL) {
       actGeo = AddGeoRef (addLon, addLat, term -> node);
       actGeo -> name = AddSynonim (term, termName -> str);
       actGeo -> collec = g_strdup (collecName);
       actGeo -> catalog = catalogName;
       actGeo -> url = urlName;
       actGeo -> reference = referenceName;
       actGeo -> identifier = identifierName;
       actGeo -> sampleDate = sampleName;
       actGeo -> collector = collectorName;
       recordName = NULL;
       catalogName = NULL;
       urlName = NULL;
       referenceName = NULL;
       identifierName = NULL;
       sampleName = NULL;
       collectorName = NULL;
     }
     else {
       g_string_append_printf (outBuffer, "\n<Folder>");
       g_string_append_printf (outBuffer, "\n\t<name>%s</name>", termName -> str);
       g_string_append (outBuffer, "\n\t<SpecimenInfo>");
       g_string_append_printf (outBuffer,"\n\t\t<name>%s</name>", termName -> str);
       if (collecName != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<collection>%s</collection>", collecName);
       if (catalogName != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<catalogCode>%s</catalogCode>", catalogName);
       if (urlName != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<url>%s</url>", urlName);
       if (referenceName != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<reference>%s</reference>", referenceName);
       if (identifierName != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<identifiedBy>%s</identifiedBy>", identifierName);
       if (sampleName != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<collectionDate>%s</collectionDate>", sampleName);
       if (collectorName != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<collector>%s</collector>", collectorName);
       g_string_append (outBuffer, "\n\t\t<Point>");
       g_string_append (outBuffer, "\n\t\t\t<coordinates>");
       numStr = g_strdup_printf ("%.6f", addLon);
       CheckRealStr (numStr);
       g_string_append_printf (outBuffer, "%s,", numStr);
       g_free (numStr);
       numStr = g_strdup_printf ("%.6f", addLat);
       CheckRealStr (numStr);
       g_string_append (outBuffer, numStr);
       g_free (numStr);
       g_string_append (outBuffer, "</coordinates>");
       g_string_append (outBuffer, "\n\t\t</Point>\n\t</SpecimenInfo>");
       g_string_append (outBuffer, "\n</Folder>");
     }
   }
   if (termName != NULL)
     g_string_free (termName, TRUE);
   term = NULL;
   termName = NULL;
   FreeNames ();
   addLon = addLat = 360.00;
 }
 return TRUE;
}

void StartNuFeed () {
 term = NULL;
 unReadList = NULL;
 termName = NULL;
 actGeo = NULL;
 FreeNames ();
 outBuffer = g_string_sized_new (1024);
 g_string_append (outBuffer, "<?xml version=\"1.0\"?>");
 g_string_append_printf (outBuffer, "\n<!-- Unreaded records from file %s -->", in -> name);
 g_string_append (outBuffer, "\n\n<Vip>");
 addLon = 360.00;
 addLat = 360.00;
}

void FreeNames () {
 if (collecName != NULL)
   g_free (collecName);
 if (catalogName != NULL)
   g_free (catalogName);
 if (urlName != NULL)
   g_free (urlName);
 if (referenceName != NULL)
   g_free (referenceName);
 if (identifierName != NULL)
   g_free (identifierName);
 if (sampleName != NULL)
   g_free (sampleName);
 recordName = NULL;
 collecName = NULL;
 catalogName = NULL;
 urlName = NULL;
 referenceName = NULL;
 identifierName = NULL;
 sampleName = NULL;
 collectorName = NULL;
}

void FinishUnRead () {
 GString* unReadBuffer;
 GList* nav;
 GString* name;

 if (unReadList != NULL) {
   g_string_append (outBuffer, "\n\n</Vip>\n");
   WriteOutLog ("unfeed.xml", outBuffer);
   unReadBuffer = g_string_sized_new (1024);
   g_string_append (unReadBuffer, "<?xml version=\"1.0\"?>");
   g_string_append (unReadBuffer, "\n\n<!-- Unmatched names -->\n");
   g_string_append (unReadBuffer, "\n\n<Vip>\n");
   for (nav = unReadList; nav != NULL; nav = g_list_next (nav)) {
     name = (GString*) nav -> data;
     g_string_append_printf (unReadBuffer, "\n<Folder>\n\t<name>%s</name>", name -> str);
     g_string_append_printf (unReadBuffer, "\n\t<synonym>%s</synonym>", name -> str);
     g_string_append (unReadBuffer, "\n</Folder>");
   }
   g_string_append (unReadBuffer, "\n\n</Vip>\n");
   g_string_append (unReadBuffer, "\n");
   WriteOutLog ("unmatch.xml", unReadBuffer);
   g_string_free (unReadBuffer, TRUE);
   CleanTokenStack (unReadList);
 }
 g_string_free (outBuffer, TRUE);
 if (termName != NULL)
   g_string_free (termName, TRUE);
}

void CleanTokenStack (GList* toClean) {
 GList* nav;
 for (nav = toClean; nav != NULL; nav = g_list_next (nav))
   g_string_free (nav -> data, TRUE);
 g_list_free (toClean);
}

SFLAG ReadRealTuple (INSTREAM* from, SREAL* r1, SREAL* r2) {
 if ((!g_ascii_isdigit (from -> nextChar)) && (from -> nextChar != '-')) return FALSE;
 ReadReal (from, r1);
 if (from -> nextChar == ',')
   ReadChar (from, NULL);
 SkipSpaces (from);
 if ((!g_ascii_isdigit (from -> nextChar)) && (from -> nextChar != '-')) return FALSE;
 ReadReal (from, r2);
 return TRUE;
}
void SetCommandLine (gchar* text) {
 if (cmdLine == NULL)
   cmdLine = CreateInStream ();
 else
   ResetInStream (cmdLine);
 if (tempParserString == NULL)
   tempParserString = g_string_new (VOIDSTRING);
 g_string_printf (tempParserString, text);
 SetInStreamToString (cmdLine, tempParserString);
}

SFLAG GetCmdNumber (SINTEGER* dest) {
 if (cmdLine -> nextChar == 0) return FALSE;
 SkipSpaces (cmdLine);
 if ((!g_ascii_isdigit (cmdLine -> nextChar)) && (cmdLine -> nextChar != '-')) return FALSE;
 ReadInt (cmdLine, dest);
 return TRUE;
}

SFLAG GetCmdReal (SREAL* dest) {
 if (cmdLine -> nextChar == 0) return FALSE;
 SkipSpaces (cmdLine);
 if ((!g_ascii_isdigit (cmdLine -> nextChar)) && (cmdLine -> nextChar != '-')) return FALSE;
 ReadReal (cmdLine, dest);
 return TRUE;
}

SFLAG IsXYDFile () {
 SkipSpaces (in);
 if (ReadString (in, tempParserString) == EOF_REACHED) {
   in -> line = 0;
   in -> actPos = 0;
   SkipSpaces (in);
   return FALSE;
 }
 if (!g_strcmp0 (tempParserString -> str, "spp")) {
   if (ReadInt (in, NULL) == BAD_IN_FORMAT) {
     in -> line = 0;
     in -> actPos = 0;
     SkipSpaces (in);
     return FALSE;
   }
   return TRUE;
 }
 in -> line = 0;
 in -> actPos = 0;
 SkipSpaces (in);
 return FALSE;
}
/* search.c is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */

#include "biogeo.h"
#include "bitfield.h"
#include "grid.h"
#include "tessellation.h"

#define JUST_COUNT_COST     1
#define FULL_OPTIMIZATION   2
#define SEARCH_OPTIMIZATION 3

/* Code specificly used for searches and optimizations */

static SREAL OptimizeNode (GNode* node, SFLAG mode);
static void OptimizeAll (GPtrArray* recArray, SFLAG optimize);

static void CopyRec (RECONS* dest, RECONS* src);
static GPtrArray* CreateRecArrayCopy (GPtrArray* src);
static void CopyRecArray (GPtrArray* dest, GPtrArray* src);

static SFLAG CompareRec (GPtrArray* probe, GList* recList);
static SFLAG CheckReconstruction (PHYLOTREE* tree, GPtrArray* recArray, SCOUNTER* stored, SCOUNTER* bestHits, SFLAG add);

static SREAL IncrementalDownPass (GNode* startNode);

static void MeasureOverlap (RECONS* actRec);

static void NodeRemoval (PHYLOTREE* actTree, GPtrArray* sectorNodeArray);
static void BlindAnnealing (PHYLOTREE* actTree, GPtrArray* recArray, SCOUNTER* stored);
static void GibbsAnnealing (PHYLOTREE* actTree, GPtrArray* recArray, GPtrArray* sectorArray, INTARRAY* prob, SCOUNTER* stored);
static void NixonAnnealing (PHYLOTREE* actTree, GPtrArray* recArray, GPtrArray* sectorArray, SCOUNTER* stored);

static void UpdateRecCounts (GPtrArray* recArray);

static void SetDefaultBremer (GPtrArray* recArray);
static void UpdateBremer (GNode* startNode, SREAL target);
static void SetBremer (GPtrArray* recArray);

static GPtrArray* tmpNodeArray = NULL;

void SetRec (GPtrArray* recArray) {
 SCOUNTER i;
 RECONS* actRec;

 for (i = recArray -> len - 1; i >= 0; -- i) {
   actRec = (RECONS*) g_ptr_array_index (recArray, i);
   actRec -> node -> rec = actRec;
 }
}

void MakeOptimization (SFLAG justCount) {
 GList* navTree;
 GList* navRec;
 PHYLOTREE* tree;
 GNode* navNode;
 PHYLONODE* actNode;
 RECONS* actRec;
 GPtrArray* rec;

 for (navTree = trees; navTree != NULL; navTree = g_list_next (navTree)) {
   tree = (PHYLOTREE*) navTree -> data;
   if (tree -> recOR == NULL) {
     /* There are no reconstruction, first creates the array */
     tree -> recOR = g_ptr_array_new ();
     for (navNode = tree -> root; navNode != NULL; ) {
       actNode = (PHYLONODE*) navNode -> data;
       if (actNode -> rec == NULL) {
         actRec = CreateRecons (actNode);
         g_ptr_array_add (tree -> recOR, actRec);
         actNode -> rec = actRec;
         if (g_node_first_child (navNode) != NULL) {
           navNode = g_node_first_child (navNode);
           continue;
         }
       }
       if (g_node_next_sibling (navNode) == NULL)
         navNode = navNode -> parent;
       else
         navNode = g_node_next_sibling (navNode);
     }
     actNode = (PHYLONODE*) tree -> root -> data;
     OptimizeAll (tree -> recOR, TRUE);
     tree -> best = tree -> worst = actNode -> rec -> cost;
     SetDefaultBremer (tree -> recOR);
   }
   else {
     /* If there are a reconstruction, updates the costs of the reconstructions */
     SetRec (tree -> recOR);
     if (justCount == TRUE)
       OptimizeAll (tree -> recOR, FALSE);
     else
       OptimizeAll (tree -> recOR, TRUE);
     actNode = (PHYLONODE*) tree -> root -> data;
     tree -> best = tree -> worst = actNode -> rec -> cost;
     SetDefaultBremer (tree -> recOR);
     if (tree -> recs != NULL) {
       for (navRec = tree -> recs; navRec != NULL; navRec = g_list_next (navRec)) {
         rec = (GPtrArray*) navRec -> data;
         SetRec (rec);
         if (justCount == TRUE)
           OptimizeAll (rec, FALSE);
         else
           OptimizeAll (rec, TRUE);
         SetBremer (rec);
         actNode = (PHYLONODE*) tree -> root -> data;
         if (actNode -> rec -> cost < tree -> best)
           tree -> best = actNode -> rec -> cost;
         if (actNode -> rec -> cost > tree -> worst)
           tree -> worst = tree -> worst;
       }
     }
     SetRec (tree -> recOR);
   }
 }
}

SREAL OptimizeNode (GNode* node, SFLAG mode) {
 RECONS* actRec;
 PHYLONODE* actNode;
 PHYLONODE* descNode;
 PHYLONODE* left;
 PHYLONODE* right;
 GNode* navDesc;
 SCOUNTER ageDif;
 SREAL timeFraq;

 actNode = (PHYLONODE*) node -> data;
 actRec = actNode -> rec;
 actRec -> cost = 0;
 actRec -> numVics = actRec -> numNoVics = actRec -> numRemovNodes = 0;
 actRec -> nodeFlag &= REMOV_NODE;        /* If the node is a removed, then nodeFlag is equal to REMOV_NODE, otherwise nodeFlag is 0 */
 if (actRec -> triangles != NULL)
   DestroyTriangleList (actRec -> triangles);
 actRec -> triangles = NULL;
 if (g_node_first_child (actNode -> node) == NULL) {
   if (mode == FULL_OPTIMIZATION) {
     CopyBitField (actRec -> obs, actNode -> term -> obs);
     if (CountBits (actRec -> obs) > 0) {
       DoFill (actRec -> cells, actRec -> obs, actRec -> node -> fill);
       actRec -> numCells = CountBits (actRec -> cells);
       actRec -> numObs = CountBits (actRec -> obs);
       actRec -> maxAge = actNode -> minAge;
     }
     else {         /* An empty set "optimization" */
       ResetBitField (actRec -> cells);
       actRec -> numCells = actRec -> numObs = 0;
     }
   }
   if (actRec -> numCells == 0) return 0;   /* Empty set exit */
   if (actRec -> nodeFlag & REMOV_NODE) {
     actRec -> numRemovNodes = 1;
     actRec -> cost = params . costRemov;
   }
   return actRec -> cost;
 }
 if (mode & (FULL_OPTIMIZATION | SEARCH_OPTIMIZATION)) {
   ResetBitField (actRec -> obs);
   ResetBitField (actRec -> cells);
   actRec -> numObs = 0;
   actRec -> numCells = 0;
   actRec -> active = 0;
   actRec -> maxAge = actNode -> minAge;
   actRec -> setA = NULL;
   actRec -> setB = NULL;
   if (!(params . search & USE_LATEOVER))
     actRec -> removSet = NULL;
 }
 left = right = NULL;
 ageDif = 0;
 for (navDesc = g_node_first_child (actNode -> node); navDesc != NULL; navDesc = g_node_next_sibling (navDesc)) {
   descNode = (PHYLONODE*) navDesc -> data;
   actRec -> numVics += descNode -> rec -> numVics;
   actRec -> numNoVics += descNode -> rec -> numNoVics;
   actRec -> numRemovNodes += descNode -> rec -> numRemovNodes;
   actRec -> cost += descNode -> rec -> cost;
   if (mode & (FULL_OPTIMIZATION | SEARCH_OPTIMIZATION)) {
     if ((descNode -> rec -> numCells == 0) || (descNode -> rec -> nodeFlag & REMOV_NODE)) continue;
     if (descNode != actRec -> removSet) {
       BitFieldORsum (actRec -> obs, descNode -> rec -> obs);
       BitFieldORsum (actRec -> cells, descNode -> rec -> cells);
       if ((actRec -> maxAge < descNode -> rec -> maxAge) && (descNode -> rec -> maxAge != NO_AGE))
         actRec -> maxAge = descNode -> rec -> maxAge;
     }
     ++ actRec -> active;
     if (left == NULL)
       left = descNode;
     else if (right == NULL)
       right = descNode;
   }
 }
 if (mode & (FULL_OPTIMIZATION | SEARCH_OPTIMIZATION)) {
   actRec -> numCells = CountBits (actRec -> cells);
   if (actRec -> numCells == 0) {
     /* This is an empty set */
     actRec -> numObs = 0;
     actRec -> cost = 0;
     return actRec -> cost;
   }
   actRec -> numObs = CountBits (actRec -> obs);
 }
 if (actRec -> nodeFlag & REMOV_NODE) {
   ++ actRec -> numRemovNodes;
   actRec -> cost += params . costRemov;
 }
 if (actRec -> removSet != NULL) {
   ++ actRec -> numRemovNodes;
   actRec -> cost += params . costPartRemov;
 }
 if (actRec -> active == 1) {
   actRec -> nodeFlag &= ~ VIC_NODE;
   if (params . costRemov == 0) {
     ++ actRec -> numNoVics;
     actRec -> cost += params . costNoVic;
   }
   if ((params . search & CONSTR_SEARCH) && (actNode -> constrained))
     actRec -> cost += 10000;
   return actRec -> cost;
 }
 else if (actRec -> active != 2) {
   actRec -> nodeFlag &= ~ VIC_NODE;
   ++ actRec -> numNoVics;
   actRec -> cost += params . costNoVic;
   if ((params . search & CONSTR_SEARCH) && (actNode -> constrained))
     actRec -> cost += 10000;
   return actRec -> cost;
 }
 if (mode & (FULL_OPTIMIZATION | SEARCH_OPTIMIZATION)) {
   actRec -> timeLag = actNode -> minAge - actRec -> maxAge;
   /* Preserves the order, indep of tree sorting */
   if (left -> id < right -> id) {
     actRec -> setA = left;
     actRec -> setB = right;
   }
   else {
     actRec -> setB = left;
     actRec -> setA = right;
   }
   MeasureOverlap (actRec);
 }
 if ((actRec -> overlap <= params . overlap) && (actRec -> overlap >= 0)) {
   if (params . mode & USE_TIME) {
     if (actRec -> timeLag > params . maxTime) {
       actRec -> nodeFlag &= ~ VIC_NODE;
       ++ actRec -> numNoVics;
       actRec -> cost += params . costNoVic;
       if ((params . search & CONSTR_SEARCH) && (actNode -> constrained))
         actRec -> cost += 10000;
       return actRec -> cost;
     }
     if (actRec -> timeLag > params . timeLag) {
       if (params . mode & TIME_REAL) {
         timeFraq = (SREAL) (actRec -> timeLag - params . timeLag) / params . timeDiff;
         actRec -> cost += timeFraq * params . costTimeLag;
       }
       else
         actRec -> cost += params . costTimeLag;
       actRec -> nodeFlag |= TIME_LAG;
     }
   }
   /* Reset the age of the node */
   actRec -> maxAge = actNode -> minAge;
   actRec -> nodeFlag |= VIC_NODE;
   ++ actRec -> numVics;
   if (params . mode & OVER_REAL)
     actRec -> cost += actRec -> overlap;
 }
 else {
   actRec -> nodeFlag &= ~ VIC_NODE;
   ++ actRec -> numNoVics;
   actRec -> cost += params . costNoVic;
   if ((params . search & CONSTR_SEARCH) && (actNode -> constrained))
     actRec -> cost += 10000;
 }
 return actRec -> cost;
}

void OptimizeAll (GPtrArray* recArray, SFLAG optimize) {
 SINTEGER i;
 RECONS* actRec;

 for (i = recArray -> len - 1; i >= 0; -- i) {
   actRec = (RECONS*) g_ptr_array_index (recArray, i);
   OptimizeNode (actRec -> node -> node, FULL_OPTIMIZATION);
 }
}

void SetDefaultBremer (GPtrArray* recArray) {
 SREAL orVal;
 SINTEGER i;
 RECONS* actRec;

 actRec = (RECONS*) g_ptr_array_index (recArray, 0);
 orVal = actRec -> cost;
 for (i = 0; i < recArray -> len; ++ i) {
   actRec = (RECONS*) g_ptr_array_index (recArray, i);
   if (actRec -> nodeFlag & VIC_NODE)
     actRec -> node -> bremer = G_MAXDOUBLE;
   else
     actRec -> node -> bremer = orVal;
 }
}

void SetBremer (GPtrArray* recArray) {
 SREAL actVal;
 SINTEGER i;
 RECONS* actRec;

 actRec = (RECONS*) g_ptr_array_index (recArray, 0);
 actVal = actRec -> cost;
 for (i = 0; i < recArray -> len; ++ i) {
   actRec = (RECONS*) g_ptr_array_index (recArray, i);
   if (actRec -> nodeFlag & VIC_NODE) continue;
   if (actRec -> node -> bremer > actVal)
     actRec -> node -> bremer = actVal;
 }
}

void UpdateBremer (GNode* startNode, SREAL target) {
 GNode* navNode;
 PHYLONODE* actNode;

 for (navNode = startNode; navNode != NULL; navNode = navNode -> parent) {
   actNode = (PHYLONODE*) navNode -> data;
   if (actNode -> rec -> nodeFlag & VIC_NODE) continue;
   if (actNode -> bremer > target)
     actNode -> bremer = target;
 }
 if (actNode -> rec -> cost < actNode -> tree -> best)
   actNode -> tree -> best = actNode -> rec -> cost;
 if (actNode -> rec -> cost > actNode -> tree -> worst)
   actNode -> tree -> worst = actNode -> tree -> worst;
}

SREAL IncrementalDownPass (GNode* startNode) {
 GNode* navNode;
 SREAL toRetCost;

 toRetCost = G_MAXDOUBLE;
 for (navNode = startNode; navNode != NULL; navNode = navNode -> parent)
   toRetCost = OptimizeNode (navNode, SEARCH_OPTIMIZATION);
 UpdateBremer (startNode, toRetCost);
 return toRetCost;
}

void MeasureOverlap (RECONS* actRec) {
 SREAL divA, divB, maxDiv;
 GNode* navDesc;
 PHYLONODE* descNode;

 actRec -> overlap = 0;
 divA = (SREAL) BitFieldCommon (actRec -> setB -> rec -> cells, actRec -> setA -> rec -> obs) / actRec -> setA -> rec -> numObs;
 divB = (SREAL) BitFieldCommon (actRec -> setA -> rec -> cells, actRec -> setB -> rec -> obs) / actRec -> setB -> rec -> numObs;
 actRec -> overlap = (divA > divB) ? divA : divB;
 /* Check if there is no overlap between a node and the descendants of its sister
    This is to remove the effect of some spurious disjunction, which produce a double detection of
    the same barrier. As is overlap against a more deepest node, it "accepts" up to the double of
    the params . overlap setting */
 if ((actRec -> overlap > 0) && (actRec -> overlap <= params . overlap)) {
   for (navDesc = g_node_first_child (actRec -> setA -> node); navDesc != NULL; navDesc = g_node_next_sibling (navDesc)) {
     descNode = (PHYLONODE*) navDesc -> data;
     if ((descNode -> rec -> numCells == 0) || (descNode -> rec -> nodeFlag & REMOV_NODE)) continue;
     if (descNode == actRec -> setA -> rec -> removSet) continue;
     divA = (SREAL) BitFieldCommon (actRec -> setB -> rec -> cells, descNode -> rec -> obs) / descNode -> rec -> numObs;
     divB = (SREAL) BitFieldCommon (descNode -> rec -> cells, actRec -> setB -> rec -> obs) / actRec -> setB -> rec -> numObs;
     maxDiv = ((divA > divB) ? divA : divB) / 2;
     if (maxDiv > actRec -> overlap)
       actRec -> overlap = maxDiv;
   }
   for (navDesc = g_node_first_child (actRec -> setB -> node); navDesc != NULL; navDesc = g_node_next_sibling (navDesc)) {
     descNode = (PHYLONODE*) navDesc -> data;
     if ((descNode -> rec -> numCells == 0) || (descNode -> rec -> nodeFlag & REMOV_NODE)) continue;
     if (descNode == actRec -> setB -> rec -> removSet) continue;
     divB = (SREAL) BitFieldCommon (actRec -> setA -> rec -> cells, descNode -> rec -> obs) / descNode -> rec -> numObs;
     divA = (SREAL) BitFieldCommon (descNode -> rec -> cells, actRec -> setA -> rec -> obs) / actRec -> setA -> rec -> numObs;
     maxDiv = ((divA > divB) ? divA : divB) / 2;
     if (maxDiv > actRec -> overlap)
       actRec -> overlap = maxDiv;
   }
 }
}

void NodeRemoval (PHYLOTREE* actTree, GPtrArray* sectorNodeArray) {
 SREAL bestScore, iniScore, actScore;
 SFLAG foundBetter, foundEqual, foundOverlap;
 SFLAG removNode, partRemov;
 SFLAG tmpNodeFlag;
 SFLAG update;
 GNode* navNode;
 PHYLONODE* actNode;
 PHYLONODE* tmpNode;
 PHYLONODE* tmpRemov;
 RECONS* actRec;
 SCOUNTER k, bestPos;
 SCOUNTER randNum, randSel;

 foundBetter = TRUE;
 while (foundBetter == TRUE) {
   foundBetter = FALSE;
   foundOverlap = FALSE;
   foundEqual = FALSE;
   ShufflePtrArray (sectorNodeArray);
   actNode = (PHYLONODE*) actTree -> root -> data;
   iniScore = bestScore = actNode -> rec -> cost;
   k = bestPos = 0;
   update = TRUE;
   removNode = partRemov = FALSE;
   for (k = 0; k < sectorNodeArray -> len; ++ k) {
     navNode = (GNode*) g_ptr_array_index (sectorNodeArray, k);
     if (navNode -> parent == NULL) continue;
     actNode = (PHYLONODE*) navNode -> data;
     actRec = actNode -> rec;
     if (actRec -> numCells == 0) continue;
     tmpNode = (PHYLONODE*) navNode -> parent -> data;
     tmpRemov = NULL;
     if (((tmpNode -> rec -> removSet == actNode) || (actRec -> nodeFlag & REMOV_NODE)) &&
         (params . algorithm & FLIP_NODES)) {
       /* Make the backup of the node */
       tmpNodeFlag = actRec -> nodeFlag;
       /* Make the change */
       if (actRec -> nodeFlag & REMOV_NODE)
         actRec -> nodeFlag &= ~REMOV_NODE;
       else {
         tmpRemov = tmpNode -> rec -> removSet;
         tmpNode -> rec -> removSet = NULL;
       }
       actScore = IncrementalDownPass (navNode);
       if (actScore < bestScore) {
         foundBetter = TRUE;
         foundEqual = FALSE;
         bestScore = actScore;
         bestPos = k;
         removNode = partRemov = FALSE;
         if (params . algorithm & FAST_SEARCH) {
           update = FALSE;
           break;
         }
       }
       else if (actScore == bestScore) {
         randNum = g_random_int_range (0, 100);
         if ((foundBetter != TRUE) && (foundEqual != TRUE)) {
           bestPos = k;
           removNode = partRemov = FALSE;
           foundEqual = TRUE;
         }
         else if (randNum < params . acceptEqual) {
           bestPos = k;
           removNode = partRemov = FALSE;
         }
       }
     }
     else {
       if ((tmpNode -> rec -> nodeFlag & REMOV_NODE) || (tmpNode -> rec -> active < 2) ||
           (tmpNode -> rec -> removSet != NULL)) continue;
       if (g_node_first_child (navNode) != NULL) {
         if (actRec -> active < 2) continue;
       }
       /* Make the backup of the node */
       tmpNodeFlag = actRec -> nodeFlag;

       randSel = 0;
       if (params . search & USE_LATEOVER)
         randSel = g_random_int_range (0, 100);
       if ((params . search & USE_LATEOVER) && (randSel >= 50)) {
         tmpNode -> rec -> removSet = actNode;
         actScore = IncrementalDownPass (navNode);
         if (actScore < bestScore) {
           foundBetter = TRUE;
           foundEqual = FALSE;
           bestScore = actScore;
           bestPos = k;
           partRemov = TRUE;
           removNode = FALSE;
           if (params . algorithm & FAST_SEARCH) {
             update = FALSE;
             break;
           }
         }
         else if (actScore == bestScore) {
           randNum = g_random_int_range (0, 100);
           if ((foundBetter != TRUE) && (foundEqual != TRUE)) {
             bestPos = k;
             partRemov = TRUE;
             removNode = FALSE;
             foundEqual = TRUE;
           }
           else if (randNum < params . acceptEqual) {
             bestPos = k;
             partRemov = TRUE;
             removNode = FALSE;
           }
         }
         tmpNode -> rec -> removSet = NULL;
       }
       actRec -> nodeFlag |= REMOV_NODE;
       actScore = IncrementalDownPass (navNode);
       if (actScore < bestScore) {
         foundBetter = TRUE;
         foundEqual = FALSE;
         bestScore = actScore;
         bestPos = k;
         removNode = TRUE;
         partRemov = FALSE;
         if (params . algorithm & FAST_SEARCH) {
           update = FALSE;
           break;
         }
       }
       else if (actScore == bestScore) {
         randNum = g_random_int_range (0, 100);
         if ((foundBetter != TRUE) && (foundEqual != TRUE)) {
           bestPos = k;
           removNode = TRUE;
           partRemov = FALSE;
           foundEqual = TRUE;
         }
         else if (randNum < params . acceptEqual) {
           bestPos = k;
           removNode = TRUE;
           partRemov = FALSE;
         }
       }
       if ((params . search & USE_LATEOVER) && (randSel < 50)) {
         tmpNode -> rec -> removSet = actNode;
         actRec -> nodeFlag = tmpNodeFlag;
         actScore = IncrementalDownPass (navNode);
         if (actScore < bestScore) {
           foundBetter = TRUE;
           foundEqual = FALSE;
           bestScore = actScore;
           bestPos = k;
           partRemov = TRUE;
           removNode = FALSE;
           if (params . algorithm & FAST_SEARCH) {
             update = FALSE;
             break;
           }
         }
         else if (actScore == bestScore) {
           randNum = g_random_int_range (0, 100);
           if ((foundBetter != TRUE) && (foundEqual != TRUE)) {
             bestPos = k;
             partRemov = TRUE;
             removNode = FALSE;
             foundEqual = TRUE;
           }
           else if (randNum < params . acceptEqual) {
             bestPos = k;
             partRemov = TRUE;
             removNode = FALSE;
           }
         }
         tmpNode -> rec -> removSet = NULL;
       }
     }
     /* Restores the reconstruction */
     actRec -> nodeFlag = tmpNodeFlag;
     if (tmpRemov != NULL)
       tmpNode -> rec -> removSet = tmpRemov;
     IncrementalDownPass (navNode);
   }
   if ((foundEqual == TRUE) && (foundBetter == FALSE)) {
     randNum = g_random_int_range (0, 100);
     if (randNum < params . acceptEqual)
       foundBetter = TRUE;
   }
   if (foundBetter == TRUE) {
     if (update == TRUE) {
       navNode = g_ptr_array_index (sectorNodeArray, bestPos);
       actNode = (PHYLONODE*) navNode -> data;
       actRec = actNode -> rec;
       if (removNode == TRUE)
         actRec -> nodeFlag |= REMOV_NODE;
       else if (partRemov == TRUE) {
         tmpNode = (PHYLONODE*) navNode -> parent -> data;
         tmpNode -> rec -> removSet = actNode;
       }
       else {
         if (actRec -> nodeFlag & REMOV_NODE)
           actRec -> nodeFlag &= ~ REMOV_NODE;
         else {
           tmpNode = (PHYLONODE*) navNode -> parent -> data;
           tmpNode -> rec -> removSet = NULL;
         }
       }
       bestScore = IncrementalDownPass (navNode);
     }
   }
 }
}

void BlindAnnealing (PHYLOTREE* actTree, GPtrArray* recArray, SCOUNTER* stored) {
 SREAL actScore, prevScore, diff;
 SCOUNTER actGen, toSel, object;
 GNode* navNode;
 PHYLONODE* actNode;
 PHYLONODE* tmpNode;
 PHYLONODE* rootNode;
 SFLAG tmpNodeFlag;
 RECONS* actRec;
 SCOUNTER bestHits;

 rootNode = (PHYLONODE*) actTree -> root -> data;
 bestHits = 0;
 for (actGen = 0; actGen < params . generations; ++ actGen) {
   prevScore = rootNode -> rec -> cost;
   toSel = g_random_int_range (0, tmpNodeArray -> len);
   navNode = (GNode*) g_ptr_array_index (tmpNodeArray, toSel);
   if (navNode -> parent == NULL) continue;
   /* If the node must be eliminated, check if elimination is valid */
   actNode = (PHYLONODE*) navNode -> data;
   actRec = actNode -> rec;
   if (actRec -> numCells == 0) continue;
   if (!(actRec -> nodeFlag & REMOV_NODE)) {
     tmpNode = (PHYLONODE*) navNode -> parent -> data;
     if ((tmpNode -> rec -> nodeFlag & REMOV_NODE) || (tmpNode -> rec -> active < 2)) continue;
     if (g_node_first_child (navNode) != NULL) {
       if (actRec -> active < 2) continue;
     }
     if (tmpNode -> rec -> removSet != NULL) {
       if (tmpNode -> rec -> removSet == actNode) continue;
       if (tmpNode -> rec -> active == 2) continue;
     }
   }
   /* Make the backup of the node */
   tmpNodeFlag = actRec -> nodeFlag;
   /* Make the change */
   ChangeNode (navNode);
   actScore = rootNode -> rec -> cost;
   diff = actScore - prevScore;
   if (diff >= 0)
     object = (params . acceptEqual * 1000) / (diff + 1);
   else {
     CheckReconstruction (actTree, recArray, stored, &bestHits, TRUE);
     object = ((diff - 1) * (params . acceptEqual * 1000)) / (diff - 2);
   }
   toSel = g_random_int_range (0, 1000);
   if (toSel >= object) {
     /* Restores the reconstruction */
     actRec -> nodeFlag = tmpNodeFlag;
     IncrementalDownPass (navNode);
   }
 }
}

void GibbsAnnealing (PHYLOTREE* actTree, GPtrArray* recArray, GPtrArray* sectorArray, INTARRAY* prob, SCOUNTER* stored) {
 SREAL actScore, prevScore, diff;
 SCOUNTER actGen, toSel, object, sumProb, i;
 GNode* navNode;
 PHYLONODE* actNode;
 PHYLONODE* tmpNode;
 PHYLONODE* rootNode;
 SFLAG tmpNodeFlag;
 RECONS* actRec;
 SCOUNTER bestHits;

 rootNode = (PHYLONODE*) actTree -> root -> data;
 bestHits = 0;
 for (actGen = 0; actGen < params . generations; ++ actGen) {
   prevScore = rootNode -> rec -> cost;
   prob -> len = 0;
   sumProb = 0;
   /* Make the node list */
   if (sectorArray -> len > 0)
     g_ptr_array_remove_range (sectorArray, 0, sectorArray -> len);
   while (sectorArray -> len < params . sectSize) {
     toSel = g_random_int_range (0, tmpNodeArray -> len);
     navNode = (GNode*) g_ptr_array_index (tmpNodeArray, toSel);
     if (navNode -> parent == NULL) continue;
     /* If the node must be eliminated, check if elimination is valid */
     actNode = (PHYLONODE*) navNode -> data;
     actRec = actNode -> rec;
     if (actRec -> numCells == 0) continue;
     if (!(actRec -> nodeFlag & REMOV_NODE)) {
       tmpNode = (PHYLONODE*) navNode -> parent -> data;
       if ((tmpNode -> rec -> nodeFlag & REMOV_NODE) || (tmpNode -> rec -> active < 2)) continue;
       if (tmpNode -> rec -> removSet != NULL) {
         if (tmpNode -> rec -> removSet == actNode) continue;
         if (tmpNode -> rec -> active == 2) continue;
       }
       if (g_node_first_child (navNode) != NULL) {
         if (actRec -> active < 2) continue;
       }
     }
     g_ptr_array_add (sectorArray, navNode);
   }
   /* Remove each element on the list */
   for (i = 0; i < params . sectSize; ++ i) {
     navNode = (GNode*) g_ptr_array_index (sectorArray, i);
     actNode = (PHYLONODE*) navNode -> data;
     actRec = actNode -> rec;
     /* Make the backup of the node */
     tmpNodeFlag = actRec -> nodeFlag;
     /* Make the change */
     ChangeNode (navNode);
     actScore = rootNode -> rec -> cost;
     diff = actScore - prevScore;
     if (diff >= 0)
       object = (params . acceptEqual * 1000) / (diff + 1);
     else {
       diff *= -1;
       object = (params . acceptEqual * 1000 * diff) / (diff + 1);
     }
     AddToArray (prob, object);
     sumProb += object;
     /* Restores the reconstruction */
     actRec -> nodeFlag = tmpNodeFlag;
     IncrementalDownPass (navNode);
   }
   toSel = g_random_int_range (0, sumProb);
   diff = 0;
   for (i = 0; i < prob -> len; ++ i) {
     diff += GetFromArray (prob, i);
     if (diff > toSel) break;
   }
   navNode = (GNode*) g_ptr_array_index (sectorArray, i);
   ChangeNode (navNode);
   CheckReconstruction (actTree, recArray, stored, &bestHits, TRUE);
 }
}

void NixonAnnealing (PHYLOTREE* actTree, GPtrArray* recArray, GPtrArray* sectorArray, SCOUNTER* stored) {
 SCOUNTER actGen, toSel, i;
 GNode* navNode;
 PHYLONODE* actNode;
 PHYLONODE* tmpNode;
 PHYLONODE* rootNode;
 RECONS* actRec;
 SCOUNTER bestHits;

 rootNode = (PHYLONODE*) actTree -> root -> data;
 bestHits = 0;
 for (actGen = 0; actGen < params . generations; ++ actGen) {
   if (tmpNodeArray -> len > 0)
     g_ptr_array_remove_range (tmpNodeArray, 0, tmpNodeArray -> len);
   CopyPtrArray (tmpNodeArray, actTree -> nodeArray);
   /* Make the node list */
   if (sectorArray -> len > 0)
     g_ptr_array_remove_range (sectorArray, 0, sectorArray -> len);
   while (sectorArray -> len < params . sectSize) {
     toSel = g_random_int_range (0, tmpNodeArray -> len);
     navNode = (GNode*) g_ptr_array_index (tmpNodeArray, toSel);
     if (navNode -> parent == NULL) continue;
     /* Check if elimination is valid */
     actNode = (PHYLONODE*) navNode -> data;
     actRec = actNode -> rec;
     if (actRec -> numCells == 0) continue;
     if (actRec -> nodeFlag & REMOV_NODE) continue;
     tmpNode = (PHYLONODE*) navNode -> parent -> data;
     if ((tmpNode -> rec -> nodeFlag & REMOV_NODE) || (tmpNode -> rec -> active < 2)) continue;
     if (g_node_first_child (navNode) != NULL) {
       if (actRec -> active < 2) continue;
     }
     if (tmpNode -> rec -> removSet != NULL) {
       if (tmpNode -> rec -> removSet == actNode) continue;
       if (tmpNode -> rec -> active == 2) continue;
     }
     g_ptr_array_add (sectorArray, navNode);
     g_ptr_array_remove_index_fast (tmpNodeArray, toSel);
   }
   /* Remove all element on the list */
   for (i = 0; i < params . sectSize; ++ i) {
     navNode = (GNode*) g_ptr_array_index (sectorArray, i);
     actNode = (PHYLONODE*) navNode -> data;
     ChangeNode (navNode);
   }
   NodeRemoval (actTree, tmpNodeArray);
   /* Restore the nodes */
   for (i = 0; i < params . sectSize; ++ i) {
     navNode = (GNode*) g_ptr_array_index (sectorArray, i);
     actNode = (PHYLONODE*) navNode -> data;
     ChangeNode (navNode);
   }
   if (tmpNodeArray -> len > 0)
     g_ptr_array_remove_range (tmpNodeArray, 0, tmpNodeArray -> len);
   CopyPtrArray (tmpNodeArray, actTree -> nodeArray);
   NodeRemoval (actTree, tmpNodeArray);
   CheckReconstruction (actTree, recArray, stored, &bestHits, TRUE);
 }
}

void CopyRec (RECONS* dest, RECONS* src) {
 dest -> node = src -> node;
 CopyBitField (dest -> cells, src -> cells);
 CopyBitField (dest -> obs, src -> obs);
 dest -> setA = src -> setA;
 dest -> setB = src -> setB;
 dest -> removSet = src -> removSet;
 dest -> nodeFlag = src -> nodeFlag;
 dest -> numVics = src -> numVics;
 dest -> numRemovNodes = src -> numRemovNodes;
 dest -> numNoVics = src -> numNoVics;
 dest -> cost = src -> cost;
 dest -> numCells = src -> numCells;
 dest -> numObs = src -> numObs;
 dest -> active = src -> active;
 dest -> overlap = src -> overlap;
 dest -> maxAge = src -> maxAge;
 dest -> timeLag = src -> timeLag;
 dest -> triangles = NULL;
}

GPtrArray* CreateRecArrayCopy (GPtrArray* src) {
 SCOUNTER i;
 GPtrArray* dest;
 RECONS* nuRec;

 dest = g_ptr_array_sized_new (src -> len);
 for (i = 0; i < src -> len; ++ i) {
   nuRec = (RECONS*) g_slice_alloc (sizeof (RECONS));
   nuRec -> cells = CreateBitField ();
   nuRec -> obs = CreateBitField ();
   nuRec -> triangles = NULL;
   CopyRec (nuRec, (RECONS*) g_ptr_array_index (src, i));
   g_ptr_array_add (dest, nuRec);
 }
 return dest;
}

void CopyRecArray (GPtrArray* dest, GPtrArray* src) {
 SCOUNTER i;

 for (i = 0; i < src -> len; ++ i)
   CopyRec ((RECONS*) g_ptr_array_index (dest, i), (RECONS*) g_ptr_array_index (src, i));
}

SFLAG CompareRec (GPtrArray* probe, GList* recList) {
 GList* navList;
 GPtrArray* recArray;
 SFLAG isEqual;
 RECONS* probeRec;
 RECONS* actRec;
 SCOUNTER i;

 for (navList = recList; navList != NULL; navList = g_list_next (navList)) {
   recArray = (GPtrArray*) navList -> data;
   isEqual = TRUE;
   for (i = 0; i < probe -> len; ++ i) {
     probeRec = (RECONS*) g_ptr_array_index (probe, i);
     actRec = (RECONS*) g_ptr_array_index (recArray, i);
     if (probeRec -> cost != actRec -> cost) {
       isEqual = FALSE;
       break;
     }
     if (probeRec -> nodeFlag != actRec -> nodeFlag) {
       isEqual = FALSE;
       break;
     }
     if (probeRec -> numVics != actRec -> numVics) {
       isEqual = FALSE;
       break;
     }
     if (probeRec -> numRemovNodes != actRec -> numRemovNodes) {
       isEqual = FALSE;
       break;
     }
   }
   if (isEqual == TRUE) return TRUE;
 }
 return FALSE;
}

SFLAG CheckReconstruction (PHYLOTREE* tree, GPtrArray* recArray, SCOUNTER* stored, SCOUNTER* bestHits, SFLAG add) {
 RECONS* actRec;
 SREAL actScore;
 SFLAG addToRec;
 GList* lastLink;
 GPtrArray* recToAdd;

 actRec = (RECONS*) g_ptr_array_index (recArray, 0);
 actScore = actRec -> cost;
 addToRec = params . keepAll;
 if (actScore <= (tree -> best + params . extra)) {
   if (actScore < tree -> best) {
     tree -> best = actScore;
     addToRec = TRUE;
     *bestHits = 1;
     FilterTree (tree, params . extra);
   }
   else {
     addToRec = TRUE;
     if (actScore == tree -> best)
       ++ *bestHits;
   }
 }
 /* Check is the rec to store is different from stored recs */
 if ((addToRec == TRUE) && (tree -> recs != NULL)) {
   if (CompareRec (recArray, tree -> recs) == TRUE)
     addToRec = FALSE;
 }
 /* Add the new reconstruction */
 if ((addToRec == TRUE) && (add == TRUE)) {
   recToAdd = CreateRecArrayCopy (recArray);
   lastLink = InsertListElement (NULL, recToAdd);
   if (tree -> recs == NULL)
     tree -> recs = lastLink;
   else
     PutAtListEnd (tree -> recs, lastLink);
   ++ *stored;
 }
 return addToRec;
}

void SearchMain (PROGRESSSTRUCT* progress) {
 GList* navTree;
 PHYLOTREE* tree;
 GPtrArray* sectorArray;
 GPtrArray* recArray;
 RECONS* actRec;
 SCOUNTER stored, bestHits;
 SCOUNTER i, j, k;
 INTARRAY* prob;

 progress -> actBar1 = j = 0;
 prob = NULL;
 if (params . algorithm & USE_SECTORS)
   sectorArray = g_ptr_array_sized_new (params . sectSize);
 else
   sectorArray = NULL;
 if (params . algorithm & GIBBS_ANNEAL) {
   if (sectorArray == NULL)
     sectorArray = g_ptr_array_sized_new (params . sectSize);
   prob = CreateIntArray (params . sectSize);
 }
 for (navTree = GETTREES; navTree != NULL; navTree = g_list_next (navTree)) {
   if (progress -> isActive == FALSE) {
     g_string_append_printf (outLog,"\n\n**** Search interrupted ****");
     break;
   }
   tree = (PHYLOTREE*) navTree -> data;
   tmpNodeArray = g_ptr_array_sized_new (tree -> nodeArray -> len);
   actRec = (RECONS*) g_ptr_array_index (tree -> recOR, 0);
   if (tree -> recs == NULL)
     tree -> best = 2 * actRec -> cost;
   stored = 0;
   bestHits = 0;
   recArray = CreateRecArrayCopy (tree -> recOR);
   for (i = 0; i < params . its; ++ i) {
     if (progress -> isActive == FALSE) break;
     progress -> actBar2 = i;
     UpdateOnCalculus ();
     if (tmpNodeArray -> len > 0)
       g_ptr_array_remove_range (tmpNodeArray, 0, tmpNodeArray -> len);
     CopyPtrArray (tmpNodeArray, tree -> nodeArray);
     CopyRecArray (recArray, tree -> recOR);
     SetRec (recArray);
     if (params . algorithm & USE_SECTORS) {
       for (k = 0; k < (2 * tmpNodeArray -> len); k += params . sectSize) {
         ShufflePtrArray (tmpNodeArray);
         if (sectorArray -> len > 0)
           g_ptr_array_remove_range (sectorArray, 0, sectorArray -> len);
         CopyPtrArrayPart (sectorArray, tmpNodeArray, 0, params . sectSize);
         NodeRemoval (tree, sectorArray);
         if (params . algorithm & FIRST_SECTOR) break;
       }
     }
     if (params . algorithm & BLIND_ANNEAL) {
       NodeRemoval (tree, tmpNodeArray);
       BlindAnnealing (tree, recArray, &stored);
     }
     else if (params . algorithm & GIBBS_ANNEAL) {
       NodeRemoval (tree, tmpNodeArray);
       GibbsAnnealing (tree, recArray, sectorArray, prob, &stored);
     }
     else if (params . algorithm & NIXON_ANNEAL) {
       NodeRemoval (tree, tmpNodeArray);
       NixonAnnealing (tree, recArray, sectorArray, &stored);
     }
     NodeRemoval (tree, tmpNodeArray);
     CheckReconstruction (tree, recArray, &stored, &bestHits, TRUE);
   }
   g_string_append_printf (outLog,"\n\nTree %ld (%s)\nBest score: %0.4f\nHits: %ld of %ld iterations\nStored reconstructions: %ld",
                                   j, tree -> name, tree -> best, bestHits, i, stored);
   tree -> stored += stored;
   g_ptr_array_free (recArray, TRUE);
   SetRec (tree -> recOR);
   tree -> actRec = NULL;
   g_ptr_array_free (tmpNodeArray, TRUE);
   ++ j;
   progress -> actBar1 = j;
 }
 if (sectorArray != NULL)
   g_ptr_array_free (sectorArray, TRUE);
 if (prob != NULL)
   DestroyIntArray (prob);
}

void SetNodeAsRemoved (GNode* node) {
 PHYLONODE* actNode;
 RECONS* actRec;

 actNode = (PHYLONODE*) node -> data;
 actRec = actNode -> rec;
 actRec -> nodeFlag |= REMOV_NODE;
 IncrementalDownPass (node);
}

void SetNodeAsPartialRemov (GNode* node) {
 PHYLONODE* ancNode;

 if ((params . search & USE_LATEOVER) && (node -> parent != NULL)) {
   ancNode = (PHYLONODE*) node -> parent -> data;
   ancNode -> rec -> removSet = (PHYLONODE*) node -> data;
   IncrementalDownPass (node);
 }
}

void ChangeNode (GNode* node) {
 PHYLONODE* actNode;
 RECONS* actRec;

 actNode = (PHYLONODE*) node -> data;
 actRec = actNode -> rec;
 if (actRec -> nodeFlag & REMOV_NODE) {
   actRec -> nodeFlag &= ~ REMOV_NODE;
   IncrementalDownPass (node);
 }
 else
   SetNodeAsRemoved (node);
}

void UpdateRecCounts (GPtrArray* recArray) {
 PHYLONODE* actNode;
 PHYLONODE* descNode;
 RECONS* actRec;
 GNode* navDesc;
 SCOUNTER i;

 for (i = recArray -> len - 1; i >= 0; -- i) {
   actRec = (RECONS*) g_ptr_array_index (recArray, i);
   actNode = actRec -> node;
   actRec -> cost = 0;
   actRec -> numVics = actRec -> numNoVics = 0;
   if (g_node_first_child (actNode -> node == NULL)) continue;
   for (navDesc = g_node_first_child (actNode -> node); navDesc != NULL; navDesc = g_node_next_sibling (navDesc)) {
     descNode = (PHYLONODE*) navDesc -> data;
     actRec -> numVics += descNode -> rec -> numVics;
     actRec -> numNoVics += descNode -> rec -> numNoVics;
   }
   if (actRec -> nodeFlag & VIC_NODE)
     ++ actRec -> numVics;
   else
     ++ actRec -> numNoVics;
 }
}

void MakeConsensus () {
 GList* navTree;
 GList* navRec;
 GPtrArray* tmpRec;
 SCOUNTER i;
 PHYLOTREE* tree;
 RECONS* actRec;
 RECONS* srcRec;

 for (navTree = GETTREES; navTree != NULL; navTree = g_list_next (navTree)) {
   tree = (PHYLOTREE*) navTree -> data;
   if (tree -> recCons != NULL) {
     SetRec (tree -> recCons);
     continue;
   }
   tmpRec = (GPtrArray*) tree -> recs -> data;
   tree -> recCons = CreateRecArrayCopy (tmpRec);
   for (navRec = g_list_next (tree -> recs); navRec != NULL; navRec = g_list_next (navRec)) {
     tmpRec = (GPtrArray*) navRec -> data;
     for (i = 0; i < tmpRec -> len; ++ i) {
       actRec = (RECONS*) g_ptr_array_index (tree -> recCons, i);
       srcRec = (RECONS*) g_ptr_array_index (tmpRec, i);
       BitFieldANDsum (actRec -> cells, srcRec -> cells);
       BitFieldANDsum (actRec -> obs, srcRec -> obs);
       if ((actRec -> setA != srcRec -> setA) || (actRec -> setB != srcRec -> setB)) {
         actRec -> setA = NULL;
         actRec -> setB = NULL;
         actRec -> nodeFlag &= ~ VIC_NODE;
       }
       if (actRec -> removSet != srcRec -> removSet)
         actRec -> removSet = NULL;
       if ((actRec -> nodeFlag & VIC_NODE) && (!(srcRec -> nodeFlag & VIC_NODE)))
         actRec -> nodeFlag &= ~ VIC_NODE;
       if ((actRec -> nodeFlag & REMOV_NODE) && (!(srcRec -> nodeFlag & REMOV_NODE)))
         actRec -> nodeFlag &= ~ REMOV_NODE;
       if (actRec -> numRemovNodes < srcRec -> numRemovNodes)
         actRec -> numRemovNodes = srcRec -> numRemovNodes;
     }
   }
   tree -> suppNodes = 0;
   SetRec (tree -> recCons);
   UpdateRecCounts (tree -> recCons);
 }
}

GList* CopyActRec (PHYLOTREE* tree) {
 GPtrArray* recArray;
 GList* toRet;

 if (tree -> actRec == NULL)
   recArray = CreateRecArrayCopy (tree -> recOR);
 else
   recArray = CreateRecArrayCopy ((GPtrArray*) tree -> actRec -> data);
 toRet = InsertListElement (NULL, recArray);
 if (tree -> recs == NULL)
   tree -> recs = toRet;
 else
   PutAtListEnd (tree -> recs, toRet);
 ++ tree -> stored;
 return toRet;
}

void SetConstrains (SFLAG toCons) {
 PHYLOTREE* tree;
 GList* navTree;
 RECONS* actRec;
 SCOUNTER i;

 for (navTree = GETTREES; navTree != NULL; navTree = g_list_next (navTree)) {
   tree = (PHYLOTREE*) navTree -> data;
   for (i = tree -> recOR -> len - 1; i >= 0; -- i) {
     actRec = (RECONS*) g_ptr_array_index (tree -> recOR, i);
     if ((toCons == TRUE) && (actRec -> nodeFlag & VIC_NODE))
       actRec -> node -> constrained = TRUE;
     else
       actRec -> node -> constrained = FALSE;
   }
 }
}
/* tessellation.c is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */

#include "tessellation.h"
#include <math.h>

static GList* StartTriangleList (SINTEGER maxCols);
static GList* RemoveInvalid (GList* triangleList);
static SFLAG IsValid (DELAUNAYTRIANGLE* tri);
static void RemoveNeighbor (DELAUNAYTRIANGLE* toDel, DELAUNAYTRIANGLE* dest);

static void InsertPoint (GList* triangleList, SCOUNTER x, SCOUNTER y, SFLAG state);

static DELAUNAYTRIANGLE* CreateTriangle (SINTEGER colA, SINTEGER rowA,
                                         SINTEGER colB, SINTEGER rowB, SINTEGER colC, SINTEGER rowC,
                                         SFLAG stateA, SFLAG stateB, SFLAG stateC);
static void ReDoTriangle (DELAUNAYTRIANGLE* reDo, SINTEGER colA, SINTEGER rowA,
                                         SINTEGER colB, SINTEGER rowB, SINTEGER colC, SINTEGER rowC,
                                         SFLAG stateA, SFLAG stateB, SFLAG stateC);
static void CalculateCircle (SINTEGER colA, SINTEGER rowA, SINTEGER colB, SINTEGER rowB, SINTEGER colC, SINTEGER rowC,
                      SREAL* cX, SREAL* cY, SREAL* r);

static SFLAG IsCellInside (SINTEGER pCol, SINTEGER pRow, DELAUNAYTRIANGLE* probe);

static void SetNeighbor (DELAUNAYTRIANGLE* triA, DELAUNAYTRIANGLE* triB);
static void ReplaceNeighbors (DELAUNAYTRIANGLE* dest, DELAUNAYTRIANGLE* src);
static void SearchCommonSegment (DELAUNAYTRIANGLE* triA, DELAUNAYTRIANGLE* triB, SCOUNTER* segmentA, SCOUNTER* segmentB);

static void IsLegal (DELAUNAYTRIANGLE* src, DELAUNAYTRIANGLE* tri);
static SREAL CalcDist (SREAL x1, SREAL x2, SREAL y1, SREAL y2);

GList* MakeVoronoi (BITFIELD* iniA, BITFIELD* iniB) {
 INTARRAY* cellList;
 GList* triangles;
 SFLAG state;
 SCOUNTER i, actCell;
 SINTEGER row, col;
 SREAL sX;
 SINTEGER maxCols;

 cellList = CreateIntArray (GETNUMCELLS);
 for (i = 0; i < GETNUMCELLS; ++ i) {
   if (IsBitOn (iniA, i) == TRUE)
     AddToArray (cellList, i);
   else if (IsBitOn (iniB, i) == TRUE)
     AddToArray (cellList, i);
 }
 /* Randomize the list */
 ShuffleIntArray (cellList);
 /* Add points to the triangulation */
 maxCols = 360.00 / GETGRIDSIZE;
 triangles = StartTriangleList (maxCols);
 for (i = 0; i < GetArraySize (cellList); ++ i) {
   actCell = GetFromArray (cellList, i);
   state = STATE_NOCELL;
   if (IsBitOn (iniA, actCell) == TRUE)
     state = STATE_SETA;
   else if (IsBitOn (iniB, actCell) == TRUE)
     state = STATE_SETB;
   if (state != STATE_NOCELL) {
     GetColRowFromCell (actCell, &row, &col);
     InsertPoint (triangles, col, row, state);
     sX = (col * GETGRIDSIZE) + GETGRIDSTARTX;
     if (sX <= maxCols / 4)
       InsertPoint (triangles, col + maxCols, row, state);
     else if (sX >= (3 * maxCols / 4))
       InsertPoint (triangles, col - maxCols, row, state);
   }
 }
 /* Destroy the list */
 DestroyIntArray (cellList);
 triangles = RemoveInvalid (triangles);
 return triangles;
}

void DestroyTriangleList (GList* triangleList) {
 DestroyList (triangleList, sizeof (DELAUNAYTRIANGLE));
}

GList* RemoveInvalid (GList* triangleList) {
 GList* head;
 GList* nav;
 DELAUNAYTRIANGLE* actTri;

 head = NULL;
 for (nav = triangleList; nav != NULL;) {
   actTri = (DELAUNAYTRIANGLE*) nav -> data;
   if (IsValid (actTri)) {
     if (head == NULL)
       head = nav;
     nav = g_list_next (nav);
     continue;
   }
   nav = g_list_next (nav);
   RemoveNeighbor (actTri, actTri -> neighbor [0]);
   RemoveNeighbor (actTri, actTri -> neighbor [1]);
   RemoveNeighbor (actTri, actTri -> neighbor [2]);
   DestroyLinkElement (actTri -> list, sizeof (DELAUNAYTRIANGLE));
 }
 return head;
}

SFLAG IsValid (DELAUNAYTRIANGLE* tri) {
 SCOUNTER i;
 SCOUNTER minNo, maxNo, stA, stB;

 minNo = maxNo = stA = stB = 0;
 for (i = 0; i < 3; ++ i) {
   if ((tri -> col [i] < 0) || (tri -> row [i] < 0))
     ++ minNo;
   if ((tri -> col [i] >= GETNUMCOLS) || (tri -> row [i] >= GETNUMROWS))
     ++ maxNo;
   if (tri -> state [i] == STATE_SETA)
     ++ stA;
   else if (tri -> state [i] == STATE_SETB)
     ++ stB;
 }
 if ((minNo == 3) || (maxNo == 3)) return FALSE;
 if ((stA == 0) || (stB == 0)) return FALSE;
 return TRUE;
}

void RemoveNeighbor (DELAUNAYTRIANGLE* toDel, DELAUNAYTRIANGLE* dest) {
 SCOUNTER n;

 if ((toDel == NULL) || (dest == NULL)) return;
 for (n = 0; n < 3;  ++ n) {
   if (dest -> neighbor [n] != toDel) continue;
   dest -> neighbor [n] = NULL;
   break;
 }
}

void InsertPoint (GList* triangleList, SCOUNTER x, SCOUNTER y, SFLAG state) {
 GList* nav;
 GList* nuList;
 DELAUNAYTRIANGLE* destA;
 DELAUNAYTRIANGLE* destB;
 DELAUNAYTRIANGLE* probe;
 DELAUNAYTRIANGLE* nuTriA;
 DELAUNAYTRIANGLE* nuTriB;
 DELAUNAYTRIANGLE* nuTriC;
 DELAUNAYTRIANGLE* nuTriD;
 SCOUNTER i;
 SCOUNTER segmentA, segmentB;

 destA = destB = NULL;
 segmentA = segmentB = 0;
 /* Look for the initial triangle */
 for (nav = triangleList; nav != NULL; nav = g_list_next (nav)) {
   probe = (DELAUNAYTRIANGLE*) nav -> data;
   if (IsCellInside (x, y, probe)) {
     destA = probe;
     break;
   }
 }
 /* check in neighbors (for the case of incidence in and edge */
 for (i = 0; i < 3; ++ i) {
   if (destA -> neighbor [i] == NULL) continue;
   if (IsCellInside (x, y, destA -> neighbor [i])) {
     destB = destA -> neighbor [i];
     break;
   }
 }
 if (destB != NULL) {
   SearchCommonSegment (destA, destB, &segmentA, &segmentB);
   if (segmentA == 0) {
     nuTriA = CreateTriangle (x,y,  destA -> col [0], destA -> row [0],  destA -> col [1], destA -> row [1],
                              state, destA -> state [0], destA -> state [1]);
     nuTriB = CreateTriangle (x,y,  destA -> col [0], destA -> row [0],  destA -> col [2], destA -> row [2],
                              state, destA -> state [0], destA -> state [2]);
   }
   else if (segmentA == 1) {
     nuTriA = CreateTriangle (x,y,  destA -> col [1], destA -> row [1],  destA -> col [2], destA -> row [2],
                              state, destA -> state [1], destA -> state [2]);
     nuTriB = CreateTriangle (x,y,  destA -> col [1], destA -> row [1],  destA -> col [0], destA -> row [0],
                              state, destA -> state [1], destA -> state [0]);
   }
   else {
     nuTriA = CreateTriangle (x,y,  destA -> col [2], destA -> row [2],  destA -> col [1], destA -> row [1],
                              state, destA -> state [2], destA -> state [1]);
     nuTriB = CreateTriangle (x,y,  destA -> col [2], destA -> row [2],  destA -> col [0], destA -> row [0],
                              state, destA -> state [2], destA -> state [0]);
   }
   if (segmentB == 0) {
     nuTriC = CreateTriangle (x,y,  destB -> col [0], destB -> row [0],  destB -> col [1], destB -> row [1],
                              state, destB -> state [0], destB -> state [1]);
     nuTriD = CreateTriangle (x,y,  destB -> col [0], destB -> row [0],  destB -> col [2], destB -> row [2],
                              state, destB -> state [0], destB -> state [2]);
   }
   else if (segmentB == 1) {
     nuTriC = CreateTriangle (x,y,  destB -> col [1], destB -> row [1],  destB -> col [2], destB -> row [2],
                              state, destB -> state [1], destB -> state [2]);
     nuTriD = CreateTriangle (x,y,  destB -> col [1], destB -> row [1],  destB -> col [0], destB -> row [0],
                              state, destB -> state [1], destB -> state [0]);
   }
   else {
     nuTriC = CreateTriangle (x,y,  destB -> col [2], destB -> row [2],  destB -> col [1], destB -> row [1],
                              state, destB -> state [2], destB -> state [1]);
     nuTriD = CreateTriangle (x,y,  destB -> col [2], destB -> row [2],  destB -> col [0], destB -> row [0],
                              state, destB -> state [2], destB -> state [0]);
   }
   SetNeighbor (nuTriA, nuTriB);
   SetNeighbor (nuTriA, nuTriC);
   SetNeighbor (nuTriA, nuTriD);
   SetNeighbor (nuTriB, nuTriC);
   SetNeighbor (nuTriB, nuTriD);
   SetNeighbor (nuTriC, nuTriD);
   ReplaceNeighbors (nuTriA, destA);
   ReplaceNeighbors (nuTriB, destA);
   ReplaceNeighbors (nuTriC, destB);
   ReplaceNeighbors (nuTriD, destB);
   nuTriA -> list = destA -> list;
   nuTriA -> list -> data = nuTriA;
   nuTriC -> list = destB -> list;
   nuTriC -> list -> data = nuTriC;
   nuList = g_list_append (NULL, nuTriB);
   nuTriB -> list = nuList;
   nuList = g_list_append (NULL, nuTriD);
   nuTriD -> list = nuList;
   PutAtListEnd (nuTriC -> list, nuTriB -> list);
   PutAtListEnd (nuTriB -> list, nuTriD -> list);
   g_slice_free1 (sizeof (DELAUNAYTRIANGLE), destA);
   g_slice_free1 (sizeof (DELAUNAYTRIANGLE), destB);
   IsLegal (NULL, nuTriA);
   IsLegal (NULL, nuTriB);
   IsLegal (NULL, nuTriC);
   IsLegal (NULL, nuTriD);
 }
 else if (destA != NULL) {
   nuTriA = CreateTriangle (x,y,  destA -> col [0], destA -> row [0],  destA -> col [1], destA -> row [1],
                              state, destA -> state [0], destA -> state [1]);
   nuTriB = CreateTriangle (x,y,  destA -> col [1], destA -> row [1],  destA -> col [2], destA -> row [2],
                              state, destA -> state [1], destA -> state [2]);
   nuTriC = CreateTriangle (x,y,  destA -> col [0], destA -> row [0],  destA -> col [2], destA -> row [2],
                              state, destA -> state [0], destA -> state [2]);
   SetNeighbor (nuTriA, nuTriB);
   SetNeighbor (nuTriA, nuTriC);
   SetNeighbor (nuTriB, nuTriC);
   ReplaceNeighbors (nuTriA, destA);
   ReplaceNeighbors (nuTriB, destA);
   ReplaceNeighbors (nuTriC, destA);
   nuTriA -> list = destA -> list;
   nuTriA -> list -> data = nuTriA;
   nuList = g_list_append (NULL, nuTriB);
   nuTriB -> list = nuList;
   nuList = g_list_append (NULL, nuTriC);
   nuTriC -> list = nuList;
   PutAtListEnd (nuTriA -> list, nuTriB -> list);
   PutAtListEnd (nuTriB -> list, nuTriC -> list);
   g_slice_free1 (sizeof (DELAUNAYTRIANGLE), destA);
   IsLegal (NULL, nuTriA);
   IsLegal (NULL, nuTriB);
   IsLegal (NULL, nuTriC);
 }
}

void ReplaceNeighbors (DELAUNAYTRIANGLE* dest, DELAUNAYTRIANGLE* src) {
 SCOUNTER a, b, b1, b2, segmentB;

 if ((dest == NULL) || (src == NULL)) return;
 b1 = b2 = -1;
 for (a = 0; a < 3; ++ a) {
   for (b = 0; b < 3; ++ b) {
     if ((dest -> col [a] == src -> col [b]) && (dest -> row [a] == src -> row [b])) {
       if (b1 == -1) {
         b1 = b;
       }
       else if (b2 == -1) {
         b2 = b;
       }
       break;
     }
   }
 }
 if (b2 == -1) return;
 if ((b1 != 0) && (b2 != 0))
   segmentB = 0;
 else if ((b1 != 1) && (b2 != 1))
   segmentB = 1;
 else
   segmentB = 2;
 SetNeighbor (dest, src -> neighbor [segmentB]);
}

void SetNeighbor (DELAUNAYTRIANGLE* triA, DELAUNAYTRIANGLE* triB) {
 SCOUNTER a, b, a1, a2, b1, b2, segmentA, segmentB;

 if ((triA == NULL) || (triB == NULL)) return;
 a1 = a2 = b1 = b2 = -1;
 for (a = 0; a < 3; ++ a) {
   for (b = 0; b < 3; ++ b) {
     if ((triA -> col [a] == triB -> col [b]) && (triA -> row [a] == triB -> row [b])) {
       if (a1 == -1) {
         a1 = a;
         b1 = b;
       }
       else if (a2 == -1) {
         a2 = a;
         b2 = b;
       }
       break;
     }
   }
 }
 if (a2 == -1) return;
 if ((a1 != 0) && (a2 != 0))
   segmentA = 0;
 else if ((a1 != 1) && (a2 != 1))
   segmentA = 1;
 else
   segmentA = 2;
 if ((b1 != 0) && (b2 != 0))
   segmentB = 0;
 else if ((b1 != 1) && (b2 != 1))
   segmentB = 1;
 else
   segmentB = 2;
 triA -> neighbor [segmentA] = triB;
 triB -> neighbor [segmentB] = triA;
}

void SearchCommonSegment (DELAUNAYTRIANGLE* triA, DELAUNAYTRIANGLE* triB, SCOUNTER* segmentA, SCOUNTER* segmentB) {
 SCOUNTER a, b;

 for (a = 0; a < 3; ++ a) {
   if (triA -> neighbor [a] != triB) continue;
   for (b = 0; b < 3; ++ b) {
     if (triB -> neighbor [b] != triA) continue;
     if (segmentA != NULL)
       *segmentA = a;
     *segmentB = b;
     break;
   }
   break;
 }
}

void IsLegal (DELAUNAYTRIANGLE* src, DELAUNAYTRIANGLE* tri) {
 DELAUNAYTRIANGLE* nei;
 DELAUNAYTRIANGLE* tmpA [3];
 DELAUNAYTRIANGLE* tmpN [3];
 SCOUNTER a, n, segmentN, tRow, tCol;
 SREAL dist;
 SFLAG isLegal;
 SFLAG tState;

 segmentN = 0;
 for (;;) {
   isLegal = TRUE;
   for (n = 0; n < 3;  ++ n) {
     if (tri -> neighbor [n] == NULL) continue;
     if (tri -> neighbor [n] == src) continue;
     nei = tri -> neighbor [n];
     SearchCommonSegment (tri, nei, NULL, &segmentN);
     dist = CalcDist (tri -> cX, nei -> col [segmentN], tri -> cY, nei -> row [segmentN]) + 0.000005;
     if (dist < tri -> r) {
       /* Illegal triangle */
       isLegal = FALSE;
       for (a = 0; a < 3; ++ a) {   /* Save the neghbors of both triangles */
         tmpN [a] = nei -> neighbor [a];
         tmpA [a] = tri -> neighbor [a];
       }
       /* ReDo the triangles */
       if (n == 0) {
         tCol = tri -> col [1];
         tRow = tri -> row [1];
         tState = tri -> state [1];
         ReDoTriangle (tri, tri -> col [0], tri -> row [0], nei -> col [segmentN], nei -> row [segmentN],
                            tri -> col [2], tri -> row [2], tri -> state [0], nei -> state [segmentN], tri -> state [2]);
         ReDoTriangle (nei, tri -> col [0], tri -> row [0], nei -> col [segmentN], nei -> row [segmentN], tCol, tRow,
                            tri -> state [0], nei -> state [segmentN], tState);
       }
       else if (n == 1) {
         tCol = tri -> col [0];
         tRow = tri -> row [0];
         tState = tri -> state [0];
         ReDoTriangle (tri, tri -> col [1], tri -> row [1], nei -> col [segmentN], nei -> row [segmentN],
                            tri -> col [2], tri -> row [2], tri -> state [1], nei -> state [segmentN], tri -> state [2]);
         ReDoTriangle (nei, tri -> col [1], tri -> row [1], nei -> col [segmentN], nei -> row [segmentN], tCol, tRow,
                            tri -> state [1], nei -> state [segmentN], tState);
       }
       else if (n == 2) {
         tCol = tri -> col [0];
         tRow = tri -> row [0];
         tState = tri -> state [0];
         ReDoTriangle (tri, tri -> col [1], tri -> row [1], nei -> col [segmentN], nei -> row [segmentN],
                            tri -> col [2], tri -> row [2], tri -> state [1], nei -> state [segmentN], tri -> state [2]);
         ReDoTriangle (nei, tri -> col [2], tri -> row [2], nei -> col [segmentN], nei -> row [segmentN], tCol, tRow,
                            tri -> state [2], nei -> state [segmentN], tState);
       }
       for (a = 0; a < 3; ++ a) {   /* Set the neoghbors */
         if (tmpN [a] != tri)
           SetNeighbor (tri, tmpN [a]);
         SetNeighbor (tri, tmpA [a]);
         SetNeighbor (nei, tmpN [a]);
         if (tmpA [a] != nei)
           SetNeighbor (nei, tmpA [a]);
       }
       IsLegal (tri, nei);
       break;
     }
   }
   if (isLegal == TRUE) break;
 }
}

GList* StartTriangleList (SINTEGER maxCols) {
 GList* list;
 DELAUNAYTRIANGLE* nuTri;
 SINTEGER sX, sY, startX, startY;

 startX = GETGRIDSTARTX / GETGRIDSIZE;
 startY = GETGRIDSTARTY / GETGRIDSIZE;
 sX = (GETNUMCOLS / 2) + startX;
 sY = (GETNUMROWS / 2) + startY;
 nuTri = CreateTriangle (sX, (-maxCols * 20) + sY,  (-maxCols * 17) + sX, (maxCols * 10) + sY,
                         (maxCols * 17) + sX, (maxCols * 10) + sY,
                         STATE_NOCELL, STATE_NOCELL, STATE_NOCELL);
 list = g_list_append (NULL, nuTri);
 nuTri -> list = list;
 return list;
}

DELAUNAYTRIANGLE* CreateTriangle (SINTEGER colA, SINTEGER rowA,
                                         SINTEGER colB, SINTEGER rowB, SINTEGER colC, SINTEGER rowC,
                                         SFLAG stateA, SFLAG stateB, SFLAG stateC) {
 DELAUNAYTRIANGLE* nuTri;
 SCOUNTER i;

 nuTri = g_slice_alloc (sizeof (DELAUNAYTRIANGLE));
 nuTri -> col [0] = colA;
 nuTri -> row [0] = rowA;
 nuTri -> state [0] = stateA;
 nuTri -> col [1] = colB;
 nuTri -> row [1] = rowB;
 nuTri -> state [1] = stateB;
 nuTri -> col [2] = colC;
 nuTri -> row [2] = rowC;
 nuTri -> state [2] = stateC;
 for (i = 0; i < 3; ++ i)
   nuTri -> neighbor [i] = NULL;
 CalculateCircle (colA, rowA, colB, rowB, colC, rowC, &nuTri -> cX, &nuTri -> cY, &nuTri -> r);
 nuTri -> list = NULL;
 return nuTri;
}

void ReDoTriangle (DELAUNAYTRIANGLE* reDo, SINTEGER colA, SINTEGER rowA,
                                         SINTEGER colB, SINTEGER rowB, SINTEGER colC, SINTEGER rowC,
                                         SFLAG stateA, SFLAG stateB, SFLAG stateC) {
 SCOUNTER i;

 reDo -> col [0] = colA;
 reDo -> row [0] = rowA;
 reDo -> state [0] = stateA;
 reDo -> col [1] = colB;
 reDo -> row [1] = rowB;
 reDo -> state [1] = stateB;
 reDo -> col [2] = colC;
 reDo -> row [2] = rowC;
 reDo -> state [2] = stateC;
 for (i = 0; i < 3; ++ i)
   reDo -> neighbor [i] = NULL;
 CalculateCircle (colA, rowA, colB, rowB, colC, rowC, &reDo -> cX, &reDo -> cY, &reDo -> r);
}


void CalculateCircle (SINTEGER colA, SINTEGER rowA, SINTEGER colB, SINTEGER rowB, SINTEGER colC, SINTEGER rowC,
                      SREAL* cX, SREAL* cY, SREAL* r) {
 SREAL aX, aY, bX, bY, aN, bN, deno;

 aY = rowA - rowC;
 aX = colA - colC;
 bY = rowB - rowC;
 bX = colB - colC;
 deno = 2 * ((aX * bY) - (aY * bX));
 if ((deno>=-0.0000001) && (deno <= 0.0000001)) {
   /* Colineal or unsorted points */
  *cX = *cY = *r = -1;
   return;
 }
 aN = (aX * aX) + (aY * aY);
 bN = (bX * bX) + (bY * bY);
 *cX = ((aN * bY) - (bN * aY)) / deno;
 *cY = ((bN * aX) - (aN * bX)) / deno;
 *r = CalcDist (*cX, aX, *cY, aY);
 *cX += colC;
 *cY += rowC;
}

SFLAG IsCellInside (SINTEGER pCol, SINTEGER pRow, DELAUNAYTRIANGLE* probe) {
 SREAL v0v0, v0v1, v0v2, v1v1, v1v2;
 SREAL deno;
 SREAL u, v;

 v0v0 = ((probe -> row [2] - probe -> row [0]) * (probe -> row [2] - probe -> row [0])) +
        ((probe -> col [2] - probe -> col [0]) * (probe -> col [2] - probe -> col [0]));
 v0v1 = ((probe -> row [2] - probe -> row [0]) * (probe -> row [1] - probe -> row [0])) +
        ((probe -> col [2] - probe -> col [0]) * (probe -> col [1] - probe -> col [0]));
 v0v2 = ((probe -> row [2] - probe -> row [0]) * (pRow - probe -> row [0])) +
        ((probe -> col [2] - probe -> col [0]) * (pCol - probe -> col [0]));
 v1v1 = ((probe -> row [1] - probe -> row [0]) * (probe -> row [1] - probe -> row [0])) +
        ((probe -> col [1] - probe -> col [0]) * (probe -> col [1] - probe -> col [0]));
 v1v2 = ((probe -> row [1] - probe -> row [0]) * (pRow - probe -> row [0])) +
        ((probe -> col [1] - probe -> col [0]) * (pCol - probe -> col [0]));
 deno = (v0v0 * v1v1) - (v0v1 * v0v1);
 u = ((v1v1 * v0v2) - (v0v1 * v1v2)) / deno;
 v = ((v0v0 * v1v2) - (v0v1 * v0v2)) / deno;
 if ((u < 0) || (v < 0) || ((u + v) > 1.0000001)) return FALSE;
 return TRUE;
}

#include <stdio.h>
SREAL CalcDist (SREAL x1, SREAL x2, SREAL y1, SREAL y2) {
 SREAL toRet;
 toRet = sqrt (((x1 - x2) * (x1 - x2)) + ((y1 - y2) * (y1 - y2)));
 printf ("%f ", toRet);
 return toRet;
}
/* tview.c is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */

#include "wins.h"
#include "dialogs.h"
#include "grid.h"
#include "parser.h"
#include "out.h"
#include <string.h>

typedef struct tagTREEVIEWWINDATA TREEVIEWWINDATA;
struct tagTREEVIEWWINDATA {
 GtkWidget* drawable;
 GdkPixmap* pixMap;

 GdkGC* blackPen;
 GdkGC* greenPen;
 GdkGC* whitePen;
 GdkGC* greyPen;
 GdkGC* bluePen;
 GdkGC* redPen;

 GtkWidget* menuFormat;
 GtkWidget* menuRagged;
 GtkWidget* menuChrono;
 GtkWidget* menuToTerm;
 GtkWidget* menuToRoot;
 GtkWidget* menuFile;

 GtkWidget* menuRecons;
 GtkWidget* menuSupported;
 GtkWidget* menuCommon;
 GtkWidget* menuBremer;

 gint clientX;
 gint clientY;
 SCOUNTER width;
 SCOUNTER height;
 SINTEGER originX;
 SINTEGER originY;
 SINTEGER endClient;
 SINTEGER maxX;
 SINTEGER maxY;
 gint prevX;
 gint prevY;

 SFLAG alignment;
 SFLAG modeRec;
};

static gint OnDelete (GtkWidget* treeWin, GdkEvent* event, gpointer data);
static gint OnExpose (GtkWidget* drawable, GdkEventExpose* event, gpointer data);
static gint OnConfigure (GtkWidget* drawable, GdkEvent* event, gpointer data);

static gint OnKey (GtkWidget* treeWin, GdkEventKey* event, gpointer data);
static gint OnMouseWheel (GtkWidget* drawable, GdkEventScroll* event, gpointer data);
static gint OnMotion (GtkWidget* drawable, GdkEventMotion* event, gpointer data);
static gint OnMouseButton (GtkWidget* drawable, GdkEventButton* event, gpointer data);

static void OnAlign (GtkWidget* menuItem, gpointer data);
#define TREE_TOROOT  1
#define TREE_TOTERM  2
#define TREE_RAGGED  4
#define TREE_ALIGN   8
#define TREE_CHRONO 16

#define MENU_TOROOT "toRoot"
#define MENU_TOTERM "toTerm"
#define MENU_RAGGED "ragged"
#define MENU_ALIGN  "align"
#define MENU_CHRONO "chrono"

static void OnMenuFitToWin (GtkWidget* menuItem, gpointer data);
static void OnMenuFitToTerm (GtkWidget* menuItem, gpointer data);

static void OnMenuSaveTree (GtkMenuItem* saveTreeMenu, gpointer data);
static void OnMenuSaveGraph (GtkMenuItem* saveGraphMenu, gpointer data);
static void OnMenuTreeName (GtkMenuItem* nameMenu, gpointer data);

static void OnMenuSortUp (GtkWidget* menuSort, gpointer data);
static void OnMenuSortDown (GtkWidget* menuSort, gpointer data);
static void OnMenuSortMix (GtkWidget* menuSort, gpointer data);
static void OnMenuSortDef (GtkWidget* menuSort, gpointer data);
static void OnMenuOptions (GtkWidget* menuOption, gpointer data);

static void DrawTree (SFLAG redraw);
static void PaintTree ();
static void PutTreeOnScreen (PHYLOTREE* tree);

static GNode* SearchNearestPoint (gint x, gint y);
static void OnChangeNodeData (GNode* node);

static TREEVIEWWINDATA tWinData;

#define EDIT_TREE_REC 1
#define VIEW_TREE_REC 0

void StartTreeViewWindow () {
 GtkWidget* gBox;
 GtkWidget* menuBar;
 GtkWidget* menu;
 GtkWidget* menuItem;
 GtkAccelGroup* accelGroup;
 GSList* group;

 GdkColor blueColor = {0, 0x0000, 0x0000, 0xFFFF};
 GdkColor greenColor = {0, 0x0000, 0xFFFF, 0x0000};
 GdkColor redColor = {0, 0xFFFF, 0x0000, 0x0000};
 GdkColor greyColor = {0, 0x6000, 0x6000, 0x6000};

 tWinData . pixMap = NULL;
 gBox = gtk_vbox_new (FALSE, 0);
 treeViewWindow = CreateWindow ("VIP: Tree view window", 250, 0, 400, 600, G_CALLBACK (OnDelete), gBox);

 tWinData . modeRec = VIEW_TREE_REC;

 /* --- Sets menues --- */
 accelGroup = gtk_accel_group_new ();
 gtk_window_add_accel_group (GTK_WINDOW (treeViewWindow), accelGroup);
 menuBar = gtk_menu_bar_new ();
 gtk_box_pack_start (GTK_BOX (gBox), menuBar, FALSE, TRUE, 0);
 gtk_widget_show (menuBar);
 /* File */
 menu = gtk_menu_new ();
 CreateMenuItem (menu, "Tree _name...", accelGroup, GDK_I, GDK_CONTROL_MASK, G_CALLBACK (OnMenuTreeName), NULL);
 CreateMenuSeparator (menu);
 CreateMenuItem (menu, "Save _tree...", accelGroup, GDK_T, GDK_CONTROL_MASK | GDK_SHIFT_MASK, G_CALLBACK (OnMenuSaveTree), NULL);
 CreateMenuItem (menu, "Save tree _graphic...", accelGroup, GDK_G, GDK_CONTROL_MASK | GDK_SHIFT_MASK, G_CALLBACK (OnMenuSaveGraph), NULL);
 tWinData . menuFile = AddMenuToBar (menu, "_File", menuBar);
 gtk_widget_set_sensitive (tWinData . menuFile, FALSE);
 /* Format */
 menu = gtk_menu_new ();
 group = NULL;
 menuItem = CreateMenuItemRadio (menu, "_Aligned terms", accelGroup, GDK_A, GDK_CONTROL_MASK,
                                             G_CALLBACK (OnAlign), MENU_ALIGN, group);
 group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (menuItem));
 gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (menuItem), TRUE);
 tWinData . menuRagged = CreateMenuItemRadio (menu, "_Ragged terms", accelGroup, GDK_R, GDK_CONTROL_MASK,
                                              G_CALLBACK (OnAlign), MENU_RAGGED, group);
 group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (tWinData . menuRagged));
 tWinData . menuChrono = CreateMenuItemRadio (menu, "_Chrono tree", accelGroup, GDK_C, GDK_CONTROL_MASK,
                                              G_CALLBACK (OnAlign), MENU_CHRONO, group);
 gtk_widget_set_sensitive (tWinData . menuChrono, FALSE);
 CreateMenuSeparator (menu);
 group = NULL;
 tWinData . menuToTerm = CreateMenuItemRadio (menu, "Pull nodes to t_erms", accelGroup, GDK_E, GDK_CONTROL_MASK,
                                                 G_CALLBACK (OnAlign), MENU_TOTERM, group);
 group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (tWinData . menuToTerm));
 gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (tWinData . menuToTerm), TRUE);
 tWinData . menuToRoot = CreateMenuItemRadio (menu, "Pull nodes to r_oot", accelGroup, GDK_O, GDK_CONTROL_MASK,
                                              G_CALLBACK (OnAlign), MENU_TOROOT, group);
 CreateMenuSeparator (menu);
 CreateMenuItem (menu, "Sort _up", accelGroup, GDK_U, GDK_CONTROL_MASK, G_CALLBACK (OnMenuSortUp), NULL);
 CreateMenuItem (menu, "Sort _down", accelGroup, GDK_D, GDK_CONTROL_MASK, G_CALLBACK (OnMenuSortDown), NULL);
 CreateMenuItem (menu, "_Mixed sorting", accelGroup, GDK_M, GDK_CONTROL_MASK, G_CALLBACK (OnMenuSortMix), NULL);
 CreateMenuItem (menu, "_Default sorting", accelGroup, GDK_G, GDK_CONTROL_MASK, G_CALLBACK (OnMenuSortDef), NULL);
 CreateMenuSeparator (menu);
 CreateMenuItem (menu, "Fit tree to _window", accelGroup, GDK_F, GDK_CONTROL_MASK, G_CALLBACK (OnMenuFitToWin), NULL);
 CreateMenuItem (menu, "Fit tree to _terms", accelGroup, GDK_T, GDK_CONTROL_MASK, G_CALLBACK (OnMenuFitToTerm), NULL);
 tWinData . menuFormat = AddMenuToBar (menu, "Forma_t", menuBar);
 gtk_widget_set_sensitive (tWinData . menuFormat, FALSE);
 /* Reconstruction */
 menu = gtk_menu_new ();
 group = NULL;
 menuItem = CreateMenuItemRadio (menu, "_All reconstruction", accelGroup, GDK_L, GDK_CONTROL_MASK,
                                             G_CALLBACK (OnMenuOptions), NULL, group);
 group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (menuItem));
 gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (menuItem), TRUE);
 tWinData . menuSupported = CreateMenuItemRadio (menu, "_Supported reconstructions (on consensus)", accelGroup, GDK_S, GDK_CONTROL_MASK,
                                                 G_CALLBACK (OnMenuOptions), NULL, group);
 group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (tWinData . menuSupported));
 tWinData . menuCommon = CreateMenuItemRadio (menu, "_Common reconstructions (on consensus)", accelGroup, GDK_N, GDK_CONTROL_MASK,
                                                 G_CALLBACK (OnMenuOptions), NULL, group);
 tWinData . menuRecons = AddMenuToBar (menu, "_Reconstruction", menuBar);
 CreateMenuSeparator (menu);
 tWinData . menuBremer = CreateMenuItemCheck (menu, "_Bremer\'s support", accelGroup, GDK_B, GDK_CONTROL_MASK,
                                                    G_CALLBACK (OnMenuOptions), NULL);
 gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (tWinData . menuBremer), FALSE);
 gtk_widget_set_sensitive (tWinData . menuRecons, FALSE);


 /* --- Drawing area --- */
 tWinData . alignment = TREE_TOTERM | TREE_ALIGN;
 tWinData . drawable = gtk_drawing_area_new ();
 gtk_box_pack_start (GTK_BOX (gBox), tWinData . drawable, TRUE, TRUE, 0);
 gtk_widget_show (tWinData . drawable);
 g_signal_connect (G_OBJECT (tWinData . drawable), "expose_event", G_CALLBACK (OnExpose), NULL);
 g_signal_connect (G_OBJECT (tWinData . drawable), "configure_event", G_CALLBACK (OnConfigure), NULL);

 gtk_widget_add_events (treeViewWindow, GDK_KEY_PRESS_MASK);
 g_signal_connect (G_OBJECT (treeViewWindow), "key_press_event", G_CALLBACK (OnKey), NULL);
 gtk_widget_add_events (tWinData . drawable, GDK_POINTER_MOTION_MASK | GDK_BUTTON_PRESS_MASK | GDK_SCROLL_MASK);
 g_signal_connect (G_OBJECT (tWinData . drawable), "motion_notify_event", G_CALLBACK (OnMotion), NULL);
 g_signal_connect (G_OBJECT (tWinData . drawable), "button_press_event", G_CALLBACK (OnMouseButton), NULL);
 g_signal_connect (G_OBJECT (tWinData . drawable), "scroll_event", G_CALLBACK (OnMouseWheel), NULL);

 gtk_widget_show (treeViewWindow);

 tWinData . blackPen = gdk_gc_new (tWinData . drawable -> window);
 gdk_gc_copy (tWinData . blackPen, tWinData . drawable -> style -> black_gc);

 tWinData . greenPen = gdk_gc_new (tWinData . drawable -> window);
 gdk_gc_copy (tWinData . greenPen, tWinData . drawable -> style -> black_gc);
 gdk_gc_set_rgb_fg_color (tWinData . greenPen, &greenColor);

 tWinData . greyPen = gdk_gc_new (tWinData . drawable -> window);
 gdk_gc_copy (tWinData . greyPen, tWinData . drawable -> style -> black_gc);
 gdk_gc_set_rgb_fg_color (tWinData . greyPen, &greyColor);

 tWinData . whitePen = gdk_gc_new (tWinData . drawable -> window);
 gdk_gc_copy (tWinData . whitePen, tWinData . drawable -> style -> white_gc);

 tWinData . bluePen = gdk_gc_new (tWinData . drawable -> window);
 gdk_gc_copy (tWinData . bluePen, tWinData . drawable -> style -> black_gc);
 gdk_gc_set_rgb_fg_color (tWinData . bluePen, &blueColor);

 tWinData . redPen = gdk_gc_new (tWinData . drawable -> window);
 gdk_gc_copy (tWinData . redPen, tWinData . drawable -> style -> black_gc);
 gdk_gc_set_rgb_fg_color (tWinData . redPen, &redColor);
}

void OpenNewTree () {
 PHYLOTREE* tree;

 if ((activeTree == NULL) && (GETTREES != NULL)) {
   activeTree = GETTREES;
   tree = (PHYLOTREE*) activeTree -> data;
   if (activeNode == NULL)
     activeNode = tree -> root;
   gtk_window_set_title (GTK_WINDOW (treeViewWindow), tree -> name);
   if (tree -> x == 0)
    tree -> x = tWinData . width = tWinData . clientX;
   else
     tWinData . width = tree -> x;
   if (tree -> y == 0)
    tree -> y = tWinData . height = tWinData . clientY;
   else
     tWinData . height = tree -> y;
   tWinData . originX = tWinData . originY = 0;
   tWinData . endClient = 50;
   DrawTree (TRUE);
   if ((tree -> maxTreeAge > 0) && (tree -> maxTreeAge != NO_AGE))
     gtk_widget_set_sensitive (tWinData . menuChrono, TRUE);
   gtk_widget_set_sensitive (tWinData . menuFormat, TRUE);
   if (tree -> recCons != NULL)
     gtk_widget_set_sensitive (tWinData . menuRecons, TRUE);
   else
     gtk_widget_set_sensitive (tWinData . menuRecons, FALSE);
   gtk_widget_set_sensitive (tWinData . menuFile, TRUE);
 }
}

void OnChangeTree (GList* nuTree) {
 PHYLOTREE* tree;

 if (nuTree != NULL) {
   tree = (PHYLOTREE*) nuTree -> data;
   activeTree = nuTree;
   gtk_window_set_title (GTK_WINDOW (treeViewWindow), tree -> name);
   if (tree -> x == 0)
    tree -> x = tWinData . width = tWinData . clientX;
   else
     tWinData . width = tree -> x;
   if (tree -> y == 0)
    tree -> y = tWinData . height = tWinData . clientY;
   else
     tWinData . height = tree -> y;
   tWinData . originX = tWinData . originY = 0;
   tWinData . endClient = 50;
   gtk_widget_set_sensitive (tWinData . menuFormat, TRUE);
   if (tree -> recCons != NULL)
     gtk_widget_set_sensitive (tWinData . menuRecons, TRUE);
   else
     gtk_widget_set_sensitive (tWinData . menuRecons, FALSE);
   if ((tree -> maxTreeAge > 0) && (tree -> maxTreeAge != NO_AGE))
     gtk_widget_set_sensitive (tWinData . menuChrono, TRUE);
   else
     gtk_widget_set_sensitive (tWinData . menuChrono, FALSE);
 }
 else {
   activeTree = NULL;
   gtk_window_set_title (GTK_WINDOW (treeViewWindow), "VIP: Tree view window");
   gtk_widget_set_sensitive (tWinData . menuFormat, FALSE);
   gtk_widget_set_sensitive (tWinData . menuRecons, FALSE);
   gtk_widget_set_sensitive (tWinData . menuFile, FALSE);
 }
 DrawTree (TRUE);
}

void OnUpdateTree () {
 PHYLOTREE* tree;

 if (activeTree == NULL) return;
 tree = (PHYLOTREE*) activeTree -> data;
 if (tree -> recCons != NULL)
   gtk_widget_set_sensitive (tWinData . menuRecons, TRUE);
 else
   gtk_widget_set_sensitive (tWinData . menuRecons, FALSE);
 DrawTree (FALSE);
}

void SetRecEdition (SFLAG mode) {
 tWinData . modeRec = (mode == TRUE) ? EDIT_TREE_REC : VIEW_TREE_REC;
}

gint OnDelete (GtkWidget* treeWin, GdkEvent* event, gpointer data) {
 gtk_window_iconify (GTK_WINDOW (treeViewWindow));
 return TRUE;
}

gint OnExpose (GtkWidget* drawable, GdkEventExpose* event, gpointer data) {
 gdk_draw_pixmap (tWinData . drawable -> window, tWinData . drawable -> style -> fg_gc [GTK_WIDGET_STATE (tWinData . drawable)],
     tWinData . pixMap, event -> area . x, event -> area . y, event -> area . x, event -> area . y,
     event -> area . width, event -> area . height);
 return FALSE;
}

gint OnConfigure (GtkWidget* drawable, GdkEvent* event, gpointer data) {
 tWinData . clientX = tWinData . drawable -> allocation . width;
 tWinData . clientY = tWinData . drawable -> allocation . height;
 if (activeTree == NULL) {
   tWinData . width = tWinData . clientX;
   tWinData . height = tWinData . clientY;
 }
 if (tWinData . pixMap)
   gdk_pixmap_unref (tWinData . pixMap);
 tWinData . pixMap = gdk_pixmap_new (tWinData . drawable -> window, tWinData . clientX, tWinData . clientY, -1);
 DrawTree (FALSE);
 return TRUE;
}

gint OnMouseButton (GtkWidget* drawable, GdkEventButton* event, gpointer data) {
 GNode* actNode;
 PHYLONODE* toChange;
 PHYLOTREE* tree;
 SCOUNTER actFill, nuFill;
 ENTRYDIALOG* radiusDialog;
 gchar* entry;


 if (activeTree == NULL) return FALSE;
 if ((event -> type == GDK_BUTTON_PRESS) && (event -> button == 1) &&
     (!(event -> state & GDK_SHIFT_MASK)) && (!(event -> state & GDK_CONTROL_MASK))) {
   actNode = SearchNearestPoint (event -> x + tWinData . originX, event -> y + tWinData . originY);
   if ((actNode != NULL) && (actNode != activeNode)) {
     activeNode = actNode;
     DrawTree (FALSE);
     ChangeGeo ();
     OnChangeRec ();
   }
 }
 else if ((event -> type == GDK_BUTTON_PRESS) && (event -> button == 3) &&
          (!(event -> state & GDK_SHIFT_MASK)) && (event -> state & GDK_CONTROL_MASK)) {
   tree = (PHYLOTREE*) activeTree -> data;
   if (tree -> actRec == NULL) return FALSE;
   actNode = SearchNearestPoint (event -> x + tWinData . originX, event -> y + tWinData . originY);
   if (actNode != NULL) {
     if (tWinData . modeRec == EDIT_TREE_REC) {
       ChangeNode (actNode);
       DrawTree (FALSE);
       ChangeGeo ();
       OnChangeRec ();
     }
   }
 }
 else if ((event -> type == GDK_BUTTON_PRESS) && (event -> button == 3) &&
          (event -> state & GDK_SHIFT_MASK) && (!(event -> state & GDK_CONTROL_MASK))) {
   actNode = SearchNearestPoint (event -> x + tWinData . originX, event -> y + tWinData . originY);
   if (actNode != NULL) {
     toChange = (PHYLONODE*) actNode -> data;
     actFill = (toChange -> fill > GETFILL) ? GETFILL : toChange -> fill;
     radiusDialog = CounterEntryDialog (GTK_WINDOW (treeViewWindow), actFill, "Set fill", "New fill value");
     if (gtk_dialog_run (GTK_DIALOG (radiusDialog -> dialog)) == GTK_RESPONSE_ACCEPT) {
       entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (radiusDialog -> entry)));
       SetCommandLine (entry);
       g_free (entry);
       if (GetCmdNumber (&nuFill) != FALSE) {
         if (nuFill != actFill)
           PutFillValue (toChange, nuFill);
       }
     }
     DestroyEntryDialog (radiusDialog);
   }
 }
 else if ((event -> type == GDK_BUTTON_PRESS) && (event -> button == 3) &&
          (!(event -> state & GDK_SHIFT_MASK)) && (!(event -> state & GDK_CONTROL_MASK))) {
   actNode = SearchNearestPoint (event -> x + tWinData . originX, event -> y + tWinData . originY);
   if (actNode != NULL)
     OnChangeNodeData (actNode);
 }
 return FALSE;
}

gint OnMouseWheel (GtkWidget* drawable, GdkEventScroll* event, gpointer data) {
 SFLAG updatePos;

 if (activeTree == NULL) return FALSE;
 updatePos = FALSE;
 switch (event -> direction){
  case GDK_SCROLL_UP: tWinData . originY -= 5;
       updatePos = TRUE;
       break;
  case GDK_SCROLL_DOWN: tWinData . originY += 5;
       updatePos = TRUE;
       break;
  default: break;
 }
 if (updatePos == TRUE) {
   if (tWinData . originX < - (tWinData . clientX - 10))
     tWinData . originX = - (tWinData . clientX - 10);
   if (tWinData . originX > (tWinData . maxX - 5))
     tWinData . originX = tWinData . maxX - 5;
   if (tWinData . originY < - (tWinData . clientY - 10))
     tWinData . originY = - (tWinData . clientY - 10);
   if (tWinData . originY > (tWinData . maxY - 5))
     tWinData . originY = tWinData . maxY - 5;
   DrawTree (FALSE);
 }
 return FALSE;
}

gint OnKey (GtkWidget* treeWin, GdkEventKey* event, gpointer data) {
 SCOUNTER nuSz;
 SFLAG updatePos;
 SFLAG draw;
 SFLAG nuTree;
 SFLAG nuNode;
 PHYLOTREE* tree;

 if (activeTree == NULL) return FALSE;
 updatePos = FALSE;
 draw = FALSE;
 nuTree = FALSE;
 nuNode = FALSE;
 switch (event -> keyval) {
  case GDK_plus:
  case GDK_KP_Add: if (event -> state & GDK_CONTROL_MASK) {
         nuSz = (tWinData . height / 100);
         if (nuSz < 5)
           nuSz = 5;
         tWinData . height += nuSz;
         draw = TRUE;
       }
       else if (event -> state & GDK_SHIFT_MASK) {
         nuSz = (tWinData . width / 100);
         if (nuSz < 5)
           nuSz = 5;
         tWinData . width += nuSz;
         draw = TRUE;
       }
       break;
  case GDK_minus:
  case GDK_KP_Subtract: if ((event -> state & GDK_CONTROL_MASK) && (tWinData . height > 5)) {
         nuSz = (tWinData . height / 100);
         if (nuSz < 5)
           nuSz = 5;
         tWinData . height -= nuSz;
         if (tWinData . height < 5)
           tWinData . height = 5;
         draw = TRUE;
       }
       else if ((event -> state & GDK_SHIFT_MASK) && (tWinData . width > 5)) {
         nuSz = (tWinData . width / 100);
         if (nuSz < 5)
           nuSz = 5;
         tWinData . width -= nuSz;
         if (tWinData . width < 5)
           tWinData . width = 5;
         draw = TRUE;
       }
       break;
  case GDK_Up:
  case GDK_KP_Up: if (event -> state & GDK_CONTROL_MASK) {
         nuSz = (tWinData . height / 100);
         if (nuSz < 5)
           nuSz = 5;
         tWinData . height -= nuSz;
         if (tWinData . height < 5)
           tWinData . height = 5;
         draw = TRUE;
       }
       else {
         tWinData . originY -= 5;
         updatePos = TRUE;
       }
       break;
  case GDK_Down:
  case GDK_KP_Down: if ((event -> state & GDK_CONTROL_MASK) && (tWinData . height > 5)) {
         nuSz = (tWinData . height / 100);
         if (nuSz < 5)
           nuSz = 5;
         tWinData . height += nuSz;
         draw = TRUE;
       }
       else {
         tWinData . originY += 5;
         updatePos = TRUE;
       }
       break;
  case GDK_Left:
  case GDK_KP_Left: if (event -> state & GDK_CONTROL_MASK) {
         nuSz = (tWinData . width / 100);
         if (nuSz < 5)
           nuSz = 5;
         tWinData . width -= nuSz;
         if (tWinData . width < 5)
           tWinData . width = 5;
         draw = TRUE;
       }
       else {
         tWinData . originX -= 5;
         updatePos = TRUE;
       }
       break;
  case GDK_Right:
  case GDK_KP_Right: if ((event -> state & GDK_CONTROL_MASK) && (tWinData . width > 5)) {
         nuSz = (tWinData . width / 100);
         if (nuSz < 5)
           nuSz = 5;
         tWinData . width += nuSz;
         draw = TRUE;
       }
       else {
         tWinData . originX += 5;
         updatePos = TRUE;
       }
       break;
  case GDK_Page_Up:
  case GDK_KP_Page_Up: tWinData . originY -= tWinData . clientY;
       updatePos = TRUE;
       break;
  case GDK_Page_Down:
  case GDK_KP_Page_Down: tWinData . originY += tWinData . clientY;
       updatePos = TRUE;
       break;
  case GDK_Home:
  case GDK_KP_Home: if (event -> state & GDK_CONTROL_MASK) {
         if (GETTREES != activeTree) {
           activeTree = GETTREES;
           nuTree = TRUE;
         }
       }
       else {
         tWinData . originX = tWinData . originY = 0;
         updatePos = TRUE;
       }
       break;
  case GDK_End:
  case GDK_KP_End: if (event -> state & GDK_CONTROL_MASK) {
         if (GETLASTTREE != activeTree) {
           activeTree = GETLASTTREE;
           nuTree = TRUE;
         }
       }
       else {
         tWinData . originX = tWinData . maxX - 5;
         tWinData . originY = tWinData . maxY - 5;
         updatePos = TRUE;
       }
       break;
  case GDK_space:
  case GDK_Return:
  case GDK_KP_Enter: if (event -> state & GDK_CONTROL_MASK) {
         activeNode = GetNextNode (activeNode);
         nuNode = TRUE;
       }
       else if (g_list_next (activeTree) != NULL) {
         activeTree = g_list_next (activeTree);
         nuTree = TRUE;
       }
       break;
  case GDK_BackSpace: if (event -> state & GDK_CONTROL_MASK) {
         activeNode = GetPrevNode (activeNode);
         nuNode = TRUE;
       }
       else if (g_list_previous (activeTree) != NULL) {
         activeTree = g_list_previous (activeTree);
         nuTree = TRUE;
       }
       break;
 }
 if (nuTree == TRUE) {
   tree = (PHYLOTREE*) activeTree -> data;
   if ((tree -> maxTreeAge > 0) && (tree -> maxTreeAge != NO_AGE))
     gtk_widget_set_sensitive (tWinData . menuChrono, TRUE);
   else {
     gtk_widget_set_sensitive (tWinData . menuChrono, FALSE);
     if (tWinData . alignment & TREE_CHRONO) {
       gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (tWinData . menuRagged), TRUE);
       gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (tWinData . menuToRoot), TRUE);
       tWinData . alignment = TREE_RAGGED | TREE_TOROOT;
     }
   }
   gtk_window_set_title (GTK_WINDOW (treeViewWindow), tree -> name);
   if (tree -> x == 0) {
    tWinData . width = tWinData . clientX;
    tree -> x = tWinData . width;
   }
   else
     tWinData . width = tree -> x;
   if (tree -> y == 0) {
    tWinData . height = tWinData . clientY;
    tree -> y = tWinData . height;
   }
   else
     tWinData . height = tree -> y;
   tWinData . originX = tWinData . originY = 0;
   tWinData . width = tWinData . clientX;
   tWinData . height = tWinData . clientY;
   tWinData . endClient = 50;
   DrawTree (TRUE);
   OnChangeRec ();
 }
 else if (nuNode == TRUE) {
   DrawTree (FALSE);
   ChangeGeo ();
 }
 else if (updatePos == TRUE) {
   if (tWinData . originX < - (tWinData . clientX - 10))
     tWinData . originX = - (tWinData . clientX - 10);
   if (tWinData . originX > (tWinData . maxX - 5))
     tWinData . originX = tWinData . maxX - 5;
   if (tWinData . originY < - (tWinData . clientY - 10))
     tWinData . originY = - (tWinData . clientY - 10);
   if (tWinData . originY > (tWinData . maxY - 5))
     tWinData . originY = tWinData . maxY - 5;
   DrawTree (draw);
 }
 else if (draw == TRUE) {
   tree = (PHYLOTREE*) activeTree -> data;
   tree -> x = tWinData . width;
   tree -> y = tWinData . height;
   DrawTree (draw);
 }
 return FALSE;
}

gint OnMotion (GtkWidget* drawable, GdkEventMotion* event, gpointer data) {
 SINTEGER nu;
 if (activeTree == NULL) return FALSE;
 if (event -> state & GDK_BUTTON1_MASK) {
   if ((tWinData . prevX != 0) || (tWinData . prevY != 0)) {
     nu = event -> x - tWinData . prevX;
     tWinData . originX -= nu;
     if (tWinData . originX < - (tWinData . clientX - 10))
       tWinData . originX = - (tWinData . clientX - 10);
     else if (tWinData . originX > (tWinData . maxX - 5))
       tWinData . originX = tWinData . maxX - 5;
     nu = event -> y - tWinData . prevY;
     tWinData . originY -= nu;
     if (tWinData . originY < - (tWinData . clientY - 10))
       tWinData . originY = - (tWinData . clientY - 10);
     else if (tWinData . originY > (tWinData . maxY - 5))
       tWinData . originY = tWinData . maxY - 5;
     DrawTree (FALSE);
   }
   tWinData . prevX = event -> x;
   tWinData . prevY = event -> y;
 }
 else {
   tWinData . prevX = 0;
   tWinData . prevY = 0;
 }
 return TRUE;
}

void OnMenuTreeName (GtkMenuItem* nameMenu, gpointer data) {
 ENTRYDIALOG* nameDialog;
 PHYLOTREE* tree;
 gchar* entry;

 if (activeTree == NULL) return;
 tree = (PHYLOTREE*) activeTree -> data;
 nameDialog = TextEntryDialog (GTK_WINDOW (mapWindow), tree -> name, "Set tree name", "Tree");
 if (gtk_dialog_run (GTK_DIALOG (nameDialog -> dialog)) == GTK_RESPONSE_ACCEPT) {
   entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (nameDialog -> entry)));
   if (strlen (entry) > 0) {
     if (tree -> name != NULL)
       g_free (tree -> name);
     tree -> name = g_strdup (entry);
     gtk_window_set_title (GTK_WINDOW (treeViewWindow), tree -> name);
   }
   g_free (entry);
 }
 DestroyEntryDialog (nameDialog);
}

void OnMenuSaveGraph (GtkMenuItem* saveAsMenu, gpointer data) {
 GtkWidget* dialog;
 GtkFileFilter* svgRawFilter;
 GtkFileFilter* svgRecFilter;
 GtkFileFilter* svgConFilter;
 GtkFileFilter* userFilter;
 gchar* fileName;
 PHYLOTREE* tree;
 PHYLONODE* actNode;

 if (activeTree == NULL) return;
 tree = (PHYLOTREE*) activeTree -> data;
 fileName = NULL;
 dialog = gtk_file_chooser_dialog_new ("Save File", GTK_WINDOW (GETMAINWIN), GTK_FILE_CHOOSER_ACTION_SAVE,
                                       GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT, NULL);
 gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (dialog), TRUE);
 gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (dialog), "unitled");
 if (tree -> recOR != NULL) {
   svgRecFilter = CreateFileFilter ("Tree with reconstruction in a SVG file (*.svg)", "*.svg", dialog);
   if (tree -> recCons != NULL) {
     actNode = (PHYLONODE*) tree -> root -> data;
     if (actNode -> rec == g_ptr_array_index (tree -> recCons, 0))
       svgConFilter = CreateFileFilter ("Tree with Bremer values in a SVG file (*.svg)", "*.svg", dialog);
     else
       svgConFilter = NULL;
   }
 }
 else {
   svgRecFilter = NULL;
   svgConFilter = NULL;
 }
 svgRawFilter = CreateFileFilter ("Just tree in a SVG file (*.svg)", "*.svg", dialog);
 if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT) {
   fileName = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
   userFilter = gtk_file_chooser_get_filter (GTK_FILE_CHOOSER (dialog));
   if (userFilter == svgRawFilter) {
     fileName = CheckExtention (fileName, "svg");
     WriteSVGFile (fileName, (PHYLOTREE*) activeTree -> data, SVG_NO_REC);
   }
   if ((userFilter == svgRecFilter) && (svgRecFilter != NULL)) {
     fileName = CheckExtention (fileName, "svg");
     WriteSVGFile (fileName, (PHYLOTREE*) activeTree -> data, SVG_WITH_REC);
   }
   if ((userFilter == svgConFilter) && (svgConFilter != NULL)) {
     fileName = CheckExtention (fileName, "svg");
     WriteSVGFile (fileName, (PHYLOTREE*) activeTree -> data, SVG_WITH_SUPP | SVG_WITH_REC);
   }
   g_free (fileName);
 }
 gtk_widget_destroy (dialog);
}

void OnMenuSaveTree (GtkMenuItem* saveAsMenu, gpointer data) {
 GtkWidget* dialog;
 GtkFileFilter* cladoFilter;
 GtkFileFilter* phyRawFilter;
 GtkFileFilter* tntFilter;
 GtkFileFilter* nexusFilter;
 GtkFileFilter* nameFilter;
 GtkFileFilter* userFilter;
 gchar* fileName;

 fileName = NULL;
 dialog = gtk_file_chooser_dialog_new ("Save File", GTK_WINDOW (GETMAINWIN), GTK_FILE_CHOOSER_ACTION_SAVE,
                                       GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT, NULL);
 gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (dialog), TRUE);
 gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (dialog), "unitled");
 cladoFilter = CreateFileFilter ("Phylogeny (just cladogram) in XML (*.xml)", "*.xml", dialog);
 phyRawFilter = CreateFileFilter ("Phylogeny in phyloXML (*.xml)", "*.xml", dialog);
 tntFilter = CreateFileFilter ("Phylogeny for TNT (*.tnt)", "*.tnt", dialog);
 nexusFilter = CreateFileFilter ("Phylogeny (translate) in Nexus (*.nex)", "*.nex", dialog);
 nameFilter = CreateFileFilter ("Phylogeny (just tree) in Nexus (*.nex)", "*.nex", dialog);
 if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT) {
   fileName = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
   userFilter = gtk_file_chooser_get_filter (GTK_FILE_CHOOSER (dialog));
   if (userFilter == phyRawFilter) {
     fileName = CheckExtention (fileName, "xml");
     WritePhyTree (fileName, (PHYLOTREE*) activeTree -> data);
   }
   else if (userFilter == cladoFilter) {
     fileName = CheckExtention (fileName, "xml");
     WriteCladoTree (fileName, (PHYLOTREE*) activeTree -> data);
   }
   else if (userFilter == tntFilter) {
     fileName = CheckExtention (fileName, "tnt");
     WriteNewickFile (fileName, (PHYLOTREE*) activeTree -> data, USE_TNT);
   }
   else if (userFilter == nexusFilter) {
     fileName = CheckExtention (fileName, "nex");
     WriteNewickFile (fileName, (PHYLOTREE*) activeTree -> data, USE_NEXUS);
   }
   else if (userFilter == nameFilter) {
     fileName = CheckExtention (fileName, "nex");
     WriteNewickFile (fileName, (PHYLOTREE*) activeTree -> data, USE_NEXUS | USE_NAMES);
   }
   g_free (fileName);
 }
 gtk_widget_destroy (dialog);
}

void OnAlign (GtkWidget* menuItem, gpointer data) {
 SFLAG nuSelection;

 nuSelection = tWinData . alignment;
 if (!g_strcmp0 (MENU_TOROOT, (gchar*) data)) {
   nuSelection &= ~ TREE_TOTERM;
   nuSelection |= TREE_TOROOT;
 }
 else if (!g_strcmp0 (MENU_TOTERM, (gchar*) data)) {
   nuSelection &= ~ TREE_TOROOT;
   nuSelection |= TREE_TOTERM;
 }
 else if (!g_strcmp0 (MENU_RAGGED, (gchar*) data)) {
   tWinData . alignment = TREE_RAGGED | TREE_TOROOT;
   gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (tWinData . menuToRoot), TRUE);
   tWinData . alignment = nuSelection;
   nuSelection = TREE_RAGGED | TREE_TOROOT;
   gtk_widget_set_sensitive (tWinData . menuToRoot, TRUE);
   gtk_widget_set_sensitive (tWinData . menuToTerm, FALSE);
 }
 else if (!g_strcmp0 (MENU_ALIGN, (gchar*) data)) {
   nuSelection &=  (TREE_TOROOT | TREE_TOTERM);
   nuSelection |= TREE_ALIGN;
   gtk_widget_set_sensitive (tWinData . menuToRoot, TRUE);
   gtk_widget_set_sensitive (tWinData . menuToTerm, TRUE);
 }
 else if (!g_strcmp0 (MENU_CHRONO, (gchar*) data)) {
   nuSelection = TREE_CHRONO;
   gtk_widget_set_sensitive (tWinData . menuToTerm, FALSE);
   gtk_widget_set_sensitive (tWinData . menuToRoot, FALSE);
 }
 if (tWinData . alignment == nuSelection) return;
 tWinData . alignment = nuSelection;
 DrawTree (TRUE);
}

void OnMenuFitToWin (GtkWidget* menuItem, gpointer data) {
 PHYLOTREE* tree;

 tWinData . width = tWinData . clientX;
 tWinData . height = tWinData . clientY;
 tWinData . originX = tWinData . originY = 0;
 tWinData . endClient = 50;
 tree = (PHYLOTREE*) activeTree -> data;
 tree -> x = tWinData . width;
 tree -> y = tWinData . height;
 DrawTree (TRUE);
}

void OnMenuFitToTerm (GtkWidget* menuItem, gpointer data) {
 PHYLONODE* root;
 PHYLOTREE* tree;

 tree = (PHYLOTREE*) activeTree -> data;
 root = (PHYLONODE*) tree -> root -> data;
 tWinData . width = 40 * root -> level + 10;
 tWinData . height = (10 * (2 * (root -> numTerms - 1))) + 10;
 tWinData . originX = tWinData . originY = 0;
 tWinData . endClient = 50;
 tree -> x = tWinData . width;
 tree -> y = tWinData . height;
 DrawTree (TRUE);
}

void OnMenuSortUp (GtkWidget* menuSort, gpointer data) {
 SortNodes (TREE_TERMS_UP);
 DrawTree (TRUE);
}

void OnMenuSortDown (GtkWidget* menuSort, gpointer data) {
 SortNodes (TREE_TERMS_DOWN);
 DrawTree (TRUE);
}

void OnMenuSortMix (GtkWidget* menuSort, gpointer data) {
 SortNodes (TREE_TERMS_MIX);
 DrawTree (TRUE);
}
void OnMenuSortDef (GtkWidget* menuSort, gpointer data) {
 SortNodes (TREE_TERMS_DEF);
 DrawTree (TRUE);
}

void OnMenuOptions (GtkWidget* menuOption, gpointer data) {
 PHYLOTREE* tree;

 if (activeTree == NULL) return;
 tree = (PHYLOTREE*) activeTree -> data;
 if (tree -> recOR == NULL) return;
 DrawTree (TRUE);
}

/* Prepare the actual tree, or clean the screen */
void DrawTree (SFLAG redraw) {
 GdkRectangle updateRec;

 if (tWinData . pixMap == NULL)
   tWinData . pixMap = gdk_pixmap_new (tWinData . drawable -> window, tWinData . clientX, tWinData . clientY, -1);
 gdk_draw_rectangle (tWinData . pixMap, tWinData . drawable -> style -> white_gc, TRUE, 0, 0, tWinData . clientX, tWinData . clientY);
 if (activeTree != NULL) {
   if (redraw == TRUE)
     PutTreeOnScreen ((PHYLOTREE*) activeTree -> data);
   PaintTree ();
 }
 updateRec . x = 0;
 updateRec . y = 0;
 updateRec . width = tWinData . clientX;
 updateRec . height = tWinData . clientY;
 gdk_window_invalidate_rect (tWinData . drawable -> window, &updateRec, TRUE);
}

/* Calculates tree' screen coordinates */
void PutTreeOnScreen (PHYLOTREE* tree) {
 SREAL sizeX, sizeY;
 PHYLONODE* actNode;
 PHYLONODE* ancNode;
 GNode* navNode;
 SCOUNTER marginX, marginY, ageSector;
 PangoLayout* lay;

 marginX = 5;
 marginY = 10;
 ageSector = tree -> maxTreeAge;

 actNode = (PHYLONODE*) tree -> root -> data;
 if (tWinData . alignment & TREE_CHRONO)
   sizeX = (SREAL) (tWinData . width - (2 * marginX) - tWinData . endClient) / ageSector;
 else
   sizeX = (SREAL) (tWinData . width - (2 * marginX) - tWinData . endClient) / actNode -> level;
 sizeY = (SREAL) (tWinData . height - (2 * marginY)) / (2 * (actNode -> numTerms - 1));
 tWinData . maxX = tWinData . maxY = 0;
 for (navNode = tree -> root; navNode != NULL; ) {
   actNode = (PHYLONODE*) navNode -> data;
   actNode -> actY = actNode -> y * sizeY + marginY;
   if (actNode -> actY > tWinData . maxY)
     tWinData . maxY = actNode -> actY;
   actNode -> topY = actNode -> startY * sizeY + marginY;
   actNode -> downY = actNode -> endY * sizeY + marginY;
   if (g_node_first_child (navNode) == NULL) {
     if ((actNode -> layout == NULL) && (actNode -> name != NULL)) {
       lay = gtk_widget_create_pango_layout (tWinData . drawable, actNode -> name);
       actNode -> layout = lay;
     }
     if (tWinData . alignment & TREE_CHRONO) {
       ancNode = (PHYLONODE*) navNode -> parent -> data;
       actNode -> actX = actNode -> ageX * sizeX + marginX;
       if (actNode -> actX <= ancNode -> actX)        /* This is to make a slight margin between ancestors and nodes */
         actNode -> actX = ancNode -> actX + 2;
     }
     else if (tWinData . alignment & TREE_ALIGN)
       actNode -> actX = actNode -> maxX * sizeX + marginX;
     else
       actNode -> actX = actNode -> minX * sizeX + marginX;
     if (actNode -> actX > tWinData . maxX)
       tWinData . maxX = actNode -> actX;
   }
   else {
     if (actNode -> layout == NULL)
       actNode -> layout = gtk_widget_create_pango_layout (tWinData . drawable, "");
     if (tWinData . alignment & TREE_CHRONO) {
       actNode -> actX = actNode -> ageX * sizeX + marginX;
       if (navNode -> parent != NULL) {
         ancNode = (PHYLONODE*) navNode -> parent -> data;
         if (actNode -> actX <= ancNode -> actX)       /* This is to make a slight margin between ancestors and nodes */
           actNode -> actX = ancNode -> actX + 2;
       }
     }
     else if (tWinData . alignment & TREE_TOROOT)
       actNode -> actX = actNode -> minX * sizeX + marginX;
     else
       actNode -> actX = actNode -> maxX * sizeX + marginX;
     navNode = g_node_first_child (navNode);
     continue;
   }
   while (navNode != NULL) {
     if (g_node_next_sibling (navNode) != NULL) {
       navNode = g_node_next_sibling (navNode);
       break;
     }
     navNode = navNode -> parent;
   }
 }
}

/* Paint a tree on a pixMap */
void PaintTree () {
 PHYLOTREE* tree;
 PHYLONODE* actNode;
 PHYLONODE* tmpNode;
 GNode* navNode;
 GList* navList;
 SINTEGER startX;
 SINTEGER endX;
 SINTEGER startY;
 SINTEGER endY;
 gchar* text;
 SFLAG onConsensus;

 gdk_draw_rectangle (tWinData . pixMap, tWinData . drawable -> style -> white_gc, TRUE, 0, 0, tWinData . clientX, tWinData . clientY);
 tree = (PHYLOTREE*) activeTree -> data;
 onConsensus = FALSE;
 if (tree -> recCons != NULL) {
   actNode = (PHYLONODE*) tree -> root -> data;
   if (actNode -> rec == g_ptr_array_index (tree -> recCons, 0))
     onConsensus = TRUE;
 }
 /* Draw nodes */
 for (navNode = tree -> root; navNode != NULL; ) {
   actNode = (PHYLONODE*) navNode -> data;
   startX = actNode -> actX - tWinData . originX;
   startY = actNode -> actY - tWinData . originY;
   if (navNode -> parent != NULL) {
     tmpNode = (PHYLONODE*) navNode -> parent -> data;
     endX = tmpNode -> actX - tWinData . originX;
   }
   else
     endX = startX - 2;
   /* Draws the horizonthal line */
   if ((startY >= 0) && (startY <= tWinData . clientY)) {
     if (endX < 0)
       endX = 0;
     if ((startX <= tWinData . clientX) && (startX >= 0)) {
       gdk_draw_line (tWinData . pixMap, tWinData . drawable -> style -> black_gc, startX, startY, endX, startY);
       if ((g_node_first_child (navNode) == NULL) && (actNode -> name != NULL))
         gdk_draw_layout (tWinData . pixMap, tWinData . drawable -> style -> black_gc, startX + 5, startY - 5,
                           (PangoLayout*) actNode -> layout);
     }
     else if ((startX >= tWinData . clientX) && (endX <= tWinData . clientX))
       gdk_draw_line (tWinData . pixMap, tWinData . drawable -> style -> black_gc, tWinData . clientX, startY, endX, startY);
   }
   if (g_node_first_child (navNode) != NULL) {
     startX = actNode -> actX - tWinData . originX;
     startY = actNode -> topY - tWinData . originY;
     endY = actNode -> downY - tWinData . originY;
     /* Draws the vertical line */
     if ((startX <= tWinData . clientX) && (startX >= 0)) {
       if (endY > tWinData . clientY)
         endY = tWinData . clientY;
       if ((startY <= tWinData . clientY) && (startY >= 0))
         gdk_draw_line (tWinData . pixMap, tWinData . drawable -> style -> black_gc, startX, startY, startX, endY);
       else if ((startY <= 0) && (endY >= 0))
         gdk_draw_line (tWinData . pixMap, tWinData . drawable -> style -> black_gc, startX, 0, startX, endY);
     }
     navNode = g_node_first_child (navNode);
     continue;
   }
   /* Find next node */
   while (navNode != NULL) {
     if (g_node_next_sibling (navNode) != NULL) {
       navNode = g_node_next_sibling (navNode);
       break;
     }
     navNode = navNode -> parent;
   }
 }
 /* Draw results from reconstruction */
 if (tree -> recOR != NULL) {
   actNode = (PHYLONODE*) tree -> root -> data;
   if ((gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (tWinData . menuCommon)) == TRUE) && (onConsensus == TRUE)) {
     actNode = (PHYLONODE*) activeNode -> data;
     if (actNode -> alloSet != NULL) {
       for (navList = actNode -> alloSet -> nodeList; navList != NULL; navList = g_list_next (navList)) {
         actNode = (PHYLONODE*) navList -> data;
         if (actNode -> tree != tree) continue;
         startX = actNode -> actX - tWinData . originX;
         startY = actNode -> actY - tWinData . originY;
         if ((startX <= tWinData . clientX) && (startX >= 0) && (startY <= tWinData . clientY) && (startY >= 0)) {
           if (actNode -> rec -> nodeFlag & REMOV_NODE) {
             gdk_draw_rectangle (tWinData . pixMap, tWinData . whitePen, TRUE, startX - 2, startY - 2, 5, 5);
             gdk_draw_rectangle (tWinData . pixMap, tWinData . blackPen, FALSE, startX - 2, startY - 2, 5, 5);
           }
           else if (actNode -> rec -> nodeFlag & TIME_LAG) {
             gdk_draw_rectangle (tWinData . pixMap, tWinData . greyPen, TRUE, startX - 2, startY - 2, 5, 5);
             gdk_draw_rectangle (tWinData . pixMap, tWinData . blackPen, FALSE, startX - 2, startY - 2, 5, 5);
           }
           else if (actNode -> rec -> removSet != NULL) {
             gdk_draw_rectangle (tWinData . pixMap, tWinData . whitePen, TRUE, startX - 2, startY - 2, 5, 5);
             gdk_draw_rectangle (tWinData . pixMap, tWinData . blackPen, FALSE, startX - 2, startY - 2, 5, 5);
             gdk_draw_line (tWinData . pixMap, tWinData . drawable -> style -> black_gc,
                              startX -2, startY - 2, startX + 2, startY + 2);
             tmpNode = actNode -> rec -> removSet;
             endY = tmpNode -> actY - tWinData . originY - 2;
             gdk_draw_rectangle (tWinData . pixMap, tWinData . whitePen, TRUE, startX - 2, endY, 5, 5);
             gdk_draw_arc (tWinData . pixMap, tWinData . blackPen, FALSE, startX - 2, endY, 5, 5, 0, (360 * 64));
           }
           else
             gdk_draw_rectangle (tWinData . pixMap, tWinData . blackPen, TRUE, startX - 2, startY - 2, 5, 5);
         }
       }
     }
   }
   else if ((gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (tWinData . menuSupported)) == TRUE) && (onConsensus == TRUE)) {
     for (navNode = tree -> root; navNode != NULL; ) {
       actNode = (PHYLONODE*) navNode -> data;
       if ((actNode -> rec -> nodeFlag & VIC_NODE) && (actNode -> alloSet != NULL) && (actNode -> alloSet -> numNodes > 0)) {
         startX = actNode -> actX - tWinData . originX;
         startY = actNode -> actY - tWinData . originY;
         if ((startX <= tWinData . clientX) && (startX >= 0) && (startY <= tWinData . clientY) && (startY >= 0)) {
           if (actNode -> rec -> nodeFlag & REMOV_NODE) {
             gdk_draw_rectangle (tWinData . pixMap, tWinData . whitePen, TRUE, startX - 2, startY - 2, 5, 5);
             gdk_draw_rectangle (tWinData . pixMap, tWinData . blackPen, FALSE, startX - 2, startY - 2, 5, 5);
           }
           else if (actNode -> rec -> nodeFlag & TIME_LAG) {
             gdk_draw_rectangle (tWinData . pixMap, tWinData . greyPen, TRUE, startX - 2, startY - 2, 5, 5);
             gdk_draw_rectangle (tWinData . pixMap, tWinData . blackPen, FALSE, startX - 2, startY - 2, 5, 5);
           }
           else if (actNode -> rec -> removSet != NULL) {
             gdk_draw_rectangle (tWinData . pixMap, tWinData . whitePen, TRUE, startX - 2, startY - 2, 5, 5);
             gdk_draw_rectangle (tWinData . pixMap, tWinData . blackPen, FALSE, startX - 2, startY - 2, 5, 5);
             gdk_draw_line (tWinData . pixMap, tWinData . drawable -> style -> black_gc,
                              startX -2, startY - 2, startX + 2, startY + 2);
             tmpNode = actNode -> rec -> removSet;
             endY = tmpNode -> actY - tWinData . originY - 2;
             gdk_draw_rectangle (tWinData . pixMap, tWinData . whitePen, TRUE, startX - 2, endY, 5, 5);
             gdk_draw_arc (tWinData . pixMap, tWinData . blackPen, FALSE, startX - 2, endY, 5, 5, 0, (360 * 64));
           }
           else
             gdk_draw_rectangle (tWinData . pixMap, tWinData . blackPen, TRUE, startX - 2, startY - 2, 5, 5);
         }
       }
       if (g_node_first_child (navNode) != NULL) {
         navNode = g_node_first_child (navNode);
         continue;
       }
       /* Find next node */
       while (navNode != NULL) {
         if (g_node_next_sibling (navNode) != NULL) {
           navNode = g_node_next_sibling (navNode);
           break;
         }
         navNode = navNode -> parent;
       }
     }
   }
   else {
     for (navNode = tree -> root; navNode != NULL; ) {
       actNode = (PHYLONODE*) navNode -> data;
       startX = actNode -> actX - tWinData . originX;
       startY = actNode -> actY - tWinData . originY;
       if ((startX <= tWinData . clientX) && (startX >= 0) && (startY <= tWinData . clientY) && (startY >= 0)) {
         if (actNode -> rec -> nodeFlag & VIC_NODE) {
           if (actNode -> rec -> nodeFlag & REMOV_NODE) {
             gdk_draw_rectangle (tWinData . pixMap, tWinData . whitePen, TRUE, startX - 2, startY - 2, 5, 5);
             gdk_draw_rectangle (tWinData . pixMap, tWinData . blackPen, FALSE, startX - 2, startY - 2, 5, 5);
           }
           else if (actNode -> rec -> nodeFlag & TIME_LAG) {
             gdk_draw_rectangle (tWinData . pixMap, tWinData . greyPen, TRUE, startX - 2, startY - 2, 5, 5);
             gdk_draw_rectangle (tWinData . pixMap, tWinData . blackPen, FALSE, startX - 2, startY - 2, 5, 5);
           }
           else if (actNode -> rec -> removSet != NULL) {
             gdk_draw_rectangle (tWinData . pixMap, tWinData . whitePen, TRUE, startX - 2, startY - 2, 5, 5);
             gdk_draw_rectangle (tWinData . pixMap, tWinData . blackPen, FALSE, startX - 2, startY - 2, 5, 5);
             gdk_draw_line (tWinData . pixMap, tWinData . drawable -> style -> black_gc,
                              startX -2, startY - 2, startX + 2, startY + 2);
             tmpNode = actNode -> rec -> removSet;
             endY = tmpNode -> actY - tWinData . originY - 2;
             gdk_draw_rectangle (tWinData . pixMap, tWinData . whitePen, TRUE, startX - 2, endY, 5, 5);
             gdk_draw_arc (tWinData . pixMap, tWinData . blackPen, FALSE, startX - 2, endY, 5, 5, 0, (360 * 64));
           }
           else
             gdk_draw_rectangle (tWinData . pixMap, tWinData . blackPen, TRUE, startX - 2, startY - 2, 5, 5);
           if ((onConsensus == TRUE) && (actNode -> layout != NULL) &&
                (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (tWinData . menuBremer)) == TRUE)) {
             if (actNode -> bremer > actNode -> tree -> worst)
               text = g_strdup_printf ("?> %.2f", actNode -> tree -> worst - actNode -> tree -> best);
             else
               text = g_strdup_printf ("%.2f", actNode -> bremer - actNode -> tree -> best);
             CheckRealStr (text);
             pango_layout_set_text (actNode -> layout, text, -1);
             g_free (text);
             gdk_draw_layout (tWinData . pixMap, tWinData . drawable -> style -> black_gc, startX + 5, startY - 5,
                           (PangoLayout*) actNode -> layout);
           }
         }
         else if (actNode -> rec -> nodeFlag & REMOV_NODE) {
           gdk_draw_rectangle (tWinData . pixMap, tWinData . whitePen, TRUE, startX - 2, startY - 2, 5, 5);
           gdk_draw_arc (tWinData . pixMap, tWinData . blackPen, FALSE, startX - 2, startY - 2, 5, 5, 0, (360 * 64));
         }
       }
       if (g_node_first_child (navNode) != NULL) {
         navNode = g_node_first_child (navNode);
         continue;
       }
       /* Find next node */
       while (navNode != NULL) {
         if (g_node_next_sibling (navNode) != NULL) {
           navNode = g_node_next_sibling (navNode);
           break;
         }
         navNode = navNode -> parent;
       }
     }
   }
 }
 /* Draw the position of the selected node */
 actNode = (PHYLONODE*) activeNode -> data;
 if (actNode -> tree == tree) {
   startX = actNode -> actX - tWinData . originX;
   startY = actNode -> actY - tWinData . originY;
   if ((startX <= tWinData . clientX) && (startX >= 0) && (startY <= tWinData . clientY) && (startY >= 0)) {
     gdk_draw_rectangle (tWinData . pixMap, tWinData . greenPen, TRUE, startX - 2, startY - 2, 5, 5);
     gdk_draw_rectangle (tWinData . pixMap, tWinData . blackPen, FALSE, startX - 2, startY - 2, 5, 5);
   }
   if ((actNode -> rec != NULL) && (actNode -> rec -> nodeFlag & VIC_NODE)) {
     startX = actNode -> rec -> setA -> actX - tWinData . originX;
     startY = actNode -> rec -> setA -> actY - tWinData . originY;
     if ((startX <= tWinData . clientX) && (startX >= 0) && (startY <= tWinData . clientY) && (startY >= 0)) {
       gdk_draw_rectangle (tWinData . pixMap, tWinData . bluePen, TRUE, startX - 2, startY - 2, 5, 5);
       gdk_draw_rectangle (tWinData . pixMap, tWinData . blackPen, FALSE, startX - 2, startY - 2, 5, 5);
     }
     startX = actNode -> rec -> setB -> actX - tWinData . originX;
     startY = actNode -> rec -> setB -> actY - tWinData . originY;
     if ((startX <= tWinData . clientX) && (startX >= 0) && (startY <= tWinData . clientY) && (startY >= 0)) {
       gdk_draw_rectangle (tWinData . pixMap, tWinData . redPen, TRUE, startX - 2, startY - 2, 5, 5);
       gdk_draw_rectangle (tWinData . pixMap, tWinData . blackPen, FALSE, startX - 2, startY - 2, 5, 5);
     }
   }
 }
}

GNode* SearchNearestPoint (gint x, gint y) {
 GNode* navNode;
 GNode* toRet;
 PHYLONODE* actNode;
 PHYLOTREE* tree;
 SINTEGER difX, difY, bestDif, actDif;

 toRet = NULL;
 bestDif = 7;
 tree = (PHYLOTREE*) activeTree -> data;
 for (navNode = tree -> root; navNode != NULL; ) {
   actNode = (PHYLONODE*) navNode -> data;
   difX = x - actNode -> actX;
   difX = (difX >= 0) ? difX : - difX;
   /* The x-coordinate of the node exceeds the search coordinate, then, no descendant of the node need to
      be considered */
   if ((difX > 3) && (actNode -> actX > x)) {
     while (navNode != NULL) {
       if (g_node_next_sibling (navNode) != NULL) {
         navNode = g_node_next_sibling (navNode);
         break;
       }
       navNode = navNode -> parent;
     }
     continue;
   }
   difY = y - actNode -> actY;
   difY = (difY >= 0) ? difY : - difY;
   actDif = difX + difY;
   if (actDif < bestDif) {
     bestDif = actDif;
     toRet = navNode;
   }
   if (actDif == 0) break;
   navNode = GetNextNode (navNode);
   /* There are no more nodes to see */
   if (navNode == tree -> root) break;
 }
 return toRet;
}

void OnChangeNodeData (GNode* node) {
 NODEDATADIALOG* dataDialog;
 PHYLONODE* actNode;
 gchar* text;
 SFLAG mode;
 SFLAG change;
 SCOUNTER age;
 PangoLayout* lay;

 mode = (GetGridState () == TRUE) ? FALSE : TRUE;
 dataDialog = NodeDataDialog (node, mode);
 change = FALSE;
 if ((gtk_dialog_run (GTK_DIALOG (dataDialog -> dialog)) == GTK_RESPONSE_ACCEPT) && (mode == TRUE)) {
   actNode = (PHYLONODE*) node -> data;
   text = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (dataDialog -> nameEntry)));
   if (actNode -> name != NULL)
     g_free (actNode -> name);
   actNode -> name = g_strdup (text);
   g_free (text);
   if (actNode -> term != NULL) {
     actNode -> term -> name = AddSynonim (actNode -> term, actNode -> name);
   }
   if (activeNode == node)
     OnChangeRec ();
   if (actNode -> layout != NULL) {
     lay = (PangoLayout*) actNode -> layout;
     pango_layout_set_text (lay, actNode -> name, -1);
   }
   text = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (dataDialog -> ageEntry)));
   SetCommandLine (text);
   g_free (text);
   if (GetCmdNumber (&age) != FALSE) {
     if ((age >= 0) && (SetNodeAge (node, age) == TRUE)) {
       gtk_widget_set_sensitive (tWinData . menuChrono, TRUE);
       if (tWinData . alignment & TREE_CHRONO)
         change = TRUE;
     }
   }
   DrawTree (change);
 }
 DestroyNodeDataDialog (dataDialog);
}
/* win_app.c is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */

#include "wins.h"
#include "biogeo.h"
#include "parser.h"
#include "progbar.h"

void OnError () {
 GtkWidget* dialog;

 dialog = gtk_message_dialog_new (GTK_WINDOW (GETMAINWIN), GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_ERROR, GTK_BUTTONS_CLOSE,
                                  errorMessage);
 gtk_dialog_run (GTK_DIALOG (dialog));
 gtk_widget_destroy (dialog);
 RESETERROR;
}

void UpdateOnCalculus () {
 while (gtk_events_pending ())
   gtk_main_iteration ();
}

int main (int argc, char** argv) {
 GTimeVal actTime;

 gtk_init (&argc, &argv);
 g_get_current_time (&actTime);
 g_random_set_seed (actTime . tv_usec);
 GETMAINWIN = NULL;
 outLog = g_string_sized_new (1024);
 StartBioGeo ();
 StartTreeViewWindow ();
 StartMapWindow ();
 StartMainWindow ();

 gtk_main ();
 return 0;
}

GtkWidget* CreateWindow (gchar* name, gint x, gint y, gint sizeX, gint sizeY, GCallback OnDelete, GtkWidget* box) {
 GtkWidget* window;

 window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
 gtk_window_set_title (GTK_WINDOW (window), name);
 gtk_window_move (GTK_WINDOW (window), x, y);
 gtk_window_set_default_size (GTK_WINDOW (window), sizeX, sizeY);
 gtk_container_border_width (GTK_CONTAINER (window), 0);
 g_signal_connect (G_OBJECT (window), "delete_event", G_CALLBACK (OnDelete), NULL);
 if (box != NULL) {
   gtk_container_add (GTK_CONTAINER (window), box);
   gtk_widget_show (box);
 }
 return window;
}

GtkWidget* AddMenuToBar (GtkWidget* menu, gchar* name, GtkWidget* menuBar) {
 GtkWidget* menuItem;

 menuItem = gtk_menu_item_new_with_mnemonic (name);
 gtk_widget_show (menuItem);
 gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuItem), menu);
 gtk_menu_bar_append (GTK_MENU_BAR (menuBar), menuItem);
 return menuItem;
}

void CreateMenuSeparator (GtkWidget* menu) {
 GtkWidget* menuItem;

 menuItem = gtk_menu_item_new ();
 gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuItem);
 gtk_widget_show (menuItem);
}

GtkWidget* CreateMenuItem (GtkWidget* menu, gchar* name, GtkAccelGroup* accelGroup, guint key,
                           GdkModifierType modifier, GCallback func, gpointer data) {
 GtkWidget* menuItem;

 menuItem = gtk_menu_item_new_with_mnemonic (name);
 gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuItem);
 g_signal_connect (G_OBJECT (menuItem), "activate", G_CALLBACK (func), data);
 gtk_widget_show (menuItem);
 if (accelGroup != NULL)
   gtk_widget_add_accelerator (menuItem, "activate", accelGroup, key, modifier, GTK_ACCEL_VISIBLE);
 return menuItem;
}

GtkWidget* CreateMenuItemCheck (GtkWidget* menu, gchar* name, GtkAccelGroup* accelGroup, guint key,
                                GdkModifierType modifier, GCallback func, gpointer data) {
 GtkWidget* menuItem;

 menuItem = gtk_check_menu_item_new_with_mnemonic (name);
 gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuItem);
 g_signal_connect (G_OBJECT (menuItem), "activate", G_CALLBACK (func), data);
 gtk_widget_show (menuItem);
 if (accelGroup != NULL)
   gtk_widget_add_accelerator (menuItem, "activate", accelGroup, key, modifier, GTK_ACCEL_VISIBLE);
 return menuItem;
}

GtkWidget* CreateMenuItemRadio (GtkWidget* menu, gchar* name, GtkAccelGroup* accelGroup, guint key,
                                GdkModifierType modifier, GCallback func, gpointer data, GSList* group) {
 GtkWidget* menuItem;

 menuItem = gtk_radio_menu_item_new_with_mnemonic (group, name);
 gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuItem);
 g_signal_connect (G_OBJECT (menuItem), "activate", G_CALLBACK (func), data);
 gtk_widget_show (menuItem);
 if (accelGroup != NULL)
   gtk_widget_add_accelerator (menuItem, "activate", accelGroup, key, modifier, GTK_ACCEL_VISIBLE);
 return menuItem;
}

GtkFileFilter* CreateFileFilter (gchar* name, gchar* pattern, GtkWidget* fileChooser) {
 GtkFileFilter* filter;

 filter = gtk_file_filter_new ();
 gtk_file_filter_set_name (filter, name);
 gtk_file_filter_add_pattern (filter, pattern);
 gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (fileChooser), filter);
 return filter;
}

GtkWidget* CreateEntry (gchar* label, GtkBox* box) {
 GtkWidget* entryLabel;
 GtkWidget* toRet;

 entryLabel = gtk_label_new (label);
 gtk_box_pack_start (GTK_BOX (box), entryLabel, TRUE, TRUE, 0);
 gtk_widget_show (entryLabel);
 toRet = gtk_entry_new ();
 gtk_box_pack_start (GTK_BOX (box), toRet, TRUE, TRUE, 0);
 gtk_widget_show (toRet);
 return toRet;
}
/* biogeo.h is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */

#ifndef BIOGEOGRAPHY_HEADER_INCLUDED
#define BIOGEOGRAPHY_HEADER_INCLUDED

/* ------ Data types for biogeographic data ------ */

#include "ghead.h"
#include "bitfield.h"
#include "progbar.h"

/* Tree data */
typedef struct tagPHYLOTREE PHYLOTREE;
typedef struct tagPHYLONODE PHYLONODE;

/* Terminal data */
typedef struct tagTERMINAL TERMINAL;

/* Georeferenced data */
typedef struct tagGEOREF GEOREF;

/* Reconstructions */
typedef struct tagRECONS RECONS;

/* Allopatry sets */
typedef struct tagALLOPATRY ALLOPATRY;

/* Search Parameters */
typedef struct tagSEARCHPARAM SEARCHPARAM;

/* Global variables */

#define NO_AGE  50000

/* Tree list */
GList* trees;
GList* lastTree;
SCOUNTER numTrees;

#define GETTREES       trees
#define SETTREES(x)    trees = (x)
#define GETLASTTREE    lastTree
#define SETLASTTREE(x) lastTree = (x)

#define GETNUMTREES         numTrees
#define SETNUMTREES(x)      numTrees = (x)
#define INCREASENUMTREES(x) numTrees += (x)
#define DECREASENUMTREES(x) numTrees -= (x)

/* Intialization, Creation, and destruction */
void StartBioGeo ();
void DestroyBioGeo ();
void DestroyTree (PHYLOTREE* tree);

PHYLOTREE* CreateTree (gchar* name, SCOUNTER id);
GNode* CreateNode (PHYLOTREE* tree, GNode* anc, GNode* lastSibling, SCOUNTER id);
void PrepareTree (PHYLOTREE* tree);
PHYLOTREE* GetTreeFromID (SCOUNTER id);

#define TREE_TERMS_UP   1
#define TREE_TERMS_DOWN 2
#define TREE_TERMS_MIX  3
#define TREE_TERMS_DEF  4
void SortNodes (SFLAG mode);

SFLAG SetNodeAge (GNode* node, SCOUNTER age);
PHYLONODE* GetNodeFromID (PHYLOTREE* tree, SCOUNTER id);
void PutFillValue (PHYLONODE* node, SCOUNTER fillValue);

/* Terminal list */
GList* spp;
GList* lastSpp;
SCOUNTER numSpp;

#define GETSPP        spp
#define SETSPP(x)     spp = (x)
#define GETLASTSPP    lastSpp
#define SETLASTSPP(x) lastSpp = (x)

#define GETNUMSPP         numSpp
#define SETNUMSPP(x)      numSpp = (x)
#define INCREASENUMSPP(x) numSpp += (x)
#define DECREASENUMSPP(x) numSpp -= (x)

TERMINAL* CreateTerminal (GNode* node);
void DestroyTermList (GList* list);

TERMINAL* GetTermFromName (gchar* name);
gchar* AddSynonim (TERMINAL* actTerm, gchar* syn);

/* Georeference list */
GList* geoRefs;
GList* lastGeoRef;
SCOUNTER numGeoRefs;

#define GETGEOREF        geoRefs
#define SETGEOREF(x)     geoRefs = (x)
#define GETLASTGEOREF    lastGeoRef
#define SETLASTGEOREF(x) lastGeoRef = (x)

#define GETNUMGEOREFS         numGeoRefs
#define SETNUMGEOREFS(x)      numGeoRefs = (x)
#define INCREASENUMGEOREFS(x) numGeoRefs += (x)
#define DECREASENUMGEOREFS(x) numGeoRefs -= (x)

GList* CreateGeoRef (GList* last, GList* term, SREAL lon, SREAL lat);
GEOREF* AddGeoRef (SREAL lon, SREAL lat, GNode* node);
void DeleteGeoRef (GEOREF* toDel);
void DestroyGeoList (GList* toDel);

void UpdateLimits (GEOREF* geo);

SCOUNTER GetNumTermsWithGeo ();
void UpdateWithData (TERMINAL* actTerm);

/* Reconstructions */
RECONS* CreateRecons (PHYLONODE* node);
void DestroySingleRec (GPtrArray* recArray);
void DestroyRecList (GList* recList);
void DestroyRecs ();
void FilterRecs (SREAL extra);
void FilterTree (PHYLOTREE* actTree, SREAL extra);
GList* CopyActRec (PHYLOTREE* tree);

/* Allopatry sets */
void RemoveAlloSets ();
void DestroyAlloSet (ALLOPATRY* alloSet);

/* General */
SREAL worldStart;

SREAL maxAzimuth;
SREAL minAzimuth;
SREAL maxInclination;
SREAL minInclination;

#define GETMAXAZM    maxAzimuth
#define GETMINAZM    minAzimuth
#define GETMAXINC    maxInclination
#define GETMININC    minInclination
#define SETMAXAZM(x) maxAzimuth = (x)
#define SETMINAZM(x) minAzimuth = (x)
#define SETMAXINC(x) maxInclination = (x)
#define SETMININC(x) minInclination = (x)

#define FROMLATTOINC(x) 90.00000000 - (x)
#define FROMINCTOLAT(x) 90.00000000 - (x)
#define FROMLONTOAZM(x) ((x) > worldStart) ? (x) - worldStart : (x) + 360.00000000 - worldStart
#define FROMAZMTOLON(x) (((x) + worldStart) < 180.00000001) ? (x) + worldStart : (x) - 360.00000000 + worldStart

SEARCHPARAM params;
#define USE_INTS  0
#define OVER_REAL 1
#define USE_TIME  2
#define TIME_REAL 4

#define NORMAL_SEARCH 0
#define USE_LATEOVER  1
#define CONSTR_SEARCH 2

#define VIC_NODE   1
#define REMOV_NODE 2
#define TIME_LAG   4

#define PAGE_SEARCH   1
#define FAST_SEARCH   2
#define FIRST_SECTOR  4
#define FULL_SECTOR   8
#define USE_SECTORS   12
#define BLIND_ANNEAL  16
#define GIBBS_ANNEAL  32
#define FLIP_NODES    64
#define NIXON_ANNEAL 128

/* ------ Searching functions ------ */

void SetRec (GPtrArray* recArray);
void MakeOptimization (SFLAG justCount);
void SearchMain (PROGRESSSTRUCT* progress);

void ChangeNode (GNode* node);
void SetNodeAsRemoved (GNode* node);
void SetNodeAsPartialRemov (GNode* node);

void SetConstrains (SFLAG toCons);

void MakeConsensus ();
void MakeCommonBarriers ();


/* ------ Type definitions ------ */

struct tagPHYLOTREE {
 gchar* name;
 SCOUNTER id;
 GNode* root;
 GList* list;
 GPtrArray* nodeArray;      /*GNode pointers */
 GPtrArray* recOR;
 GPtrArray* recCons;
 GList* recs;
 GList* actRec;
 SCOUNTER stored;
 SCOUNTER suppNodes;
 //Graphic part
 SCOUNTER maxTreeAge;
 SCOUNTER x;
 SCOUNTER y;
 SREAL best;
 SREAL worst;
};

struct tagPHYLONODE {
 gchar* name;
 GNode* node;
 TERMINAL* term;
 PHYLOTREE* tree;
 SCOUNTER minAge;
 SCOUNTER id;           /* Simple id, used for reconstructions */
 RECONS* rec;           /* Actual reconstruction */
 SREAL bremer;          /* "Bremer" support for the node */
 ALLOPATRY* alloSet;    /* Nodes with the same allopatric pattern */
 SCOUNTER withData;
 SCOUNTER fill;
 SFLAG constrained;
 /* Graphic part */
 SCOUNTER numTerms;
 SINTEGER size;
 SCOUNTER level;
 SCOUNTER nested;
 /* Graphic part, virtual screen */
 SINTEGER ageX;
 SINTEGER minX;
 SINTEGER maxX;
 SINTEGER y;
 SINTEGER startY;
 SINTEGER endY;
 /* Graphic part, actual screen */
 SINTEGER actX;
 SINTEGER actY;
 SINTEGER topY;
 SINTEGER downY;
 gpointer layout;
};

struct tagTERMINAL {
 gchar* name;
 GNode* node;
 GList* list;
 GList* firstGeo;
 GList* synonyms;
 SCOUNTER numGeo;
 BITFIELD* obs;
};

struct tagGEOREF {
 gchar* name;
 SREAL lat;
 SREAL lon;
 GList* list;
 SCOUNTER actCell;
 /* Standarized georeference */
 SREAL inclination;     /* Zenith from equatorial plane to north pole */
 SREAL azimuth;         /* Reference plane, the equatorial plane */
 GList* sp;
 /* Graphic part, actual screen */
 SINTEGER actX;
 SINTEGER actY;
 /* Other info */
 gchar* collec;         /* Collection name */
 gchar* collector;
 gchar* catalog;        /* Catalog id */
 gchar* url;            /* Internet url */
 gchar* reference;      /* Bibliographic reference */
 gchar* identifier;     /* The person that made the identification */
 gchar* sampleDate;     /* Sampling date */
};

struct tagRECONS {
 PHYLONODE* node;
 BITFIELD* cells;       /* Cells asigned to the node */
 BITFIELD* obs;         /* Cells with "physical" points */
 PHYLONODE* setA;
 PHYLONODE* setB;
 PHYLONODE* removSet;
 SFLAG nodeFlag;
 /* Data for incremental optimization */
 SCOUNTER numVics;
 SCOUNTER numRemovNodes;
 SCOUNTER numNoVics;
 SREAL cost;
 /* Data for node cost calculation */
 SCOUNTER numCells;
 SCOUNTER numObs;
 SCOUNTER active;
 SREAL overlap;
 SCOUNTER timeLag;
 SCOUNTER maxAge;
 /* Barrier */
 GList* triangles;
};

struct tagALLOPATRY {
 GList* list;
 BITFIELD* baseA;
 BITFIELD* baseB;
 BITFIELD* setA;
 BITFIELD* setB;
 SCOUNTER numNodes;
 GList* nodeList;
 /* Barrier */
 GList* triangles;
};

struct tagSEARCHPARAM {
 SREAL costNoVic;
 SREAL costRemov;
 SREAL costPartRemov;
 SREAL costTimeLag;
 SREAL overlap;
 SREAL extra;
 SINTEGER timeLag;
 SINTEGER maxTime;
 SCOUNTER timeDiff;
 SFLAG mode;
 SFLAG search;
 SINTEGER acceptEqual;
 SFLAG algorithm;
 SFLAG keepAll;
 SCOUNTER its;      /* Iterations */
 SCOUNTER sectSize;
 SCOUNTER abandon;
 SCOUNTER generations;
};

#endif
/* bitfield.h is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */


#ifndef BIT_FIELD_HEADER_INCLUDED
#define BIT_FIELD_HEADER_INCLUDED

/* ------ Bit field header ------ */
/* A set of bitfield operations */

#include "ghead.h"

typedef guint16 BITFIELD;
#define BITS_ON_FIELD 16

/* ------ Globals ------ */
gsize maxFieldSize;
#define GETBITFIELDSIZE maxFieldSize

void StartBitFields (SCOUNTER bits);
void EndBitFields ();

BITFIELD* CreateBitField ();
void ResetBitField (BITFIELD* a);

void PutBitOn (BITFIELD* a, SCOUNTER bit);
SFLAG IsBitOn (BITFIELD* a, SCOUNTER bit);
void CopyBitField (BITFIELD* dest, BITFIELD* src);

SFLAG BitFieldAnd (BITFIELD* a1, BITFIELD* a2);

void BitFieldORsum (BITFIELD* dest, BITFIELD* src);
void BitFieldANDsum (BITFIELD* dest, BITFIELD* src);
void BitFieldRemove (BITFIELD* dest, BITFIELD* toRemov);

void BitFieldAddNotRemoved (BITFIELD* dest, BITFIELD* src, BITFIELD* toRemov);

SCOUNTER CountBits (BITFIELD* a);
SCOUNTER BitFieldCommon (BITFIELD* a, BITFIELD* b);

#endif
/* dialogs.h is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */

#ifndef DIALOGS_HEADER_INCLUDED
#define DIALOGS_HEADER_INCLUDED

#include <gtk/gtk.h>
#include "progbar.h"

typedef struct tagGRIDDIALOG GRIDDIALOG;
struct tagGRIDDIALOG {
 GtkWidget* dialog;
 GtkWidget* sizeEntry;
 GtkWidget* fillEntry;
 GtkWidget* mooreRadio;
 GtkWidget* vonneRadio;
};

GRIDDIALOG* GridDialog (SREAL size, SINTEGER fill, SFLAG editable);
void DestroyGridDialog (GRIDDIALOG* grid);

typedef struct tagENTRYDIALOG ENTRYDIALOG;
struct tagENTRYDIALOG {
 GtkWidget* dialog;
 GtkWidget* entry;
};

ENTRYDIALOG* CounterEntryDialog (GtkWindow* parent, SCOUNTER count, gchar* name, gchar* caption);
ENTRYDIALOG* RealEntryDialog (GtkWindow* parent, SREAL number, gchar* name, gchar* caption);
ENTRYDIALOG* TextEntryDialog (GtkWindow* parent, gchar* text, gchar* name, gchar* caption);
void DestroyEntryDialog (ENTRYDIALOG* toDel);

typedef struct tagCENTREDIALOG CENTREDIALOG;
struct tagCENTREDIALOG {
 GtkWidget* dialog;
 GtkWidget* latEntry;
 GtkWidget* lonEntry;
};

CENTREDIALOG* CentreDialog ();
void DestroyCentreDialog (CENTREDIALOG* centre);

typedef struct tagMAPLIMITSDIALOG MAPLIMITSDIALOG;
struct tagMAPLIMITSDIALOG {
 GtkWidget* dialog;
 GtkWidget* northEntry;
 GtkWidget* southEntry;
 GtkWidget* eastEntry;
 GtkWidget* westEntry;
};

MAPLIMITSDIALOG* MapLimitsDialog (SREAL n, SREAL s, SREAL e, SREAL w);
void DestroyMapLimitsDialog (MAPLIMITSDIALOG* mapLimits);

typedef struct tagSEARCHDIALOG SEARCHDIALOG;
struct tagSEARCHDIALOG {
 GtkWidget* dialog;
 GtkWidget* iterEntry;
 GtkWidget* equalEntry;
 GtkWidget* pageCheck;
 GtkWidget* flipCheck;
 GtkWidget* consCheck;
 GtkWidget* noSecRadio;
 GtkWidget* startSecRadio;
 GtkWidget* secRadio;
 GtkWidget* secSizeEntry;
 GtkWidget* noAnnealRadio;
 GtkWidget* blindAnnealRadio;
 GtkWidget* gibbsAnnealRadio;
 GtkWidget* nixonAnnealRadio;
 GtkWidget* abandonEntry;
 GtkWidget* generationEntry;
 GtkWidget* keepCheck;
 GtkWidget* allCheck;
};

SEARCHDIALOG* SearchDialog ();
void DestroySearchDialog (SEARCHDIALOG* searchDialog);

typedef struct tagSEARCHPROGDLG SEARCHPROGDLG;
struct tagSEARCHPROGDLG {
 GtkWidget* dialog;
 GtkWidget* treeProgress;
 GtkWidget* iterProgress;
 GtkWidget* stopButton;
 SFLAG progressUp;
 PROGRESSSTRUCT* progressData;
};

SEARCHPROGDLG* SearchProgress (PROGRESSSTRUCT* progress, SFLAG mode);
void DestroyProgressSearchDialog (SEARCHPROGDLG* toDel);

gboolean UpdateSearchProgress (gpointer data);

typedef struct tagPARSINGPROGDLG PARSINGPROGDLG;
struct tagPARSINGPROGDLG {
 GtkWidget* dialog;
 GtkWidget* fileProgress;
 SFLAG progressUp;
 PROGRESSSTRUCT* progressData;
};

PARSINGPROGDLG* ParsingProgress (PROGRESSSTRUCT* progress);
void DestroyParsingProgressDialog (PARSINGPROGDLG* toDel);

gboolean UpdateParsingProgress (gpointer data);

typedef struct tagGEOREFBROWSERDLG GEOREFBROWSERDLG;
struct tagGEOREFBROWSERDLG {
 GtkWidget* dialog;
 GtkWidget* spEntry;
 GtkWidget* nameEntry;
 GtkWidget* latEntry;
 GtkWidget* lonEntry;
 GtkWidget* collecEntry;
 GtkWidget* catalogEntry;
 GtkWidget* urlEntry;
 GtkWidget* referenceEntry;
 GtkWidget* identifierEntry;
 GtkWidget* sampleEntry;
 GtkWidget* collectorEntry;
 GtkWidget* nextButton;
 GtkWidget* prevButton;
 GtkWidget* applyButton;
 GtkWidget* delButton;
 GEOREF* actGeo;
};

GEOREFBROWSERDLG* GeoRefBrowser (GEOREF* geo, SFLAG mode);

typedef struct tagNODEDATADIALOG NODEDATADIALOG;
struct tagNODEDATADIALOG {
 GtkWidget* dialog;
 GtkWidget* nameEntry;
 GtkWidget* ageEntry;
};

NODEDATADIALOG* NodeDataDialog (GNode* node, SFLAG mode);
void DestroyNodeDataDialog (NODEDATADIALOG* toDel);

#endif
/* ghead.h is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */

#ifndef GENERAL_HEADER_INCLUDED
#define GENERAL_HEADER_INCLUDED

/* ------ General header ------ */
/* The general header is based on glib, as the main project
   is under GTK+.
   But the general part, just uses the non GUI (i.e. glib)
   part of GTK+.
   */
#include <glib.h>

/* ------ Simple types ------ */
/* Synonims of the general types */

#define SINTEGER glong
#define SREAL gdouble
/* Use scounter to declare explicit unsigned values, it is negative for domain errors */
#define SCOUNTER SINTEGER

typedef guint16 SFLAG;      /* Control type */

/* Constant used to start strings */
#define VOIDSTRING "NoStr"

/* ------ Complex types ------ */

/* --- Arrays --- */
/* I use my own array function and structures because the strange behaviuor of the
   glib arrays. */

typedef struct tagINTARRAY INTARRAY;
struct tagINTARRAY {
 SINTEGER* data;
 SCOUNTER len;
 SCOUNTER max;
};

INTARRAY* CreateIntArray (SCOUNTER size);
#define AddToArray(x,y)         (x) -> data [(x) -> len ++] = (y)
#define InsertValToArray(x,y,i) (x) -> data [(i)] = (y)
#define GetFromArray(x,i)       (x) -> data [(i)]
#define GetArraySize(x)         (x) -> len
#define GetArrayMax(x)          (x) -> max
void ShuffleIntArray (INTARRAY* array);
SCOUNTER SearchIndexValue (INTARRAY* array, SINTEGER val);
void DestroyIntArray (INTARRAY* array);

void CopyPtrArray (GPtrArray* dest, GPtrArray* src);
void CopyPtrArrayPart (GPtrArray* dest, GPtrArray* src, SCOUNTER start, SCOUNTER end);
void ShufflePtrArray (GPtrArray* array);
void ShuffleSinglePtrArray (GPtrArray* array, SCOUNTER pos);

GList* NameOnList (gchar* name, GList* list);

/* --- Functions for linked structures --- */
/* Some home-made functions for linked structures, as the standar implementation of glib
   states that does not guarantee to keep the original pointer on additions, removals,
   etc. on lists */

/* Insert an chain-element after prev */
GList* InsertListElement (GList* prev, gpointer data);
/* Deletes a list */
void DestroyList (GList* toDel, gsize sz);
void DestroyLinkElement (GList* toDel, gsize sz);
void DestroyLink (GList* toDel);
/* Put a list after last element of the destination list */
void PutAtListEnd (GList* dest, GList* toPut);
void AddChildren (GNode* toAdd, GNode* anc, GNode* lastSibling);
/* Get next valid node */
GNode* GetNextNode (GNode* actNode);
GNode* GetNextTerm (GNode* actNode);
GNode* GetPrevNode (GNode* actNode);
SFLAG IsDesc (GNode* probe, GNode* anc);

/* ------ Output log */
GString* outLog;

/* ------ Error interface ------ */
/* Used only for user errors (e.g. "bad file format") */
gchar* errorMessage;

/* This function must be defined in the user interface, and must include the RESETERROR macro */
#define RESETERROR g_free (errorMessage)
void OnError ();

#endif
/* grid.h is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */


#ifndef GRID_HEADER_INCLUDED
#define GRID_HEADER_INCLUDED

#include "ghead.h"
#include "bitfield.h"
#include "biogeo.h"

typedef struct tagQUADTREE QUADTREE;
struct tagQUADTREE {
 SINTEGER firstRow;
 SINTEGER lastRow;
 SINTEGER firstCol;
 SINTEGER lastCol;
 SCOUNTER cell;
 SINTEGER size;
 SFLAG state;
 SFLAG obsCell;
 GNode* node;
};

/* ------ Global variables ------ */
SCOUNTER numCells;
SCOUNTER numRows;
SCOUNTER numCols;

#define GETNUMCELLS numCells
#define GETNUMCOLS  numCols
#define GETNUMROWS  numRows

SREAL actualGridSize;
SREAL gridStartX;
SREAL gridStartY;
SCOUNTER actualFill;

SFLAG gridMode;

#define GETGRIDSIZE   actualGridSize
#define GETGRIDSTARTX gridStartX
#define GETGRIDSTARTY gridStartY
#define GETFILL       actualFill

#define SETGRIDSIZE(x)   actualGridSize = (x)
#define SETFILL(x)       actualFill = (x)

#define MOORE_NEIGHBORS      0x01
#define VONNEUMANN_NEIGHBORS 0x02

#define GETGRIDMODE     gridMode
#define SETGRIDMODE(x)  gridMode = (x)

/* ------ Functions ------ */
SCOUNTER GetCell (SINTEGER row, SINTEGER col);
void CreateGrid ();
void DestroyGrid ();

SFLAG GetGridState ();

void MapIntoQTree (BITFIELD* setA, BITFIELD* setB);
SFLAG GetRowColState (SINTEGER row, SINTEGER col);
void GetColRowFromCell (SCOUNTER cell, SINTEGER* row, SINTEGER* col);

SCOUNTER GetCellIfObs (SINTEGER row, SINTEGER col);

void DoFill (BITFIELD* dest, BITFIELD* obs, SCOUNTER fillValue);
void GetObsSet (BITFIELD* obs, GNode* refNode);

SCOUNTER CalculatePhyloDiv (SCOUNTER cell, SCOUNTER max);

#endif
/* in.h is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */

#ifndef LOW_LEVEL_INPUT_HEADER_INCLUDED
#define LOW_LEVEL_INPUT_HEADER_INCLUDED

/* ------ Low level input ------ */
/* Structures and functions to control user generated flow (strings, numbers, etc.) */

#include "ghead.h"

typedef struct tagINSTREAM INSTREAM;

/* INSTREAM states */
#define OPEN_STREAM 1
#define EOF_REACHED 0

#define BAD_IN_FORMAT 2
#define END_CHAR_FOUND 4
#define END_OF_IN_LINE 8

/* Global reading modes */
#define READ_AS_IS 0
#define ALL_LOWER  1
SFLAG readingMode;

#define SETREADINGMODE(x) readingMode = (x)

/* Create and destroy a stream */
INSTREAM* CreateInStream ();
void DestroyInStream (INSTREAM* in);
void ResetInStream (INSTREAM* in);

/* Set the stream */
SFLAG SetInStreamToFile (INSTREAM* in, gchar* name);
SFLAG SetInStreamToString (INSTREAM* in, GString* src);

/* Reading the input */
SFLAG ReadString (INSTREAM* in, GString* dest);
SFLAG ReadStringUntil (INSTREAM* in, GString* dest, gchar end);
SFLAG PeekString (INSTREAM* in, GString* dest);
SFLAG ReadBoundComment (INSTREAM* in, GString* dest, gchar start, gchar end);
SFLAG ReadInt (INSTREAM* in, SINTEGER* number);
SFLAG ReadReal (INSTREAM* in, SREAL* number);
SFLAG ReadChar (INSTREAM* in, gchar* dest);

void SkipSpaces (INSTREAM* in);

/* ------ Type definitions ------ */

struct tagINSTREAM {
 gchar* stream;
 gchar* name;
 SCOUNTER actPos;
 SCOUNTER line;
 gsize len;
 SFLAG state;
 gchar nextChar;
};

#endif
/* inxml.h is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */

#ifndef INPUT_STREAM_FOR_KML_LIKE_FILES_HEADER_INCLUDED
#define INPUT_STREAM_FOR_KML_LIKE_FILES_HEADER_INCLUDED

/* ------ xml file reading ------ */
/* Functions to parse and read a sub-set of XML files, used as input */

#include "ghead.h"
#include "in.h"

#define XML_ERROR           0x0000
#define XML_TOKEN_OPEN      0x0001
#define XML_TOKEN_CLOSE     0x0002
#define XML_TOKEN_COMMENT   0x0004
#define XML_TOKEN_CONTENT   0x0008
#define XML_TOKEN_CDATA     0x0010
#define XML_TOKEN_NOCONTENT 0x0020
#define XML_TOKEN_EMPTY     0x0040
#define XML_WITH_PARAMS     0x0100
#define XML_NO_PARAMS       0x0200
#define XML_AUTO_CLOSE      0x0400

typedef struct tagXMLPARSER XMLPARSER;

/* Reading the XML input */
SFLAG IsXMLFile (INSTREAM* in);
SFLAG ReadToken (INSTREAM* in, GString* dest, SFLAG* tokenState);
SFLAG ReadXMLTokenParam (INSTREAM* in, GString* paramName, GString* paramVal, SFLAG* tokenState);
SFLAG ReadXMLContent (INSTREAM* in, GString* dest, SFLAG* textState);

SFLAG SkipParams (INSTREAM* in);

SFLAG ReadXMLFile (XMLPARSER* xmlParser);

void RemoveFromTokenList (XMLPARSER* xmlParser);
SFLAG IgnoreOpenTag (XMLPARSER* xmlParser);
GString* GetPrevToken (XMLPARSER* xmlParser);

/* Definitions */

struct tagXMLPARSER {
 SFLAG (*OpenTag) ();
 SFLAG (*EndTag) ();
 SFLAG (*OnError) ();

 GString* tempParserString;
 INSTREAM* in;
 gchar* fileHeader;
 GList* tokenStack;
 SFLAG tokenState;
};

#endif
/* out.h is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */

#ifndef OUT_HEADER_INCLUDED
#define OUT_HEADER_INCLUDED

#include "ghead.h"
#include "biogeo.h"

#define USE_NEXUS 1
#define USE_TNT   2
#define USE_NAMES 4

#define SVG_NO_REC    0
#define SVG_WITH_REC  1
#define SVG_WITH_SUPP 2

void WriteXMLFile (gchar* fileName);
void WriteKMLFile (gchar* fileName);
void WriteXYDFile (gchar* fileName);
void WriteTABTable (gchar* fileName);
void WriteLstFile (gchar* fileName);
void WriteCllFile (gchar* fileName);
void WriteVRXFile (gchar* fileName);
void WriteXMLWithData (gchar* fileName);
void WritePhyFile (gchar* fileName);
void WriteCladoFile (gchar* fileName);

void WriteOutLog (gchar* fileName, GString* outBuffer);

void WriteSVGFile (gchar* fileName, PHYLOTREE* tree, SFLAG withRec);
void WritePhyTree (gchar* fileName, PHYLOTREE* tree);
void WriteCladoTree (gchar* fileName, PHYLOTREE* tree);

void WriteKMLRecons (gchar* fileName, GNode* refNode, SFLAG modeRecords, SFLAG modeRemoved);
void WriteKMLTree (gchar* fileName, PHYLOTREE* tree, SFLAG modeRecords, SFLAG modeRemoved);
void WriteNewickFile (gchar* fileName, PHYLOTREE* tree, SFLAG mode);

void WriteSpRichness (gchar* fileName);
void WriteRichKMLFile (gchar* fileName);
void WritePhyloDiv (gchar* fileName);
void WritePDKMLFile (gchar* fileName);
void WriteNDMMatrix (gchar* fileName);

void CheckRealStr (gchar* num);

gchar* CheckExtention (gchar* name, gchar* extention);

#endif
/* parser.h is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */

#ifndef PARSER_HEADER_INCLUDED
#define PARSER_HEADER_INCLUDED

/* ------ High level input ------ */
/* Parser functions for high level control of user generated input */

#include "ghead.h"
#include "progbar.h"

/* Read an input file */
SFLAG ReadFile (gchar* name, PROGRESSSTRUCT* progress);
SFLAG FeedFile (gchar* name, PROGRESSSTRUCT* progress);
SFLAG ReadResults (gchar* name, PROGRESSSTRUCT* progress);

/* Set a command line */
void SetCommandLine (gchar* text);
SFLAG GetCmdNumber (SINTEGER* dest);
SFLAG GetCmdReal (SREAL* dest);

#endif
/* progbar.h is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */

#ifndef PROGRESS_BAR_HEADER_INCLUDED
#define PROGRESS_BAR_HEADER_INCLUDED

#include "ghead.h"

/* Search progress */
typedef struct tagPROGRESSSTRUCT PROGRESSSTRUCT;

/* This function must be defined in the user interface, used for updating during calculations */
void UpdateOnCalculus ();

/* --- Definitions --- */
struct tagPROGRESSSTRUCT {
 SCOUNTER actBar1;
 SCOUNTER numBar1;
 SCOUNTER actBar2;
 SCOUNTER numBar2;
 SFLAG isActive;
};

#endif
/* tessellation.h is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */

#ifndef TESSELLATION_HEADER_INCLUDED
#define TESSELLATION_HEADER_INCLUDED

/* ------ Data types for the Voronoi "line" used to mark disjunctions ------ */

#include "biogeo.h"
#include "grid.h"
#include "bitfield.h"

#define STATE_NOCELL 0
#define STATE_SETA   1
#define STATE_SETB   2

/* Triangle use the trigonometric notation, with segment a as the
   segment in front point A (i.e. the segment between B-C) */
typedef struct tagDELAUNAYTRIANGLE DELAUNAYTRIANGLE;
struct tagDELAUNAYTRIANGLE {
 GList* list;
 SINTEGER col [3];
 SINTEGER row [3];
 SFLAG state [3];
 SREAL cX;
 SREAL cY;
 SREAL r;
 DELAUNAYTRIANGLE* neighbor [3];
};

GList* MakeVoronoi (BITFIELD* iniA, BITFIELD* iniB);
void DestroyTriangleList (GList* triangleList);

#endif
/* wins.h is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */

#ifndef WINDOWED_INTERFACE_HEADER_INCLUDED
#define WINDOWED_INTERFACE_HEADER_INCLUDED

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

/* ------ Windowed interface ------ */
/* General functions and structures used by the graphic user interface. */

#include "biogeo.h"
#include "grid.h"

/* ------ General functions ------ */
GtkWidget* CreateWindow (gchar* name, gint x, gint y, gint sizeX, gint sizeY, GCallback OnDelete, GtkWidget* box);

GtkWidget* AddMenuToBar (GtkWidget* menu, gchar* name, GtkWidget* menuBar);
GtkWidget* CreateMenuItem (GtkWidget* menu, gchar* name, GtkAccelGroup* accelGroup, guint key,
                           GdkModifierType modifier, GCallback func, gpointer data);
GtkWidget* CreateMenuItemCheck (GtkWidget* menu, gchar* name, GtkAccelGroup* accelGroup, guint key,
                                GdkModifierType modifier, GCallback func, gpointer data);
GtkWidget* CreateMenuItemRadio (GtkWidget* menu, gchar* name, GtkAccelGroup* accelGroup, guint key,
                                GdkModifierType modifier, GCallback func, gpointer data, GSList* group);
void CreateMenuSeparator (GtkWidget* menu);

GtkWidget* CreateEntry (gchar* label, GtkBox* box);

GtkFileFilter* CreateFileFilter (gchar* name, gchar* pattern, GtkWidget* fileChooser);

/* ------ Global windows ------ */

/* --- Main window --- */
GtkWidget* mainWindow;
#define GETMAINWIN mainWindow

void StartMainWindow ();
void OnChangeRec ();

/* --- Treee View window --- */
GtkWidget* treeViewWindow;
#define GETTREEWIN treeViewWindow

void StartTreeViewWindow ();
void OpenNewTree ();
void OnChangeTree (GList* nuTree);
void OnUpdateTree ();
void SetRecEdition (SFLAG mode);

/* --- Map window --- */
GtkWidget* mapWindow;
#define GETMAPWIN mapWindow

void StartMapWindow ();
void ChangeGeo ();
void ActivateGrid (SFLAG mode);
void SetEdition (SFLAG mode);

void PaintNewGeoRef (GEOREF* geo);

/* ------ Other globals ------ */
GNode* activeNode;
GList* activeTree;

#endif
